"""Defines various tool bars displayed at the bottom of apps."""

# Define toolbar buffer names

SEARCH_BAR_BUFFER = "SEARCH_BAR_BUFFER"
COMMAND_BAR_BUFFER = "COMMAND_BAR_BUFFER"

BAR_BUFFERS = {
    SEARCH_BAR_BUFFER,
    COMMAND_BAR_BUFFER,
}
