"""Define collections of generic key-bindings which do not belong to widgets."""

from . import basic, completion, micro, mouse, page_navigation, vi

__all__ = ["basic", "completion", "micro", "mouse", "page_navigation", "vi"]
