"""Register additional key escape sequences."""

from enum import Enum
from typing import TYPE_CHECKING, cast

from prompt_toolkit.input.ansi_escape_sequences import ANSI_SEQUENCES

if TYPE_CHECKING:
    from prompt_toolkit.keys import Keys


class MoreKeys(str, Enum):
    """Additional key definitions."""

    # Special terminal response keys
    ColorsResponse = "<colors-response>"
    PixelSizeResponse = "<pixel-size-response>"
    KittyGraphicsStatusResponse = "<kitty-graphics-status-response>"
    SixelGraphicsStatusResponse = "<sixel-graphics-status-response>"
    ItermGraphicsStatusResponse = "<iterm-graphics-status-response>"
    SgrPixelStatusResponse = "<sgr-pixel-status-response>"
    ClipboardDataResponse = "<clipboard-data-response>"

    # Regular key-presses

    AltTab = "A-tab"
    AltEnter = "A-enter"
    AltEscape = "A-escape"
    AltSpace = "A-space"
    AltTilde = "A-~"
    AltRightcurlybracket = "A-}"
    AltVerticalline = "A-|"
    AltLeftcurlybracket = "A-{"
    AltLatinsmallletterz = "A-z"
    AltLatinsmalllettery = "A-y"
    AltLatinsmallletterx = "A-x"
    AltLatinsmallletterw = "A-w"
    AltLatinsmallletterv = "A-v"
    AltLatinsmallletteru = "A-u"
    AltLatinsmalllettert = "A-t"
    AltLatinsmallletters = "A-s"
    AltLatinsmallletterr = "A-r"
    AltLatinsmallletterq = "A-q"
    AltLatinsmallletterp = "A-p"
    AltLatinsmalllettero = "A-o"
    AltLatinsmalllettern = "A-n"
    AltLatinsmallletterm = "A-m"
    AltLatinsmallletterl = "A-l"
    AltLatinsmallletterk = "A-k"
    AltLatinsmallletterj = "A-j"
    AltLatinsmallletteri = "A-i"
    AltLatinsmallletterh = "A-h"
    AltLatinsmallletterg = "A-g"
    AltLatinsmallletterf = "A-f"
    AltLatinsmalllettere = "A-e"
    AltLatinsmallletterd = "A-d"
    AltLatinsmallletterc = "A-c"
    AltLatinsmallletterb = "A-b"
    AltLatinsmalllettera = "A-a"
    AltGraveaccent = "A-`"
    AltLowline = "A-_"
    AltCircumflexaccent = "A-^"
    AltRightsquarebracket = "A-]"
    AltReversesolidus = "A-\\"
    AltLeftsquarebracket = "A-["
    AltLatincapitalletterz = "A-Z"
    AltLatincapitallettery = "A-Y"
    AltLatincapitalletterx = "A-X"
    AltLatincapitalletterw = "A-W"
    AltLatincapitalletterv = "A-V"
    AltLatincapitalletteru = "A-U"
    AltLatincapitallettert = "A-T"
    AltLatincapitalletters = "A-S"
    AltLatincapitalletterr = "A-R"
    AltLatincapitalletterq = "A-Q"
    AltLatincapitalletterp = "A-P"
    AltLatincapitallettero = "A-O"
    AltLatincapitallettern = "A-N"
    AltLatincapitalletterm = "A-M"
    AltLatincapitalletterl = "A-L"
    AltLatincapitalletterk = "A-K"
    AltLatincapitalletterj = "A-J"
    AltLatincapitalletteri = "A-I"
    AltLatincapitalletterh = "A-H"
    AltLatincapitalletterg = "A-G"
    AltLatincapitalletterf = "A-F"
    AltLatincapitallettere = "A-E"
    AltLatincapitalletterd = "A-D"
    AltLatincapitalletterc = "A-C"
    AltLatincapitalletterb = "A-B"
    AltLatincapitallettera = "A-A"
    AltCommercialat = "A-@"
    AltQuestionmark = "A-?"
    AltGreaterthansign = "A->"
    AltEqualssign = "A-="
    AltLessthansign = "A-<"
    AltSemicolon = "A-;"
    AltColon = "A-:"
    AltDigitnine = "A-9"
    AltDigiteight = "A-8"
    AltDigitseven = "A-7"
    AltDigitsix = "A-6"
    AltDigitfive = "A-5"
    AltDigitfour = "A-4"
    AltDigitthree = "A-3"
    AltDigittwo = "A-2"
    AltDigitone = "A-1"
    AltDigitzero = "A-0"
    AltSolidus = "A-/"
    AltFullstop = "A-."
    AltHyphenminus = "A--"
    AltComma = "A-,"
    AltPlussign = "A-+"
    AltAsterisk = "A-*"
    AltRightparenthesis = "A-)"
    AltLeftparenthesis = "A-("
    AltApostrophe = "A-'"
    AltAmpersand = "A-&"
    AltPercentsign = "A-%"
    AltDollarsign = "A-$"
    AltNumbersign = "A-#"
    AltQuotationmark = 'A-"'
    AltExclamationmark = "A-!"
    AltBackspace = "A-backspace"
    Escape = "escape"
    Insert = "insert"
    Delete = "delete"
    Pageup = "pageup"
    Pagedown = "pagedown"
    Up = "up"
    Down = "down"
    Right = "right"
    Left = "left"
    Home = "home"
    End = "end"
    Printscreen = "printscreen"
    Pause = "pause"
    Menu = "menu"
    F1 = "f1"
    F2 = "f2"
    F3 = "f3"
    F4 = "f4"
    F5 = "f5"
    F6 = "f6"
    F7 = "f7"
    F8 = "f8"
    F9 = "f9"
    F10 = "f10"
    F11 = "f11"
    F12 = "f12"
    Backspace = "backspace"
    Enter = "enter"
    Tab = "tab"
    ShiftTab = "s-tab"
    ShiftEnter = "s-enter"
    ShiftEscape = "s-escape"
    ShiftSpace = "s-space"
    ShiftTilde = "s-~"
    ShiftRightcurlybracket = "s-}"
    ShiftVerticalline = "s-|"
    ShiftLeftcurlybracket = "s-{"
    ShiftLatinsmallletterz = "s-z"
    ShiftLatinsmalllettery = "s-y"
    ShiftLatinsmallletterx = "s-x"
    ShiftLatinsmallletterw = "s-w"
    ShiftLatinsmallletterv = "s-v"
    ShiftLatinsmallletteru = "s-u"
    ShiftLatinsmalllettert = "s-t"
    ShiftLatinsmallletters = "s-s"
    ShiftLatinsmallletterr = "s-r"
    ShiftLatinsmallletterq = "s-q"
    ShiftLatinsmallletterp = "s-p"
    ShiftLatinsmalllettero = "s-o"
    ShiftLatinsmalllettern = "s-n"
    ShiftLatinsmallletterm = "s-m"
    ShiftLatinsmallletterl = "s-l"
    ShiftLatinsmallletterk = "s-k"
    ShiftLatinsmallletterj = "s-j"
    ShiftLatinsmallletteri = "s-i"
    ShiftLatinsmallletterh = "s-h"
    ShiftLatinsmallletterg = "s-g"
    ShiftLatinsmallletterf = "s-f"
    ShiftLatinsmalllettere = "s-e"
    ShiftLatinsmallletterd = "s-d"
    ShiftLatinsmallletterc = "s-c"
    ShiftLatinsmallletterb = "s-b"
    ShiftLatinsmalllettera = "s-a"
    ShiftGraveaccent = "s-`"
    ShiftLowline = "s-_"
    ShiftCircumflexaccent = "s-^"
    ShiftRightsquarebracket = "s-]"
    ShiftReversesolidus = "s-\\"
    ShiftLeftsquarebracket = "s-["
    ShiftLatincapitalletterz = "s-Z"
    ShiftLatincapitallettery = "s-Y"
    ShiftLatincapitalletterx = "s-X"
    ShiftLatincapitalletterw = "s-W"
    ShiftLatincapitalletterv = "s-V"
    ShiftLatincapitalletteru = "s-U"
    ShiftLatincapitallettert = "s-T"
    ShiftLatincapitalletters = "s-S"
    ShiftLatincapitalletterr = "s-R"
    ShiftLatincapitalletterq = "s-Q"
    ShiftLatincapitalletterp = "s-P"
    ShiftLatincapitallettero = "s-O"
    ShiftLatincapitallettern = "s-N"
    ShiftLatincapitalletterm = "s-M"
    ShiftLatincapitalletterl = "s-L"
    ShiftLatincapitalletterk = "s-K"
    ShiftLatincapitalletterj = "s-J"
    ShiftLatincapitalletteri = "s-I"
    ShiftLatincapitalletterh = "s-H"
    ShiftLatincapitalletterg = "s-G"
    ShiftLatincapitalletterf = "s-F"
    ShiftLatincapitallettere = "s-E"
    ShiftLatincapitalletterd = "s-D"
    ShiftLatincapitalletterc = "s-C"
    ShiftLatincapitalletterb = "s-B"
    ShiftLatincapitallettera = "s-A"
    ShiftCommercialat = "s-@"
    ShiftQuestionmark = "s-?"
    ShiftGreaterthansign = "s->"
    ShiftEqualssign = "s-="
    ShiftLessthansign = "s-<"
    ShiftSemicolon = "s-;"
    ShiftColon = "s-:"
    ShiftDigitnine = "s-9"
    ShiftDigiteight = "s-8"
    ShiftDigitseven = "s-7"
    ShiftDigitsix = "s-6"
    ShiftDigitfive = "s-5"
    ShiftDigitfour = "s-4"
    ShiftDigitthree = "s-3"
    ShiftDigittwo = "s-2"
    ShiftDigitone = "s-1"
    ShiftDigitzero = "s-0"
    ShiftSolidus = "s-/"
    ShiftFullstop = "s-."
    ShiftHyphenminus = "s--"
    ShiftComma = "s-,"
    ShiftPlussign = "s-+"
    ShiftAsterisk = "s-*"
    ShiftRightparenthesis = "s-)"
    ShiftLeftparenthesis = "s-("
    ShiftApostrophe = "s-'"
    ShiftAmpersand = "s-&"
    ShiftPercentsign = "s-%"
    ShiftDollarsign = "s-$"
    ShiftNumbersign = "s-#"
    ShiftQuotationmark = 's-"'
    ShiftExclamationmark = "s-!"
    ShiftBackspace = "s-backspace"
    ShiftInsert = "s-insert"
    ShiftDelete = "s-delete"
    ShiftPageup = "s-pageup"
    ShiftPagedown = "s-pagedown"
    ShiftUp = "s-up"
    ShiftDown = "s-down"
    ShiftRight = "s-right"
    ShiftLeft = "s-left"
    ShiftHome = "s-home"
    ShiftEnd = "s-end"
    ShiftPrintscreen = "s-printscreen"
    ShiftPause = "s-pause"
    ShiftMenu = "s-menu"
    ShiftF1 = "s-f1"
    ShiftF2 = "s-f2"
    ShiftF3 = "s-f3"
    ShiftF4 = "s-f4"
    ShiftF5 = "s-f5"
    ShiftF6 = "s-f6"
    ShiftF7 = "s-f7"
    ShiftF8 = "s-f8"
    ShiftF9 = "s-f9"
    ShiftF10 = "s-f10"
    ShiftF11 = "s-f11"
    ShiftF12 = "s-f12"
    AltInsert = "A-insert"
    AltDelete = "A-delete"
    AltPageup = "A-pageup"
    AltPagedown = "A-pagedown"
    AltUp = "A-up"
    AltDown = "A-down"
    AltRight = "A-right"
    AltLeft = "A-left"
    AltHome = "A-home"
    AltEnd = "A-end"
    AltPrintscreen = "A-printscreen"
    AltPause = "A-pause"
    AltMenu = "A-menu"
    AltF1 = "A-f1"
    AltF2 = "A-f2"
    AltF3 = "A-f3"
    AltF4 = "A-f4"
    AltF5 = "A-f5"
    AltF6 = "A-f6"
    AltF7 = "A-f7"
    AltF8 = "A-f8"
    AltF9 = "A-f9"
    AltF10 = "A-f10"
    AltF11 = "A-f11"
    AltF12 = "A-f12"
    ControlTab = "c-tab"
    ControlEnter = "c-enter"
    ControlEscape = "c-escape"
    ControlSpace = "c-space"
    ControlTilde = "c-~"
    ControlRightcurlybracket = "c-}"
    ControlVerticalline = "c-|"
    ControlLeftcurlybracket = "c-{"
    ControlLatinsmallletterz = "c-z"
    ControlLatinsmalllettery = "c-y"
    ControlLatinsmallletterx = "c-x"
    ControlLatinsmallletterw = "c-w"
    ControlLatinsmallletterv = "c-v"
    ControlLatinsmallletteru = "c-u"
    ControlLatinsmalllettert = "c-t"
    ControlLatinsmallletters = "c-s"
    ControlLatinsmallletterr = "c-r"
    ControlLatinsmallletterq = "c-q"
    ControlLatinsmallletterp = "c-p"
    ControlLatinsmalllettero = "c-o"
    ControlLatinsmalllettern = "c-n"
    ControlLatinsmallletterm = "c-m"
    ControlLatinsmallletterl = "c-l"
    ControlLatinsmallletterk = "c-k"
    ControlLatinsmallletterj = "c-j"
    ControlLatinsmallletteri = "c-i"
    ControlLatinsmallletterh = "c-h"
    ControlLatinsmallletterg = "c-g"
    ControlLatinsmallletterf = "c-f"
    ControlLatinsmalllettere = "c-e"
    ControlLatinsmallletterd = "c-d"
    ControlLatinsmallletterc = "c-c"
    ControlLatinsmallletterb = "c-b"
    ControlLatinsmalllettera = "c-a"
    ControlGraveaccent = "c-`"
    ControlLowline = "c-_"
    ControlCircumflexaccent = "c-^"
    ControlRightsquarebracket = "c-]"
    ControlReversesolidus = "c-\\"
    ControlLeftsquarebracket = "c-["
    ControlLatincapitalletterz = "c-Z"
    ControlLatincapitallettery = "c-Y"
    ControlLatincapitalletterx = "c-X"
    ControlLatincapitalletterw = "c-W"
    ControlLatincapitalletterv = "c-V"
    ControlLatincapitalletteru = "c-U"
    ControlLatincapitallettert = "c-T"
    ControlLatincapitalletters = "c-S"
    ControlLatincapitalletterr = "c-R"
    ControlLatincapitalletterq = "c-Q"
    ControlLatincapitalletterp = "c-P"
    ControlLatincapitallettero = "c-O"
    ControlLatincapitallettern = "c-N"
    ControlLatincapitalletterm = "c-M"
    ControlLatincapitalletterl = "c-L"
    ControlLatincapitalletterk = "c-K"
    ControlLatincapitalletterj = "c-J"
    ControlLatincapitalletteri = "c-I"
    ControlLatincapitalletterh = "c-H"
    ControlLatincapitalletterg = "c-G"
    ControlLatincapitalletterf = "c-F"
    ControlLatincapitallettere = "c-E"
    ControlLatincapitalletterd = "c-D"
    ControlLatincapitalletterc = "c-C"
    ControlLatincapitalletterb = "c-B"
    ControlLatincapitallettera = "c-A"
    ControlCommercialat = "c-@"
    ControlQuestionmark = "c-?"
    ControlGreaterthansign = "c->"
    ControlEqualssign = "c-="
    ControlLessthansign = "c-<"
    ControlSemicolon = "c-;"
    ControlColon = "c-:"
    ControlDigitnine = "c-9"
    ControlDigiteight = "c-8"
    ControlDigitseven = "c-7"
    ControlDigitsix = "c-6"
    ControlDigitfive = "c-5"
    ControlDigitfour = "c-4"
    ControlDigitthree = "c-3"
    ControlDigittwo = "c-2"
    ControlDigitone = "c-1"
    ControlDigitzero = "c-0"
    ControlSolidus = "c-/"
    ControlFullstop = "c-."
    ControlHyphenminus = "c--"
    ControlComma = "c-,"
    ControlPlussign = "c-+"
    ControlAsterisk = "c-*"
    ControlRightparenthesis = "c-)"
    ControlLeftparenthesis = "c-("
    ControlApostrophe = "c-'"
    ControlAmpersand = "c-&"
    ControlPercentsign = "c-%"
    ControlDollarsign = "c-$"
    ControlNumbersign = "c-#"
    ControlQuotationmark = 'c-"'
    ControlExclamationmark = "c-!"
    ControlBackspace = "c-backspace"
    ControlInsert = "c-insert"
    ControlDelete = "c-delete"
    ControlPageup = "c-pageup"
    ControlPagedown = "c-pagedown"
    ControlUp = "c-up"
    ControlDown = "c-down"
    ControlRight = "c-right"
    ControlLeft = "c-left"
    ControlHome = "c-home"
    ControlEnd = "c-end"
    ControlPrintscreen = "c-printscreen"
    ControlPause = "c-pause"
    ControlMenu = "c-menu"
    ControlF1 = "c-f1"
    ControlF2 = "c-f2"
    ControlF3 = "c-f3"
    ControlF4 = "c-f4"
    ControlF5 = "c-f5"
    ControlF6 = "c-f6"
    ControlF7 = "c-f7"
    ControlF8 = "c-f8"
    ControlF9 = "c-f9"
    ControlF10 = "c-f10"
    ControlF11 = "c-f11"
    ControlF12 = "c-f12"
    AltShiftTab = "A-s-tab"
    AltShiftEnter = "A-s-enter"
    AltShiftEscape = "A-s-escape"
    AltShiftSpace = "A-s-space"
    AltShiftTilde = "A-s-~"
    AltShiftRightcurlybracket = "A-s-}"
    AltShiftVerticalline = "A-s-|"
    AltShiftLeftcurlybracket = "A-s-{"
    AltShiftLatinsmallletterz = "A-s-z"
    AltShiftLatinsmalllettery = "A-s-y"
    AltShiftLatinsmallletterx = "A-s-x"
    AltShiftLatinsmallletterw = "A-s-w"
    AltShiftLatinsmallletterv = "A-s-v"
    AltShiftLatinsmallletteru = "A-s-u"
    AltShiftLatinsmalllettert = "A-s-t"
    AltShiftLatinsmallletters = "A-s-s"
    AltShiftLatinsmallletterr = "A-s-r"
    AltShiftLatinsmallletterq = "A-s-q"
    AltShiftLatinsmallletterp = "A-s-p"
    AltShiftLatinsmalllettero = "A-s-o"
    AltShiftLatinsmalllettern = "A-s-n"
    AltShiftLatinsmallletterm = "A-s-m"
    AltShiftLatinsmallletterl = "A-s-l"
    AltShiftLatinsmallletterk = "A-s-k"
    AltShiftLatinsmallletterj = "A-s-j"
    AltShiftLatinsmallletteri = "A-s-i"
    AltShiftLatinsmallletterh = "A-s-h"
    AltShiftLatinsmallletterg = "A-s-g"
    AltShiftLatinsmallletterf = "A-s-f"
    AltShiftLatinsmalllettere = "A-s-e"
    AltShiftLatinsmallletterd = "A-s-d"
    AltShiftLatinsmallletterc = "A-s-c"
    AltShiftLatinsmallletterb = "A-s-b"
    AltShiftLatinsmalllettera = "A-s-a"
    AltShiftGraveaccent = "A-s-`"
    AltShiftLowline = "A-s-_"
    AltShiftCircumflexaccent = "A-s-^"
    AltShiftRightsquarebracket = "A-s-]"
    AltShiftReversesolidus = "A-s-\\"
    AltShiftLeftsquarebracket = "A-s-["
    AltShiftLatincapitalletterz = "A-s-Z"
    AltShiftLatincapitallettery = "A-s-Y"
    AltShiftLatincapitalletterx = "A-s-X"
    AltShiftLatincapitalletterw = "A-s-W"
    AltShiftLatincapitalletterv = "A-s-V"
    AltShiftLatincapitalletteru = "A-s-U"
    AltShiftLatincapitallettert = "A-s-T"
    AltShiftLatincapitalletters = "A-s-S"
    AltShiftLatincapitalletterr = "A-s-R"
    AltShiftLatincapitalletterq = "A-s-Q"
    AltShiftLatincapitalletterp = "A-s-P"
    AltShiftLatincapitallettero = "A-s-O"
    AltShiftLatincapitallettern = "A-s-N"
    AltShiftLatincapitalletterm = "A-s-M"
    AltShiftLatincapitalletterl = "A-s-L"
    AltShiftLatincapitalletterk = "A-s-K"
    AltShiftLatincapitalletterj = "A-s-J"
    AltShiftLatincapitalletteri = "A-s-I"
    AltShiftLatincapitalletterh = "A-s-H"
    AltShiftLatincapitalletterg = "A-s-G"
    AltShiftLatincapitalletterf = "A-s-F"
    AltShiftLatincapitallettere = "A-s-E"
    AltShiftLatincapitalletterd = "A-s-D"
    AltShiftLatincapitalletterc = "A-s-C"
    AltShiftLatincapitalletterb = "A-s-B"
    AltShiftLatincapitallettera = "A-s-A"
    AltShiftCommercialat = "A-s-@"
    AltShiftQuestionmark = "A-s-?"
    AltShiftGreaterthansign = "A-s->"
    AltShiftEqualssign = "A-s-="
    AltShiftLessthansign = "A-s-<"
    AltShiftSemicolon = "A-s-;"
    AltShiftColon = "A-s-:"
    AltShiftDigitnine = "A-s-9"
    AltShiftDigiteight = "A-s-8"
    AltShiftDigitseven = "A-s-7"
    AltShiftDigitsix = "A-s-6"
    AltShiftDigitfive = "A-s-5"
    AltShiftDigitfour = "A-s-4"
    AltShiftDigitthree = "A-s-3"
    AltShiftDigittwo = "A-s-2"
    AltShiftDigitone = "A-s-1"
    AltShiftDigitzero = "A-s-0"
    AltShiftSolidus = "A-s-/"
    AltShiftFullstop = "A-s-."
    AltShiftHyphenminus = "A-s--"
    AltShiftComma = "A-s-,"
    AltShiftPlussign = "A-s-+"
    AltShiftAsterisk = "A-s-*"
    AltShiftRightparenthesis = "A-s-)"
    AltShiftLeftparenthesis = "A-s-("
    AltShiftApostrophe = "A-s-'"
    AltShiftAmpersand = "A-s-&"
    AltShiftPercentsign = "A-s-%"
    AltShiftDollarsign = "A-s-$"
    AltShiftNumbersign = "A-s-#"
    AltShiftQuotationmark = 'A-s-"'
    AltShiftExclamationmark = "A-s-!"
    AltShiftBackspace = "A-s-backspace"
    AltShiftInsert = "A-s-insert"
    AltShiftDelete = "A-s-delete"
    AltShiftPageup = "A-s-pageup"
    AltShiftPagedown = "A-s-pagedown"
    AltShiftUp = "A-s-up"
    AltShiftDown = "A-s-down"
    AltShiftRight = "A-s-right"
    AltShiftLeft = "A-s-left"
    AltShiftHome = "A-s-home"
    AltShiftEnd = "A-s-end"
    AltShiftPrintscreen = "A-s-printscreen"
    AltShiftPause = "A-s-pause"
    AltShiftMenu = "A-s-menu"
    AltShiftF1 = "A-s-f1"
    AltShiftF2 = "A-s-f2"
    AltShiftF3 = "A-s-f3"
    AltShiftF4 = "A-s-f4"
    AltShiftF5 = "A-s-f5"
    AltShiftF6 = "A-s-f6"
    AltShiftF7 = "A-s-f7"
    AltShiftF8 = "A-s-f8"
    AltShiftF9 = "A-s-f9"
    AltShiftF10 = "A-s-f10"
    AltShiftF11 = "A-s-f11"
    AltShiftF12 = "A-s-f12"
    ControlShiftTab = "c-s-tab"
    ControlShiftEnter = "c-s-enter"
    ControlShiftEscape = "c-s-escape"
    ControlShiftSpace = "c-s-space"
    ControlShiftTilde = "c-s-~"
    ControlShiftRightcurlybracket = "c-s-}"
    ControlShiftVerticalline = "c-s-|"
    ControlShiftLeftcurlybracket = "c-s-{"
    ControlShiftLatinsmallletterz = "c-s-z"
    ControlShiftLatinsmalllettery = "c-s-y"
    ControlShiftLatinsmallletterx = "c-s-x"
    ControlShiftLatinsmallletterw = "c-s-w"
    ControlShiftLatinsmallletterv = "c-s-v"
    ControlShiftLatinsmallletteru = "c-s-u"
    ControlShiftLatinsmalllettert = "c-s-t"
    ControlShiftLatinsmallletters = "c-s-s"
    ControlShiftLatinsmallletterr = "c-s-r"
    ControlShiftLatinsmallletterq = "c-s-q"
    ControlShiftLatinsmallletterp = "c-s-p"
    ControlShiftLatinsmalllettero = "c-s-o"
    ControlShiftLatinsmalllettern = "c-s-n"
    ControlShiftLatinsmallletterm = "c-s-m"
    ControlShiftLatinsmallletterl = "c-s-l"
    ControlShiftLatinsmallletterk = "c-s-k"
    ControlShiftLatinsmallletterj = "c-s-j"
    ControlShiftLatinsmallletteri = "c-s-i"
    ControlShiftLatinsmallletterh = "c-s-h"
    ControlShiftLatinsmallletterg = "c-s-g"
    ControlShiftLatinsmallletterf = "c-s-f"
    ControlShiftLatinsmalllettere = "c-s-e"
    ControlShiftLatinsmallletterd = "c-s-d"
    ControlShiftLatinsmallletterc = "c-s-c"
    ControlShiftLatinsmallletterb = "c-s-b"
    ControlShiftLatinsmalllettera = "c-s-a"
    ControlShiftGraveaccent = "c-s-`"
    ControlShiftLowline = "c-s-_"
    ControlShiftCircumflexaccent = "c-s-^"
    ControlShiftRightsquarebracket = "c-s-]"
    ControlShiftReversesolidus = "c-s-\\"
    ControlShiftLeftsquarebracket = "c-s-["
    ControlShiftLatincapitalletterz = "c-s-Z"
    ControlShiftLatincapitallettery = "c-s-Y"
    ControlShiftLatincapitalletterx = "c-s-X"
    ControlShiftLatincapitalletterw = "c-s-W"
    ControlShiftLatincapitalletterv = "c-s-V"
    ControlShiftLatincapitalletteru = "c-s-U"
    ControlShiftLatincapitallettert = "c-s-T"
    ControlShiftLatincapitalletters = "c-s-S"
    ControlShiftLatincapitalletterr = "c-s-R"
    ControlShiftLatincapitalletterq = "c-s-Q"
    ControlShiftLatincapitalletterp = "c-s-P"
    ControlShiftLatincapitallettero = "c-s-O"
    ControlShiftLatincapitallettern = "c-s-N"
    ControlShiftLatincapitalletterm = "c-s-M"
    ControlShiftLatincapitalletterl = "c-s-L"
    ControlShiftLatincapitalletterk = "c-s-K"
    ControlShiftLatincapitalletterj = "c-s-J"
    ControlShiftLatincapitalletteri = "c-s-I"
    ControlShiftLatincapitalletterh = "c-s-H"
    ControlShiftLatincapitalletterg = "c-s-G"
    ControlShiftLatincapitalletterf = "c-s-F"
    ControlShiftLatincapitallettere = "c-s-E"
    ControlShiftLatincapitalletterd = "c-s-D"
    ControlShiftLatincapitalletterc = "c-s-C"
    ControlShiftLatincapitalletterb = "c-s-B"
    ControlShiftLatincapitallettera = "c-s-A"
    ControlShiftCommercialat = "c-s-@"
    ControlShiftQuestionmark = "c-s-?"
    ControlShiftGreaterthansign = "c-s->"
    ControlShiftEqualssign = "c-s-="
    ControlShiftLessthansign = "c-s-<"
    ControlShiftSemicolon = "c-s-;"
    ControlShiftColon = "c-s-:"
    ControlShiftDigitnine = "c-s-9"
    ControlShiftDigiteight = "c-s-8"
    ControlShiftDigitseven = "c-s-7"
    ControlShiftDigitsix = "c-s-6"
    ControlShiftDigitfive = "c-s-5"
    ControlShiftDigitfour = "c-s-4"
    ControlShiftDigitthree = "c-s-3"
    ControlShiftDigittwo = "c-s-2"
    ControlShiftDigitone = "c-s-1"
    ControlShiftDigitzero = "c-s-0"
    ControlShiftSolidus = "c-s-/"
    ControlShiftFullstop = "c-s-."
    ControlShiftHyphenminus = "c-s--"
    ControlShiftComma = "c-s-,"
    ControlShiftPlussign = "c-s-+"
    ControlShiftAsterisk = "c-s-*"
    ControlShiftRightparenthesis = "c-s-)"
    ControlShiftLeftparenthesis = "c-s-("
    ControlShiftApostrophe = "c-s-'"
    ControlShiftAmpersand = "c-s-&"
    ControlShiftPercentsign = "c-s-%"
    ControlShiftDollarsign = "c-s-$"
    ControlShiftNumbersign = "c-s-#"
    ControlShiftQuotationmark = 'c-s-"'
    ControlShiftExclamationmark = "c-s-!"
    ControlShiftBackspace = "c-s-backspace"
    ControlShiftInsert = "c-s-insert"
    ControlShiftDelete = "c-s-delete"
    ControlShiftPageup = "c-s-pageup"
    ControlShiftPagedown = "c-s-pagedown"
    ControlShiftUp = "c-s-up"
    ControlShiftDown = "c-s-down"
    ControlShiftRight = "c-s-right"
    ControlShiftLeft = "c-s-left"
    ControlShiftHome = "c-s-home"
    ControlShiftEnd = "c-s-end"
    ControlShiftPrintscreen = "c-s-printscreen"
    ControlShiftPause = "c-s-pause"
    ControlShiftMenu = "c-s-menu"
    ControlShiftF1 = "c-s-f1"
    ControlShiftF2 = "c-s-f2"
    ControlShiftF3 = "c-s-f3"
    ControlShiftF4 = "c-s-f4"
    ControlShiftF5 = "c-s-f5"
    ControlShiftF6 = "c-s-f6"
    ControlShiftF7 = "c-s-f7"
    ControlShiftF8 = "c-s-f8"
    ControlShiftF9 = "c-s-f9"
    ControlShiftF10 = "c-s-f10"
    ControlShiftF11 = "c-s-f11"
    ControlShiftF12 = "c-s-f12"
    ControlAltTab = "c-A-tab"
    ControlAltEnter = "c-A-enter"
    ControlAltEscape = "c-A-escape"
    ControlAltSpace = "c-A-space"
    ControlAltTilde = "c-A-~"
    ControlAltRightcurlybracket = "c-A-}"
    ControlAltVerticalline = "c-A-|"
    ControlAltLeftcurlybracket = "c-A-{"
    ControlAltLatinsmallletterz = "c-A-z"
    ControlAltLatinsmalllettery = "c-A-y"
    ControlAltLatinsmallletterx = "c-A-x"
    ControlAltLatinsmallletterw = "c-A-w"
    ControlAltLatinsmallletterv = "c-A-v"
    ControlAltLatinsmallletteru = "c-A-u"
    ControlAltLatinsmalllettert = "c-A-t"
    ControlAltLatinsmallletters = "c-A-s"
    ControlAltLatinsmallletterr = "c-A-r"
    ControlAltLatinsmallletterq = "c-A-q"
    ControlAltLatinsmallletterp = "c-A-p"
    ControlAltLatinsmalllettero = "c-A-o"
    ControlAltLatinsmalllettern = "c-A-n"
    ControlAltLatinsmallletterm = "c-A-m"
    ControlAltLatinsmallletterl = "c-A-l"
    ControlAltLatinsmallletterk = "c-A-k"
    ControlAltLatinsmallletterj = "c-A-j"
    ControlAltLatinsmallletteri = "c-A-i"
    ControlAltLatinsmallletterh = "c-A-h"
    ControlAltLatinsmallletterg = "c-A-g"
    ControlAltLatinsmallletterf = "c-A-f"
    ControlAltLatinsmalllettere = "c-A-e"
    ControlAltLatinsmallletterd = "c-A-d"
    ControlAltLatinsmallletterc = "c-A-c"
    ControlAltLatinsmallletterb = "c-A-b"
    ControlAltLatinsmalllettera = "c-A-a"
    ControlAltGraveaccent = "c-A-`"
    ControlAltLowline = "c-A-_"
    ControlAltCircumflexaccent = "c-A-^"
    ControlAltRightsquarebracket = "c-A-]"
    ControlAltReversesolidus = "c-A-\\"
    ControlAltLeftsquarebracket = "c-A-["
    ControlAltLatincapitalletterz = "c-A-Z"
    ControlAltLatincapitallettery = "c-A-Y"
    ControlAltLatincapitalletterx = "c-A-X"
    ControlAltLatincapitalletterw = "c-A-W"
    ControlAltLatincapitalletterv = "c-A-V"
    ControlAltLatincapitalletteru = "c-A-U"
    ControlAltLatincapitallettert = "c-A-T"
    ControlAltLatincapitalletters = "c-A-S"
    ControlAltLatincapitalletterr = "c-A-R"
    ControlAltLatincapitalletterq = "c-A-Q"
    ControlAltLatincapitalletterp = "c-A-P"
    ControlAltLatincapitallettero = "c-A-O"
    ControlAltLatincapitallettern = "c-A-N"
    ControlAltLatincapitalletterm = "c-A-M"
    ControlAltLatincapitalletterl = "c-A-L"
    ControlAltLatincapitalletterk = "c-A-K"
    ControlAltLatincapitalletterj = "c-A-J"
    ControlAltLatincapitalletteri = "c-A-I"
    ControlAltLatincapitalletterh = "c-A-H"
    ControlAltLatincapitalletterg = "c-A-G"
    ControlAltLatincapitalletterf = "c-A-F"
    ControlAltLatincapitallettere = "c-A-E"
    ControlAltLatincapitalletterd = "c-A-D"
    ControlAltLatincapitalletterc = "c-A-C"
    ControlAltLatincapitalletterb = "c-A-B"
    ControlAltLatincapitallettera = "c-A-A"
    ControlAltCommercialat = "c-A-@"
    ControlAltQuestionmark = "c-A-?"
    ControlAltGreaterthansign = "c-A->"
    ControlAltEqualssign = "c-A-="
    ControlAltLessthansign = "c-A-<"
    ControlAltSemicolon = "c-A-;"
    ControlAltColon = "c-A-:"
    ControlAltDigitnine = "c-A-9"
    ControlAltDigiteight = "c-A-8"
    ControlAltDigitseven = "c-A-7"
    ControlAltDigitsix = "c-A-6"
    ControlAltDigitfive = "c-A-5"
    ControlAltDigitfour = "c-A-4"
    ControlAltDigitthree = "c-A-3"
    ControlAltDigittwo = "c-A-2"
    ControlAltDigitone = "c-A-1"
    ControlAltDigitzero = "c-A-0"
    ControlAltSolidus = "c-A-/"
    ControlAltFullstop = "c-A-."
    ControlAltHyphenminus = "c-A--"
    ControlAltComma = "c-A-,"
    ControlAltPlussign = "c-A-+"
    ControlAltAsterisk = "c-A-*"
    ControlAltRightparenthesis = "c-A-)"
    ControlAltLeftparenthesis = "c-A-("
    ControlAltApostrophe = "c-A-'"
    ControlAltAmpersand = "c-A-&"
    ControlAltPercentsign = "c-A-%"
    ControlAltDollarsign = "c-A-$"
    ControlAltNumbersign = "c-A-#"
    ControlAltQuotationmark = 'c-A-"'
    ControlAltExclamationmark = "c-A-!"
    ControlAltBackspace = "c-A-backspace"
    ControlAltInsert = "c-A-insert"
    ControlAltDelete = "c-A-delete"
    ControlAltPageup = "c-A-pageup"
    ControlAltPagedown = "c-A-pagedown"
    ControlAltUp = "c-A-up"
    ControlAltDown = "c-A-down"
    ControlAltRight = "c-A-right"
    ControlAltLeft = "c-A-left"
    ControlAltHome = "c-A-home"
    ControlAltEnd = "c-A-end"
    ControlAltPrintscreen = "c-A-printscreen"
    ControlAltPause = "c-A-pause"
    ControlAltMenu = "c-A-menu"
    ControlAltF1 = "c-A-f1"
    ControlAltF2 = "c-A-f2"
    ControlAltF3 = "c-A-f3"
    ControlAltF4 = "c-A-f4"
    ControlAltF5 = "c-A-f5"
    ControlAltF6 = "c-A-f6"
    ControlAltF7 = "c-A-f7"
    ControlAltF8 = "c-A-f8"
    ControlAltF9 = "c-A-f9"
    ControlAltF10 = "c-A-f10"
    ControlAltF11 = "c-A-f11"
    ControlAltF12 = "c-A-f12"
    ControlAltShiftTab = "c-A-s-tab"
    ControlAltShiftEnter = "c-A-s-enter"
    ControlAltShiftEscape = "c-A-s-escape"
    ControlAltShiftSpace = "c-A-s-space"
    ControlAltShiftTilde = "c-A-s-~"
    ControlAltShiftRightcurlybracket = "c-A-s-}"
    ControlAltShiftVerticalline = "c-A-s-|"
    ControlAltShiftLeftcurlybracket = "c-A-s-{"
    ControlAltShiftLatinsmallletterz = "c-A-s-z"
    ControlAltShiftLatinsmalllettery = "c-A-s-y"
    ControlAltShiftLatinsmallletterx = "c-A-s-x"
    ControlAltShiftLatinsmallletterw = "c-A-s-w"
    ControlAltShiftLatinsmallletterv = "c-A-s-v"
    ControlAltShiftLatinsmallletteru = "c-A-s-u"
    ControlAltShiftLatinsmalllettert = "c-A-s-t"
    ControlAltShiftLatinsmallletters = "c-A-s-s"
    ControlAltShiftLatinsmallletterr = "c-A-s-r"
    ControlAltShiftLatinsmallletterq = "c-A-s-q"
    ControlAltShiftLatinsmallletterp = "c-A-s-p"
    ControlAltShiftLatinsmalllettero = "c-A-s-o"
    ControlAltShiftLatinsmalllettern = "c-A-s-n"
    ControlAltShiftLatinsmallletterm = "c-A-s-m"
    ControlAltShiftLatinsmallletterl = "c-A-s-l"
    ControlAltShiftLatinsmallletterk = "c-A-s-k"
    ControlAltShiftLatinsmallletterj = "c-A-s-j"
    ControlAltShiftLatinsmallletteri = "c-A-s-i"
    ControlAltShiftLatinsmallletterh = "c-A-s-h"
    ControlAltShiftLatinsmallletterg = "c-A-s-g"
    ControlAltShiftLatinsmallletterf = "c-A-s-f"
    ControlAltShiftLatinsmalllettere = "c-A-s-e"
    ControlAltShiftLatinsmallletterd = "c-A-s-d"
    ControlAltShiftLatinsmallletterc = "c-A-s-c"
    ControlAltShiftLatinsmallletterb = "c-A-s-b"
    ControlAltShiftLatinsmalllettera = "c-A-s-a"
    ControlAltShiftGraveaccent = "c-A-s-`"
    ControlAltShiftLowline = "c-A-s-_"
    ControlAltShiftCircumflexaccent = "c-A-s-^"
    ControlAltShiftRightsquarebracket = "c-A-s-]"
    ControlAltShiftReversesolidus = "c-A-s-\\"
    ControlAltShiftLeftsquarebracket = "c-A-s-["
    ControlAltShiftLatincapitalletterz = "c-A-s-Z"
    ControlAltShiftLatincapitallettery = "c-A-s-Y"
    ControlAltShiftLatincapitalletterx = "c-A-s-X"
    ControlAltShiftLatincapitalletterw = "c-A-s-W"
    ControlAltShiftLatincapitalletterv = "c-A-s-V"
    ControlAltShiftLatincapitalletteru = "c-A-s-U"
    ControlAltShiftLatincapitallettert = "c-A-s-T"
    ControlAltShiftLatincapitalletters = "c-A-s-S"
    ControlAltShiftLatincapitalletterr = "c-A-s-R"
    ControlAltShiftLatincapitalletterq = "c-A-s-Q"
    ControlAltShiftLatincapitalletterp = "c-A-s-P"
    ControlAltShiftLatincapitallettero = "c-A-s-O"
    ControlAltShiftLatincapitallettern = "c-A-s-N"
    ControlAltShiftLatincapitalletterm = "c-A-s-M"
    ControlAltShiftLatincapitalletterl = "c-A-s-L"
    ControlAltShiftLatincapitalletterk = "c-A-s-K"
    ControlAltShiftLatincapitalletterj = "c-A-s-J"
    ControlAltShiftLatincapitalletteri = "c-A-s-I"
    ControlAltShiftLatincapitalletterh = "c-A-s-H"
    ControlAltShiftLatincapitalletterg = "c-A-s-G"
    ControlAltShiftLatincapitalletterf = "c-A-s-F"
    ControlAltShiftLatincapitallettere = "c-A-s-E"
    ControlAltShiftLatincapitalletterd = "c-A-s-D"
    ControlAltShiftLatincapitalletterc = "c-A-s-C"
    ControlAltShiftLatincapitalletterb = "c-A-s-B"
    ControlAltShiftLatincapitallettera = "c-A-s-A"
    ControlAltShiftCommercialat = "c-A-s-@"
    ControlAltShiftQuestionmark = "c-A-s-?"
    ControlAltShiftGreaterthansign = "c-A-s->"
    ControlAltShiftEqualssign = "c-A-s-="
    ControlAltShiftLessthansign = "c-A-s-<"
    ControlAltShiftSemicolon = "c-A-s-;"
    ControlAltShiftColon = "c-A-s-:"
    ControlAltShiftDigitnine = "c-A-s-9"
    ControlAltShiftDigiteight = "c-A-s-8"
    ControlAltShiftDigitseven = "c-A-s-7"
    ControlAltShiftDigitsix = "c-A-s-6"
    ControlAltShiftDigitfive = "c-A-s-5"
    ControlAltShiftDigitfour = "c-A-s-4"
    ControlAltShiftDigitthree = "c-A-s-3"
    ControlAltShiftDigittwo = "c-A-s-2"
    ControlAltShiftDigitone = "c-A-s-1"
    ControlAltShiftDigitzero = "c-A-s-0"
    ControlAltShiftSolidus = "c-A-s-/"
    ControlAltShiftFullstop = "c-A-s-."
    ControlAltShiftHyphenminus = "c-A-s--"
    ControlAltShiftComma = "c-A-s-,"
    ControlAltShiftPlussign = "c-A-s-+"
    ControlAltShiftAsterisk = "c-A-s-*"
    ControlAltShiftRightparenthesis = "c-A-s-)"
    ControlAltShiftLeftparenthesis = "c-A-s-("
    ControlAltShiftApostrophe = "c-A-s-'"
    ControlAltShiftAmpersand = "c-A-s-&"
    ControlAltShiftPercentsign = "c-A-s-%"
    ControlAltShiftDollarsign = "c-A-s-$"
    ControlAltShiftNumbersign = "c-A-s-#"
    ControlAltShiftQuotationmark = 'c-A-s-"'
    ControlAltShiftExclamationmark = "c-A-s-!"
    ControlAltShiftBackspace = "c-A-s-backspace"
    ControlAltShiftInsert = "c-A-s-insert"
    ControlAltShiftDelete = "c-A-s-delete"
    ControlAltShiftPageup = "c-A-s-pageup"
    ControlAltShiftPagedown = "c-A-s-pagedown"
    ControlAltShiftUp = "c-A-s-up"
    ControlAltShiftDown = "c-A-s-down"
    ControlAltShiftRight = "c-A-s-right"
    ControlAltShiftLeft = "c-A-s-left"
    ControlAltShiftHome = "c-A-s-home"
    ControlAltShiftEnd = "c-A-s-end"
    ControlAltShiftPrintscreen = "c-A-s-printscreen"
    ControlAltShiftPause = "c-A-s-pause"
    ControlAltShiftMenu = "c-A-s-menu"
    ControlAltShiftF1 = "c-A-s-f1"
    ControlAltShiftF2 = "c-A-s-f2"
    ControlAltShiftF3 = "c-A-s-f3"
    ControlAltShiftF4 = "c-A-s-f4"
    ControlAltShiftF5 = "c-A-s-f5"
    ControlAltShiftF6 = "c-A-s-f6"
    ControlAltShiftF7 = "c-A-s-f7"
    ControlAltShiftF8 = "c-A-s-f8"
    ControlAltShiftF9 = "c-A-s-f9"
    ControlAltShiftF10 = "c-A-s-f10"
    ControlAltShiftF11 = "c-A-s-f11"
    ControlAltShiftF12 = "c-A-s-f12"


# Update PTK's mapping of escape codes to known key-presses
ANSI_SEQUENCES.update(
    cast(
        "dict[str, Keys]",
        {
            "\x1btab": MoreKeys.AltTab,
            "\x1benter": MoreKeys.AltEnter,
            "\x1bescape": MoreKeys.AltEscape,
            "\x1bspace": MoreKeys.AltSpace,
            "\x1b~": MoreKeys.AltTilde,
            "\x1b}": MoreKeys.AltRightcurlybracket,
            "\x1b|": MoreKeys.AltVerticalline,
            "\x1b{": MoreKeys.AltLeftcurlybracket,
            "\x1bz": MoreKeys.AltLatinsmallletterz,
            "\x1by": MoreKeys.AltLatinsmalllettery,
            "\x1bx": MoreKeys.AltLatinsmallletterx,
            "\x1bw": MoreKeys.AltLatinsmallletterw,
            "\x1bv": MoreKeys.AltLatinsmallletterv,
            "\x1bu": MoreKeys.AltLatinsmallletteru,
            "\x1bt": MoreKeys.AltLatinsmalllettert,
            "\x1bs": MoreKeys.AltLatinsmallletters,
            "\x1br": MoreKeys.AltLatinsmallletterr,
            "\x1bq": MoreKeys.AltLatinsmallletterq,
            "\x1bp": MoreKeys.AltLatinsmallletterp,
            "\x1bo": MoreKeys.AltLatinsmalllettero,
            "\x1bn": MoreKeys.AltLatinsmalllettern,
            "\x1bm": MoreKeys.AltLatinsmallletterm,
            "\x1bl": MoreKeys.AltLatinsmallletterl,
            "\x1bk": MoreKeys.AltLatinsmallletterk,
            "\x1bj": MoreKeys.AltLatinsmallletterj,
            "\x1bi": MoreKeys.AltLatinsmallletteri,
            "\x1bh": MoreKeys.AltLatinsmallletterh,
            "\x1bg": MoreKeys.AltLatinsmallletterg,
            "\x1bf": MoreKeys.AltLatinsmallletterf,
            "\x1be": MoreKeys.AltLatinsmalllettere,
            "\x1bd": MoreKeys.AltLatinsmallletterd,
            "\x1bc": MoreKeys.AltLatinsmallletterc,
            "\x1bb": MoreKeys.AltLatinsmallletterb,
            "\x1ba": MoreKeys.AltLatinsmalllettera,
            "\x1b`": MoreKeys.AltGraveaccent,
            "\x1b_": MoreKeys.AltLowline,
            "\x1b^": MoreKeys.AltCircumflexaccent,
            "\x1b]": MoreKeys.AltRightsquarebracket,
            "\x1b\\": MoreKeys.AltReversesolidus,
            "\x1b[": MoreKeys.AltLeftsquarebracket,
            "\x1bZ": MoreKeys.AltLatincapitalletterz,
            "\x1bY": MoreKeys.AltLatincapitallettery,
            "\x1bX": MoreKeys.AltLatincapitalletterx,
            "\x1bW": MoreKeys.AltLatincapitalletterw,
            "\x1bV": MoreKeys.AltLatincapitalletterv,
            "\x1bU": MoreKeys.AltLatincapitalletteru,
            "\x1bT": MoreKeys.AltLatincapitallettert,
            "\x1bS": MoreKeys.AltLatincapitalletters,
            "\x1bR": MoreKeys.AltLatincapitalletterr,
            "\x1bQ": MoreKeys.AltLatincapitalletterq,
            "\x1bP": MoreKeys.AltLatincapitalletterp,
            "\x1bO": MoreKeys.AltLatincapitallettero,
            "\x1bN": MoreKeys.AltLatincapitallettern,
            "\x1bM": MoreKeys.AltLatincapitalletterm,
            "\x1bL": MoreKeys.AltLatincapitalletterl,
            "\x1bK": MoreKeys.AltLatincapitalletterk,
            "\x1bJ": MoreKeys.AltLatincapitalletterj,
            "\x1bI": MoreKeys.AltLatincapitalletteri,
            "\x1bH": MoreKeys.AltLatincapitalletterh,
            "\x1bG": MoreKeys.AltLatincapitalletterg,
            "\x1bF": MoreKeys.AltLatincapitalletterf,
            "\x1bE": MoreKeys.AltLatincapitallettere,
            "\x1bD": MoreKeys.AltLatincapitalletterd,
            "\x1bC": MoreKeys.AltLatincapitalletterc,
            "\x1bB": MoreKeys.AltLatincapitalletterb,
            "\x1bA": MoreKeys.AltLatincapitallettera,
            "\x1b@": MoreKeys.AltCommercialat,
            "\x1b?": MoreKeys.AltQuestionmark,
            "\x1b>": MoreKeys.AltGreaterthansign,
            "\x1b=": MoreKeys.AltEqualssign,
            "\x1b<": MoreKeys.AltLessthansign,
            "\x1b;": MoreKeys.AltSemicolon,
            "\x1b:": MoreKeys.AltColon,
            "\x1b9": MoreKeys.AltDigitnine,
            "\x1b8": MoreKeys.AltDigiteight,
            "\x1b7": MoreKeys.AltDigitseven,
            "\x1b6": MoreKeys.AltDigitsix,
            "\x1b5": MoreKeys.AltDigitfive,
            "\x1b4": MoreKeys.AltDigitfour,
            "\x1b3": MoreKeys.AltDigitthree,
            "\x1b2": MoreKeys.AltDigittwo,
            "\x1b1": MoreKeys.AltDigitone,
            "\x1b0": MoreKeys.AltDigitzero,
            "\x1b/": MoreKeys.AltSolidus,
            "\x1b.": MoreKeys.AltFullstop,
            "\x1b-": MoreKeys.AltHyphenminus,
            "\x1b,": MoreKeys.AltComma,
            "\x1b+": MoreKeys.AltPlussign,
            "\x1b*": MoreKeys.AltAsterisk,
            "\x1b)": MoreKeys.AltRightparenthesis,
            "\x1b(": MoreKeys.AltLeftparenthesis,
            "\x1b'": MoreKeys.AltApostrophe,
            "\x1b&": MoreKeys.AltAmpersand,
            "\x1b%": MoreKeys.AltPercentsign,
            "\x1b$": MoreKeys.AltDollarsign,
            "\x1b#": MoreKeys.AltNumbersign,
            '\x1b"': MoreKeys.AltQuotationmark,
            "\x1b!": MoreKeys.AltExclamationmark,
            "\x1bbackspace": MoreKeys.AltBackspace,
            "\x1b[27u": MoreKeys.Escape,
            "\x1b[2~": MoreKeys.Insert,
            "\x1b[3~": MoreKeys.Delete,
            "\x1b[5~": MoreKeys.Pageup,
            "\x1b[6~": MoreKeys.Pagedown,
            "\x1b[A": MoreKeys.Up,
            "\x1b[B": MoreKeys.Down,
            "\x1b[C": MoreKeys.Right,
            "\x1b[D": MoreKeys.Left,
            "\x1b[H": MoreKeys.Home,
            "\x1b[F": MoreKeys.End,
            "\x1b[57361u": MoreKeys.Printscreen,
            "\x1b[57362u": MoreKeys.Pause,
            "\x1b[57363u": MoreKeys.Menu,
            "\x1b[P": MoreKeys.F1,
            "\x1b[Q": MoreKeys.F2,
            "\x1b[13~": MoreKeys.F3,
            "\x1b[S": MoreKeys.F4,
            "\x1b[15~": MoreKeys.F5,
            "\x1b[17~": MoreKeys.F6,
            "\x1b[18~": MoreKeys.F7,
            "\x1b[19~": MoreKeys.F8,
            "\x1b[20~": MoreKeys.F9,
            "\x1b[21~": MoreKeys.F10,
            "\x1b[23~": MoreKeys.F11,
            "\x1b[24~": MoreKeys.F12,
            "\x1b[127u": MoreKeys.Backspace,
            "\x1b[13u": MoreKeys.Enter,
            "\x1b[9u": MoreKeys.Tab,
            "\x1b[9;2u": MoreKeys.ShiftTab,
            "\x1b[27;2;9~": MoreKeys.ShiftTab,
            "\x1b[13;2u": MoreKeys.ShiftEnter,
            "\x1b[27;2;13~": MoreKeys.ShiftEnter,
            "\x1b[27;2u": MoreKeys.ShiftEscape,
            "\x1b[27;2;27~": MoreKeys.ShiftEscape,
            "\x1b[32;2u": MoreKeys.ShiftSpace,
            "\x1b[27;2;32~": MoreKeys.ShiftSpace,
            "\x1b[126;2u": MoreKeys.ShiftTilde,
            "\x1b[27;2;126~": MoreKeys.ShiftTilde,
            "\x1b[125;2u": MoreKeys.ShiftRightcurlybracket,
            "\x1b[27;2;125~": MoreKeys.ShiftRightcurlybracket,
            "\x1b[124;2u": MoreKeys.ShiftVerticalline,
            "\x1b[27;2;124~": MoreKeys.ShiftVerticalline,
            "\x1b[123;2u": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[27;2;123~": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[122;2u": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[27;2;122~": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[121;2u": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[27;2;121~": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[120;2u": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[27;2;120~": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[119;2u": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[27;2;119~": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[118;2u": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[27;2;118~": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[117;2u": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[27;2;117~": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[116;2u": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[27;2;116~": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[115;2u": MoreKeys.ShiftLatinsmallletters,
            "\x1b[27;2;115~": MoreKeys.ShiftLatinsmallletters,
            "\x1b[114;2u": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[27;2;114~": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[113;2u": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[27;2;113~": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[112;2u": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[27;2;112~": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[111;2u": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[27;2;111~": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[110;2u": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[27;2;110~": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[109;2u": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[27;2;109~": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[108;2u": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[27;2;108~": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[107;2u": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[27;2;107~": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[106;2u": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[27;2;106~": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[105;2u": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[27;2;105~": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[104;2u": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[27;2;104~": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[103;2u": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[27;2;103~": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[102;2u": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[27;2;102~": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[101;2u": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[27;2;101~": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[100;2u": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[27;2;100~": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[99;2u": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[27;2;99~": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[98;2u": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[27;2;98~": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[97;2u": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[27;2;97~": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[96;2u": MoreKeys.ShiftGraveaccent,
            "\x1b[27;2;96~": MoreKeys.ShiftGraveaccent,
            "\x1b[95;2u": MoreKeys.ShiftLowline,
            "\x1b[27;2;95~": MoreKeys.ShiftLowline,
            "\x1b[94;2u": MoreKeys.ShiftCircumflexaccent,
            "\x1b[27;2;94~": MoreKeys.ShiftCircumflexaccent,
            "\x1b[93;2u": MoreKeys.ShiftRightsquarebracket,
            "\x1b[27;2;93~": MoreKeys.ShiftRightsquarebracket,
            "\x1b[92;2u": MoreKeys.ShiftReversesolidus,
            "\x1b[27;2;92~": MoreKeys.ShiftReversesolidus,
            "\x1b[91;2u": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[27;2;91~": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[90;2u": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[27;2;90~": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[89;2u": MoreKeys.ShiftLatincapitallettery,
            "\x1b[27;2;89~": MoreKeys.ShiftLatincapitallettery,
            "\x1b[88;2u": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[27;2;88~": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[87;2u": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[27;2;87~": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[86;2u": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[27;2;86~": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[85;2u": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[27;2;85~": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[84;2u": MoreKeys.ShiftLatincapitallettert,
            "\x1b[27;2;84~": MoreKeys.ShiftLatincapitallettert,
            "\x1b[83;2u": MoreKeys.ShiftLatincapitalletters,
            "\x1b[27;2;83~": MoreKeys.ShiftLatincapitalletters,
            "\x1b[82;2u": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[27;2;82~": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[81;2u": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[27;2;81~": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[80;2u": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[27;2;80~": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[79;2u": MoreKeys.ShiftLatincapitallettero,
            "\x1b[27;2;79~": MoreKeys.ShiftLatincapitallettero,
            "\x1b[78;2u": MoreKeys.ShiftLatincapitallettern,
            "\x1b[27;2;78~": MoreKeys.ShiftLatincapitallettern,
            "\x1b[77;2u": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[27;2;77~": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[76;2u": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[27;2;76~": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[75;2u": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[27;2;75~": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[74;2u": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[27;2;74~": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[73;2u": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[27;2;73~": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[72;2u": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[27;2;72~": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[71;2u": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[27;2;71~": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[70;2u": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[27;2;70~": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[69;2u": MoreKeys.ShiftLatincapitallettere,
            "\x1b[27;2;69~": MoreKeys.ShiftLatincapitallettere,
            "\x1b[68;2u": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[27;2;68~": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[67;2u": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[27;2;67~": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[66;2u": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[27;2;66~": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[65;2u": MoreKeys.ShiftLatincapitallettera,
            "\x1b[27;2;65~": MoreKeys.ShiftLatincapitallettera,
            "\x1b[64;2u": MoreKeys.ShiftCommercialat,
            "\x1b[27;2;64~": MoreKeys.ShiftCommercialat,
            "\x1b[63;2u": MoreKeys.ShiftQuestionmark,
            "\x1b[27;2;63~": MoreKeys.ShiftQuestionmark,
            "\x1b[62;2u": MoreKeys.ShiftGreaterthansign,
            "\x1b[27;2;62~": MoreKeys.ShiftGreaterthansign,
            "\x1b[61;2u": MoreKeys.ShiftEqualssign,
            "\x1b[27;2;61~": MoreKeys.ShiftEqualssign,
            "\x1b[60;2u": MoreKeys.ShiftLessthansign,
            "\x1b[27;2;60~": MoreKeys.ShiftLessthansign,
            "\x1b[59;2u": MoreKeys.ShiftSemicolon,
            "\x1b[27;2;59~": MoreKeys.ShiftSemicolon,
            "\x1b[58;2u": MoreKeys.ShiftColon,
            "\x1b[27;2;58~": MoreKeys.ShiftColon,
            "\x1b[57;2u": MoreKeys.ShiftDigitnine,
            "\x1b[27;2;57~": MoreKeys.ShiftDigitnine,
            "\x1b[56;2u": MoreKeys.ShiftDigiteight,
            "\x1b[27;2;56~": MoreKeys.ShiftDigiteight,
            "\x1b[55;2u": MoreKeys.ShiftDigitseven,
            "\x1b[27;2;55~": MoreKeys.ShiftDigitseven,
            "\x1b[54;2u": MoreKeys.ShiftDigitsix,
            "\x1b[27;2;54~": MoreKeys.ShiftDigitsix,
            "\x1b[53;2u": MoreKeys.ShiftDigitfive,
            "\x1b[27;2;53~": MoreKeys.ShiftDigitfive,
            "\x1b[52;2u": MoreKeys.ShiftDigitfour,
            "\x1b[27;2;52~": MoreKeys.ShiftDigitfour,
            "\x1b[51;2u": MoreKeys.ShiftDigitthree,
            "\x1b[27;2;51~": MoreKeys.ShiftDigitthree,
            "\x1b[50;2u": MoreKeys.ShiftDigittwo,
            "\x1b[27;2;50~": MoreKeys.ShiftDigittwo,
            "\x1b[49;2u": MoreKeys.ShiftDigitone,
            "\x1b[27;2;49~": MoreKeys.ShiftDigitone,
            "\x1b[48;2u": MoreKeys.ShiftDigitzero,
            "\x1b[27;2;48~": MoreKeys.ShiftDigitzero,
            "\x1b[47;2u": MoreKeys.ShiftSolidus,
            "\x1b[27;2;47~": MoreKeys.ShiftSolidus,
            "\x1b[46;2u": MoreKeys.ShiftFullstop,
            "\x1b[27;2;46~": MoreKeys.ShiftFullstop,
            "\x1b[45;2u": MoreKeys.ShiftHyphenminus,
            "\x1b[27;2;45~": MoreKeys.ShiftHyphenminus,
            "\x1b[44;2u": MoreKeys.ShiftComma,
            "\x1b[27;2;44~": MoreKeys.ShiftComma,
            "\x1b[43;2u": MoreKeys.ShiftPlussign,
            "\x1b[27;2;43~": MoreKeys.ShiftPlussign,
            "\x1b[42;2u": MoreKeys.ShiftAsterisk,
            "\x1b[27;2;42~": MoreKeys.ShiftAsterisk,
            "\x1b[41;2u": MoreKeys.ShiftRightparenthesis,
            "\x1b[27;2;41~": MoreKeys.ShiftRightparenthesis,
            "\x1b[40;2u": MoreKeys.ShiftLeftparenthesis,
            "\x1b[27;2;40~": MoreKeys.ShiftLeftparenthesis,
            "\x1b[39;2u": MoreKeys.ShiftApostrophe,
            "\x1b[27;2;39~": MoreKeys.ShiftApostrophe,
            "\x1b[38;2u": MoreKeys.ShiftAmpersand,
            "\x1b[27;2;38~": MoreKeys.ShiftAmpersand,
            "\x1b[37;2u": MoreKeys.ShiftPercentsign,
            "\x1b[27;2;37~": MoreKeys.ShiftPercentsign,
            "\x1b[36;2u": MoreKeys.ShiftDollarsign,
            "\x1b[27;2;36~": MoreKeys.ShiftDollarsign,
            "\x1b[35;2u": MoreKeys.ShiftNumbersign,
            "\x1b[27;2;35~": MoreKeys.ShiftNumbersign,
            "\x1b[34;2u": MoreKeys.ShiftQuotationmark,
            "\x1b[27;2;34~": MoreKeys.ShiftQuotationmark,
            "\x1b[33;2u": MoreKeys.ShiftExclamationmark,
            "\x1b[27;2;33~": MoreKeys.ShiftExclamationmark,
            "\x1b[127;2u": MoreKeys.ShiftBackspace,
            "\x1b[27;2;127~": MoreKeys.ShiftBackspace,
            "\x1b[2;2~": MoreKeys.ShiftInsert,
            "\x1b[3;2~": MoreKeys.ShiftDelete,
            "\x1b[5;2~": MoreKeys.ShiftPageup,
            "\x1b[6;2~": MoreKeys.ShiftPagedown,
            "\x1b[1;2A": MoreKeys.ShiftUp,
            "\x1b[1;2B": MoreKeys.ShiftDown,
            "\x1b[1;2C": MoreKeys.ShiftRight,
            "\x1b[1;2D": MoreKeys.ShiftLeft,
            "\x1b[1;2H": MoreKeys.ShiftHome,
            "\x1b[1;2F": MoreKeys.ShiftEnd,
            "\x1b[57361;2u": MoreKeys.ShiftPrintscreen,
            "\x1b[57362;2u": MoreKeys.ShiftPause,
            "\x1b[57363;2u": MoreKeys.ShiftMenu,
            "\x1b[1;2P": MoreKeys.ShiftF1,
            "\x1b[1;2Q": MoreKeys.ShiftF2,
            "\x1b[13;2~": MoreKeys.ShiftF3,
            "\x1b[1;2S": MoreKeys.ShiftF4,
            "\x1b[15;2~": MoreKeys.ShiftF5,
            "\x1b[17;2~": MoreKeys.ShiftF6,
            "\x1b[18;2~": MoreKeys.ShiftF7,
            "\x1b[19;2~": MoreKeys.ShiftF8,
            "\x1b[20;2~": MoreKeys.ShiftF9,
            "\x1b[21;2~": MoreKeys.ShiftF10,
            "\x1b[23;2~": MoreKeys.ShiftF11,
            "\x1b[24;2~": MoreKeys.ShiftF12,
            "\x1b[9;3u": MoreKeys.AltTab,
            "\x1b[27;3;9~": MoreKeys.AltTab,
            "\x1b[13;3u": MoreKeys.AltEnter,
            "\x1b[27;3;13~": MoreKeys.AltEnter,
            "\x1b[27;3u": MoreKeys.AltEscape,
            "\x1b[27;3;27~": MoreKeys.AltEscape,
            "\x1b[32;3u": MoreKeys.AltSpace,
            "\x1b[27;3;32~": MoreKeys.AltSpace,
            "\x1b[126;3u": MoreKeys.AltTilde,
            "\x1b[27;3;126~": MoreKeys.AltTilde,
            "\x1b[125;3u": MoreKeys.AltRightcurlybracket,
            "\x1b[27;3;125~": MoreKeys.AltRightcurlybracket,
            "\x1b[124;3u": MoreKeys.AltVerticalline,
            "\x1b[27;3;124~": MoreKeys.AltVerticalline,
            "\x1b[123;3u": MoreKeys.AltLeftcurlybracket,
            "\x1b[27;3;123~": MoreKeys.AltLeftcurlybracket,
            "\x1b[122;3u": MoreKeys.AltLatinsmallletterz,
            "\x1b[27;3;122~": MoreKeys.AltLatinsmallletterz,
            "\x1b[121;3u": MoreKeys.AltLatinsmalllettery,
            "\x1b[27;3;121~": MoreKeys.AltLatinsmalllettery,
            "\x1b[120;3u": MoreKeys.AltLatinsmallletterx,
            "\x1b[27;3;120~": MoreKeys.AltLatinsmallletterx,
            "\x1b[119;3u": MoreKeys.AltLatinsmallletterw,
            "\x1b[27;3;119~": MoreKeys.AltLatinsmallletterw,
            "\x1b[118;3u": MoreKeys.AltLatinsmallletterv,
            "\x1b[27;3;118~": MoreKeys.AltLatinsmallletterv,
            "\x1b[117;3u": MoreKeys.AltLatinsmallletteru,
            "\x1b[27;3;117~": MoreKeys.AltLatinsmallletteru,
            "\x1b[116;3u": MoreKeys.AltLatinsmalllettert,
            "\x1b[27;3;116~": MoreKeys.AltLatinsmalllettert,
            "\x1b[115;3u": MoreKeys.AltLatinsmallletters,
            "\x1b[27;3;115~": MoreKeys.AltLatinsmallletters,
            "\x1b[114;3u": MoreKeys.AltLatinsmallletterr,
            "\x1b[27;3;114~": MoreKeys.AltLatinsmallletterr,
            "\x1b[113;3u": MoreKeys.AltLatinsmallletterq,
            "\x1b[27;3;113~": MoreKeys.AltLatinsmallletterq,
            "\x1b[112;3u": MoreKeys.AltLatinsmallletterp,
            "\x1b[27;3;112~": MoreKeys.AltLatinsmallletterp,
            "\x1b[111;3u": MoreKeys.AltLatinsmalllettero,
            "\x1b[27;3;111~": MoreKeys.AltLatinsmalllettero,
            "\x1b[110;3u": MoreKeys.AltLatinsmalllettern,
            "\x1b[27;3;110~": MoreKeys.AltLatinsmalllettern,
            "\x1b[109;3u": MoreKeys.AltLatinsmallletterm,
            "\x1b[27;3;109~": MoreKeys.AltLatinsmallletterm,
            "\x1b[108;3u": MoreKeys.AltLatinsmallletterl,
            "\x1b[27;3;108~": MoreKeys.AltLatinsmallletterl,
            "\x1b[107;3u": MoreKeys.AltLatinsmallletterk,
            "\x1b[27;3;107~": MoreKeys.AltLatinsmallletterk,
            "\x1b[106;3u": MoreKeys.AltLatinsmallletterj,
            "\x1b[27;3;106~": MoreKeys.AltLatinsmallletterj,
            "\x1b[105;3u": MoreKeys.AltLatinsmallletteri,
            "\x1b[27;3;105~": MoreKeys.AltLatinsmallletteri,
            "\x1b[104;3u": MoreKeys.AltLatinsmallletterh,
            "\x1b[27;3;104~": MoreKeys.AltLatinsmallletterh,
            "\x1b[103;3u": MoreKeys.AltLatinsmallletterg,
            "\x1b[27;3;103~": MoreKeys.AltLatinsmallletterg,
            "\x1b[102;3u": MoreKeys.AltLatinsmallletterf,
            "\x1b[27;3;102~": MoreKeys.AltLatinsmallletterf,
            "\x1b[101;3u": MoreKeys.AltLatinsmalllettere,
            "\x1b[27;3;101~": MoreKeys.AltLatinsmalllettere,
            "\x1b[100;3u": MoreKeys.AltLatinsmallletterd,
            "\x1b[27;3;100~": MoreKeys.AltLatinsmallletterd,
            "\x1b[99;3u": MoreKeys.AltLatinsmallletterc,
            "\x1b[27;3;99~": MoreKeys.AltLatinsmallletterc,
            "\x1b[98;3u": MoreKeys.AltLatinsmallletterb,
            "\x1b[27;3;98~": MoreKeys.AltLatinsmallletterb,
            "\x1b[97;3u": MoreKeys.AltLatinsmalllettera,
            "\x1b[27;3;97~": MoreKeys.AltLatinsmalllettera,
            "\x1b[96;3u": MoreKeys.AltGraveaccent,
            "\x1b[27;3;96~": MoreKeys.AltGraveaccent,
            "\x1b[95;3u": MoreKeys.AltLowline,
            "\x1b[27;3;95~": MoreKeys.AltLowline,
            "\x1b[94;3u": MoreKeys.AltCircumflexaccent,
            "\x1b[27;3;94~": MoreKeys.AltCircumflexaccent,
            "\x1b[93;3u": MoreKeys.AltRightsquarebracket,
            "\x1b[27;3;93~": MoreKeys.AltRightsquarebracket,
            "\x1b[92;3u": MoreKeys.AltReversesolidus,
            "\x1b[27;3;92~": MoreKeys.AltReversesolidus,
            "\x1b[91;3u": MoreKeys.AltLeftsquarebracket,
            "\x1b[27;3;91~": MoreKeys.AltLeftsquarebracket,
            "\x1b[90;3u": MoreKeys.AltLatincapitalletterz,
            "\x1b[27;3;90~": MoreKeys.AltLatincapitalletterz,
            "\x1b[89;3u": MoreKeys.AltLatincapitallettery,
            "\x1b[27;3;89~": MoreKeys.AltLatincapitallettery,
            "\x1b[88;3u": MoreKeys.AltLatincapitalletterx,
            "\x1b[27;3;88~": MoreKeys.AltLatincapitalletterx,
            "\x1b[87;3u": MoreKeys.AltLatincapitalletterw,
            "\x1b[27;3;87~": MoreKeys.AltLatincapitalletterw,
            "\x1b[86;3u": MoreKeys.AltLatincapitalletterv,
            "\x1b[27;3;86~": MoreKeys.AltLatincapitalletterv,
            "\x1b[85;3u": MoreKeys.AltLatincapitalletteru,
            "\x1b[27;3;85~": MoreKeys.AltLatincapitalletteru,
            "\x1b[84;3u": MoreKeys.AltLatincapitallettert,
            "\x1b[27;3;84~": MoreKeys.AltLatincapitallettert,
            "\x1b[83;3u": MoreKeys.AltLatincapitalletters,
            "\x1b[27;3;83~": MoreKeys.AltLatincapitalletters,
            "\x1b[82;3u": MoreKeys.AltLatincapitalletterr,
            "\x1b[27;3;82~": MoreKeys.AltLatincapitalletterr,
            "\x1b[81;3u": MoreKeys.AltLatincapitalletterq,
            "\x1b[27;3;81~": MoreKeys.AltLatincapitalletterq,
            "\x1b[80;3u": MoreKeys.AltLatincapitalletterp,
            "\x1b[27;3;80~": MoreKeys.AltLatincapitalletterp,
            "\x1b[79;3u": MoreKeys.AltLatincapitallettero,
            "\x1b[27;3;79~": MoreKeys.AltLatincapitallettero,
            "\x1b[78;3u": MoreKeys.AltLatincapitallettern,
            "\x1b[27;3;78~": MoreKeys.AltLatincapitallettern,
            "\x1b[77;3u": MoreKeys.AltLatincapitalletterm,
            "\x1b[27;3;77~": MoreKeys.AltLatincapitalletterm,
            "\x1b[76;3u": MoreKeys.AltLatincapitalletterl,
            "\x1b[27;3;76~": MoreKeys.AltLatincapitalletterl,
            "\x1b[75;3u": MoreKeys.AltLatincapitalletterk,
            "\x1b[27;3;75~": MoreKeys.AltLatincapitalletterk,
            "\x1b[74;3u": MoreKeys.AltLatincapitalletterj,
            "\x1b[27;3;74~": MoreKeys.AltLatincapitalletterj,
            "\x1b[73;3u": MoreKeys.AltLatincapitalletteri,
            "\x1b[27;3;73~": MoreKeys.AltLatincapitalletteri,
            "\x1b[72;3u": MoreKeys.AltLatincapitalletterh,
            "\x1b[27;3;72~": MoreKeys.AltLatincapitalletterh,
            "\x1b[71;3u": MoreKeys.AltLatincapitalletterg,
            "\x1b[27;3;71~": MoreKeys.AltLatincapitalletterg,
            "\x1b[70;3u": MoreKeys.AltLatincapitalletterf,
            "\x1b[27;3;70~": MoreKeys.AltLatincapitalletterf,
            "\x1b[69;3u": MoreKeys.AltLatincapitallettere,
            "\x1b[27;3;69~": MoreKeys.AltLatincapitallettere,
            "\x1b[68;3u": MoreKeys.AltLatincapitalletterd,
            "\x1b[27;3;68~": MoreKeys.AltLatincapitalletterd,
            "\x1b[67;3u": MoreKeys.AltLatincapitalletterc,
            "\x1b[27;3;67~": MoreKeys.AltLatincapitalletterc,
            "\x1b[66;3u": MoreKeys.AltLatincapitalletterb,
            "\x1b[27;3;66~": MoreKeys.AltLatincapitalletterb,
            "\x1b[65;3u": MoreKeys.AltLatincapitallettera,
            "\x1b[27;3;65~": MoreKeys.AltLatincapitallettera,
            "\x1b[64;3u": MoreKeys.AltCommercialat,
            "\x1b[27;3;64~": MoreKeys.AltCommercialat,
            "\x1b[63;3u": MoreKeys.AltQuestionmark,
            "\x1b[27;3;63~": MoreKeys.AltQuestionmark,
            "\x1b[62;3u": MoreKeys.AltGreaterthansign,
            "\x1b[27;3;62~": MoreKeys.AltGreaterthansign,
            "\x1b[61;3u": MoreKeys.AltEqualssign,
            "\x1b[27;3;61~": MoreKeys.AltEqualssign,
            "\x1b[60;3u": MoreKeys.AltLessthansign,
            "\x1b[27;3;60~": MoreKeys.AltLessthansign,
            "\x1b[59;3u": MoreKeys.AltSemicolon,
            "\x1b[27;3;59~": MoreKeys.AltSemicolon,
            "\x1b[58;3u": MoreKeys.AltColon,
            "\x1b[27;3;58~": MoreKeys.AltColon,
            "\x1b[57;3u": MoreKeys.AltDigitnine,
            "\x1b[27;3;57~": MoreKeys.AltDigitnine,
            "\x1b[56;3u": MoreKeys.AltDigiteight,
            "\x1b[27;3;56~": MoreKeys.AltDigiteight,
            "\x1b[55;3u": MoreKeys.AltDigitseven,
            "\x1b[27;3;55~": MoreKeys.AltDigitseven,
            "\x1b[54;3u": MoreKeys.AltDigitsix,
            "\x1b[27;3;54~": MoreKeys.AltDigitsix,
            "\x1b[53;3u": MoreKeys.AltDigitfive,
            "\x1b[27;3;53~": MoreKeys.AltDigitfive,
            "\x1b[52;3u": MoreKeys.AltDigitfour,
            "\x1b[27;3;52~": MoreKeys.AltDigitfour,
            "\x1b[51;3u": MoreKeys.AltDigitthree,
            "\x1b[27;3;51~": MoreKeys.AltDigitthree,
            "\x1b[50;3u": MoreKeys.AltDigittwo,
            "\x1b[27;3;50~": MoreKeys.AltDigittwo,
            "\x1b[49;3u": MoreKeys.AltDigitone,
            "\x1b[27;3;49~": MoreKeys.AltDigitone,
            "\x1b[48;3u": MoreKeys.AltDigitzero,
            "\x1b[27;3;48~": MoreKeys.AltDigitzero,
            "\x1b[47;3u": MoreKeys.AltSolidus,
            "\x1b[27;3;47~": MoreKeys.AltSolidus,
            "\x1b[46;3u": MoreKeys.AltFullstop,
            "\x1b[27;3;46~": MoreKeys.AltFullstop,
            "\x1b[45;3u": MoreKeys.AltHyphenminus,
            "\x1b[27;3;45~": MoreKeys.AltHyphenminus,
            "\x1b[44;3u": MoreKeys.AltComma,
            "\x1b[27;3;44~": MoreKeys.AltComma,
            "\x1b[43;3u": MoreKeys.AltPlussign,
            "\x1b[27;3;43~": MoreKeys.AltPlussign,
            "\x1b[42;3u": MoreKeys.AltAsterisk,
            "\x1b[27;3;42~": MoreKeys.AltAsterisk,
            "\x1b[41;3u": MoreKeys.AltRightparenthesis,
            "\x1b[27;3;41~": MoreKeys.AltRightparenthesis,
            "\x1b[40;3u": MoreKeys.AltLeftparenthesis,
            "\x1b[27;3;40~": MoreKeys.AltLeftparenthesis,
            "\x1b[39;3u": MoreKeys.AltApostrophe,
            "\x1b[27;3;39~": MoreKeys.AltApostrophe,
            "\x1b[38;3u": MoreKeys.AltAmpersand,
            "\x1b[27;3;38~": MoreKeys.AltAmpersand,
            "\x1b[37;3u": MoreKeys.AltPercentsign,
            "\x1b[27;3;37~": MoreKeys.AltPercentsign,
            "\x1b[36;3u": MoreKeys.AltDollarsign,
            "\x1b[27;3;36~": MoreKeys.AltDollarsign,
            "\x1b[35;3u": MoreKeys.AltNumbersign,
            "\x1b[27;3;35~": MoreKeys.AltNumbersign,
            "\x1b[34;3u": MoreKeys.AltQuotationmark,
            "\x1b[27;3;34~": MoreKeys.AltQuotationmark,
            "\x1b[33;3u": MoreKeys.AltExclamationmark,
            "\x1b[27;3;33~": MoreKeys.AltExclamationmark,
            "\x1b[127;3u": MoreKeys.AltBackspace,
            "\x1b[27;3;127~": MoreKeys.AltBackspace,
            "\x1b[2;3~": MoreKeys.AltInsert,
            "\x1b[3;3~": MoreKeys.AltDelete,
            "\x1b[5;3~": MoreKeys.AltPageup,
            "\x1b[6;3~": MoreKeys.AltPagedown,
            "\x1b[1;3A": MoreKeys.AltUp,
            "\x1b[1;3B": MoreKeys.AltDown,
            "\x1b[1;3C": MoreKeys.AltRight,
            "\x1b[1;3D": MoreKeys.AltLeft,
            "\x1b[1;3H": MoreKeys.AltHome,
            "\x1b[1;3F": MoreKeys.AltEnd,
            "\x1b[57361;3u": MoreKeys.AltPrintscreen,
            "\x1b[57362;3u": MoreKeys.AltPause,
            "\x1b[57363;3u": MoreKeys.AltMenu,
            "\x1b[1;3P": MoreKeys.AltF1,
            "\x1b[1;3Q": MoreKeys.AltF2,
            "\x1b[13;3~": MoreKeys.AltF3,
            "\x1b[1;3S": MoreKeys.AltF4,
            "\x1b[15;3~": MoreKeys.AltF5,
            "\x1b[17;3~": MoreKeys.AltF6,
            "\x1b[18;3~": MoreKeys.AltF7,
            "\x1b[19;3~": MoreKeys.AltF8,
            "\x1b[20;3~": MoreKeys.AltF9,
            "\x1b[21;3~": MoreKeys.AltF10,
            "\x1b[23;3~": MoreKeys.AltF11,
            "\x1b[24;3~": MoreKeys.AltF12,
            "\x1b[9;5u": MoreKeys.ControlTab,
            "\x1b[27;5;9~": MoreKeys.ControlTab,
            "\x1b[13;5u": MoreKeys.ControlEnter,
            "\x1b[27;5;13~": MoreKeys.ControlEnter,
            "\x1b[27;5u": MoreKeys.ControlEscape,
            "\x1b[27;5;27~": MoreKeys.ControlEscape,
            "\x1b[32;5u": MoreKeys.ControlSpace,
            "\x1b[27;5;32~": MoreKeys.ControlSpace,
            "\x1b[126;5u": MoreKeys.ControlTilde,
            "\x1b[27;5;126~": MoreKeys.ControlTilde,
            "\x1b[125;5u": MoreKeys.ControlRightcurlybracket,
            "\x1b[27;5;125~": MoreKeys.ControlRightcurlybracket,
            "\x1b[124;5u": MoreKeys.ControlVerticalline,
            "\x1b[27;5;124~": MoreKeys.ControlVerticalline,
            "\x1b[123;5u": MoreKeys.ControlLeftcurlybracket,
            "\x1b[27;5;123~": MoreKeys.ControlLeftcurlybracket,
            "\x1b[122;5u": MoreKeys.ControlLatinsmallletterz,
            "\x1b[27;5;122~": MoreKeys.ControlLatinsmallletterz,
            "\x1b[121;5u": MoreKeys.ControlLatinsmalllettery,
            "\x1b[27;5;121~": MoreKeys.ControlLatinsmalllettery,
            "\x1b[120;5u": MoreKeys.ControlLatinsmallletterx,
            "\x1b[27;5;120~": MoreKeys.ControlLatinsmallletterx,
            "\x1b[119;5u": MoreKeys.ControlLatinsmallletterw,
            "\x1b[27;5;119~": MoreKeys.ControlLatinsmallletterw,
            "\x1b[118;5u": MoreKeys.ControlLatinsmallletterv,
            "\x1b[27;5;118~": MoreKeys.ControlLatinsmallletterv,
            "\x1b[117;5u": MoreKeys.ControlLatinsmallletteru,
            "\x1b[27;5;117~": MoreKeys.ControlLatinsmallletteru,
            "\x1b[116;5u": MoreKeys.ControlLatinsmalllettert,
            "\x1b[27;5;116~": MoreKeys.ControlLatinsmalllettert,
            "\x1b[115;5u": MoreKeys.ControlLatinsmallletters,
            "\x1b[27;5;115~": MoreKeys.ControlLatinsmallletters,
            "\x1b[114;5u": MoreKeys.ControlLatinsmallletterr,
            "\x1b[27;5;114~": MoreKeys.ControlLatinsmallletterr,
            "\x1b[113;5u": MoreKeys.ControlLatinsmallletterq,
            "\x1b[27;5;113~": MoreKeys.ControlLatinsmallletterq,
            "\x1b[112;5u": MoreKeys.ControlLatinsmallletterp,
            "\x1b[27;5;112~": MoreKeys.ControlLatinsmallletterp,
            "\x1b[111;5u": MoreKeys.ControlLatinsmalllettero,
            "\x1b[27;5;111~": MoreKeys.ControlLatinsmalllettero,
            "\x1b[110;5u": MoreKeys.ControlLatinsmalllettern,
            "\x1b[27;5;110~": MoreKeys.ControlLatinsmalllettern,
            "\x1b[109;5u": MoreKeys.ControlLatinsmallletterm,
            "\x1b[27;5;109~": MoreKeys.ControlLatinsmallletterm,
            "\x1b[108;5u": MoreKeys.ControlLatinsmallletterl,
            "\x1b[27;5;108~": MoreKeys.ControlLatinsmallletterl,
            "\x1b[107;5u": MoreKeys.ControlLatinsmallletterk,
            "\x1b[27;5;107~": MoreKeys.ControlLatinsmallletterk,
            "\x1b[106;5u": MoreKeys.ControlLatinsmallletterj,
            "\x1b[27;5;106~": MoreKeys.ControlLatinsmallletterj,
            "\x1b[105;5u": MoreKeys.ControlLatinsmallletteri,
            "\x1b[27;5;105~": MoreKeys.ControlLatinsmallletteri,
            "\x1b[104;5u": MoreKeys.ControlLatinsmallletterh,
            "\x1b[27;5;104~": MoreKeys.ControlLatinsmallletterh,
            "\x1b[103;5u": MoreKeys.ControlLatinsmallletterg,
            "\x1b[27;5;103~": MoreKeys.ControlLatinsmallletterg,
            "\x1b[102;5u": MoreKeys.ControlLatinsmallletterf,
            "\x1b[27;5;102~": MoreKeys.ControlLatinsmallletterf,
            "\x1b[101;5u": MoreKeys.ControlLatinsmalllettere,
            "\x1b[27;5;101~": MoreKeys.ControlLatinsmalllettere,
            "\x1b[100;5u": MoreKeys.ControlLatinsmallletterd,
            "\x1b[27;5;100~": MoreKeys.ControlLatinsmallletterd,
            "\x1b[99;5u": MoreKeys.ControlLatinsmallletterc,
            "\x1b[27;5;99~": MoreKeys.ControlLatinsmallletterc,
            "\x1b[98;5u": MoreKeys.ControlLatinsmallletterb,
            "\x1b[27;5;98~": MoreKeys.ControlLatinsmallletterb,
            "\x1b[97;5u": MoreKeys.ControlLatinsmalllettera,
            "\x1b[27;5;97~": MoreKeys.ControlLatinsmalllettera,
            "\x1b[96;5u": MoreKeys.ControlGraveaccent,
            "\x1b[27;5;96~": MoreKeys.ControlGraveaccent,
            "\x1b[95;5u": MoreKeys.ControlLowline,
            "\x1b[27;5;95~": MoreKeys.ControlLowline,
            "\x1b[94;5u": MoreKeys.ControlCircumflexaccent,
            "\x1b[27;5;94~": MoreKeys.ControlCircumflexaccent,
            "\x1b[93;5u": MoreKeys.ControlRightsquarebracket,
            "\x1b[27;5;93~": MoreKeys.ControlRightsquarebracket,
            "\x1b[92;5u": MoreKeys.ControlReversesolidus,
            "\x1b[27;5;92~": MoreKeys.ControlReversesolidus,
            "\x1b[91;5u": MoreKeys.ControlLeftsquarebracket,
            "\x1b[27;5;91~": MoreKeys.ControlLeftsquarebracket,
            "\x1b[90;5u": MoreKeys.ControlLatincapitalletterz,
            "\x1b[27;5;90~": MoreKeys.ControlLatincapitalletterz,
            "\x1b[89;5u": MoreKeys.ControlLatincapitallettery,
            "\x1b[27;5;89~": MoreKeys.ControlLatincapitallettery,
            "\x1b[88;5u": MoreKeys.ControlLatincapitalletterx,
            "\x1b[27;5;88~": MoreKeys.ControlLatincapitalletterx,
            "\x1b[87;5u": MoreKeys.ControlLatincapitalletterw,
            "\x1b[27;5;87~": MoreKeys.ControlLatincapitalletterw,
            "\x1b[86;5u": MoreKeys.ControlLatincapitalletterv,
            "\x1b[27;5;86~": MoreKeys.ControlLatincapitalletterv,
            "\x1b[85;5u": MoreKeys.ControlLatincapitalletteru,
            "\x1b[27;5;85~": MoreKeys.ControlLatincapitalletteru,
            "\x1b[84;5u": MoreKeys.ControlLatincapitallettert,
            "\x1b[27;5;84~": MoreKeys.ControlLatincapitallettert,
            "\x1b[83;5u": MoreKeys.ControlLatincapitalletters,
            "\x1b[27;5;83~": MoreKeys.ControlLatincapitalletters,
            "\x1b[82;5u": MoreKeys.ControlLatincapitalletterr,
            "\x1b[27;5;82~": MoreKeys.ControlLatincapitalletterr,
            "\x1b[81;5u": MoreKeys.ControlLatincapitalletterq,
            "\x1b[27;5;81~": MoreKeys.ControlLatincapitalletterq,
            "\x1b[80;5u": MoreKeys.ControlLatincapitalletterp,
            "\x1b[27;5;80~": MoreKeys.ControlLatincapitalletterp,
            "\x1b[79;5u": MoreKeys.ControlLatincapitallettero,
            "\x1b[27;5;79~": MoreKeys.ControlLatincapitallettero,
            "\x1b[78;5u": MoreKeys.ControlLatincapitallettern,
            "\x1b[27;5;78~": MoreKeys.ControlLatincapitallettern,
            "\x1b[77;5u": MoreKeys.ControlLatincapitalletterm,
            "\x1b[27;5;77~": MoreKeys.ControlLatincapitalletterm,
            "\x1b[76;5u": MoreKeys.ControlLatincapitalletterl,
            "\x1b[27;5;76~": MoreKeys.ControlLatincapitalletterl,
            "\x1b[75;5u": MoreKeys.ControlLatincapitalletterk,
            "\x1b[27;5;75~": MoreKeys.ControlLatincapitalletterk,
            "\x1b[74;5u": MoreKeys.ControlLatincapitalletterj,
            "\x1b[27;5;74~": MoreKeys.ControlLatincapitalletterj,
            "\x1b[73;5u": MoreKeys.ControlLatincapitalletteri,
            "\x1b[27;5;73~": MoreKeys.ControlLatincapitalletteri,
            "\x1b[72;5u": MoreKeys.ControlLatincapitalletterh,
            "\x1b[27;5;72~": MoreKeys.ControlLatincapitalletterh,
            "\x1b[71;5u": MoreKeys.ControlLatincapitalletterg,
            "\x1b[27;5;71~": MoreKeys.ControlLatincapitalletterg,
            "\x1b[70;5u": MoreKeys.ControlLatincapitalletterf,
            "\x1b[27;5;70~": MoreKeys.ControlLatincapitalletterf,
            "\x1b[69;5u": MoreKeys.ControlLatincapitallettere,
            "\x1b[27;5;69~": MoreKeys.ControlLatincapitallettere,
            "\x1b[68;5u": MoreKeys.ControlLatincapitalletterd,
            "\x1b[27;5;68~": MoreKeys.ControlLatincapitalletterd,
            "\x1b[67;5u": MoreKeys.ControlLatincapitalletterc,
            "\x1b[27;5;67~": MoreKeys.ControlLatincapitalletterc,
            "\x1b[66;5u": MoreKeys.ControlLatincapitalletterb,
            "\x1b[27;5;66~": MoreKeys.ControlLatincapitalletterb,
            "\x1b[65;5u": MoreKeys.ControlLatincapitallettera,
            "\x1b[27;5;65~": MoreKeys.ControlLatincapitallettera,
            "\x1b[64;5u": MoreKeys.ControlCommercialat,
            "\x1b[27;5;64~": MoreKeys.ControlCommercialat,
            "\x1b[63;5u": MoreKeys.ControlQuestionmark,
            "\x1b[27;5;63~": MoreKeys.ControlQuestionmark,
            "\x1b[62;5u": MoreKeys.ControlGreaterthansign,
            "\x1b[27;5;62~": MoreKeys.ControlGreaterthansign,
            "\x1b[61;5u": MoreKeys.ControlEqualssign,
            "\x1b[27;5;61~": MoreKeys.ControlEqualssign,
            "\x1b[60;5u": MoreKeys.ControlLessthansign,
            "\x1b[27;5;60~": MoreKeys.ControlLessthansign,
            "\x1b[59;5u": MoreKeys.ControlSemicolon,
            "\x1b[27;5;59~": MoreKeys.ControlSemicolon,
            "\x1b[58;5u": MoreKeys.ControlColon,
            "\x1b[27;5;58~": MoreKeys.ControlColon,
            "\x1b[57;5u": MoreKeys.ControlDigitnine,
            "\x1b[27;5;57~": MoreKeys.ControlDigitnine,
            "\x1b[56;5u": MoreKeys.ControlDigiteight,
            "\x1b[27;5;56~": MoreKeys.ControlDigiteight,
            "\x1b[55;5u": MoreKeys.ControlDigitseven,
            "\x1b[27;5;55~": MoreKeys.ControlDigitseven,
            "\x1b[54;5u": MoreKeys.ControlDigitsix,
            "\x1b[27;5;54~": MoreKeys.ControlDigitsix,
            "\x1b[53;5u": MoreKeys.ControlDigitfive,
            "\x1b[27;5;53~": MoreKeys.ControlDigitfive,
            "\x1b[52;5u": MoreKeys.ControlDigitfour,
            "\x1b[27;5;52~": MoreKeys.ControlDigitfour,
            "\x1b[51;5u": MoreKeys.ControlDigitthree,
            "\x1b[27;5;51~": MoreKeys.ControlDigitthree,
            "\x1b[50;5u": MoreKeys.ControlDigittwo,
            "\x1b[27;5;50~": MoreKeys.ControlDigittwo,
            "\x1b[49;5u": MoreKeys.ControlDigitone,
            "\x1b[27;5;49~": MoreKeys.ControlDigitone,
            "\x1b[48;5u": MoreKeys.ControlDigitzero,
            "\x1b[27;5;48~": MoreKeys.ControlDigitzero,
            "\x1b[47;5u": MoreKeys.ControlSolidus,
            "\x1b[27;5;47~": MoreKeys.ControlSolidus,
            "\x1b[46;5u": MoreKeys.ControlFullstop,
            "\x1b[27;5;46~": MoreKeys.ControlFullstop,
            "\x1b[45;5u": MoreKeys.ControlHyphenminus,
            "\x1b[27;5;45~": MoreKeys.ControlHyphenminus,
            "\x1b[44;5u": MoreKeys.ControlComma,
            "\x1b[27;5;44~": MoreKeys.ControlComma,
            "\x1b[43;5u": MoreKeys.ControlPlussign,
            "\x1b[27;5;43~": MoreKeys.ControlPlussign,
            "\x1b[42;5u": MoreKeys.ControlAsterisk,
            "\x1b[27;5;42~": MoreKeys.ControlAsterisk,
            "\x1b[41;5u": MoreKeys.ControlRightparenthesis,
            "\x1b[27;5;41~": MoreKeys.ControlRightparenthesis,
            "\x1b[40;5u": MoreKeys.ControlLeftparenthesis,
            "\x1b[27;5;40~": MoreKeys.ControlLeftparenthesis,
            "\x1b[39;5u": MoreKeys.ControlApostrophe,
            "\x1b[27;5;39~": MoreKeys.ControlApostrophe,
            "\x1b[38;5u": MoreKeys.ControlAmpersand,
            "\x1b[27;5;38~": MoreKeys.ControlAmpersand,
            "\x1b[37;5u": MoreKeys.ControlPercentsign,
            "\x1b[27;5;37~": MoreKeys.ControlPercentsign,
            "\x1b[36;5u": MoreKeys.ControlDollarsign,
            "\x1b[27;5;36~": MoreKeys.ControlDollarsign,
            "\x1b[35;5u": MoreKeys.ControlNumbersign,
            "\x1b[27;5;35~": MoreKeys.ControlNumbersign,
            "\x1b[34;5u": MoreKeys.ControlQuotationmark,
            "\x1b[27;5;34~": MoreKeys.ControlQuotationmark,
            "\x1b[33;5u": MoreKeys.ControlExclamationmark,
            "\x1b[27;5;33~": MoreKeys.ControlExclamationmark,
            "\x1b[127;5u": MoreKeys.ControlBackspace,
            "\x1b[27;5;127~": MoreKeys.ControlBackspace,
            "\x1b[2;5~": MoreKeys.ControlInsert,
            "\x1b[3;5~": MoreKeys.ControlDelete,
            "\x1b[5;5~": MoreKeys.ControlPageup,
            "\x1b[6;5~": MoreKeys.ControlPagedown,
            "\x1b[1;5A": MoreKeys.ControlUp,
            "\x1b[1;5B": MoreKeys.ControlDown,
            "\x1b[1;5C": MoreKeys.ControlRight,
            "\x1b[1;5D": MoreKeys.ControlLeft,
            "\x1b[1;5H": MoreKeys.ControlHome,
            "\x1b[1;5F": MoreKeys.ControlEnd,
            "\x1b[57361;5u": MoreKeys.ControlPrintscreen,
            "\x1b[57362;5u": MoreKeys.ControlPause,
            "\x1b[57363;5u": MoreKeys.ControlMenu,
            "\x1b[1;5P": MoreKeys.ControlF1,
            "\x1b[1;5Q": MoreKeys.ControlF2,
            "\x1b[13;5~": MoreKeys.ControlF3,
            "\x1b[1;5S": MoreKeys.ControlF4,
            "\x1b[15;5~": MoreKeys.ControlF5,
            "\x1b[17;5~": MoreKeys.ControlF6,
            "\x1b[18;5~": MoreKeys.ControlF7,
            "\x1b[19;5~": MoreKeys.ControlF8,
            "\x1b[20;5~": MoreKeys.ControlF9,
            "\x1b[21;5~": MoreKeys.ControlF10,
            "\x1b[23;5~": MoreKeys.ControlF11,
            "\x1b[24;5~": MoreKeys.ControlF12,
            "\x1b[27;65u": MoreKeys.Escape,
            "\x1b[2;65~": MoreKeys.Insert,
            "\x1b[3;65~": MoreKeys.Delete,
            "\x1b[5;65~": MoreKeys.Pageup,
            "\x1b[6;65~": MoreKeys.Pagedown,
            "\x1b[1;65A": MoreKeys.Up,
            "\x1b[1;65B": MoreKeys.Down,
            "\x1b[1;65C": MoreKeys.Right,
            "\x1b[1;65D": MoreKeys.Left,
            "\x1b[1;65H": MoreKeys.Home,
            "\x1b[1;65F": MoreKeys.End,
            "\x1b[57361;65u": MoreKeys.Printscreen,
            "\x1b[57362;65u": MoreKeys.Pause,
            "\x1b[57363;65u": MoreKeys.Menu,
            "\x1b[1;65P": MoreKeys.F1,
            "\x1b[1;65Q": MoreKeys.F2,
            "\x1b[13;65~": MoreKeys.F3,
            "\x1b[1;65S": MoreKeys.F4,
            "\x1b[15;65~": MoreKeys.F5,
            "\x1b[17;65~": MoreKeys.F6,
            "\x1b[18;65~": MoreKeys.F7,
            "\x1b[19;65~": MoreKeys.F8,
            "\x1b[20;65~": MoreKeys.F9,
            "\x1b[21;65~": MoreKeys.F10,
            "\x1b[23;65~": MoreKeys.F11,
            "\x1b[24;65~": MoreKeys.F12,
            "\x1b[127;65u": MoreKeys.Backspace,
            "\x1b[13;65u": MoreKeys.Enter,
            "\x1b[9;65u": MoreKeys.Tab,
            "\x1b[27;129u": MoreKeys.Escape,
            "\x1b[2;129~": MoreKeys.Insert,
            "\x1b[3;129~": MoreKeys.Delete,
            "\x1b[5;129~": MoreKeys.Pageup,
            "\x1b[6;129~": MoreKeys.Pagedown,
            "\x1b[1;129A": MoreKeys.Up,
            "\x1b[1;129B": MoreKeys.Down,
            "\x1b[1;129C": MoreKeys.Right,
            "\x1b[1;129D": MoreKeys.Left,
            "\x1b[1;129H": MoreKeys.Home,
            "\x1b[1;129F": MoreKeys.End,
            "\x1b[57361;129u": MoreKeys.Printscreen,
            "\x1b[57362;129u": MoreKeys.Pause,
            "\x1b[57363;129u": MoreKeys.Menu,
            "\x1b[1;129P": MoreKeys.F1,
            "\x1b[1;129Q": MoreKeys.F2,
            "\x1b[13;129~": MoreKeys.F3,
            "\x1b[1;129S": MoreKeys.F4,
            "\x1b[15;129~": MoreKeys.F5,
            "\x1b[17;129~": MoreKeys.F6,
            "\x1b[18;129~": MoreKeys.F7,
            "\x1b[19;129~": MoreKeys.F8,
            "\x1b[20;129~": MoreKeys.F9,
            "\x1b[21;129~": MoreKeys.F10,
            "\x1b[23;129~": MoreKeys.F11,
            "\x1b[24;129~": MoreKeys.F12,
            "\x1b[127;129u": MoreKeys.Backspace,
            "\x1b[13;129u": MoreKeys.Enter,
            "\x1b[9;129u": MoreKeys.Tab,
            "\x1b[9;4u": MoreKeys.AltShiftTab,
            "\x1b[27;4;9~": MoreKeys.AltShiftTab,
            "\x1b[13;4u": MoreKeys.AltShiftEnter,
            "\x1b[27;4;13~": MoreKeys.AltShiftEnter,
            "\x1b[27;4u": MoreKeys.AltShiftEscape,
            "\x1b[27;4;27~": MoreKeys.AltShiftEscape,
            "\x1b[32;4u": MoreKeys.AltShiftSpace,
            "\x1b[27;4;32~": MoreKeys.AltShiftSpace,
            "\x1b[126;4u": MoreKeys.AltShiftTilde,
            "\x1b[27;4;126~": MoreKeys.AltShiftTilde,
            "\x1b[125;4u": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[27;4;125~": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[124;4u": MoreKeys.AltShiftVerticalline,
            "\x1b[27;4;124~": MoreKeys.AltShiftVerticalline,
            "\x1b[123;4u": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[27;4;123~": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[122;4u": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[27;4;122~": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[121;4u": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[27;4;121~": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[120;4u": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[27;4;120~": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[119;4u": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[27;4;119~": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[118;4u": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[27;4;118~": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[117;4u": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[27;4;117~": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[116;4u": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[27;4;116~": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[115;4u": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[27;4;115~": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[114;4u": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[27;4;114~": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[113;4u": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[27;4;113~": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[112;4u": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[27;4;112~": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[111;4u": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[27;4;111~": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[110;4u": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[27;4;110~": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[109;4u": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[27;4;109~": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[108;4u": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[27;4;108~": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[107;4u": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[27;4;107~": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[106;4u": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[27;4;106~": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[105;4u": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[27;4;105~": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[104;4u": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[27;4;104~": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[103;4u": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[27;4;103~": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[102;4u": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[27;4;102~": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[101;4u": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[27;4;101~": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[100;4u": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[27;4;100~": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[99;4u": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[27;4;99~": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[98;4u": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[27;4;98~": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[97;4u": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[27;4;97~": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[96;4u": MoreKeys.AltShiftGraveaccent,
            "\x1b[27;4;96~": MoreKeys.AltShiftGraveaccent,
            "\x1b[95;4u": MoreKeys.AltShiftLowline,
            "\x1b[27;4;95~": MoreKeys.AltShiftLowline,
            "\x1b[94;4u": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[27;4;94~": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[93;4u": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[27;4;93~": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[92;4u": MoreKeys.AltShiftReversesolidus,
            "\x1b[27;4;92~": MoreKeys.AltShiftReversesolidus,
            "\x1b[91;4u": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[27;4;91~": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[90;4u": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[27;4;90~": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[89;4u": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[27;4;89~": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[88;4u": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[27;4;88~": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[87;4u": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[27;4;87~": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[86;4u": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[27;4;86~": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[85;4u": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[27;4;85~": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[84;4u": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[27;4;84~": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[83;4u": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[27;4;83~": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[82;4u": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[27;4;82~": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[81;4u": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[27;4;81~": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[80;4u": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[27;4;80~": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[79;4u": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[27;4;79~": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[78;4u": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[27;4;78~": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[77;4u": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[27;4;77~": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[76;4u": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[27;4;76~": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[75;4u": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[27;4;75~": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[74;4u": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[27;4;74~": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[73;4u": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[27;4;73~": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[72;4u": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[27;4;72~": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[71;4u": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[27;4;71~": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[70;4u": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[27;4;70~": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[69;4u": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[27;4;69~": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[68;4u": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[27;4;68~": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[67;4u": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[27;4;67~": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[66;4u": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[27;4;66~": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[65;4u": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[27;4;65~": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[64;4u": MoreKeys.AltShiftCommercialat,
            "\x1b[27;4;64~": MoreKeys.AltShiftCommercialat,
            "\x1b[63;4u": MoreKeys.AltShiftQuestionmark,
            "\x1b[27;4;63~": MoreKeys.AltShiftQuestionmark,
            "\x1b[62;4u": MoreKeys.AltShiftGreaterthansign,
            "\x1b[27;4;62~": MoreKeys.AltShiftGreaterthansign,
            "\x1b[61;4u": MoreKeys.AltShiftEqualssign,
            "\x1b[27;4;61~": MoreKeys.AltShiftEqualssign,
            "\x1b[60;4u": MoreKeys.AltShiftLessthansign,
            "\x1b[27;4;60~": MoreKeys.AltShiftLessthansign,
            "\x1b[59;4u": MoreKeys.AltShiftSemicolon,
            "\x1b[27;4;59~": MoreKeys.AltShiftSemicolon,
            "\x1b[58;4u": MoreKeys.AltShiftColon,
            "\x1b[27;4;58~": MoreKeys.AltShiftColon,
            "\x1b[57;4u": MoreKeys.AltShiftDigitnine,
            "\x1b[27;4;57~": MoreKeys.AltShiftDigitnine,
            "\x1b[56;4u": MoreKeys.AltShiftDigiteight,
            "\x1b[27;4;56~": MoreKeys.AltShiftDigiteight,
            "\x1b[55;4u": MoreKeys.AltShiftDigitseven,
            "\x1b[27;4;55~": MoreKeys.AltShiftDigitseven,
            "\x1b[54;4u": MoreKeys.AltShiftDigitsix,
            "\x1b[27;4;54~": MoreKeys.AltShiftDigitsix,
            "\x1b[53;4u": MoreKeys.AltShiftDigitfive,
            "\x1b[27;4;53~": MoreKeys.AltShiftDigitfive,
            "\x1b[52;4u": MoreKeys.AltShiftDigitfour,
            "\x1b[27;4;52~": MoreKeys.AltShiftDigitfour,
            "\x1b[51;4u": MoreKeys.AltShiftDigitthree,
            "\x1b[27;4;51~": MoreKeys.AltShiftDigitthree,
            "\x1b[50;4u": MoreKeys.AltShiftDigittwo,
            "\x1b[27;4;50~": MoreKeys.AltShiftDigittwo,
            "\x1b[49;4u": MoreKeys.AltShiftDigitone,
            "\x1b[27;4;49~": MoreKeys.AltShiftDigitone,
            "\x1b[48;4u": MoreKeys.AltShiftDigitzero,
            "\x1b[27;4;48~": MoreKeys.AltShiftDigitzero,
            "\x1b[47;4u": MoreKeys.AltShiftSolidus,
            "\x1b[27;4;47~": MoreKeys.AltShiftSolidus,
            "\x1b[46;4u": MoreKeys.AltShiftFullstop,
            "\x1b[27;4;46~": MoreKeys.AltShiftFullstop,
            "\x1b[45;4u": MoreKeys.AltShiftHyphenminus,
            "\x1b[27;4;45~": MoreKeys.AltShiftHyphenminus,
            "\x1b[44;4u": MoreKeys.AltShiftComma,
            "\x1b[27;4;44~": MoreKeys.AltShiftComma,
            "\x1b[43;4u": MoreKeys.AltShiftPlussign,
            "\x1b[27;4;43~": MoreKeys.AltShiftPlussign,
            "\x1b[42;4u": MoreKeys.AltShiftAsterisk,
            "\x1b[27;4;42~": MoreKeys.AltShiftAsterisk,
            "\x1b[41;4u": MoreKeys.AltShiftRightparenthesis,
            "\x1b[27;4;41~": MoreKeys.AltShiftRightparenthesis,
            "\x1b[40;4u": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[27;4;40~": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[39;4u": MoreKeys.AltShiftApostrophe,
            "\x1b[27;4;39~": MoreKeys.AltShiftApostrophe,
            "\x1b[38;4u": MoreKeys.AltShiftAmpersand,
            "\x1b[27;4;38~": MoreKeys.AltShiftAmpersand,
            "\x1b[37;4u": MoreKeys.AltShiftPercentsign,
            "\x1b[27;4;37~": MoreKeys.AltShiftPercentsign,
            "\x1b[36;4u": MoreKeys.AltShiftDollarsign,
            "\x1b[27;4;36~": MoreKeys.AltShiftDollarsign,
            "\x1b[35;4u": MoreKeys.AltShiftNumbersign,
            "\x1b[27;4;35~": MoreKeys.AltShiftNumbersign,
            "\x1b[34;4u": MoreKeys.AltShiftQuotationmark,
            "\x1b[27;4;34~": MoreKeys.AltShiftQuotationmark,
            "\x1b[33;4u": MoreKeys.AltShiftExclamationmark,
            "\x1b[27;4;33~": MoreKeys.AltShiftExclamationmark,
            "\x1b[127;4u": MoreKeys.AltShiftBackspace,
            "\x1b[27;4;127~": MoreKeys.AltShiftBackspace,
            "\x1b[2;4~": MoreKeys.AltShiftInsert,
            "\x1b[3;4~": MoreKeys.AltShiftDelete,
            "\x1b[5;4~": MoreKeys.AltShiftPageup,
            "\x1b[6;4~": MoreKeys.AltShiftPagedown,
            "\x1b[1;4A": MoreKeys.AltShiftUp,
            "\x1b[1;4B": MoreKeys.AltShiftDown,
            "\x1b[1;4C": MoreKeys.AltShiftRight,
            "\x1b[1;4D": MoreKeys.AltShiftLeft,
            "\x1b[1;4H": MoreKeys.AltShiftHome,
            "\x1b[1;4F": MoreKeys.AltShiftEnd,
            "\x1b[57361;4u": MoreKeys.AltShiftPrintscreen,
            "\x1b[57362;4u": MoreKeys.AltShiftPause,
            "\x1b[57363;4u": MoreKeys.AltShiftMenu,
            "\x1b[1;4P": MoreKeys.AltShiftF1,
            "\x1b[1;4Q": MoreKeys.AltShiftF2,
            "\x1b[13;4~": MoreKeys.AltShiftF3,
            "\x1b[1;4S": MoreKeys.AltShiftF4,
            "\x1b[15;4~": MoreKeys.AltShiftF5,
            "\x1b[17;4~": MoreKeys.AltShiftF6,
            "\x1b[18;4~": MoreKeys.AltShiftF7,
            "\x1b[19;4~": MoreKeys.AltShiftF8,
            "\x1b[20;4~": MoreKeys.AltShiftF9,
            "\x1b[21;4~": MoreKeys.AltShiftF10,
            "\x1b[23;4~": MoreKeys.AltShiftF11,
            "\x1b[24;4~": MoreKeys.AltShiftF12,
            "\x1b[9;6u": MoreKeys.ControlShiftTab,
            "\x1b[27;6;9~": MoreKeys.ControlShiftTab,
            "\x1b[13;6u": MoreKeys.ControlShiftEnter,
            "\x1b[27;6;13~": MoreKeys.ControlShiftEnter,
            "\x1b[27;6u": MoreKeys.ControlShiftEscape,
            "\x1b[27;6;27~": MoreKeys.ControlShiftEscape,
            "\x1b[32;6u": MoreKeys.ControlShiftSpace,
            "\x1b[27;6;32~": MoreKeys.ControlShiftSpace,
            "\x1b[126;6u": MoreKeys.ControlShiftTilde,
            "\x1b[27;6;126~": MoreKeys.ControlShiftTilde,
            "\x1b[125;6u": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[27;6;125~": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[124;6u": MoreKeys.ControlShiftVerticalline,
            "\x1b[27;6;124~": MoreKeys.ControlShiftVerticalline,
            "\x1b[123;6u": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[27;6;123~": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[122;6u": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[27;6;122~": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[121;6u": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[27;6;121~": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[120;6u": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[27;6;120~": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[119;6u": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[27;6;119~": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[118;6u": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[27;6;118~": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[117;6u": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[27;6;117~": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[116;6u": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[27;6;116~": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[115;6u": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[27;6;115~": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[114;6u": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[27;6;114~": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[113;6u": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[27;6;113~": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[112;6u": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[27;6;112~": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[111;6u": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[27;6;111~": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[110;6u": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[27;6;110~": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[109;6u": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[27;6;109~": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[108;6u": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[27;6;108~": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[107;6u": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[27;6;107~": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[106;6u": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[27;6;106~": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[105;6u": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[27;6;105~": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[104;6u": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[27;6;104~": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[103;6u": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[27;6;103~": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[102;6u": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[27;6;102~": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[101;6u": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[27;6;101~": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[100;6u": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[27;6;100~": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[99;6u": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[27;6;99~": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[98;6u": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[27;6;98~": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[97;6u": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[27;6;97~": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[96;6u": MoreKeys.ControlShiftGraveaccent,
            "\x1b[27;6;96~": MoreKeys.ControlShiftGraveaccent,
            "\x1b[95;6u": MoreKeys.ControlShiftLowline,
            "\x1b[27;6;95~": MoreKeys.ControlShiftLowline,
            "\x1b[94;6u": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[27;6;94~": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[93;6u": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[27;6;93~": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[92;6u": MoreKeys.ControlShiftReversesolidus,
            "\x1b[27;6;92~": MoreKeys.ControlShiftReversesolidus,
            "\x1b[91;6u": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[27;6;91~": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[90;6u": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[27;6;90~": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[89;6u": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[27;6;89~": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[88;6u": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[27;6;88~": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[87;6u": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[27;6;87~": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[86;6u": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[27;6;86~": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[85;6u": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[27;6;85~": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[84;6u": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[27;6;84~": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[83;6u": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[27;6;83~": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[82;6u": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[27;6;82~": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[81;6u": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[27;6;81~": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[80;6u": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[27;6;80~": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[79;6u": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[27;6;79~": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[78;6u": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[27;6;78~": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[77;6u": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[27;6;77~": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[76;6u": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[27;6;76~": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[75;6u": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[27;6;75~": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[74;6u": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[27;6;74~": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[73;6u": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[27;6;73~": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[72;6u": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[27;6;72~": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[71;6u": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[27;6;71~": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[70;6u": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[27;6;70~": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[69;6u": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[27;6;69~": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[68;6u": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[27;6;68~": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[67;6u": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[27;6;67~": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[66;6u": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[27;6;66~": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[65;6u": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[27;6;65~": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[64;6u": MoreKeys.ControlShiftCommercialat,
            "\x1b[27;6;64~": MoreKeys.ControlShiftCommercialat,
            "\x1b[63;6u": MoreKeys.ControlShiftQuestionmark,
            "\x1b[27;6;63~": MoreKeys.ControlShiftQuestionmark,
            "\x1b[62;6u": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[27;6;62~": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[61;6u": MoreKeys.ControlShiftEqualssign,
            "\x1b[27;6;61~": MoreKeys.ControlShiftEqualssign,
            "\x1b[60;6u": MoreKeys.ControlShiftLessthansign,
            "\x1b[27;6;60~": MoreKeys.ControlShiftLessthansign,
            "\x1b[59;6u": MoreKeys.ControlShiftSemicolon,
            "\x1b[27;6;59~": MoreKeys.ControlShiftSemicolon,
            "\x1b[58;6u": MoreKeys.ControlShiftColon,
            "\x1b[27;6;58~": MoreKeys.ControlShiftColon,
            "\x1b[57;6u": MoreKeys.ControlShiftDigitnine,
            "\x1b[27;6;57~": MoreKeys.ControlShiftDigitnine,
            "\x1b[56;6u": MoreKeys.ControlShiftDigiteight,
            "\x1b[27;6;56~": MoreKeys.ControlShiftDigiteight,
            "\x1b[55;6u": MoreKeys.ControlShiftDigitseven,
            "\x1b[27;6;55~": MoreKeys.ControlShiftDigitseven,
            "\x1b[54;6u": MoreKeys.ControlShiftDigitsix,
            "\x1b[27;6;54~": MoreKeys.ControlShiftDigitsix,
            "\x1b[53;6u": MoreKeys.ControlShiftDigitfive,
            "\x1b[27;6;53~": MoreKeys.ControlShiftDigitfive,
            "\x1b[52;6u": MoreKeys.ControlShiftDigitfour,
            "\x1b[27;6;52~": MoreKeys.ControlShiftDigitfour,
            "\x1b[51;6u": MoreKeys.ControlShiftDigitthree,
            "\x1b[27;6;51~": MoreKeys.ControlShiftDigitthree,
            "\x1b[50;6u": MoreKeys.ControlShiftDigittwo,
            "\x1b[27;6;50~": MoreKeys.ControlShiftDigittwo,
            "\x1b[49;6u": MoreKeys.ControlShiftDigitone,
            "\x1b[27;6;49~": MoreKeys.ControlShiftDigitone,
            "\x1b[48;6u": MoreKeys.ControlShiftDigitzero,
            "\x1b[27;6;48~": MoreKeys.ControlShiftDigitzero,
            "\x1b[47;6u": MoreKeys.ControlShiftSolidus,
            "\x1b[27;6;47~": MoreKeys.ControlShiftSolidus,
            "\x1b[46;6u": MoreKeys.ControlShiftFullstop,
            "\x1b[27;6;46~": MoreKeys.ControlShiftFullstop,
            "\x1b[45;6u": MoreKeys.ControlShiftHyphenminus,
            "\x1b[27;6;45~": MoreKeys.ControlShiftHyphenminus,
            "\x1b[44;6u": MoreKeys.ControlShiftComma,
            "\x1b[27;6;44~": MoreKeys.ControlShiftComma,
            "\x1b[43;6u": MoreKeys.ControlShiftPlussign,
            "\x1b[27;6;43~": MoreKeys.ControlShiftPlussign,
            "\x1b[42;6u": MoreKeys.ControlShiftAsterisk,
            "\x1b[27;6;42~": MoreKeys.ControlShiftAsterisk,
            "\x1b[41;6u": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[27;6;41~": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[40;6u": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[27;6;40~": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[39;6u": MoreKeys.ControlShiftApostrophe,
            "\x1b[27;6;39~": MoreKeys.ControlShiftApostrophe,
            "\x1b[38;6u": MoreKeys.ControlShiftAmpersand,
            "\x1b[27;6;38~": MoreKeys.ControlShiftAmpersand,
            "\x1b[37;6u": MoreKeys.ControlShiftPercentsign,
            "\x1b[27;6;37~": MoreKeys.ControlShiftPercentsign,
            "\x1b[36;6u": MoreKeys.ControlShiftDollarsign,
            "\x1b[27;6;36~": MoreKeys.ControlShiftDollarsign,
            "\x1b[35;6u": MoreKeys.ControlShiftNumbersign,
            "\x1b[27;6;35~": MoreKeys.ControlShiftNumbersign,
            "\x1b[34;6u": MoreKeys.ControlShiftQuotationmark,
            "\x1b[27;6;34~": MoreKeys.ControlShiftQuotationmark,
            "\x1b[33;6u": MoreKeys.ControlShiftExclamationmark,
            "\x1b[27;6;33~": MoreKeys.ControlShiftExclamationmark,
            "\x1b[127;6u": MoreKeys.ControlShiftBackspace,
            "\x1b[27;6;127~": MoreKeys.ControlShiftBackspace,
            "\x1b[2;6~": MoreKeys.ControlShiftInsert,
            "\x1b[3;6~": MoreKeys.ControlShiftDelete,
            "\x1b[5;6~": MoreKeys.ControlShiftPageup,
            "\x1b[6;6~": MoreKeys.ControlShiftPagedown,
            "\x1b[1;6A": MoreKeys.ControlShiftUp,
            "\x1b[1;6B": MoreKeys.ControlShiftDown,
            "\x1b[1;6C": MoreKeys.ControlShiftRight,
            "\x1b[1;6D": MoreKeys.ControlShiftLeft,
            "\x1b[1;6H": MoreKeys.ControlShiftHome,
            "\x1b[1;6F": MoreKeys.ControlShiftEnd,
            "\x1b[57361;6u": MoreKeys.ControlShiftPrintscreen,
            "\x1b[57362;6u": MoreKeys.ControlShiftPause,
            "\x1b[57363;6u": MoreKeys.ControlShiftMenu,
            "\x1b[1;6P": MoreKeys.ControlShiftF1,
            "\x1b[1;6Q": MoreKeys.ControlShiftF2,
            "\x1b[13;6~": MoreKeys.ControlShiftF3,
            "\x1b[1;6S": MoreKeys.ControlShiftF4,
            "\x1b[15;6~": MoreKeys.ControlShiftF5,
            "\x1b[17;6~": MoreKeys.ControlShiftF6,
            "\x1b[18;6~": MoreKeys.ControlShiftF7,
            "\x1b[19;6~": MoreKeys.ControlShiftF8,
            "\x1b[20;6~": MoreKeys.ControlShiftF9,
            "\x1b[21;6~": MoreKeys.ControlShiftF10,
            "\x1b[23;6~": MoreKeys.ControlShiftF11,
            "\x1b[24;6~": MoreKeys.ControlShiftF12,
            "\x1b[9;66u": MoreKeys.ShiftTab,
            "\x1b[27;66;9~": MoreKeys.ShiftTab,
            "\x1b[13;66u": MoreKeys.ShiftEnter,
            "\x1b[27;66;13~": MoreKeys.ShiftEnter,
            "\x1b[27;66u": MoreKeys.ShiftEscape,
            "\x1b[27;66;27~": MoreKeys.ShiftEscape,
            "\x1b[32;66u": MoreKeys.ShiftSpace,
            "\x1b[27;66;32~": MoreKeys.ShiftSpace,
            "\x1b[126;66u": MoreKeys.ShiftTilde,
            "\x1b[27;66;126~": MoreKeys.ShiftTilde,
            "\x1b[125;66u": MoreKeys.ShiftRightcurlybracket,
            "\x1b[27;66;125~": MoreKeys.ShiftRightcurlybracket,
            "\x1b[124;66u": MoreKeys.ShiftVerticalline,
            "\x1b[27;66;124~": MoreKeys.ShiftVerticalline,
            "\x1b[123;66u": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[27;66;123~": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[122;66u": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[27;66;122~": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[121;66u": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[27;66;121~": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[120;66u": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[27;66;120~": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[119;66u": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[27;66;119~": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[118;66u": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[27;66;118~": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[117;66u": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[27;66;117~": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[116;66u": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[27;66;116~": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[115;66u": MoreKeys.ShiftLatinsmallletters,
            "\x1b[27;66;115~": MoreKeys.ShiftLatinsmallletters,
            "\x1b[114;66u": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[27;66;114~": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[113;66u": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[27;66;113~": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[112;66u": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[27;66;112~": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[111;66u": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[27;66;111~": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[110;66u": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[27;66;110~": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[109;66u": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[27;66;109~": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[108;66u": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[27;66;108~": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[107;66u": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[27;66;107~": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[106;66u": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[27;66;106~": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[105;66u": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[27;66;105~": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[104;66u": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[27;66;104~": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[103;66u": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[27;66;103~": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[102;66u": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[27;66;102~": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[101;66u": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[27;66;101~": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[100;66u": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[27;66;100~": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[99;66u": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[27;66;99~": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[98;66u": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[27;66;98~": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[97;66u": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[27;66;97~": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[96;66u": MoreKeys.ShiftGraveaccent,
            "\x1b[27;66;96~": MoreKeys.ShiftGraveaccent,
            "\x1b[95;66u": MoreKeys.ShiftLowline,
            "\x1b[27;66;95~": MoreKeys.ShiftLowline,
            "\x1b[94;66u": MoreKeys.ShiftCircumflexaccent,
            "\x1b[27;66;94~": MoreKeys.ShiftCircumflexaccent,
            "\x1b[93;66u": MoreKeys.ShiftRightsquarebracket,
            "\x1b[27;66;93~": MoreKeys.ShiftRightsquarebracket,
            "\x1b[92;66u": MoreKeys.ShiftReversesolidus,
            "\x1b[27;66;92~": MoreKeys.ShiftReversesolidus,
            "\x1b[91;66u": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[27;66;91~": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[90;66u": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[27;66;90~": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[89;66u": MoreKeys.ShiftLatincapitallettery,
            "\x1b[27;66;89~": MoreKeys.ShiftLatincapitallettery,
            "\x1b[88;66u": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[27;66;88~": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[87;66u": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[27;66;87~": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[86;66u": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[27;66;86~": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[85;66u": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[27;66;85~": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[84;66u": MoreKeys.ShiftLatincapitallettert,
            "\x1b[27;66;84~": MoreKeys.ShiftLatincapitallettert,
            "\x1b[83;66u": MoreKeys.ShiftLatincapitalletters,
            "\x1b[27;66;83~": MoreKeys.ShiftLatincapitalletters,
            "\x1b[82;66u": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[27;66;82~": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[81;66u": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[27;66;81~": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[80;66u": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[27;66;80~": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[79;66u": MoreKeys.ShiftLatincapitallettero,
            "\x1b[27;66;79~": MoreKeys.ShiftLatincapitallettero,
            "\x1b[78;66u": MoreKeys.ShiftLatincapitallettern,
            "\x1b[27;66;78~": MoreKeys.ShiftLatincapitallettern,
            "\x1b[77;66u": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[27;66;77~": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[76;66u": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[27;66;76~": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[75;66u": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[27;66;75~": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[74;66u": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[27;66;74~": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[73;66u": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[27;66;73~": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[72;66u": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[27;66;72~": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[71;66u": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[27;66;71~": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[70;66u": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[27;66;70~": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[69;66u": MoreKeys.ShiftLatincapitallettere,
            "\x1b[27;66;69~": MoreKeys.ShiftLatincapitallettere,
            "\x1b[68;66u": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[27;66;68~": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[67;66u": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[27;66;67~": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[66;66u": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[27;66;66~": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[65;66u": MoreKeys.ShiftLatincapitallettera,
            "\x1b[27;66;65~": MoreKeys.ShiftLatincapitallettera,
            "\x1b[64;66u": MoreKeys.ShiftCommercialat,
            "\x1b[27;66;64~": MoreKeys.ShiftCommercialat,
            "\x1b[63;66u": MoreKeys.ShiftQuestionmark,
            "\x1b[27;66;63~": MoreKeys.ShiftQuestionmark,
            "\x1b[62;66u": MoreKeys.ShiftGreaterthansign,
            "\x1b[27;66;62~": MoreKeys.ShiftGreaterthansign,
            "\x1b[61;66u": MoreKeys.ShiftEqualssign,
            "\x1b[27;66;61~": MoreKeys.ShiftEqualssign,
            "\x1b[60;66u": MoreKeys.ShiftLessthansign,
            "\x1b[27;66;60~": MoreKeys.ShiftLessthansign,
            "\x1b[59;66u": MoreKeys.ShiftSemicolon,
            "\x1b[27;66;59~": MoreKeys.ShiftSemicolon,
            "\x1b[58;66u": MoreKeys.ShiftColon,
            "\x1b[27;66;58~": MoreKeys.ShiftColon,
            "\x1b[57;66u": MoreKeys.ShiftDigitnine,
            "\x1b[27;66;57~": MoreKeys.ShiftDigitnine,
            "\x1b[56;66u": MoreKeys.ShiftDigiteight,
            "\x1b[27;66;56~": MoreKeys.ShiftDigiteight,
            "\x1b[55;66u": MoreKeys.ShiftDigitseven,
            "\x1b[27;66;55~": MoreKeys.ShiftDigitseven,
            "\x1b[54;66u": MoreKeys.ShiftDigitsix,
            "\x1b[27;66;54~": MoreKeys.ShiftDigitsix,
            "\x1b[53;66u": MoreKeys.ShiftDigitfive,
            "\x1b[27;66;53~": MoreKeys.ShiftDigitfive,
            "\x1b[52;66u": MoreKeys.ShiftDigitfour,
            "\x1b[27;66;52~": MoreKeys.ShiftDigitfour,
            "\x1b[51;66u": MoreKeys.ShiftDigitthree,
            "\x1b[27;66;51~": MoreKeys.ShiftDigitthree,
            "\x1b[50;66u": MoreKeys.ShiftDigittwo,
            "\x1b[27;66;50~": MoreKeys.ShiftDigittwo,
            "\x1b[49;66u": MoreKeys.ShiftDigitone,
            "\x1b[27;66;49~": MoreKeys.ShiftDigitone,
            "\x1b[48;66u": MoreKeys.ShiftDigitzero,
            "\x1b[27;66;48~": MoreKeys.ShiftDigitzero,
            "\x1b[47;66u": MoreKeys.ShiftSolidus,
            "\x1b[27;66;47~": MoreKeys.ShiftSolidus,
            "\x1b[46;66u": MoreKeys.ShiftFullstop,
            "\x1b[27;66;46~": MoreKeys.ShiftFullstop,
            "\x1b[45;66u": MoreKeys.ShiftHyphenminus,
            "\x1b[27;66;45~": MoreKeys.ShiftHyphenminus,
            "\x1b[44;66u": MoreKeys.ShiftComma,
            "\x1b[27;66;44~": MoreKeys.ShiftComma,
            "\x1b[43;66u": MoreKeys.ShiftPlussign,
            "\x1b[27;66;43~": MoreKeys.ShiftPlussign,
            "\x1b[42;66u": MoreKeys.ShiftAsterisk,
            "\x1b[27;66;42~": MoreKeys.ShiftAsterisk,
            "\x1b[41;66u": MoreKeys.ShiftRightparenthesis,
            "\x1b[27;66;41~": MoreKeys.ShiftRightparenthesis,
            "\x1b[40;66u": MoreKeys.ShiftLeftparenthesis,
            "\x1b[27;66;40~": MoreKeys.ShiftLeftparenthesis,
            "\x1b[39;66u": MoreKeys.ShiftApostrophe,
            "\x1b[27;66;39~": MoreKeys.ShiftApostrophe,
            "\x1b[38;66u": MoreKeys.ShiftAmpersand,
            "\x1b[27;66;38~": MoreKeys.ShiftAmpersand,
            "\x1b[37;66u": MoreKeys.ShiftPercentsign,
            "\x1b[27;66;37~": MoreKeys.ShiftPercentsign,
            "\x1b[36;66u": MoreKeys.ShiftDollarsign,
            "\x1b[27;66;36~": MoreKeys.ShiftDollarsign,
            "\x1b[35;66u": MoreKeys.ShiftNumbersign,
            "\x1b[27;66;35~": MoreKeys.ShiftNumbersign,
            "\x1b[34;66u": MoreKeys.ShiftQuotationmark,
            "\x1b[27;66;34~": MoreKeys.ShiftQuotationmark,
            "\x1b[33;66u": MoreKeys.ShiftExclamationmark,
            "\x1b[27;66;33~": MoreKeys.ShiftExclamationmark,
            "\x1b[127;66u": MoreKeys.ShiftBackspace,
            "\x1b[27;66;127~": MoreKeys.ShiftBackspace,
            "\x1b[2;66~": MoreKeys.ShiftInsert,
            "\x1b[3;66~": MoreKeys.ShiftDelete,
            "\x1b[5;66~": MoreKeys.ShiftPageup,
            "\x1b[6;66~": MoreKeys.ShiftPagedown,
            "\x1b[1;66A": MoreKeys.ShiftUp,
            "\x1b[1;66B": MoreKeys.ShiftDown,
            "\x1b[1;66C": MoreKeys.ShiftRight,
            "\x1b[1;66D": MoreKeys.ShiftLeft,
            "\x1b[1;66H": MoreKeys.ShiftHome,
            "\x1b[1;66F": MoreKeys.ShiftEnd,
            "\x1b[57361;66u": MoreKeys.ShiftPrintscreen,
            "\x1b[57362;66u": MoreKeys.ShiftPause,
            "\x1b[57363;66u": MoreKeys.ShiftMenu,
            "\x1b[1;66P": MoreKeys.ShiftF1,
            "\x1b[1;66Q": MoreKeys.ShiftF2,
            "\x1b[13;66~": MoreKeys.ShiftF3,
            "\x1b[1;66S": MoreKeys.ShiftF4,
            "\x1b[15;66~": MoreKeys.ShiftF5,
            "\x1b[17;66~": MoreKeys.ShiftF6,
            "\x1b[18;66~": MoreKeys.ShiftF7,
            "\x1b[19;66~": MoreKeys.ShiftF8,
            "\x1b[20;66~": MoreKeys.ShiftF9,
            "\x1b[21;66~": MoreKeys.ShiftF10,
            "\x1b[23;66~": MoreKeys.ShiftF11,
            "\x1b[24;66~": MoreKeys.ShiftF12,
            "\x1b[9;130u": MoreKeys.ShiftTab,
            "\x1b[27;130;9~": MoreKeys.ShiftTab,
            "\x1b[13;130u": MoreKeys.ShiftEnter,
            "\x1b[27;130;13~": MoreKeys.ShiftEnter,
            "\x1b[27;130u": MoreKeys.ShiftEscape,
            "\x1b[27;130;27~": MoreKeys.ShiftEscape,
            "\x1b[32;130u": MoreKeys.ShiftSpace,
            "\x1b[27;130;32~": MoreKeys.ShiftSpace,
            "\x1b[126;130u": MoreKeys.ShiftTilde,
            "\x1b[27;130;126~": MoreKeys.ShiftTilde,
            "\x1b[125;130u": MoreKeys.ShiftRightcurlybracket,
            "\x1b[27;130;125~": MoreKeys.ShiftRightcurlybracket,
            "\x1b[124;130u": MoreKeys.ShiftVerticalline,
            "\x1b[27;130;124~": MoreKeys.ShiftVerticalline,
            "\x1b[123;130u": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[27;130;123~": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[122;130u": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[27;130;122~": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[121;130u": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[27;130;121~": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[120;130u": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[27;130;120~": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[119;130u": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[27;130;119~": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[118;130u": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[27;130;118~": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[117;130u": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[27;130;117~": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[116;130u": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[27;130;116~": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[115;130u": MoreKeys.ShiftLatinsmallletters,
            "\x1b[27;130;115~": MoreKeys.ShiftLatinsmallletters,
            "\x1b[114;130u": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[27;130;114~": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[113;130u": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[27;130;113~": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[112;130u": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[27;130;112~": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[111;130u": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[27;130;111~": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[110;130u": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[27;130;110~": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[109;130u": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[27;130;109~": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[108;130u": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[27;130;108~": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[107;130u": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[27;130;107~": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[106;130u": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[27;130;106~": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[105;130u": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[27;130;105~": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[104;130u": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[27;130;104~": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[103;130u": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[27;130;103~": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[102;130u": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[27;130;102~": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[101;130u": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[27;130;101~": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[100;130u": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[27;130;100~": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[99;130u": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[27;130;99~": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[98;130u": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[27;130;98~": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[97;130u": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[27;130;97~": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[96;130u": MoreKeys.ShiftGraveaccent,
            "\x1b[27;130;96~": MoreKeys.ShiftGraveaccent,
            "\x1b[95;130u": MoreKeys.ShiftLowline,
            "\x1b[27;130;95~": MoreKeys.ShiftLowline,
            "\x1b[94;130u": MoreKeys.ShiftCircumflexaccent,
            "\x1b[27;130;94~": MoreKeys.ShiftCircumflexaccent,
            "\x1b[93;130u": MoreKeys.ShiftRightsquarebracket,
            "\x1b[27;130;93~": MoreKeys.ShiftRightsquarebracket,
            "\x1b[92;130u": MoreKeys.ShiftReversesolidus,
            "\x1b[27;130;92~": MoreKeys.ShiftReversesolidus,
            "\x1b[91;130u": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[27;130;91~": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[90;130u": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[27;130;90~": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[89;130u": MoreKeys.ShiftLatincapitallettery,
            "\x1b[27;130;89~": MoreKeys.ShiftLatincapitallettery,
            "\x1b[88;130u": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[27;130;88~": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[87;130u": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[27;130;87~": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[86;130u": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[27;130;86~": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[85;130u": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[27;130;85~": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[84;130u": MoreKeys.ShiftLatincapitallettert,
            "\x1b[27;130;84~": MoreKeys.ShiftLatincapitallettert,
            "\x1b[83;130u": MoreKeys.ShiftLatincapitalletters,
            "\x1b[27;130;83~": MoreKeys.ShiftLatincapitalletters,
            "\x1b[82;130u": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[27;130;82~": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[81;130u": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[27;130;81~": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[80;130u": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[27;130;80~": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[79;130u": MoreKeys.ShiftLatincapitallettero,
            "\x1b[27;130;79~": MoreKeys.ShiftLatincapitallettero,
            "\x1b[78;130u": MoreKeys.ShiftLatincapitallettern,
            "\x1b[27;130;78~": MoreKeys.ShiftLatincapitallettern,
            "\x1b[77;130u": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[27;130;77~": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[76;130u": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[27;130;76~": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[75;130u": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[27;130;75~": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[74;130u": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[27;130;74~": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[73;130u": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[27;130;73~": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[72;130u": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[27;130;72~": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[71;130u": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[27;130;71~": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[70;130u": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[27;130;70~": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[69;130u": MoreKeys.ShiftLatincapitallettere,
            "\x1b[27;130;69~": MoreKeys.ShiftLatincapitallettere,
            "\x1b[68;130u": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[27;130;68~": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[67;130u": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[27;130;67~": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[66;130u": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[27;130;66~": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[65;130u": MoreKeys.ShiftLatincapitallettera,
            "\x1b[27;130;65~": MoreKeys.ShiftLatincapitallettera,
            "\x1b[64;130u": MoreKeys.ShiftCommercialat,
            "\x1b[27;130;64~": MoreKeys.ShiftCommercialat,
            "\x1b[63;130u": MoreKeys.ShiftQuestionmark,
            "\x1b[27;130;63~": MoreKeys.ShiftQuestionmark,
            "\x1b[62;130u": MoreKeys.ShiftGreaterthansign,
            "\x1b[27;130;62~": MoreKeys.ShiftGreaterthansign,
            "\x1b[61;130u": MoreKeys.ShiftEqualssign,
            "\x1b[27;130;61~": MoreKeys.ShiftEqualssign,
            "\x1b[60;130u": MoreKeys.ShiftLessthansign,
            "\x1b[27;130;60~": MoreKeys.ShiftLessthansign,
            "\x1b[59;130u": MoreKeys.ShiftSemicolon,
            "\x1b[27;130;59~": MoreKeys.ShiftSemicolon,
            "\x1b[58;130u": MoreKeys.ShiftColon,
            "\x1b[27;130;58~": MoreKeys.ShiftColon,
            "\x1b[57;130u": MoreKeys.ShiftDigitnine,
            "\x1b[27;130;57~": MoreKeys.ShiftDigitnine,
            "\x1b[56;130u": MoreKeys.ShiftDigiteight,
            "\x1b[27;130;56~": MoreKeys.ShiftDigiteight,
            "\x1b[55;130u": MoreKeys.ShiftDigitseven,
            "\x1b[27;130;55~": MoreKeys.ShiftDigitseven,
            "\x1b[54;130u": MoreKeys.ShiftDigitsix,
            "\x1b[27;130;54~": MoreKeys.ShiftDigitsix,
            "\x1b[53;130u": MoreKeys.ShiftDigitfive,
            "\x1b[27;130;53~": MoreKeys.ShiftDigitfive,
            "\x1b[52;130u": MoreKeys.ShiftDigitfour,
            "\x1b[27;130;52~": MoreKeys.ShiftDigitfour,
            "\x1b[51;130u": MoreKeys.ShiftDigitthree,
            "\x1b[27;130;51~": MoreKeys.ShiftDigitthree,
            "\x1b[50;130u": MoreKeys.ShiftDigittwo,
            "\x1b[27;130;50~": MoreKeys.ShiftDigittwo,
            "\x1b[49;130u": MoreKeys.ShiftDigitone,
            "\x1b[27;130;49~": MoreKeys.ShiftDigitone,
            "\x1b[48;130u": MoreKeys.ShiftDigitzero,
            "\x1b[27;130;48~": MoreKeys.ShiftDigitzero,
            "\x1b[47;130u": MoreKeys.ShiftSolidus,
            "\x1b[27;130;47~": MoreKeys.ShiftSolidus,
            "\x1b[46;130u": MoreKeys.ShiftFullstop,
            "\x1b[27;130;46~": MoreKeys.ShiftFullstop,
            "\x1b[45;130u": MoreKeys.ShiftHyphenminus,
            "\x1b[27;130;45~": MoreKeys.ShiftHyphenminus,
            "\x1b[44;130u": MoreKeys.ShiftComma,
            "\x1b[27;130;44~": MoreKeys.ShiftComma,
            "\x1b[43;130u": MoreKeys.ShiftPlussign,
            "\x1b[27;130;43~": MoreKeys.ShiftPlussign,
            "\x1b[42;130u": MoreKeys.ShiftAsterisk,
            "\x1b[27;130;42~": MoreKeys.ShiftAsterisk,
            "\x1b[41;130u": MoreKeys.ShiftRightparenthesis,
            "\x1b[27;130;41~": MoreKeys.ShiftRightparenthesis,
            "\x1b[40;130u": MoreKeys.ShiftLeftparenthesis,
            "\x1b[27;130;40~": MoreKeys.ShiftLeftparenthesis,
            "\x1b[39;130u": MoreKeys.ShiftApostrophe,
            "\x1b[27;130;39~": MoreKeys.ShiftApostrophe,
            "\x1b[38;130u": MoreKeys.ShiftAmpersand,
            "\x1b[27;130;38~": MoreKeys.ShiftAmpersand,
            "\x1b[37;130u": MoreKeys.ShiftPercentsign,
            "\x1b[27;130;37~": MoreKeys.ShiftPercentsign,
            "\x1b[36;130u": MoreKeys.ShiftDollarsign,
            "\x1b[27;130;36~": MoreKeys.ShiftDollarsign,
            "\x1b[35;130u": MoreKeys.ShiftNumbersign,
            "\x1b[27;130;35~": MoreKeys.ShiftNumbersign,
            "\x1b[34;130u": MoreKeys.ShiftQuotationmark,
            "\x1b[27;130;34~": MoreKeys.ShiftQuotationmark,
            "\x1b[33;130u": MoreKeys.ShiftExclamationmark,
            "\x1b[27;130;33~": MoreKeys.ShiftExclamationmark,
            "\x1b[127;130u": MoreKeys.ShiftBackspace,
            "\x1b[27;130;127~": MoreKeys.ShiftBackspace,
            "\x1b[2;130~": MoreKeys.ShiftInsert,
            "\x1b[3;130~": MoreKeys.ShiftDelete,
            "\x1b[5;130~": MoreKeys.ShiftPageup,
            "\x1b[6;130~": MoreKeys.ShiftPagedown,
            "\x1b[1;130A": MoreKeys.ShiftUp,
            "\x1b[1;130B": MoreKeys.ShiftDown,
            "\x1b[1;130C": MoreKeys.ShiftRight,
            "\x1b[1;130D": MoreKeys.ShiftLeft,
            "\x1b[1;130H": MoreKeys.ShiftHome,
            "\x1b[1;130F": MoreKeys.ShiftEnd,
            "\x1b[57361;130u": MoreKeys.ShiftPrintscreen,
            "\x1b[57362;130u": MoreKeys.ShiftPause,
            "\x1b[57363;130u": MoreKeys.ShiftMenu,
            "\x1b[1;130P": MoreKeys.ShiftF1,
            "\x1b[1;130Q": MoreKeys.ShiftF2,
            "\x1b[13;130~": MoreKeys.ShiftF3,
            "\x1b[1;130S": MoreKeys.ShiftF4,
            "\x1b[15;130~": MoreKeys.ShiftF5,
            "\x1b[17;130~": MoreKeys.ShiftF6,
            "\x1b[18;130~": MoreKeys.ShiftF7,
            "\x1b[19;130~": MoreKeys.ShiftF8,
            "\x1b[20;130~": MoreKeys.ShiftF9,
            "\x1b[21;130~": MoreKeys.ShiftF10,
            "\x1b[23;130~": MoreKeys.ShiftF11,
            "\x1b[24;130~": MoreKeys.ShiftF12,
            "\x1b[9;7u": MoreKeys.ControlAltTab,
            "\x1b[27;7;9~": MoreKeys.ControlAltTab,
            "\x1b[13;7u": MoreKeys.ControlAltEnter,
            "\x1b[27;7;13~": MoreKeys.ControlAltEnter,
            "\x1b[27;7u": MoreKeys.ControlAltEscape,
            "\x1b[27;7;27~": MoreKeys.ControlAltEscape,
            "\x1b[32;7u": MoreKeys.ControlAltSpace,
            "\x1b[27;7;32~": MoreKeys.ControlAltSpace,
            "\x1b[126;7u": MoreKeys.ControlAltTilde,
            "\x1b[27;7;126~": MoreKeys.ControlAltTilde,
            "\x1b[125;7u": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[27;7;125~": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[124;7u": MoreKeys.ControlAltVerticalline,
            "\x1b[27;7;124~": MoreKeys.ControlAltVerticalline,
            "\x1b[123;7u": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[27;7;123~": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[122;7u": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[27;7;122~": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[121;7u": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[27;7;121~": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[120;7u": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[27;7;120~": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[119;7u": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[27;7;119~": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[118;7u": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[27;7;118~": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[117;7u": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[27;7;117~": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[116;7u": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[27;7;116~": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[115;7u": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[27;7;115~": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[114;7u": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[27;7;114~": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[113;7u": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[27;7;113~": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[112;7u": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[27;7;112~": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[111;7u": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[27;7;111~": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[110;7u": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[27;7;110~": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[109;7u": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[27;7;109~": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[108;7u": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[27;7;108~": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[107;7u": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[27;7;107~": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[106;7u": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[27;7;106~": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[105;7u": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[27;7;105~": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[104;7u": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[27;7;104~": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[103;7u": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[27;7;103~": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[102;7u": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[27;7;102~": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[101;7u": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[27;7;101~": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[100;7u": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[27;7;100~": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[99;7u": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[27;7;99~": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[98;7u": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[27;7;98~": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[97;7u": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[27;7;97~": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[96;7u": MoreKeys.ControlAltGraveaccent,
            "\x1b[27;7;96~": MoreKeys.ControlAltGraveaccent,
            "\x1b[95;7u": MoreKeys.ControlAltLowline,
            "\x1b[27;7;95~": MoreKeys.ControlAltLowline,
            "\x1b[94;7u": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[27;7;94~": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[93;7u": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[27;7;93~": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[92;7u": MoreKeys.ControlAltReversesolidus,
            "\x1b[27;7;92~": MoreKeys.ControlAltReversesolidus,
            "\x1b[91;7u": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[27;7;91~": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[90;7u": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[27;7;90~": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[89;7u": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[27;7;89~": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[88;7u": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[27;7;88~": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[87;7u": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[27;7;87~": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[86;7u": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[27;7;86~": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[85;7u": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[27;7;85~": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[84;7u": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[27;7;84~": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[83;7u": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[27;7;83~": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[82;7u": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[27;7;82~": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[81;7u": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[27;7;81~": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[80;7u": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[27;7;80~": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[79;7u": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[27;7;79~": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[78;7u": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[27;7;78~": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[77;7u": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[27;7;77~": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[76;7u": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[27;7;76~": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[75;7u": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[27;7;75~": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[74;7u": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[27;7;74~": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[73;7u": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[27;7;73~": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[72;7u": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[27;7;72~": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[71;7u": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[27;7;71~": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[70;7u": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[27;7;70~": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[69;7u": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[27;7;69~": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[68;7u": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[27;7;68~": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[67;7u": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[27;7;67~": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[66;7u": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[27;7;66~": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[65;7u": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[27;7;65~": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[64;7u": MoreKeys.ControlAltCommercialat,
            "\x1b[27;7;64~": MoreKeys.ControlAltCommercialat,
            "\x1b[63;7u": MoreKeys.ControlAltQuestionmark,
            "\x1b[27;7;63~": MoreKeys.ControlAltQuestionmark,
            "\x1b[62;7u": MoreKeys.ControlAltGreaterthansign,
            "\x1b[27;7;62~": MoreKeys.ControlAltGreaterthansign,
            "\x1b[61;7u": MoreKeys.ControlAltEqualssign,
            "\x1b[27;7;61~": MoreKeys.ControlAltEqualssign,
            "\x1b[60;7u": MoreKeys.ControlAltLessthansign,
            "\x1b[27;7;60~": MoreKeys.ControlAltLessthansign,
            "\x1b[59;7u": MoreKeys.ControlAltSemicolon,
            "\x1b[27;7;59~": MoreKeys.ControlAltSemicolon,
            "\x1b[58;7u": MoreKeys.ControlAltColon,
            "\x1b[27;7;58~": MoreKeys.ControlAltColon,
            "\x1b[57;7u": MoreKeys.ControlAltDigitnine,
            "\x1b[27;7;57~": MoreKeys.ControlAltDigitnine,
            "\x1b[56;7u": MoreKeys.ControlAltDigiteight,
            "\x1b[27;7;56~": MoreKeys.ControlAltDigiteight,
            "\x1b[55;7u": MoreKeys.ControlAltDigitseven,
            "\x1b[27;7;55~": MoreKeys.ControlAltDigitseven,
            "\x1b[54;7u": MoreKeys.ControlAltDigitsix,
            "\x1b[27;7;54~": MoreKeys.ControlAltDigitsix,
            "\x1b[53;7u": MoreKeys.ControlAltDigitfive,
            "\x1b[27;7;53~": MoreKeys.ControlAltDigitfive,
            "\x1b[52;7u": MoreKeys.ControlAltDigitfour,
            "\x1b[27;7;52~": MoreKeys.ControlAltDigitfour,
            "\x1b[51;7u": MoreKeys.ControlAltDigitthree,
            "\x1b[27;7;51~": MoreKeys.ControlAltDigitthree,
            "\x1b[50;7u": MoreKeys.ControlAltDigittwo,
            "\x1b[27;7;50~": MoreKeys.ControlAltDigittwo,
            "\x1b[49;7u": MoreKeys.ControlAltDigitone,
            "\x1b[27;7;49~": MoreKeys.ControlAltDigitone,
            "\x1b[48;7u": MoreKeys.ControlAltDigitzero,
            "\x1b[27;7;48~": MoreKeys.ControlAltDigitzero,
            "\x1b[47;7u": MoreKeys.ControlAltSolidus,
            "\x1b[27;7;47~": MoreKeys.ControlAltSolidus,
            "\x1b[46;7u": MoreKeys.ControlAltFullstop,
            "\x1b[27;7;46~": MoreKeys.ControlAltFullstop,
            "\x1b[45;7u": MoreKeys.ControlAltHyphenminus,
            "\x1b[27;7;45~": MoreKeys.ControlAltHyphenminus,
            "\x1b[44;7u": MoreKeys.ControlAltComma,
            "\x1b[27;7;44~": MoreKeys.ControlAltComma,
            "\x1b[43;7u": MoreKeys.ControlAltPlussign,
            "\x1b[27;7;43~": MoreKeys.ControlAltPlussign,
            "\x1b[42;7u": MoreKeys.ControlAltAsterisk,
            "\x1b[27;7;42~": MoreKeys.ControlAltAsterisk,
            "\x1b[41;7u": MoreKeys.ControlAltRightparenthesis,
            "\x1b[27;7;41~": MoreKeys.ControlAltRightparenthesis,
            "\x1b[40;7u": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[27;7;40~": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[39;7u": MoreKeys.ControlAltApostrophe,
            "\x1b[27;7;39~": MoreKeys.ControlAltApostrophe,
            "\x1b[38;7u": MoreKeys.ControlAltAmpersand,
            "\x1b[27;7;38~": MoreKeys.ControlAltAmpersand,
            "\x1b[37;7u": MoreKeys.ControlAltPercentsign,
            "\x1b[27;7;37~": MoreKeys.ControlAltPercentsign,
            "\x1b[36;7u": MoreKeys.ControlAltDollarsign,
            "\x1b[27;7;36~": MoreKeys.ControlAltDollarsign,
            "\x1b[35;7u": MoreKeys.ControlAltNumbersign,
            "\x1b[27;7;35~": MoreKeys.ControlAltNumbersign,
            "\x1b[34;7u": MoreKeys.ControlAltQuotationmark,
            "\x1b[27;7;34~": MoreKeys.ControlAltQuotationmark,
            "\x1b[33;7u": MoreKeys.ControlAltExclamationmark,
            "\x1b[27;7;33~": MoreKeys.ControlAltExclamationmark,
            "\x1b[127;7u": MoreKeys.ControlAltBackspace,
            "\x1b[27;7;127~": MoreKeys.ControlAltBackspace,
            "\x1b[2;7~": MoreKeys.ControlAltInsert,
            "\x1b[3;7~": MoreKeys.ControlAltDelete,
            "\x1b[5;7~": MoreKeys.ControlAltPageup,
            "\x1b[6;7~": MoreKeys.ControlAltPagedown,
            "\x1b[1;7A": MoreKeys.ControlAltUp,
            "\x1b[1;7B": MoreKeys.ControlAltDown,
            "\x1b[1;7C": MoreKeys.ControlAltRight,
            "\x1b[1;7D": MoreKeys.ControlAltLeft,
            "\x1b[1;7H": MoreKeys.ControlAltHome,
            "\x1b[1;7F": MoreKeys.ControlAltEnd,
            "\x1b[57361;7u": MoreKeys.ControlAltPrintscreen,
            "\x1b[57362;7u": MoreKeys.ControlAltPause,
            "\x1b[57363;7u": MoreKeys.ControlAltMenu,
            "\x1b[1;7P": MoreKeys.ControlAltF1,
            "\x1b[1;7Q": MoreKeys.ControlAltF2,
            "\x1b[13;7~": MoreKeys.ControlAltF3,
            "\x1b[1;7S": MoreKeys.ControlAltF4,
            "\x1b[15;7~": MoreKeys.ControlAltF5,
            "\x1b[17;7~": MoreKeys.ControlAltF6,
            "\x1b[18;7~": MoreKeys.ControlAltF7,
            "\x1b[19;7~": MoreKeys.ControlAltF8,
            "\x1b[20;7~": MoreKeys.ControlAltF9,
            "\x1b[21;7~": MoreKeys.ControlAltF10,
            "\x1b[23;7~": MoreKeys.ControlAltF11,
            "\x1b[24;7~": MoreKeys.ControlAltF12,
            "\x1b[9;67u": MoreKeys.AltTab,
            "\x1b[27;67;9~": MoreKeys.AltTab,
            "\x1b[13;67u": MoreKeys.AltEnter,
            "\x1b[27;67;13~": MoreKeys.AltEnter,
            "\x1b[27;67u": MoreKeys.AltEscape,
            "\x1b[27;67;27~": MoreKeys.AltEscape,
            "\x1b[32;67u": MoreKeys.AltSpace,
            "\x1b[27;67;32~": MoreKeys.AltSpace,
            "\x1b[126;67u": MoreKeys.AltTilde,
            "\x1b[27;67;126~": MoreKeys.AltTilde,
            "\x1b[125;67u": MoreKeys.AltRightcurlybracket,
            "\x1b[27;67;125~": MoreKeys.AltRightcurlybracket,
            "\x1b[124;67u": MoreKeys.AltVerticalline,
            "\x1b[27;67;124~": MoreKeys.AltVerticalline,
            "\x1b[123;67u": MoreKeys.AltLeftcurlybracket,
            "\x1b[27;67;123~": MoreKeys.AltLeftcurlybracket,
            "\x1b[122;67u": MoreKeys.AltLatinsmallletterz,
            "\x1b[27;67;122~": MoreKeys.AltLatinsmallletterz,
            "\x1b[121;67u": MoreKeys.AltLatinsmalllettery,
            "\x1b[27;67;121~": MoreKeys.AltLatinsmalllettery,
            "\x1b[120;67u": MoreKeys.AltLatinsmallletterx,
            "\x1b[27;67;120~": MoreKeys.AltLatinsmallletterx,
            "\x1b[119;67u": MoreKeys.AltLatinsmallletterw,
            "\x1b[27;67;119~": MoreKeys.AltLatinsmallletterw,
            "\x1b[118;67u": MoreKeys.AltLatinsmallletterv,
            "\x1b[27;67;118~": MoreKeys.AltLatinsmallletterv,
            "\x1b[117;67u": MoreKeys.AltLatinsmallletteru,
            "\x1b[27;67;117~": MoreKeys.AltLatinsmallletteru,
            "\x1b[116;67u": MoreKeys.AltLatinsmalllettert,
            "\x1b[27;67;116~": MoreKeys.AltLatinsmalllettert,
            "\x1b[115;67u": MoreKeys.AltLatinsmallletters,
            "\x1b[27;67;115~": MoreKeys.AltLatinsmallletters,
            "\x1b[114;67u": MoreKeys.AltLatinsmallletterr,
            "\x1b[27;67;114~": MoreKeys.AltLatinsmallletterr,
            "\x1b[113;67u": MoreKeys.AltLatinsmallletterq,
            "\x1b[27;67;113~": MoreKeys.AltLatinsmallletterq,
            "\x1b[112;67u": MoreKeys.AltLatinsmallletterp,
            "\x1b[27;67;112~": MoreKeys.AltLatinsmallletterp,
            "\x1b[111;67u": MoreKeys.AltLatinsmalllettero,
            "\x1b[27;67;111~": MoreKeys.AltLatinsmalllettero,
            "\x1b[110;67u": MoreKeys.AltLatinsmalllettern,
            "\x1b[27;67;110~": MoreKeys.AltLatinsmalllettern,
            "\x1b[109;67u": MoreKeys.AltLatinsmallletterm,
            "\x1b[27;67;109~": MoreKeys.AltLatinsmallletterm,
            "\x1b[108;67u": MoreKeys.AltLatinsmallletterl,
            "\x1b[27;67;108~": MoreKeys.AltLatinsmallletterl,
            "\x1b[107;67u": MoreKeys.AltLatinsmallletterk,
            "\x1b[27;67;107~": MoreKeys.AltLatinsmallletterk,
            "\x1b[106;67u": MoreKeys.AltLatinsmallletterj,
            "\x1b[27;67;106~": MoreKeys.AltLatinsmallletterj,
            "\x1b[105;67u": MoreKeys.AltLatinsmallletteri,
            "\x1b[27;67;105~": MoreKeys.AltLatinsmallletteri,
            "\x1b[104;67u": MoreKeys.AltLatinsmallletterh,
            "\x1b[27;67;104~": MoreKeys.AltLatinsmallletterh,
            "\x1b[103;67u": MoreKeys.AltLatinsmallletterg,
            "\x1b[27;67;103~": MoreKeys.AltLatinsmallletterg,
            "\x1b[102;67u": MoreKeys.AltLatinsmallletterf,
            "\x1b[27;67;102~": MoreKeys.AltLatinsmallletterf,
            "\x1b[101;67u": MoreKeys.AltLatinsmalllettere,
            "\x1b[27;67;101~": MoreKeys.AltLatinsmalllettere,
            "\x1b[100;67u": MoreKeys.AltLatinsmallletterd,
            "\x1b[27;67;100~": MoreKeys.AltLatinsmallletterd,
            "\x1b[99;67u": MoreKeys.AltLatinsmallletterc,
            "\x1b[27;67;99~": MoreKeys.AltLatinsmallletterc,
            "\x1b[98;67u": MoreKeys.AltLatinsmallletterb,
            "\x1b[27;67;98~": MoreKeys.AltLatinsmallletterb,
            "\x1b[97;67u": MoreKeys.AltLatinsmalllettera,
            "\x1b[27;67;97~": MoreKeys.AltLatinsmalllettera,
            "\x1b[96;67u": MoreKeys.AltGraveaccent,
            "\x1b[27;67;96~": MoreKeys.AltGraveaccent,
            "\x1b[95;67u": MoreKeys.AltLowline,
            "\x1b[27;67;95~": MoreKeys.AltLowline,
            "\x1b[94;67u": MoreKeys.AltCircumflexaccent,
            "\x1b[27;67;94~": MoreKeys.AltCircumflexaccent,
            "\x1b[93;67u": MoreKeys.AltRightsquarebracket,
            "\x1b[27;67;93~": MoreKeys.AltRightsquarebracket,
            "\x1b[92;67u": MoreKeys.AltReversesolidus,
            "\x1b[27;67;92~": MoreKeys.AltReversesolidus,
            "\x1b[91;67u": MoreKeys.AltLeftsquarebracket,
            "\x1b[27;67;91~": MoreKeys.AltLeftsquarebracket,
            "\x1b[90;67u": MoreKeys.AltLatincapitalletterz,
            "\x1b[27;67;90~": MoreKeys.AltLatincapitalletterz,
            "\x1b[89;67u": MoreKeys.AltLatincapitallettery,
            "\x1b[27;67;89~": MoreKeys.AltLatincapitallettery,
            "\x1b[88;67u": MoreKeys.AltLatincapitalletterx,
            "\x1b[27;67;88~": MoreKeys.AltLatincapitalletterx,
            "\x1b[87;67u": MoreKeys.AltLatincapitalletterw,
            "\x1b[27;67;87~": MoreKeys.AltLatincapitalletterw,
            "\x1b[86;67u": MoreKeys.AltLatincapitalletterv,
            "\x1b[27;67;86~": MoreKeys.AltLatincapitalletterv,
            "\x1b[85;67u": MoreKeys.AltLatincapitalletteru,
            "\x1b[27;67;85~": MoreKeys.AltLatincapitalletteru,
            "\x1b[84;67u": MoreKeys.AltLatincapitallettert,
            "\x1b[27;67;84~": MoreKeys.AltLatincapitallettert,
            "\x1b[83;67u": MoreKeys.AltLatincapitalletters,
            "\x1b[27;67;83~": MoreKeys.AltLatincapitalletters,
            "\x1b[82;67u": MoreKeys.AltLatincapitalletterr,
            "\x1b[27;67;82~": MoreKeys.AltLatincapitalletterr,
            "\x1b[81;67u": MoreKeys.AltLatincapitalletterq,
            "\x1b[27;67;81~": MoreKeys.AltLatincapitalletterq,
            "\x1b[80;67u": MoreKeys.AltLatincapitalletterp,
            "\x1b[27;67;80~": MoreKeys.AltLatincapitalletterp,
            "\x1b[79;67u": MoreKeys.AltLatincapitallettero,
            "\x1b[27;67;79~": MoreKeys.AltLatincapitallettero,
            "\x1b[78;67u": MoreKeys.AltLatincapitallettern,
            "\x1b[27;67;78~": MoreKeys.AltLatincapitallettern,
            "\x1b[77;67u": MoreKeys.AltLatincapitalletterm,
            "\x1b[27;67;77~": MoreKeys.AltLatincapitalletterm,
            "\x1b[76;67u": MoreKeys.AltLatincapitalletterl,
            "\x1b[27;67;76~": MoreKeys.AltLatincapitalletterl,
            "\x1b[75;67u": MoreKeys.AltLatincapitalletterk,
            "\x1b[27;67;75~": MoreKeys.AltLatincapitalletterk,
            "\x1b[74;67u": MoreKeys.AltLatincapitalletterj,
            "\x1b[27;67;74~": MoreKeys.AltLatincapitalletterj,
            "\x1b[73;67u": MoreKeys.AltLatincapitalletteri,
            "\x1b[27;67;73~": MoreKeys.AltLatincapitalletteri,
            "\x1b[72;67u": MoreKeys.AltLatincapitalletterh,
            "\x1b[27;67;72~": MoreKeys.AltLatincapitalletterh,
            "\x1b[71;67u": MoreKeys.AltLatincapitalletterg,
            "\x1b[27;67;71~": MoreKeys.AltLatincapitalletterg,
            "\x1b[70;67u": MoreKeys.AltLatincapitalletterf,
            "\x1b[27;67;70~": MoreKeys.AltLatincapitalletterf,
            "\x1b[69;67u": MoreKeys.AltLatincapitallettere,
            "\x1b[27;67;69~": MoreKeys.AltLatincapitallettere,
            "\x1b[68;67u": MoreKeys.AltLatincapitalletterd,
            "\x1b[27;67;68~": MoreKeys.AltLatincapitalletterd,
            "\x1b[67;67u": MoreKeys.AltLatincapitalletterc,
            "\x1b[27;67;67~": MoreKeys.AltLatincapitalletterc,
            "\x1b[66;67u": MoreKeys.AltLatincapitalletterb,
            "\x1b[27;67;66~": MoreKeys.AltLatincapitalletterb,
            "\x1b[65;67u": MoreKeys.AltLatincapitallettera,
            "\x1b[27;67;65~": MoreKeys.AltLatincapitallettera,
            "\x1b[64;67u": MoreKeys.AltCommercialat,
            "\x1b[27;67;64~": MoreKeys.AltCommercialat,
            "\x1b[63;67u": MoreKeys.AltQuestionmark,
            "\x1b[27;67;63~": MoreKeys.AltQuestionmark,
            "\x1b[62;67u": MoreKeys.AltGreaterthansign,
            "\x1b[27;67;62~": MoreKeys.AltGreaterthansign,
            "\x1b[61;67u": MoreKeys.AltEqualssign,
            "\x1b[27;67;61~": MoreKeys.AltEqualssign,
            "\x1b[60;67u": MoreKeys.AltLessthansign,
            "\x1b[27;67;60~": MoreKeys.AltLessthansign,
            "\x1b[59;67u": MoreKeys.AltSemicolon,
            "\x1b[27;67;59~": MoreKeys.AltSemicolon,
            "\x1b[58;67u": MoreKeys.AltColon,
            "\x1b[27;67;58~": MoreKeys.AltColon,
            "\x1b[57;67u": MoreKeys.AltDigitnine,
            "\x1b[27;67;57~": MoreKeys.AltDigitnine,
            "\x1b[56;67u": MoreKeys.AltDigiteight,
            "\x1b[27;67;56~": MoreKeys.AltDigiteight,
            "\x1b[55;67u": MoreKeys.AltDigitseven,
            "\x1b[27;67;55~": MoreKeys.AltDigitseven,
            "\x1b[54;67u": MoreKeys.AltDigitsix,
            "\x1b[27;67;54~": MoreKeys.AltDigitsix,
            "\x1b[53;67u": MoreKeys.AltDigitfive,
            "\x1b[27;67;53~": MoreKeys.AltDigitfive,
            "\x1b[52;67u": MoreKeys.AltDigitfour,
            "\x1b[27;67;52~": MoreKeys.AltDigitfour,
            "\x1b[51;67u": MoreKeys.AltDigitthree,
            "\x1b[27;67;51~": MoreKeys.AltDigitthree,
            "\x1b[50;67u": MoreKeys.AltDigittwo,
            "\x1b[27;67;50~": MoreKeys.AltDigittwo,
            "\x1b[49;67u": MoreKeys.AltDigitone,
            "\x1b[27;67;49~": MoreKeys.AltDigitone,
            "\x1b[48;67u": MoreKeys.AltDigitzero,
            "\x1b[27;67;48~": MoreKeys.AltDigitzero,
            "\x1b[47;67u": MoreKeys.AltSolidus,
            "\x1b[27;67;47~": MoreKeys.AltSolidus,
            "\x1b[46;67u": MoreKeys.AltFullstop,
            "\x1b[27;67;46~": MoreKeys.AltFullstop,
            "\x1b[45;67u": MoreKeys.AltHyphenminus,
            "\x1b[27;67;45~": MoreKeys.AltHyphenminus,
            "\x1b[44;67u": MoreKeys.AltComma,
            "\x1b[27;67;44~": MoreKeys.AltComma,
            "\x1b[43;67u": MoreKeys.AltPlussign,
            "\x1b[27;67;43~": MoreKeys.AltPlussign,
            "\x1b[42;67u": MoreKeys.AltAsterisk,
            "\x1b[27;67;42~": MoreKeys.AltAsterisk,
            "\x1b[41;67u": MoreKeys.AltRightparenthesis,
            "\x1b[27;67;41~": MoreKeys.AltRightparenthesis,
            "\x1b[40;67u": MoreKeys.AltLeftparenthesis,
            "\x1b[27;67;40~": MoreKeys.AltLeftparenthesis,
            "\x1b[39;67u": MoreKeys.AltApostrophe,
            "\x1b[27;67;39~": MoreKeys.AltApostrophe,
            "\x1b[38;67u": MoreKeys.AltAmpersand,
            "\x1b[27;67;38~": MoreKeys.AltAmpersand,
            "\x1b[37;67u": MoreKeys.AltPercentsign,
            "\x1b[27;67;37~": MoreKeys.AltPercentsign,
            "\x1b[36;67u": MoreKeys.AltDollarsign,
            "\x1b[27;67;36~": MoreKeys.AltDollarsign,
            "\x1b[35;67u": MoreKeys.AltNumbersign,
            "\x1b[27;67;35~": MoreKeys.AltNumbersign,
            "\x1b[34;67u": MoreKeys.AltQuotationmark,
            "\x1b[27;67;34~": MoreKeys.AltQuotationmark,
            "\x1b[33;67u": MoreKeys.AltExclamationmark,
            "\x1b[27;67;33~": MoreKeys.AltExclamationmark,
            "\x1b[127;67u": MoreKeys.AltBackspace,
            "\x1b[27;67;127~": MoreKeys.AltBackspace,
            "\x1b[2;67~": MoreKeys.AltInsert,
            "\x1b[3;67~": MoreKeys.AltDelete,
            "\x1b[5;67~": MoreKeys.AltPageup,
            "\x1b[6;67~": MoreKeys.AltPagedown,
            "\x1b[1;67A": MoreKeys.AltUp,
            "\x1b[1;67B": MoreKeys.AltDown,
            "\x1b[1;67C": MoreKeys.AltRight,
            "\x1b[1;67D": MoreKeys.AltLeft,
            "\x1b[1;67H": MoreKeys.AltHome,
            "\x1b[1;67F": MoreKeys.AltEnd,
            "\x1b[57361;67u": MoreKeys.AltPrintscreen,
            "\x1b[57362;67u": MoreKeys.AltPause,
            "\x1b[57363;67u": MoreKeys.AltMenu,
            "\x1b[1;67P": MoreKeys.AltF1,
            "\x1b[1;67Q": MoreKeys.AltF2,
            "\x1b[13;67~": MoreKeys.AltF3,
            "\x1b[1;67S": MoreKeys.AltF4,
            "\x1b[15;67~": MoreKeys.AltF5,
            "\x1b[17;67~": MoreKeys.AltF6,
            "\x1b[18;67~": MoreKeys.AltF7,
            "\x1b[19;67~": MoreKeys.AltF8,
            "\x1b[20;67~": MoreKeys.AltF9,
            "\x1b[21;67~": MoreKeys.AltF10,
            "\x1b[23;67~": MoreKeys.AltF11,
            "\x1b[24;67~": MoreKeys.AltF12,
            "\x1b[9;131u": MoreKeys.AltTab,
            "\x1b[27;131;9~": MoreKeys.AltTab,
            "\x1b[13;131u": MoreKeys.AltEnter,
            "\x1b[27;131;13~": MoreKeys.AltEnter,
            "\x1b[27;131u": MoreKeys.AltEscape,
            "\x1b[27;131;27~": MoreKeys.AltEscape,
            "\x1b[32;131u": MoreKeys.AltSpace,
            "\x1b[27;131;32~": MoreKeys.AltSpace,
            "\x1b[126;131u": MoreKeys.AltTilde,
            "\x1b[27;131;126~": MoreKeys.AltTilde,
            "\x1b[125;131u": MoreKeys.AltRightcurlybracket,
            "\x1b[27;131;125~": MoreKeys.AltRightcurlybracket,
            "\x1b[124;131u": MoreKeys.AltVerticalline,
            "\x1b[27;131;124~": MoreKeys.AltVerticalline,
            "\x1b[123;131u": MoreKeys.AltLeftcurlybracket,
            "\x1b[27;131;123~": MoreKeys.AltLeftcurlybracket,
            "\x1b[122;131u": MoreKeys.AltLatinsmallletterz,
            "\x1b[27;131;122~": MoreKeys.AltLatinsmallletterz,
            "\x1b[121;131u": MoreKeys.AltLatinsmalllettery,
            "\x1b[27;131;121~": MoreKeys.AltLatinsmalllettery,
            "\x1b[120;131u": MoreKeys.AltLatinsmallletterx,
            "\x1b[27;131;120~": MoreKeys.AltLatinsmallletterx,
            "\x1b[119;131u": MoreKeys.AltLatinsmallletterw,
            "\x1b[27;131;119~": MoreKeys.AltLatinsmallletterw,
            "\x1b[118;131u": MoreKeys.AltLatinsmallletterv,
            "\x1b[27;131;118~": MoreKeys.AltLatinsmallletterv,
            "\x1b[117;131u": MoreKeys.AltLatinsmallletteru,
            "\x1b[27;131;117~": MoreKeys.AltLatinsmallletteru,
            "\x1b[116;131u": MoreKeys.AltLatinsmalllettert,
            "\x1b[27;131;116~": MoreKeys.AltLatinsmalllettert,
            "\x1b[115;131u": MoreKeys.AltLatinsmallletters,
            "\x1b[27;131;115~": MoreKeys.AltLatinsmallletters,
            "\x1b[114;131u": MoreKeys.AltLatinsmallletterr,
            "\x1b[27;131;114~": MoreKeys.AltLatinsmallletterr,
            "\x1b[113;131u": MoreKeys.AltLatinsmallletterq,
            "\x1b[27;131;113~": MoreKeys.AltLatinsmallletterq,
            "\x1b[112;131u": MoreKeys.AltLatinsmallletterp,
            "\x1b[27;131;112~": MoreKeys.AltLatinsmallletterp,
            "\x1b[111;131u": MoreKeys.AltLatinsmalllettero,
            "\x1b[27;131;111~": MoreKeys.AltLatinsmalllettero,
            "\x1b[110;131u": MoreKeys.AltLatinsmalllettern,
            "\x1b[27;131;110~": MoreKeys.AltLatinsmalllettern,
            "\x1b[109;131u": MoreKeys.AltLatinsmallletterm,
            "\x1b[27;131;109~": MoreKeys.AltLatinsmallletterm,
            "\x1b[108;131u": MoreKeys.AltLatinsmallletterl,
            "\x1b[27;131;108~": MoreKeys.AltLatinsmallletterl,
            "\x1b[107;131u": MoreKeys.AltLatinsmallletterk,
            "\x1b[27;131;107~": MoreKeys.AltLatinsmallletterk,
            "\x1b[106;131u": MoreKeys.AltLatinsmallletterj,
            "\x1b[27;131;106~": MoreKeys.AltLatinsmallletterj,
            "\x1b[105;131u": MoreKeys.AltLatinsmallletteri,
            "\x1b[27;131;105~": MoreKeys.AltLatinsmallletteri,
            "\x1b[104;131u": MoreKeys.AltLatinsmallletterh,
            "\x1b[27;131;104~": MoreKeys.AltLatinsmallletterh,
            "\x1b[103;131u": MoreKeys.AltLatinsmallletterg,
            "\x1b[27;131;103~": MoreKeys.AltLatinsmallletterg,
            "\x1b[102;131u": MoreKeys.AltLatinsmallletterf,
            "\x1b[27;131;102~": MoreKeys.AltLatinsmallletterf,
            "\x1b[101;131u": MoreKeys.AltLatinsmalllettere,
            "\x1b[27;131;101~": MoreKeys.AltLatinsmalllettere,
            "\x1b[100;131u": MoreKeys.AltLatinsmallletterd,
            "\x1b[27;131;100~": MoreKeys.AltLatinsmallletterd,
            "\x1b[99;131u": MoreKeys.AltLatinsmallletterc,
            "\x1b[27;131;99~": MoreKeys.AltLatinsmallletterc,
            "\x1b[98;131u": MoreKeys.AltLatinsmallletterb,
            "\x1b[27;131;98~": MoreKeys.AltLatinsmallletterb,
            "\x1b[97;131u": MoreKeys.AltLatinsmalllettera,
            "\x1b[27;131;97~": MoreKeys.AltLatinsmalllettera,
            "\x1b[96;131u": MoreKeys.AltGraveaccent,
            "\x1b[27;131;96~": MoreKeys.AltGraveaccent,
            "\x1b[95;131u": MoreKeys.AltLowline,
            "\x1b[27;131;95~": MoreKeys.AltLowline,
            "\x1b[94;131u": MoreKeys.AltCircumflexaccent,
            "\x1b[27;131;94~": MoreKeys.AltCircumflexaccent,
            "\x1b[93;131u": MoreKeys.AltRightsquarebracket,
            "\x1b[27;131;93~": MoreKeys.AltRightsquarebracket,
            "\x1b[92;131u": MoreKeys.AltReversesolidus,
            "\x1b[27;131;92~": MoreKeys.AltReversesolidus,
            "\x1b[91;131u": MoreKeys.AltLeftsquarebracket,
            "\x1b[27;131;91~": MoreKeys.AltLeftsquarebracket,
            "\x1b[90;131u": MoreKeys.AltLatincapitalletterz,
            "\x1b[27;131;90~": MoreKeys.AltLatincapitalletterz,
            "\x1b[89;131u": MoreKeys.AltLatincapitallettery,
            "\x1b[27;131;89~": MoreKeys.AltLatincapitallettery,
            "\x1b[88;131u": MoreKeys.AltLatincapitalletterx,
            "\x1b[27;131;88~": MoreKeys.AltLatincapitalletterx,
            "\x1b[87;131u": MoreKeys.AltLatincapitalletterw,
            "\x1b[27;131;87~": MoreKeys.AltLatincapitalletterw,
            "\x1b[86;131u": MoreKeys.AltLatincapitalletterv,
            "\x1b[27;131;86~": MoreKeys.AltLatincapitalletterv,
            "\x1b[85;131u": MoreKeys.AltLatincapitalletteru,
            "\x1b[27;131;85~": MoreKeys.AltLatincapitalletteru,
            "\x1b[84;131u": MoreKeys.AltLatincapitallettert,
            "\x1b[27;131;84~": MoreKeys.AltLatincapitallettert,
            "\x1b[83;131u": MoreKeys.AltLatincapitalletters,
            "\x1b[27;131;83~": MoreKeys.AltLatincapitalletters,
            "\x1b[82;131u": MoreKeys.AltLatincapitalletterr,
            "\x1b[27;131;82~": MoreKeys.AltLatincapitalletterr,
            "\x1b[81;131u": MoreKeys.AltLatincapitalletterq,
            "\x1b[27;131;81~": MoreKeys.AltLatincapitalletterq,
            "\x1b[80;131u": MoreKeys.AltLatincapitalletterp,
            "\x1b[27;131;80~": MoreKeys.AltLatincapitalletterp,
            "\x1b[79;131u": MoreKeys.AltLatincapitallettero,
            "\x1b[27;131;79~": MoreKeys.AltLatincapitallettero,
            "\x1b[78;131u": MoreKeys.AltLatincapitallettern,
            "\x1b[27;131;78~": MoreKeys.AltLatincapitallettern,
            "\x1b[77;131u": MoreKeys.AltLatincapitalletterm,
            "\x1b[27;131;77~": MoreKeys.AltLatincapitalletterm,
            "\x1b[76;131u": MoreKeys.AltLatincapitalletterl,
            "\x1b[27;131;76~": MoreKeys.AltLatincapitalletterl,
            "\x1b[75;131u": MoreKeys.AltLatincapitalletterk,
            "\x1b[27;131;75~": MoreKeys.AltLatincapitalletterk,
            "\x1b[74;131u": MoreKeys.AltLatincapitalletterj,
            "\x1b[27;131;74~": MoreKeys.AltLatincapitalletterj,
            "\x1b[73;131u": MoreKeys.AltLatincapitalletteri,
            "\x1b[27;131;73~": MoreKeys.AltLatincapitalletteri,
            "\x1b[72;131u": MoreKeys.AltLatincapitalletterh,
            "\x1b[27;131;72~": MoreKeys.AltLatincapitalletterh,
            "\x1b[71;131u": MoreKeys.AltLatincapitalletterg,
            "\x1b[27;131;71~": MoreKeys.AltLatincapitalletterg,
            "\x1b[70;131u": MoreKeys.AltLatincapitalletterf,
            "\x1b[27;131;70~": MoreKeys.AltLatincapitalletterf,
            "\x1b[69;131u": MoreKeys.AltLatincapitallettere,
            "\x1b[27;131;69~": MoreKeys.AltLatincapitallettere,
            "\x1b[68;131u": MoreKeys.AltLatincapitalletterd,
            "\x1b[27;131;68~": MoreKeys.AltLatincapitalletterd,
            "\x1b[67;131u": MoreKeys.AltLatincapitalletterc,
            "\x1b[27;131;67~": MoreKeys.AltLatincapitalletterc,
            "\x1b[66;131u": MoreKeys.AltLatincapitalletterb,
            "\x1b[27;131;66~": MoreKeys.AltLatincapitalletterb,
            "\x1b[65;131u": MoreKeys.AltLatincapitallettera,
            "\x1b[27;131;65~": MoreKeys.AltLatincapitallettera,
            "\x1b[64;131u": MoreKeys.AltCommercialat,
            "\x1b[27;131;64~": MoreKeys.AltCommercialat,
            "\x1b[63;131u": MoreKeys.AltQuestionmark,
            "\x1b[27;131;63~": MoreKeys.AltQuestionmark,
            "\x1b[62;131u": MoreKeys.AltGreaterthansign,
            "\x1b[27;131;62~": MoreKeys.AltGreaterthansign,
            "\x1b[61;131u": MoreKeys.AltEqualssign,
            "\x1b[27;131;61~": MoreKeys.AltEqualssign,
            "\x1b[60;131u": MoreKeys.AltLessthansign,
            "\x1b[27;131;60~": MoreKeys.AltLessthansign,
            "\x1b[59;131u": MoreKeys.AltSemicolon,
            "\x1b[27;131;59~": MoreKeys.AltSemicolon,
            "\x1b[58;131u": MoreKeys.AltColon,
            "\x1b[27;131;58~": MoreKeys.AltColon,
            "\x1b[57;131u": MoreKeys.AltDigitnine,
            "\x1b[27;131;57~": MoreKeys.AltDigitnine,
            "\x1b[56;131u": MoreKeys.AltDigiteight,
            "\x1b[27;131;56~": MoreKeys.AltDigiteight,
            "\x1b[55;131u": MoreKeys.AltDigitseven,
            "\x1b[27;131;55~": MoreKeys.AltDigitseven,
            "\x1b[54;131u": MoreKeys.AltDigitsix,
            "\x1b[27;131;54~": MoreKeys.AltDigitsix,
            "\x1b[53;131u": MoreKeys.AltDigitfive,
            "\x1b[27;131;53~": MoreKeys.AltDigitfive,
            "\x1b[52;131u": MoreKeys.AltDigitfour,
            "\x1b[27;131;52~": MoreKeys.AltDigitfour,
            "\x1b[51;131u": MoreKeys.AltDigitthree,
            "\x1b[27;131;51~": MoreKeys.AltDigitthree,
            "\x1b[50;131u": MoreKeys.AltDigittwo,
            "\x1b[27;131;50~": MoreKeys.AltDigittwo,
            "\x1b[49;131u": MoreKeys.AltDigitone,
            "\x1b[27;131;49~": MoreKeys.AltDigitone,
            "\x1b[48;131u": MoreKeys.AltDigitzero,
            "\x1b[27;131;48~": MoreKeys.AltDigitzero,
            "\x1b[47;131u": MoreKeys.AltSolidus,
            "\x1b[27;131;47~": MoreKeys.AltSolidus,
            "\x1b[46;131u": MoreKeys.AltFullstop,
            "\x1b[27;131;46~": MoreKeys.AltFullstop,
            "\x1b[45;131u": MoreKeys.AltHyphenminus,
            "\x1b[27;131;45~": MoreKeys.AltHyphenminus,
            "\x1b[44;131u": MoreKeys.AltComma,
            "\x1b[27;131;44~": MoreKeys.AltComma,
            "\x1b[43;131u": MoreKeys.AltPlussign,
            "\x1b[27;131;43~": MoreKeys.AltPlussign,
            "\x1b[42;131u": MoreKeys.AltAsterisk,
            "\x1b[27;131;42~": MoreKeys.AltAsterisk,
            "\x1b[41;131u": MoreKeys.AltRightparenthesis,
            "\x1b[27;131;41~": MoreKeys.AltRightparenthesis,
            "\x1b[40;131u": MoreKeys.AltLeftparenthesis,
            "\x1b[27;131;40~": MoreKeys.AltLeftparenthesis,
            "\x1b[39;131u": MoreKeys.AltApostrophe,
            "\x1b[27;131;39~": MoreKeys.AltApostrophe,
            "\x1b[38;131u": MoreKeys.AltAmpersand,
            "\x1b[27;131;38~": MoreKeys.AltAmpersand,
            "\x1b[37;131u": MoreKeys.AltPercentsign,
            "\x1b[27;131;37~": MoreKeys.AltPercentsign,
            "\x1b[36;131u": MoreKeys.AltDollarsign,
            "\x1b[27;131;36~": MoreKeys.AltDollarsign,
            "\x1b[35;131u": MoreKeys.AltNumbersign,
            "\x1b[27;131;35~": MoreKeys.AltNumbersign,
            "\x1b[34;131u": MoreKeys.AltQuotationmark,
            "\x1b[27;131;34~": MoreKeys.AltQuotationmark,
            "\x1b[33;131u": MoreKeys.AltExclamationmark,
            "\x1b[27;131;33~": MoreKeys.AltExclamationmark,
            "\x1b[127;131u": MoreKeys.AltBackspace,
            "\x1b[27;131;127~": MoreKeys.AltBackspace,
            "\x1b[2;131~": MoreKeys.AltInsert,
            "\x1b[3;131~": MoreKeys.AltDelete,
            "\x1b[5;131~": MoreKeys.AltPageup,
            "\x1b[6;131~": MoreKeys.AltPagedown,
            "\x1b[1;131A": MoreKeys.AltUp,
            "\x1b[1;131B": MoreKeys.AltDown,
            "\x1b[1;131C": MoreKeys.AltRight,
            "\x1b[1;131D": MoreKeys.AltLeft,
            "\x1b[1;131H": MoreKeys.AltHome,
            "\x1b[1;131F": MoreKeys.AltEnd,
            "\x1b[57361;131u": MoreKeys.AltPrintscreen,
            "\x1b[57362;131u": MoreKeys.AltPause,
            "\x1b[57363;131u": MoreKeys.AltMenu,
            "\x1b[1;131P": MoreKeys.AltF1,
            "\x1b[1;131Q": MoreKeys.AltF2,
            "\x1b[13;131~": MoreKeys.AltF3,
            "\x1b[1;131S": MoreKeys.AltF4,
            "\x1b[15;131~": MoreKeys.AltF5,
            "\x1b[17;131~": MoreKeys.AltF6,
            "\x1b[18;131~": MoreKeys.AltF7,
            "\x1b[19;131~": MoreKeys.AltF8,
            "\x1b[20;131~": MoreKeys.AltF9,
            "\x1b[21;131~": MoreKeys.AltF10,
            "\x1b[23;131~": MoreKeys.AltF11,
            "\x1b[24;131~": MoreKeys.AltF12,
            "\x1b[9;69u": MoreKeys.ControlTab,
            "\x1b[27;69;9~": MoreKeys.ControlTab,
            "\x1b[13;69u": MoreKeys.ControlEnter,
            "\x1b[27;69;13~": MoreKeys.ControlEnter,
            "\x1b[27;69u": MoreKeys.ControlEscape,
            "\x1b[27;69;27~": MoreKeys.ControlEscape,
            "\x1b[32;69u": MoreKeys.ControlSpace,
            "\x1b[27;69;32~": MoreKeys.ControlSpace,
            "\x1b[126;69u": MoreKeys.ControlTilde,
            "\x1b[27;69;126~": MoreKeys.ControlTilde,
            "\x1b[125;69u": MoreKeys.ControlRightcurlybracket,
            "\x1b[27;69;125~": MoreKeys.ControlRightcurlybracket,
            "\x1b[124;69u": MoreKeys.ControlVerticalline,
            "\x1b[27;69;124~": MoreKeys.ControlVerticalline,
            "\x1b[123;69u": MoreKeys.ControlLeftcurlybracket,
            "\x1b[27;69;123~": MoreKeys.ControlLeftcurlybracket,
            "\x1b[122;69u": MoreKeys.ControlLatinsmallletterz,
            "\x1b[27;69;122~": MoreKeys.ControlLatinsmallletterz,
            "\x1b[121;69u": MoreKeys.ControlLatinsmalllettery,
            "\x1b[27;69;121~": MoreKeys.ControlLatinsmalllettery,
            "\x1b[120;69u": MoreKeys.ControlLatinsmallletterx,
            "\x1b[27;69;120~": MoreKeys.ControlLatinsmallletterx,
            "\x1b[119;69u": MoreKeys.ControlLatinsmallletterw,
            "\x1b[27;69;119~": MoreKeys.ControlLatinsmallletterw,
            "\x1b[118;69u": MoreKeys.ControlLatinsmallletterv,
            "\x1b[27;69;118~": MoreKeys.ControlLatinsmallletterv,
            "\x1b[117;69u": MoreKeys.ControlLatinsmallletteru,
            "\x1b[27;69;117~": MoreKeys.ControlLatinsmallletteru,
            "\x1b[116;69u": MoreKeys.ControlLatinsmalllettert,
            "\x1b[27;69;116~": MoreKeys.ControlLatinsmalllettert,
            "\x1b[115;69u": MoreKeys.ControlLatinsmallletters,
            "\x1b[27;69;115~": MoreKeys.ControlLatinsmallletters,
            "\x1b[114;69u": MoreKeys.ControlLatinsmallletterr,
            "\x1b[27;69;114~": MoreKeys.ControlLatinsmallletterr,
            "\x1b[113;69u": MoreKeys.ControlLatinsmallletterq,
            "\x1b[27;69;113~": MoreKeys.ControlLatinsmallletterq,
            "\x1b[112;69u": MoreKeys.ControlLatinsmallletterp,
            "\x1b[27;69;112~": MoreKeys.ControlLatinsmallletterp,
            "\x1b[111;69u": MoreKeys.ControlLatinsmalllettero,
            "\x1b[27;69;111~": MoreKeys.ControlLatinsmalllettero,
            "\x1b[110;69u": MoreKeys.ControlLatinsmalllettern,
            "\x1b[27;69;110~": MoreKeys.ControlLatinsmalllettern,
            "\x1b[109;69u": MoreKeys.ControlLatinsmallletterm,
            "\x1b[27;69;109~": MoreKeys.ControlLatinsmallletterm,
            "\x1b[108;69u": MoreKeys.ControlLatinsmallletterl,
            "\x1b[27;69;108~": MoreKeys.ControlLatinsmallletterl,
            "\x1b[107;69u": MoreKeys.ControlLatinsmallletterk,
            "\x1b[27;69;107~": MoreKeys.ControlLatinsmallletterk,
            "\x1b[106;69u": MoreKeys.ControlLatinsmallletterj,
            "\x1b[27;69;106~": MoreKeys.ControlLatinsmallletterj,
            "\x1b[105;69u": MoreKeys.ControlLatinsmallletteri,
            "\x1b[27;69;105~": MoreKeys.ControlLatinsmallletteri,
            "\x1b[104;69u": MoreKeys.ControlLatinsmallletterh,
            "\x1b[27;69;104~": MoreKeys.ControlLatinsmallletterh,
            "\x1b[103;69u": MoreKeys.ControlLatinsmallletterg,
            "\x1b[27;69;103~": MoreKeys.ControlLatinsmallletterg,
            "\x1b[102;69u": MoreKeys.ControlLatinsmallletterf,
            "\x1b[27;69;102~": MoreKeys.ControlLatinsmallletterf,
            "\x1b[101;69u": MoreKeys.ControlLatinsmalllettere,
            "\x1b[27;69;101~": MoreKeys.ControlLatinsmalllettere,
            "\x1b[100;69u": MoreKeys.ControlLatinsmallletterd,
            "\x1b[27;69;100~": MoreKeys.ControlLatinsmallletterd,
            "\x1b[99;69u": MoreKeys.ControlLatinsmallletterc,
            "\x1b[27;69;99~": MoreKeys.ControlLatinsmallletterc,
            "\x1b[98;69u": MoreKeys.ControlLatinsmallletterb,
            "\x1b[27;69;98~": MoreKeys.ControlLatinsmallletterb,
            "\x1b[97;69u": MoreKeys.ControlLatinsmalllettera,
            "\x1b[27;69;97~": MoreKeys.ControlLatinsmalllettera,
            "\x1b[96;69u": MoreKeys.ControlGraveaccent,
            "\x1b[27;69;96~": MoreKeys.ControlGraveaccent,
            "\x1b[95;69u": MoreKeys.ControlLowline,
            "\x1b[27;69;95~": MoreKeys.ControlLowline,
            "\x1b[94;69u": MoreKeys.ControlCircumflexaccent,
            "\x1b[27;69;94~": MoreKeys.ControlCircumflexaccent,
            "\x1b[93;69u": MoreKeys.ControlRightsquarebracket,
            "\x1b[27;69;93~": MoreKeys.ControlRightsquarebracket,
            "\x1b[92;69u": MoreKeys.ControlReversesolidus,
            "\x1b[27;69;92~": MoreKeys.ControlReversesolidus,
            "\x1b[91;69u": MoreKeys.ControlLeftsquarebracket,
            "\x1b[27;69;91~": MoreKeys.ControlLeftsquarebracket,
            "\x1b[90;69u": MoreKeys.ControlLatincapitalletterz,
            "\x1b[27;69;90~": MoreKeys.ControlLatincapitalletterz,
            "\x1b[89;69u": MoreKeys.ControlLatincapitallettery,
            "\x1b[27;69;89~": MoreKeys.ControlLatincapitallettery,
            "\x1b[88;69u": MoreKeys.ControlLatincapitalletterx,
            "\x1b[27;69;88~": MoreKeys.ControlLatincapitalletterx,
            "\x1b[87;69u": MoreKeys.ControlLatincapitalletterw,
            "\x1b[27;69;87~": MoreKeys.ControlLatincapitalletterw,
            "\x1b[86;69u": MoreKeys.ControlLatincapitalletterv,
            "\x1b[27;69;86~": MoreKeys.ControlLatincapitalletterv,
            "\x1b[85;69u": MoreKeys.ControlLatincapitalletteru,
            "\x1b[27;69;85~": MoreKeys.ControlLatincapitalletteru,
            "\x1b[84;69u": MoreKeys.ControlLatincapitallettert,
            "\x1b[27;69;84~": MoreKeys.ControlLatincapitallettert,
            "\x1b[83;69u": MoreKeys.ControlLatincapitalletters,
            "\x1b[27;69;83~": MoreKeys.ControlLatincapitalletters,
            "\x1b[82;69u": MoreKeys.ControlLatincapitalletterr,
            "\x1b[27;69;82~": MoreKeys.ControlLatincapitalletterr,
            "\x1b[81;69u": MoreKeys.ControlLatincapitalletterq,
            "\x1b[27;69;81~": MoreKeys.ControlLatincapitalletterq,
            "\x1b[80;69u": MoreKeys.ControlLatincapitalletterp,
            "\x1b[27;69;80~": MoreKeys.ControlLatincapitalletterp,
            "\x1b[79;69u": MoreKeys.ControlLatincapitallettero,
            "\x1b[27;69;79~": MoreKeys.ControlLatincapitallettero,
            "\x1b[78;69u": MoreKeys.ControlLatincapitallettern,
            "\x1b[27;69;78~": MoreKeys.ControlLatincapitallettern,
            "\x1b[77;69u": MoreKeys.ControlLatincapitalletterm,
            "\x1b[27;69;77~": MoreKeys.ControlLatincapitalletterm,
            "\x1b[76;69u": MoreKeys.ControlLatincapitalletterl,
            "\x1b[27;69;76~": MoreKeys.ControlLatincapitalletterl,
            "\x1b[75;69u": MoreKeys.ControlLatincapitalletterk,
            "\x1b[27;69;75~": MoreKeys.ControlLatincapitalletterk,
            "\x1b[74;69u": MoreKeys.ControlLatincapitalletterj,
            "\x1b[27;69;74~": MoreKeys.ControlLatincapitalletterj,
            "\x1b[73;69u": MoreKeys.ControlLatincapitalletteri,
            "\x1b[27;69;73~": MoreKeys.ControlLatincapitalletteri,
            "\x1b[72;69u": MoreKeys.ControlLatincapitalletterh,
            "\x1b[27;69;72~": MoreKeys.ControlLatincapitalletterh,
            "\x1b[71;69u": MoreKeys.ControlLatincapitalletterg,
            "\x1b[27;69;71~": MoreKeys.ControlLatincapitalletterg,
            "\x1b[70;69u": MoreKeys.ControlLatincapitalletterf,
            "\x1b[27;69;70~": MoreKeys.ControlLatincapitalletterf,
            "\x1b[69;69u": MoreKeys.ControlLatincapitallettere,
            "\x1b[27;69;69~": MoreKeys.ControlLatincapitallettere,
            "\x1b[68;69u": MoreKeys.ControlLatincapitalletterd,
            "\x1b[27;69;68~": MoreKeys.ControlLatincapitalletterd,
            "\x1b[67;69u": MoreKeys.ControlLatincapitalletterc,
            "\x1b[27;69;67~": MoreKeys.ControlLatincapitalletterc,
            "\x1b[66;69u": MoreKeys.ControlLatincapitalletterb,
            "\x1b[27;69;66~": MoreKeys.ControlLatincapitalletterb,
            "\x1b[65;69u": MoreKeys.ControlLatincapitallettera,
            "\x1b[27;69;65~": MoreKeys.ControlLatincapitallettera,
            "\x1b[64;69u": MoreKeys.ControlCommercialat,
            "\x1b[27;69;64~": MoreKeys.ControlCommercialat,
            "\x1b[63;69u": MoreKeys.ControlQuestionmark,
            "\x1b[27;69;63~": MoreKeys.ControlQuestionmark,
            "\x1b[62;69u": MoreKeys.ControlGreaterthansign,
            "\x1b[27;69;62~": MoreKeys.ControlGreaterthansign,
            "\x1b[61;69u": MoreKeys.ControlEqualssign,
            "\x1b[27;69;61~": MoreKeys.ControlEqualssign,
            "\x1b[60;69u": MoreKeys.ControlLessthansign,
            "\x1b[27;69;60~": MoreKeys.ControlLessthansign,
            "\x1b[59;69u": MoreKeys.ControlSemicolon,
            "\x1b[27;69;59~": MoreKeys.ControlSemicolon,
            "\x1b[58;69u": MoreKeys.ControlColon,
            "\x1b[27;69;58~": MoreKeys.ControlColon,
            "\x1b[57;69u": MoreKeys.ControlDigitnine,
            "\x1b[27;69;57~": MoreKeys.ControlDigitnine,
            "\x1b[56;69u": MoreKeys.ControlDigiteight,
            "\x1b[27;69;56~": MoreKeys.ControlDigiteight,
            "\x1b[55;69u": MoreKeys.ControlDigitseven,
            "\x1b[27;69;55~": MoreKeys.ControlDigitseven,
            "\x1b[54;69u": MoreKeys.ControlDigitsix,
            "\x1b[27;69;54~": MoreKeys.ControlDigitsix,
            "\x1b[53;69u": MoreKeys.ControlDigitfive,
            "\x1b[27;69;53~": MoreKeys.ControlDigitfive,
            "\x1b[52;69u": MoreKeys.ControlDigitfour,
            "\x1b[27;69;52~": MoreKeys.ControlDigitfour,
            "\x1b[51;69u": MoreKeys.ControlDigitthree,
            "\x1b[27;69;51~": MoreKeys.ControlDigitthree,
            "\x1b[50;69u": MoreKeys.ControlDigittwo,
            "\x1b[27;69;50~": MoreKeys.ControlDigittwo,
            "\x1b[49;69u": MoreKeys.ControlDigitone,
            "\x1b[27;69;49~": MoreKeys.ControlDigitone,
            "\x1b[48;69u": MoreKeys.ControlDigitzero,
            "\x1b[27;69;48~": MoreKeys.ControlDigitzero,
            "\x1b[47;69u": MoreKeys.ControlSolidus,
            "\x1b[27;69;47~": MoreKeys.ControlSolidus,
            "\x1b[46;69u": MoreKeys.ControlFullstop,
            "\x1b[27;69;46~": MoreKeys.ControlFullstop,
            "\x1b[45;69u": MoreKeys.ControlHyphenminus,
            "\x1b[27;69;45~": MoreKeys.ControlHyphenminus,
            "\x1b[44;69u": MoreKeys.ControlComma,
            "\x1b[27;69;44~": MoreKeys.ControlComma,
            "\x1b[43;69u": MoreKeys.ControlPlussign,
            "\x1b[27;69;43~": MoreKeys.ControlPlussign,
            "\x1b[42;69u": MoreKeys.ControlAsterisk,
            "\x1b[27;69;42~": MoreKeys.ControlAsterisk,
            "\x1b[41;69u": MoreKeys.ControlRightparenthesis,
            "\x1b[27;69;41~": MoreKeys.ControlRightparenthesis,
            "\x1b[40;69u": MoreKeys.ControlLeftparenthesis,
            "\x1b[27;69;40~": MoreKeys.ControlLeftparenthesis,
            "\x1b[39;69u": MoreKeys.ControlApostrophe,
            "\x1b[27;69;39~": MoreKeys.ControlApostrophe,
            "\x1b[38;69u": MoreKeys.ControlAmpersand,
            "\x1b[27;69;38~": MoreKeys.ControlAmpersand,
            "\x1b[37;69u": MoreKeys.ControlPercentsign,
            "\x1b[27;69;37~": MoreKeys.ControlPercentsign,
            "\x1b[36;69u": MoreKeys.ControlDollarsign,
            "\x1b[27;69;36~": MoreKeys.ControlDollarsign,
            "\x1b[35;69u": MoreKeys.ControlNumbersign,
            "\x1b[27;69;35~": MoreKeys.ControlNumbersign,
            "\x1b[34;69u": MoreKeys.ControlQuotationmark,
            "\x1b[27;69;34~": MoreKeys.ControlQuotationmark,
            "\x1b[33;69u": MoreKeys.ControlExclamationmark,
            "\x1b[27;69;33~": MoreKeys.ControlExclamationmark,
            "\x1b[127;69u": MoreKeys.ControlBackspace,
            "\x1b[27;69;127~": MoreKeys.ControlBackspace,
            "\x1b[2;69~": MoreKeys.ControlInsert,
            "\x1b[3;69~": MoreKeys.ControlDelete,
            "\x1b[5;69~": MoreKeys.ControlPageup,
            "\x1b[6;69~": MoreKeys.ControlPagedown,
            "\x1b[1;69A": MoreKeys.ControlUp,
            "\x1b[1;69B": MoreKeys.ControlDown,
            "\x1b[1;69C": MoreKeys.ControlRight,
            "\x1b[1;69D": MoreKeys.ControlLeft,
            "\x1b[1;69H": MoreKeys.ControlHome,
            "\x1b[1;69F": MoreKeys.ControlEnd,
            "\x1b[57361;69u": MoreKeys.ControlPrintscreen,
            "\x1b[57362;69u": MoreKeys.ControlPause,
            "\x1b[57363;69u": MoreKeys.ControlMenu,
            "\x1b[1;69P": MoreKeys.ControlF1,
            "\x1b[1;69Q": MoreKeys.ControlF2,
            "\x1b[13;69~": MoreKeys.ControlF3,
            "\x1b[1;69S": MoreKeys.ControlF4,
            "\x1b[15;69~": MoreKeys.ControlF5,
            "\x1b[17;69~": MoreKeys.ControlF6,
            "\x1b[18;69~": MoreKeys.ControlF7,
            "\x1b[19;69~": MoreKeys.ControlF8,
            "\x1b[20;69~": MoreKeys.ControlF9,
            "\x1b[21;69~": MoreKeys.ControlF10,
            "\x1b[23;69~": MoreKeys.ControlF11,
            "\x1b[24;69~": MoreKeys.ControlF12,
            "\x1b[9;133u": MoreKeys.ControlTab,
            "\x1b[27;133;9~": MoreKeys.ControlTab,
            "\x1b[13;133u": MoreKeys.ControlEnter,
            "\x1b[27;133;13~": MoreKeys.ControlEnter,
            "\x1b[27;133u": MoreKeys.ControlEscape,
            "\x1b[27;133;27~": MoreKeys.ControlEscape,
            "\x1b[32;133u": MoreKeys.ControlSpace,
            "\x1b[27;133;32~": MoreKeys.ControlSpace,
            "\x1b[126;133u": MoreKeys.ControlTilde,
            "\x1b[27;133;126~": MoreKeys.ControlTilde,
            "\x1b[125;133u": MoreKeys.ControlRightcurlybracket,
            "\x1b[27;133;125~": MoreKeys.ControlRightcurlybracket,
            "\x1b[124;133u": MoreKeys.ControlVerticalline,
            "\x1b[27;133;124~": MoreKeys.ControlVerticalline,
            "\x1b[123;133u": MoreKeys.ControlLeftcurlybracket,
            "\x1b[27;133;123~": MoreKeys.ControlLeftcurlybracket,
            "\x1b[122;133u": MoreKeys.ControlLatinsmallletterz,
            "\x1b[27;133;122~": MoreKeys.ControlLatinsmallletterz,
            "\x1b[121;133u": MoreKeys.ControlLatinsmalllettery,
            "\x1b[27;133;121~": MoreKeys.ControlLatinsmalllettery,
            "\x1b[120;133u": MoreKeys.ControlLatinsmallletterx,
            "\x1b[27;133;120~": MoreKeys.ControlLatinsmallletterx,
            "\x1b[119;133u": MoreKeys.ControlLatinsmallletterw,
            "\x1b[27;133;119~": MoreKeys.ControlLatinsmallletterw,
            "\x1b[118;133u": MoreKeys.ControlLatinsmallletterv,
            "\x1b[27;133;118~": MoreKeys.ControlLatinsmallletterv,
            "\x1b[117;133u": MoreKeys.ControlLatinsmallletteru,
            "\x1b[27;133;117~": MoreKeys.ControlLatinsmallletteru,
            "\x1b[116;133u": MoreKeys.ControlLatinsmalllettert,
            "\x1b[27;133;116~": MoreKeys.ControlLatinsmalllettert,
            "\x1b[115;133u": MoreKeys.ControlLatinsmallletters,
            "\x1b[27;133;115~": MoreKeys.ControlLatinsmallletters,
            "\x1b[114;133u": MoreKeys.ControlLatinsmallletterr,
            "\x1b[27;133;114~": MoreKeys.ControlLatinsmallletterr,
            "\x1b[113;133u": MoreKeys.ControlLatinsmallletterq,
            "\x1b[27;133;113~": MoreKeys.ControlLatinsmallletterq,
            "\x1b[112;133u": MoreKeys.ControlLatinsmallletterp,
            "\x1b[27;133;112~": MoreKeys.ControlLatinsmallletterp,
            "\x1b[111;133u": MoreKeys.ControlLatinsmalllettero,
            "\x1b[27;133;111~": MoreKeys.ControlLatinsmalllettero,
            "\x1b[110;133u": MoreKeys.ControlLatinsmalllettern,
            "\x1b[27;133;110~": MoreKeys.ControlLatinsmalllettern,
            "\x1b[109;133u": MoreKeys.ControlLatinsmallletterm,
            "\x1b[27;133;109~": MoreKeys.ControlLatinsmallletterm,
            "\x1b[108;133u": MoreKeys.ControlLatinsmallletterl,
            "\x1b[27;133;108~": MoreKeys.ControlLatinsmallletterl,
            "\x1b[107;133u": MoreKeys.ControlLatinsmallletterk,
            "\x1b[27;133;107~": MoreKeys.ControlLatinsmallletterk,
            "\x1b[106;133u": MoreKeys.ControlLatinsmallletterj,
            "\x1b[27;133;106~": MoreKeys.ControlLatinsmallletterj,
            "\x1b[105;133u": MoreKeys.ControlLatinsmallletteri,
            "\x1b[27;133;105~": MoreKeys.ControlLatinsmallletteri,
            "\x1b[104;133u": MoreKeys.ControlLatinsmallletterh,
            "\x1b[27;133;104~": MoreKeys.ControlLatinsmallletterh,
            "\x1b[103;133u": MoreKeys.ControlLatinsmallletterg,
            "\x1b[27;133;103~": MoreKeys.ControlLatinsmallletterg,
            "\x1b[102;133u": MoreKeys.ControlLatinsmallletterf,
            "\x1b[27;133;102~": MoreKeys.ControlLatinsmallletterf,
            "\x1b[101;133u": MoreKeys.ControlLatinsmalllettere,
            "\x1b[27;133;101~": MoreKeys.ControlLatinsmalllettere,
            "\x1b[100;133u": MoreKeys.ControlLatinsmallletterd,
            "\x1b[27;133;100~": MoreKeys.ControlLatinsmallletterd,
            "\x1b[99;133u": MoreKeys.ControlLatinsmallletterc,
            "\x1b[27;133;99~": MoreKeys.ControlLatinsmallletterc,
            "\x1b[98;133u": MoreKeys.ControlLatinsmallletterb,
            "\x1b[27;133;98~": MoreKeys.ControlLatinsmallletterb,
            "\x1b[97;133u": MoreKeys.ControlLatinsmalllettera,
            "\x1b[27;133;97~": MoreKeys.ControlLatinsmalllettera,
            "\x1b[96;133u": MoreKeys.ControlGraveaccent,
            "\x1b[27;133;96~": MoreKeys.ControlGraveaccent,
            "\x1b[95;133u": MoreKeys.ControlLowline,
            "\x1b[27;133;95~": MoreKeys.ControlLowline,
            "\x1b[94;133u": MoreKeys.ControlCircumflexaccent,
            "\x1b[27;133;94~": MoreKeys.ControlCircumflexaccent,
            "\x1b[93;133u": MoreKeys.ControlRightsquarebracket,
            "\x1b[27;133;93~": MoreKeys.ControlRightsquarebracket,
            "\x1b[92;133u": MoreKeys.ControlReversesolidus,
            "\x1b[27;133;92~": MoreKeys.ControlReversesolidus,
            "\x1b[91;133u": MoreKeys.ControlLeftsquarebracket,
            "\x1b[27;133;91~": MoreKeys.ControlLeftsquarebracket,
            "\x1b[90;133u": MoreKeys.ControlLatincapitalletterz,
            "\x1b[27;133;90~": MoreKeys.ControlLatincapitalletterz,
            "\x1b[89;133u": MoreKeys.ControlLatincapitallettery,
            "\x1b[27;133;89~": MoreKeys.ControlLatincapitallettery,
            "\x1b[88;133u": MoreKeys.ControlLatincapitalletterx,
            "\x1b[27;133;88~": MoreKeys.ControlLatincapitalletterx,
            "\x1b[87;133u": MoreKeys.ControlLatincapitalletterw,
            "\x1b[27;133;87~": MoreKeys.ControlLatincapitalletterw,
            "\x1b[86;133u": MoreKeys.ControlLatincapitalletterv,
            "\x1b[27;133;86~": MoreKeys.ControlLatincapitalletterv,
            "\x1b[85;133u": MoreKeys.ControlLatincapitalletteru,
            "\x1b[27;133;85~": MoreKeys.ControlLatincapitalletteru,
            "\x1b[84;133u": MoreKeys.ControlLatincapitallettert,
            "\x1b[27;133;84~": MoreKeys.ControlLatincapitallettert,
            "\x1b[83;133u": MoreKeys.ControlLatincapitalletters,
            "\x1b[27;133;83~": MoreKeys.ControlLatincapitalletters,
            "\x1b[82;133u": MoreKeys.ControlLatincapitalletterr,
            "\x1b[27;133;82~": MoreKeys.ControlLatincapitalletterr,
            "\x1b[81;133u": MoreKeys.ControlLatincapitalletterq,
            "\x1b[27;133;81~": MoreKeys.ControlLatincapitalletterq,
            "\x1b[80;133u": MoreKeys.ControlLatincapitalletterp,
            "\x1b[27;133;80~": MoreKeys.ControlLatincapitalletterp,
            "\x1b[79;133u": MoreKeys.ControlLatincapitallettero,
            "\x1b[27;133;79~": MoreKeys.ControlLatincapitallettero,
            "\x1b[78;133u": MoreKeys.ControlLatincapitallettern,
            "\x1b[27;133;78~": MoreKeys.ControlLatincapitallettern,
            "\x1b[77;133u": MoreKeys.ControlLatincapitalletterm,
            "\x1b[27;133;77~": MoreKeys.ControlLatincapitalletterm,
            "\x1b[76;133u": MoreKeys.ControlLatincapitalletterl,
            "\x1b[27;133;76~": MoreKeys.ControlLatincapitalletterl,
            "\x1b[75;133u": MoreKeys.ControlLatincapitalletterk,
            "\x1b[27;133;75~": MoreKeys.ControlLatincapitalletterk,
            "\x1b[74;133u": MoreKeys.ControlLatincapitalletterj,
            "\x1b[27;133;74~": MoreKeys.ControlLatincapitalletterj,
            "\x1b[73;133u": MoreKeys.ControlLatincapitalletteri,
            "\x1b[27;133;73~": MoreKeys.ControlLatincapitalletteri,
            "\x1b[72;133u": MoreKeys.ControlLatincapitalletterh,
            "\x1b[27;133;72~": MoreKeys.ControlLatincapitalletterh,
            "\x1b[71;133u": MoreKeys.ControlLatincapitalletterg,
            "\x1b[27;133;71~": MoreKeys.ControlLatincapitalletterg,
            "\x1b[70;133u": MoreKeys.ControlLatincapitalletterf,
            "\x1b[27;133;70~": MoreKeys.ControlLatincapitalletterf,
            "\x1b[69;133u": MoreKeys.ControlLatincapitallettere,
            "\x1b[27;133;69~": MoreKeys.ControlLatincapitallettere,
            "\x1b[68;133u": MoreKeys.ControlLatincapitalletterd,
            "\x1b[27;133;68~": MoreKeys.ControlLatincapitalletterd,
            "\x1b[67;133u": MoreKeys.ControlLatincapitalletterc,
            "\x1b[27;133;67~": MoreKeys.ControlLatincapitalletterc,
            "\x1b[66;133u": MoreKeys.ControlLatincapitalletterb,
            "\x1b[27;133;66~": MoreKeys.ControlLatincapitalletterb,
            "\x1b[65;133u": MoreKeys.ControlLatincapitallettera,
            "\x1b[27;133;65~": MoreKeys.ControlLatincapitallettera,
            "\x1b[64;133u": MoreKeys.ControlCommercialat,
            "\x1b[27;133;64~": MoreKeys.ControlCommercialat,
            "\x1b[63;133u": MoreKeys.ControlQuestionmark,
            "\x1b[27;133;63~": MoreKeys.ControlQuestionmark,
            "\x1b[62;133u": MoreKeys.ControlGreaterthansign,
            "\x1b[27;133;62~": MoreKeys.ControlGreaterthansign,
            "\x1b[61;133u": MoreKeys.ControlEqualssign,
            "\x1b[27;133;61~": MoreKeys.ControlEqualssign,
            "\x1b[60;133u": MoreKeys.ControlLessthansign,
            "\x1b[27;133;60~": MoreKeys.ControlLessthansign,
            "\x1b[59;133u": MoreKeys.ControlSemicolon,
            "\x1b[27;133;59~": MoreKeys.ControlSemicolon,
            "\x1b[58;133u": MoreKeys.ControlColon,
            "\x1b[27;133;58~": MoreKeys.ControlColon,
            "\x1b[57;133u": MoreKeys.ControlDigitnine,
            "\x1b[27;133;57~": MoreKeys.ControlDigitnine,
            "\x1b[56;133u": MoreKeys.ControlDigiteight,
            "\x1b[27;133;56~": MoreKeys.ControlDigiteight,
            "\x1b[55;133u": MoreKeys.ControlDigitseven,
            "\x1b[27;133;55~": MoreKeys.ControlDigitseven,
            "\x1b[54;133u": MoreKeys.ControlDigitsix,
            "\x1b[27;133;54~": MoreKeys.ControlDigitsix,
            "\x1b[53;133u": MoreKeys.ControlDigitfive,
            "\x1b[27;133;53~": MoreKeys.ControlDigitfive,
            "\x1b[52;133u": MoreKeys.ControlDigitfour,
            "\x1b[27;133;52~": MoreKeys.ControlDigitfour,
            "\x1b[51;133u": MoreKeys.ControlDigitthree,
            "\x1b[27;133;51~": MoreKeys.ControlDigitthree,
            "\x1b[50;133u": MoreKeys.ControlDigittwo,
            "\x1b[27;133;50~": MoreKeys.ControlDigittwo,
            "\x1b[49;133u": MoreKeys.ControlDigitone,
            "\x1b[27;133;49~": MoreKeys.ControlDigitone,
            "\x1b[48;133u": MoreKeys.ControlDigitzero,
            "\x1b[27;133;48~": MoreKeys.ControlDigitzero,
            "\x1b[47;133u": MoreKeys.ControlSolidus,
            "\x1b[27;133;47~": MoreKeys.ControlSolidus,
            "\x1b[46;133u": MoreKeys.ControlFullstop,
            "\x1b[27;133;46~": MoreKeys.ControlFullstop,
            "\x1b[45;133u": MoreKeys.ControlHyphenminus,
            "\x1b[27;133;45~": MoreKeys.ControlHyphenminus,
            "\x1b[44;133u": MoreKeys.ControlComma,
            "\x1b[27;133;44~": MoreKeys.ControlComma,
            "\x1b[43;133u": MoreKeys.ControlPlussign,
            "\x1b[27;133;43~": MoreKeys.ControlPlussign,
            "\x1b[42;133u": MoreKeys.ControlAsterisk,
            "\x1b[27;133;42~": MoreKeys.ControlAsterisk,
            "\x1b[41;133u": MoreKeys.ControlRightparenthesis,
            "\x1b[27;133;41~": MoreKeys.ControlRightparenthesis,
            "\x1b[40;133u": MoreKeys.ControlLeftparenthesis,
            "\x1b[27;133;40~": MoreKeys.ControlLeftparenthesis,
            "\x1b[39;133u": MoreKeys.ControlApostrophe,
            "\x1b[27;133;39~": MoreKeys.ControlApostrophe,
            "\x1b[38;133u": MoreKeys.ControlAmpersand,
            "\x1b[27;133;38~": MoreKeys.ControlAmpersand,
            "\x1b[37;133u": MoreKeys.ControlPercentsign,
            "\x1b[27;133;37~": MoreKeys.ControlPercentsign,
            "\x1b[36;133u": MoreKeys.ControlDollarsign,
            "\x1b[27;133;36~": MoreKeys.ControlDollarsign,
            "\x1b[35;133u": MoreKeys.ControlNumbersign,
            "\x1b[27;133;35~": MoreKeys.ControlNumbersign,
            "\x1b[34;133u": MoreKeys.ControlQuotationmark,
            "\x1b[27;133;34~": MoreKeys.ControlQuotationmark,
            "\x1b[33;133u": MoreKeys.ControlExclamationmark,
            "\x1b[27;133;33~": MoreKeys.ControlExclamationmark,
            "\x1b[127;133u": MoreKeys.ControlBackspace,
            "\x1b[27;133;127~": MoreKeys.ControlBackspace,
            "\x1b[2;133~": MoreKeys.ControlInsert,
            "\x1b[3;133~": MoreKeys.ControlDelete,
            "\x1b[5;133~": MoreKeys.ControlPageup,
            "\x1b[6;133~": MoreKeys.ControlPagedown,
            "\x1b[1;133A": MoreKeys.ControlUp,
            "\x1b[1;133B": MoreKeys.ControlDown,
            "\x1b[1;133C": MoreKeys.ControlRight,
            "\x1b[1;133D": MoreKeys.ControlLeft,
            "\x1b[1;133H": MoreKeys.ControlHome,
            "\x1b[1;133F": MoreKeys.ControlEnd,
            "\x1b[57361;133u": MoreKeys.ControlPrintscreen,
            "\x1b[57362;133u": MoreKeys.ControlPause,
            "\x1b[57363;133u": MoreKeys.ControlMenu,
            "\x1b[1;133P": MoreKeys.ControlF1,
            "\x1b[1;133Q": MoreKeys.ControlF2,
            "\x1b[13;133~": MoreKeys.ControlF3,
            "\x1b[1;133S": MoreKeys.ControlF4,
            "\x1b[15;133~": MoreKeys.ControlF5,
            "\x1b[17;133~": MoreKeys.ControlF6,
            "\x1b[18;133~": MoreKeys.ControlF7,
            "\x1b[19;133~": MoreKeys.ControlF8,
            "\x1b[20;133~": MoreKeys.ControlF9,
            "\x1b[21;133~": MoreKeys.ControlF10,
            "\x1b[23;133~": MoreKeys.ControlF11,
            "\x1b[24;133~": MoreKeys.ControlF12,
            "\x1b[27;193u": MoreKeys.Escape,
            "\x1b[2;193~": MoreKeys.Insert,
            "\x1b[3;193~": MoreKeys.Delete,
            "\x1b[5;193~": MoreKeys.Pageup,
            "\x1b[6;193~": MoreKeys.Pagedown,
            "\x1b[1;193A": MoreKeys.Up,
            "\x1b[1;193B": MoreKeys.Down,
            "\x1b[1;193C": MoreKeys.Right,
            "\x1b[1;193D": MoreKeys.Left,
            "\x1b[1;193H": MoreKeys.Home,
            "\x1b[1;193F": MoreKeys.End,
            "\x1b[57361;193u": MoreKeys.Printscreen,
            "\x1b[57362;193u": MoreKeys.Pause,
            "\x1b[57363;193u": MoreKeys.Menu,
            "\x1b[1;193P": MoreKeys.F1,
            "\x1b[1;193Q": MoreKeys.F2,
            "\x1b[13;193~": MoreKeys.F3,
            "\x1b[1;193S": MoreKeys.F4,
            "\x1b[15;193~": MoreKeys.F5,
            "\x1b[17;193~": MoreKeys.F6,
            "\x1b[18;193~": MoreKeys.F7,
            "\x1b[19;193~": MoreKeys.F8,
            "\x1b[20;193~": MoreKeys.F9,
            "\x1b[21;193~": MoreKeys.F10,
            "\x1b[23;193~": MoreKeys.F11,
            "\x1b[24;193~": MoreKeys.F12,
            "\x1b[127;193u": MoreKeys.Backspace,
            "\x1b[13;193u": MoreKeys.Enter,
            "\x1b[9;193u": MoreKeys.Tab,
            "\x1b[9;8u": MoreKeys.ControlAltShiftTab,
            "\x1b[27;8;9~": MoreKeys.ControlAltShiftTab,
            "\x1b[13;8u": MoreKeys.ControlAltShiftEnter,
            "\x1b[27;8;13~": MoreKeys.ControlAltShiftEnter,
            "\x1b[27;8u": MoreKeys.ControlAltShiftEscape,
            "\x1b[27;8;27~": MoreKeys.ControlAltShiftEscape,
            "\x1b[32;8u": MoreKeys.ControlAltShiftSpace,
            "\x1b[27;8;32~": MoreKeys.ControlAltShiftSpace,
            "\x1b[126;8u": MoreKeys.ControlAltShiftTilde,
            "\x1b[27;8;126~": MoreKeys.ControlAltShiftTilde,
            "\x1b[125;8u": MoreKeys.ControlAltShiftRightcurlybracket,
            "\x1b[27;8;125~": MoreKeys.ControlAltShiftRightcurlybracket,
            "\x1b[124;8u": MoreKeys.ControlAltShiftVerticalline,
            "\x1b[27;8;124~": MoreKeys.ControlAltShiftVerticalline,
            "\x1b[123;8u": MoreKeys.ControlAltShiftLeftcurlybracket,
            "\x1b[27;8;123~": MoreKeys.ControlAltShiftLeftcurlybracket,
            "\x1b[122;8u": MoreKeys.ControlAltShiftLatinsmallletterz,
            "\x1b[27;8;122~": MoreKeys.ControlAltShiftLatinsmallletterz,
            "\x1b[121;8u": MoreKeys.ControlAltShiftLatinsmalllettery,
            "\x1b[27;8;121~": MoreKeys.ControlAltShiftLatinsmalllettery,
            "\x1b[120;8u": MoreKeys.ControlAltShiftLatinsmallletterx,
            "\x1b[27;8;120~": MoreKeys.ControlAltShiftLatinsmallletterx,
            "\x1b[119;8u": MoreKeys.ControlAltShiftLatinsmallletterw,
            "\x1b[27;8;119~": MoreKeys.ControlAltShiftLatinsmallletterw,
            "\x1b[118;8u": MoreKeys.ControlAltShiftLatinsmallletterv,
            "\x1b[27;8;118~": MoreKeys.ControlAltShiftLatinsmallletterv,
            "\x1b[117;8u": MoreKeys.ControlAltShiftLatinsmallletteru,
            "\x1b[27;8;117~": MoreKeys.ControlAltShiftLatinsmallletteru,
            "\x1b[116;8u": MoreKeys.ControlAltShiftLatinsmalllettert,
            "\x1b[27;8;116~": MoreKeys.ControlAltShiftLatinsmalllettert,
            "\x1b[115;8u": MoreKeys.ControlAltShiftLatinsmallletters,
            "\x1b[27;8;115~": MoreKeys.ControlAltShiftLatinsmallletters,
            "\x1b[114;8u": MoreKeys.ControlAltShiftLatinsmallletterr,
            "\x1b[27;8;114~": MoreKeys.ControlAltShiftLatinsmallletterr,
            "\x1b[113;8u": MoreKeys.ControlAltShiftLatinsmallletterq,
            "\x1b[27;8;113~": MoreKeys.ControlAltShiftLatinsmallletterq,
            "\x1b[112;8u": MoreKeys.ControlAltShiftLatinsmallletterp,
            "\x1b[27;8;112~": MoreKeys.ControlAltShiftLatinsmallletterp,
            "\x1b[111;8u": MoreKeys.ControlAltShiftLatinsmalllettero,
            "\x1b[27;8;111~": MoreKeys.ControlAltShiftLatinsmalllettero,
            "\x1b[110;8u": MoreKeys.ControlAltShiftLatinsmalllettern,
            "\x1b[27;8;110~": MoreKeys.ControlAltShiftLatinsmalllettern,
            "\x1b[109;8u": MoreKeys.ControlAltShiftLatinsmallletterm,
            "\x1b[27;8;109~": MoreKeys.ControlAltShiftLatinsmallletterm,
            "\x1b[108;8u": MoreKeys.ControlAltShiftLatinsmallletterl,
            "\x1b[27;8;108~": MoreKeys.ControlAltShiftLatinsmallletterl,
            "\x1b[107;8u": MoreKeys.ControlAltShiftLatinsmallletterk,
            "\x1b[27;8;107~": MoreKeys.ControlAltShiftLatinsmallletterk,
            "\x1b[106;8u": MoreKeys.ControlAltShiftLatinsmallletterj,
            "\x1b[27;8;106~": MoreKeys.ControlAltShiftLatinsmallletterj,
            "\x1b[105;8u": MoreKeys.ControlAltShiftLatinsmallletteri,
            "\x1b[27;8;105~": MoreKeys.ControlAltShiftLatinsmallletteri,
            "\x1b[104;8u": MoreKeys.ControlAltShiftLatinsmallletterh,
            "\x1b[27;8;104~": MoreKeys.ControlAltShiftLatinsmallletterh,
            "\x1b[103;8u": MoreKeys.ControlAltShiftLatinsmallletterg,
            "\x1b[27;8;103~": MoreKeys.ControlAltShiftLatinsmallletterg,
            "\x1b[102;8u": MoreKeys.ControlAltShiftLatinsmallletterf,
            "\x1b[27;8;102~": MoreKeys.ControlAltShiftLatinsmallletterf,
            "\x1b[101;8u": MoreKeys.ControlAltShiftLatinsmalllettere,
            "\x1b[27;8;101~": MoreKeys.ControlAltShiftLatinsmalllettere,
            "\x1b[100;8u": MoreKeys.ControlAltShiftLatinsmallletterd,
            "\x1b[27;8;100~": MoreKeys.ControlAltShiftLatinsmallletterd,
            "\x1b[99;8u": MoreKeys.ControlAltShiftLatinsmallletterc,
            "\x1b[27;8;99~": MoreKeys.ControlAltShiftLatinsmallletterc,
            "\x1b[98;8u": MoreKeys.ControlAltShiftLatinsmallletterb,
            "\x1b[27;8;98~": MoreKeys.ControlAltShiftLatinsmallletterb,
            "\x1b[97;8u": MoreKeys.ControlAltShiftLatinsmalllettera,
            "\x1b[27;8;97~": MoreKeys.ControlAltShiftLatinsmalllettera,
            "\x1b[96;8u": MoreKeys.ControlAltShiftGraveaccent,
            "\x1b[27;8;96~": MoreKeys.ControlAltShiftGraveaccent,
            "\x1b[95;8u": MoreKeys.ControlAltShiftLowline,
            "\x1b[27;8;95~": MoreKeys.ControlAltShiftLowline,
            "\x1b[94;8u": MoreKeys.ControlAltShiftCircumflexaccent,
            "\x1b[27;8;94~": MoreKeys.ControlAltShiftCircumflexaccent,
            "\x1b[93;8u": MoreKeys.ControlAltShiftRightsquarebracket,
            "\x1b[27;8;93~": MoreKeys.ControlAltShiftRightsquarebracket,
            "\x1b[92;8u": MoreKeys.ControlAltShiftReversesolidus,
            "\x1b[27;8;92~": MoreKeys.ControlAltShiftReversesolidus,
            "\x1b[91;8u": MoreKeys.ControlAltShiftLeftsquarebracket,
            "\x1b[27;8;91~": MoreKeys.ControlAltShiftLeftsquarebracket,
            "\x1b[90;8u": MoreKeys.ControlAltShiftLatincapitalletterz,
            "\x1b[27;8;90~": MoreKeys.ControlAltShiftLatincapitalletterz,
            "\x1b[89;8u": MoreKeys.ControlAltShiftLatincapitallettery,
            "\x1b[27;8;89~": MoreKeys.ControlAltShiftLatincapitallettery,
            "\x1b[88;8u": MoreKeys.ControlAltShiftLatincapitalletterx,
            "\x1b[27;8;88~": MoreKeys.ControlAltShiftLatincapitalletterx,
            "\x1b[87;8u": MoreKeys.ControlAltShiftLatincapitalletterw,
            "\x1b[27;8;87~": MoreKeys.ControlAltShiftLatincapitalletterw,
            "\x1b[86;8u": MoreKeys.ControlAltShiftLatincapitalletterv,
            "\x1b[27;8;86~": MoreKeys.ControlAltShiftLatincapitalletterv,
            "\x1b[85;8u": MoreKeys.ControlAltShiftLatincapitalletteru,
            "\x1b[27;8;85~": MoreKeys.ControlAltShiftLatincapitalletteru,
            "\x1b[84;8u": MoreKeys.ControlAltShiftLatincapitallettert,
            "\x1b[27;8;84~": MoreKeys.ControlAltShiftLatincapitallettert,
            "\x1b[83;8u": MoreKeys.ControlAltShiftLatincapitalletters,
            "\x1b[27;8;83~": MoreKeys.ControlAltShiftLatincapitalletters,
            "\x1b[82;8u": MoreKeys.ControlAltShiftLatincapitalletterr,
            "\x1b[27;8;82~": MoreKeys.ControlAltShiftLatincapitalletterr,
            "\x1b[81;8u": MoreKeys.ControlAltShiftLatincapitalletterq,
            "\x1b[27;8;81~": MoreKeys.ControlAltShiftLatincapitalletterq,
            "\x1b[80;8u": MoreKeys.ControlAltShiftLatincapitalletterp,
            "\x1b[27;8;80~": MoreKeys.ControlAltShiftLatincapitalletterp,
            "\x1b[79;8u": MoreKeys.ControlAltShiftLatincapitallettero,
            "\x1b[27;8;79~": MoreKeys.ControlAltShiftLatincapitallettero,
            "\x1b[78;8u": MoreKeys.ControlAltShiftLatincapitallettern,
            "\x1b[27;8;78~": MoreKeys.ControlAltShiftLatincapitallettern,
            "\x1b[77;8u": MoreKeys.ControlAltShiftLatincapitalletterm,
            "\x1b[27;8;77~": MoreKeys.ControlAltShiftLatincapitalletterm,
            "\x1b[76;8u": MoreKeys.ControlAltShiftLatincapitalletterl,
            "\x1b[27;8;76~": MoreKeys.ControlAltShiftLatincapitalletterl,
            "\x1b[75;8u": MoreKeys.ControlAltShiftLatincapitalletterk,
            "\x1b[27;8;75~": MoreKeys.ControlAltShiftLatincapitalletterk,
            "\x1b[74;8u": MoreKeys.ControlAltShiftLatincapitalletterj,
            "\x1b[27;8;74~": MoreKeys.ControlAltShiftLatincapitalletterj,
            "\x1b[73;8u": MoreKeys.ControlAltShiftLatincapitalletteri,
            "\x1b[27;8;73~": MoreKeys.ControlAltShiftLatincapitalletteri,
            "\x1b[72;8u": MoreKeys.ControlAltShiftLatincapitalletterh,
            "\x1b[27;8;72~": MoreKeys.ControlAltShiftLatincapitalletterh,
            "\x1b[71;8u": MoreKeys.ControlAltShiftLatincapitalletterg,
            "\x1b[27;8;71~": MoreKeys.ControlAltShiftLatincapitalletterg,
            "\x1b[70;8u": MoreKeys.ControlAltShiftLatincapitalletterf,
            "\x1b[27;8;70~": MoreKeys.ControlAltShiftLatincapitalletterf,
            "\x1b[69;8u": MoreKeys.ControlAltShiftLatincapitallettere,
            "\x1b[27;8;69~": MoreKeys.ControlAltShiftLatincapitallettere,
            "\x1b[68;8u": MoreKeys.ControlAltShiftLatincapitalletterd,
            "\x1b[27;8;68~": MoreKeys.ControlAltShiftLatincapitalletterd,
            "\x1b[67;8u": MoreKeys.ControlAltShiftLatincapitalletterc,
            "\x1b[27;8;67~": MoreKeys.ControlAltShiftLatincapitalletterc,
            "\x1b[66;8u": MoreKeys.ControlAltShiftLatincapitalletterb,
            "\x1b[27;8;66~": MoreKeys.ControlAltShiftLatincapitalletterb,
            "\x1b[65;8u": MoreKeys.ControlAltShiftLatincapitallettera,
            "\x1b[27;8;65~": MoreKeys.ControlAltShiftLatincapitallettera,
            "\x1b[64;8u": MoreKeys.ControlAltShiftCommercialat,
            "\x1b[27;8;64~": MoreKeys.ControlAltShiftCommercialat,
            "\x1b[63;8u": MoreKeys.ControlAltShiftQuestionmark,
            "\x1b[27;8;63~": MoreKeys.ControlAltShiftQuestionmark,
            "\x1b[62;8u": MoreKeys.ControlAltShiftGreaterthansign,
            "\x1b[27;8;62~": MoreKeys.ControlAltShiftGreaterthansign,
            "\x1b[61;8u": MoreKeys.ControlAltShiftEqualssign,
            "\x1b[27;8;61~": MoreKeys.ControlAltShiftEqualssign,
            "\x1b[60;8u": MoreKeys.ControlAltShiftLessthansign,
            "\x1b[27;8;60~": MoreKeys.ControlAltShiftLessthansign,
            "\x1b[59;8u": MoreKeys.ControlAltShiftSemicolon,
            "\x1b[27;8;59~": MoreKeys.ControlAltShiftSemicolon,
            "\x1b[58;8u": MoreKeys.ControlAltShiftColon,
            "\x1b[27;8;58~": MoreKeys.ControlAltShiftColon,
            "\x1b[57;8u": MoreKeys.ControlAltShiftDigitnine,
            "\x1b[27;8;57~": MoreKeys.ControlAltShiftDigitnine,
            "\x1b[56;8u": MoreKeys.ControlAltShiftDigiteight,
            "\x1b[27;8;56~": MoreKeys.ControlAltShiftDigiteight,
            "\x1b[55;8u": MoreKeys.ControlAltShiftDigitseven,
            "\x1b[27;8;55~": MoreKeys.ControlAltShiftDigitseven,
            "\x1b[54;8u": MoreKeys.ControlAltShiftDigitsix,
            "\x1b[27;8;54~": MoreKeys.ControlAltShiftDigitsix,
            "\x1b[53;8u": MoreKeys.ControlAltShiftDigitfive,
            "\x1b[27;8;53~": MoreKeys.ControlAltShiftDigitfive,
            "\x1b[52;8u": MoreKeys.ControlAltShiftDigitfour,
            "\x1b[27;8;52~": MoreKeys.ControlAltShiftDigitfour,
            "\x1b[51;8u": MoreKeys.ControlAltShiftDigitthree,
            "\x1b[27;8;51~": MoreKeys.ControlAltShiftDigitthree,
            "\x1b[50;8u": MoreKeys.ControlAltShiftDigittwo,
            "\x1b[27;8;50~": MoreKeys.ControlAltShiftDigittwo,
            "\x1b[49;8u": MoreKeys.ControlAltShiftDigitone,
            "\x1b[27;8;49~": MoreKeys.ControlAltShiftDigitone,
            "\x1b[48;8u": MoreKeys.ControlAltShiftDigitzero,
            "\x1b[27;8;48~": MoreKeys.ControlAltShiftDigitzero,
            "\x1b[47;8u": MoreKeys.ControlAltShiftSolidus,
            "\x1b[27;8;47~": MoreKeys.ControlAltShiftSolidus,
            "\x1b[46;8u": MoreKeys.ControlAltShiftFullstop,
            "\x1b[27;8;46~": MoreKeys.ControlAltShiftFullstop,
            "\x1b[45;8u": MoreKeys.ControlAltShiftHyphenminus,
            "\x1b[27;8;45~": MoreKeys.ControlAltShiftHyphenminus,
            "\x1b[44;8u": MoreKeys.ControlAltShiftComma,
            "\x1b[27;8;44~": MoreKeys.ControlAltShiftComma,
            "\x1b[43;8u": MoreKeys.ControlAltShiftPlussign,
            "\x1b[27;8;43~": MoreKeys.ControlAltShiftPlussign,
            "\x1b[42;8u": MoreKeys.ControlAltShiftAsterisk,
            "\x1b[27;8;42~": MoreKeys.ControlAltShiftAsterisk,
            "\x1b[41;8u": MoreKeys.ControlAltShiftRightparenthesis,
            "\x1b[27;8;41~": MoreKeys.ControlAltShiftRightparenthesis,
            "\x1b[40;8u": MoreKeys.ControlAltShiftLeftparenthesis,
            "\x1b[27;8;40~": MoreKeys.ControlAltShiftLeftparenthesis,
            "\x1b[39;8u": MoreKeys.ControlAltShiftApostrophe,
            "\x1b[27;8;39~": MoreKeys.ControlAltShiftApostrophe,
            "\x1b[38;8u": MoreKeys.ControlAltShiftAmpersand,
            "\x1b[27;8;38~": MoreKeys.ControlAltShiftAmpersand,
            "\x1b[37;8u": MoreKeys.ControlAltShiftPercentsign,
            "\x1b[27;8;37~": MoreKeys.ControlAltShiftPercentsign,
            "\x1b[36;8u": MoreKeys.ControlAltShiftDollarsign,
            "\x1b[27;8;36~": MoreKeys.ControlAltShiftDollarsign,
            "\x1b[35;8u": MoreKeys.ControlAltShiftNumbersign,
            "\x1b[27;8;35~": MoreKeys.ControlAltShiftNumbersign,
            "\x1b[34;8u": MoreKeys.ControlAltShiftQuotationmark,
            "\x1b[27;8;34~": MoreKeys.ControlAltShiftQuotationmark,
            "\x1b[33;8u": MoreKeys.ControlAltShiftExclamationmark,
            "\x1b[27;8;33~": MoreKeys.ControlAltShiftExclamationmark,
            "\x1b[127;8u": MoreKeys.ControlAltShiftBackspace,
            "\x1b[27;8;127~": MoreKeys.ControlAltShiftBackspace,
            "\x1b[2;8~": MoreKeys.ControlAltShiftInsert,
            "\x1b[3;8~": MoreKeys.ControlAltShiftDelete,
            "\x1b[5;8~": MoreKeys.ControlAltShiftPageup,
            "\x1b[6;8~": MoreKeys.ControlAltShiftPagedown,
            "\x1b[1;8A": MoreKeys.ControlAltShiftUp,
            "\x1b[1;8B": MoreKeys.ControlAltShiftDown,
            "\x1b[1;8C": MoreKeys.ControlAltShiftRight,
            "\x1b[1;8D": MoreKeys.ControlAltShiftLeft,
            "\x1b[1;8H": MoreKeys.ControlAltShiftHome,
            "\x1b[1;8F": MoreKeys.ControlAltShiftEnd,
            "\x1b[57361;8u": MoreKeys.ControlAltShiftPrintscreen,
            "\x1b[57362;8u": MoreKeys.ControlAltShiftPause,
            "\x1b[57363;8u": MoreKeys.ControlAltShiftMenu,
            "\x1b[1;8P": MoreKeys.ControlAltShiftF1,
            "\x1b[1;8Q": MoreKeys.ControlAltShiftF2,
            "\x1b[13;8~": MoreKeys.ControlAltShiftF3,
            "\x1b[1;8S": MoreKeys.ControlAltShiftF4,
            "\x1b[15;8~": MoreKeys.ControlAltShiftF5,
            "\x1b[17;8~": MoreKeys.ControlAltShiftF6,
            "\x1b[18;8~": MoreKeys.ControlAltShiftF7,
            "\x1b[19;8~": MoreKeys.ControlAltShiftF8,
            "\x1b[20;8~": MoreKeys.ControlAltShiftF9,
            "\x1b[21;8~": MoreKeys.ControlAltShiftF10,
            "\x1b[23;8~": MoreKeys.ControlAltShiftF11,
            "\x1b[24;8~": MoreKeys.ControlAltShiftF12,
            "\x1b[9;68u": MoreKeys.AltShiftTab,
            "\x1b[27;68;9~": MoreKeys.AltShiftTab,
            "\x1b[13;68u": MoreKeys.AltShiftEnter,
            "\x1b[27;68;13~": MoreKeys.AltShiftEnter,
            "\x1b[27;68u": MoreKeys.AltShiftEscape,
            "\x1b[27;68;27~": MoreKeys.AltShiftEscape,
            "\x1b[32;68u": MoreKeys.AltShiftSpace,
            "\x1b[27;68;32~": MoreKeys.AltShiftSpace,
            "\x1b[126;68u": MoreKeys.AltShiftTilde,
            "\x1b[27;68;126~": MoreKeys.AltShiftTilde,
            "\x1b[125;68u": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[27;68;125~": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[124;68u": MoreKeys.AltShiftVerticalline,
            "\x1b[27;68;124~": MoreKeys.AltShiftVerticalline,
            "\x1b[123;68u": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[27;68;123~": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[122;68u": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[27;68;122~": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[121;68u": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[27;68;121~": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[120;68u": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[27;68;120~": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[119;68u": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[27;68;119~": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[118;68u": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[27;68;118~": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[117;68u": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[27;68;117~": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[116;68u": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[27;68;116~": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[115;68u": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[27;68;115~": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[114;68u": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[27;68;114~": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[113;68u": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[27;68;113~": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[112;68u": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[27;68;112~": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[111;68u": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[27;68;111~": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[110;68u": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[27;68;110~": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[109;68u": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[27;68;109~": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[108;68u": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[27;68;108~": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[107;68u": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[27;68;107~": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[106;68u": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[27;68;106~": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[105;68u": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[27;68;105~": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[104;68u": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[27;68;104~": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[103;68u": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[27;68;103~": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[102;68u": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[27;68;102~": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[101;68u": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[27;68;101~": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[100;68u": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[27;68;100~": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[99;68u": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[27;68;99~": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[98;68u": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[27;68;98~": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[97;68u": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[27;68;97~": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[96;68u": MoreKeys.AltShiftGraveaccent,
            "\x1b[27;68;96~": MoreKeys.AltShiftGraveaccent,
            "\x1b[95;68u": MoreKeys.AltShiftLowline,
            "\x1b[27;68;95~": MoreKeys.AltShiftLowline,
            "\x1b[94;68u": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[27;68;94~": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[93;68u": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[27;68;93~": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[92;68u": MoreKeys.AltShiftReversesolidus,
            "\x1b[27;68;92~": MoreKeys.AltShiftReversesolidus,
            "\x1b[91;68u": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[27;68;91~": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[90;68u": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[27;68;90~": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[89;68u": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[27;68;89~": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[88;68u": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[27;68;88~": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[87;68u": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[27;68;87~": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[86;68u": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[27;68;86~": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[85;68u": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[27;68;85~": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[84;68u": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[27;68;84~": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[83;68u": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[27;68;83~": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[82;68u": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[27;68;82~": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[81;68u": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[27;68;81~": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[80;68u": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[27;68;80~": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[79;68u": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[27;68;79~": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[78;68u": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[27;68;78~": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[77;68u": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[27;68;77~": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[76;68u": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[27;68;76~": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[75;68u": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[27;68;75~": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[74;68u": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[27;68;74~": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[73;68u": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[27;68;73~": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[72;68u": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[27;68;72~": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[71;68u": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[27;68;71~": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[70;68u": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[27;68;70~": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[69;68u": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[27;68;69~": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[68;68u": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[27;68;68~": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[67;68u": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[27;68;67~": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[66;68u": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[27;68;66~": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[65;68u": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[27;68;65~": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[64;68u": MoreKeys.AltShiftCommercialat,
            "\x1b[27;68;64~": MoreKeys.AltShiftCommercialat,
            "\x1b[63;68u": MoreKeys.AltShiftQuestionmark,
            "\x1b[27;68;63~": MoreKeys.AltShiftQuestionmark,
            "\x1b[62;68u": MoreKeys.AltShiftGreaterthansign,
            "\x1b[27;68;62~": MoreKeys.AltShiftGreaterthansign,
            "\x1b[61;68u": MoreKeys.AltShiftEqualssign,
            "\x1b[27;68;61~": MoreKeys.AltShiftEqualssign,
            "\x1b[60;68u": MoreKeys.AltShiftLessthansign,
            "\x1b[27;68;60~": MoreKeys.AltShiftLessthansign,
            "\x1b[59;68u": MoreKeys.AltShiftSemicolon,
            "\x1b[27;68;59~": MoreKeys.AltShiftSemicolon,
            "\x1b[58;68u": MoreKeys.AltShiftColon,
            "\x1b[27;68;58~": MoreKeys.AltShiftColon,
            "\x1b[57;68u": MoreKeys.AltShiftDigitnine,
            "\x1b[27;68;57~": MoreKeys.AltShiftDigitnine,
            "\x1b[56;68u": MoreKeys.AltShiftDigiteight,
            "\x1b[27;68;56~": MoreKeys.AltShiftDigiteight,
            "\x1b[55;68u": MoreKeys.AltShiftDigitseven,
            "\x1b[27;68;55~": MoreKeys.AltShiftDigitseven,
            "\x1b[54;68u": MoreKeys.AltShiftDigitsix,
            "\x1b[27;68;54~": MoreKeys.AltShiftDigitsix,
            "\x1b[53;68u": MoreKeys.AltShiftDigitfive,
            "\x1b[27;68;53~": MoreKeys.AltShiftDigitfive,
            "\x1b[52;68u": MoreKeys.AltShiftDigitfour,
            "\x1b[27;68;52~": MoreKeys.AltShiftDigitfour,
            "\x1b[51;68u": MoreKeys.AltShiftDigitthree,
            "\x1b[27;68;51~": MoreKeys.AltShiftDigitthree,
            "\x1b[50;68u": MoreKeys.AltShiftDigittwo,
            "\x1b[27;68;50~": MoreKeys.AltShiftDigittwo,
            "\x1b[49;68u": MoreKeys.AltShiftDigitone,
            "\x1b[27;68;49~": MoreKeys.AltShiftDigitone,
            "\x1b[48;68u": MoreKeys.AltShiftDigitzero,
            "\x1b[27;68;48~": MoreKeys.AltShiftDigitzero,
            "\x1b[47;68u": MoreKeys.AltShiftSolidus,
            "\x1b[27;68;47~": MoreKeys.AltShiftSolidus,
            "\x1b[46;68u": MoreKeys.AltShiftFullstop,
            "\x1b[27;68;46~": MoreKeys.AltShiftFullstop,
            "\x1b[45;68u": MoreKeys.AltShiftHyphenminus,
            "\x1b[27;68;45~": MoreKeys.AltShiftHyphenminus,
            "\x1b[44;68u": MoreKeys.AltShiftComma,
            "\x1b[27;68;44~": MoreKeys.AltShiftComma,
            "\x1b[43;68u": MoreKeys.AltShiftPlussign,
            "\x1b[27;68;43~": MoreKeys.AltShiftPlussign,
            "\x1b[42;68u": MoreKeys.AltShiftAsterisk,
            "\x1b[27;68;42~": MoreKeys.AltShiftAsterisk,
            "\x1b[41;68u": MoreKeys.AltShiftRightparenthesis,
            "\x1b[27;68;41~": MoreKeys.AltShiftRightparenthesis,
            "\x1b[40;68u": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[27;68;40~": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[39;68u": MoreKeys.AltShiftApostrophe,
            "\x1b[27;68;39~": MoreKeys.AltShiftApostrophe,
            "\x1b[38;68u": MoreKeys.AltShiftAmpersand,
            "\x1b[27;68;38~": MoreKeys.AltShiftAmpersand,
            "\x1b[37;68u": MoreKeys.AltShiftPercentsign,
            "\x1b[27;68;37~": MoreKeys.AltShiftPercentsign,
            "\x1b[36;68u": MoreKeys.AltShiftDollarsign,
            "\x1b[27;68;36~": MoreKeys.AltShiftDollarsign,
            "\x1b[35;68u": MoreKeys.AltShiftNumbersign,
            "\x1b[27;68;35~": MoreKeys.AltShiftNumbersign,
            "\x1b[34;68u": MoreKeys.AltShiftQuotationmark,
            "\x1b[27;68;34~": MoreKeys.AltShiftQuotationmark,
            "\x1b[33;68u": MoreKeys.AltShiftExclamationmark,
            "\x1b[27;68;33~": MoreKeys.AltShiftExclamationmark,
            "\x1b[127;68u": MoreKeys.AltShiftBackspace,
            "\x1b[27;68;127~": MoreKeys.AltShiftBackspace,
            "\x1b[2;68~": MoreKeys.AltShiftInsert,
            "\x1b[3;68~": MoreKeys.AltShiftDelete,
            "\x1b[5;68~": MoreKeys.AltShiftPageup,
            "\x1b[6;68~": MoreKeys.AltShiftPagedown,
            "\x1b[1;68A": MoreKeys.AltShiftUp,
            "\x1b[1;68B": MoreKeys.AltShiftDown,
            "\x1b[1;68C": MoreKeys.AltShiftRight,
            "\x1b[1;68D": MoreKeys.AltShiftLeft,
            "\x1b[1;68H": MoreKeys.AltShiftHome,
            "\x1b[1;68F": MoreKeys.AltShiftEnd,
            "\x1b[57361;68u": MoreKeys.AltShiftPrintscreen,
            "\x1b[57362;68u": MoreKeys.AltShiftPause,
            "\x1b[57363;68u": MoreKeys.AltShiftMenu,
            "\x1b[1;68P": MoreKeys.AltShiftF1,
            "\x1b[1;68Q": MoreKeys.AltShiftF2,
            "\x1b[13;68~": MoreKeys.AltShiftF3,
            "\x1b[1;68S": MoreKeys.AltShiftF4,
            "\x1b[15;68~": MoreKeys.AltShiftF5,
            "\x1b[17;68~": MoreKeys.AltShiftF6,
            "\x1b[18;68~": MoreKeys.AltShiftF7,
            "\x1b[19;68~": MoreKeys.AltShiftF8,
            "\x1b[20;68~": MoreKeys.AltShiftF9,
            "\x1b[21;68~": MoreKeys.AltShiftF10,
            "\x1b[23;68~": MoreKeys.AltShiftF11,
            "\x1b[24;68~": MoreKeys.AltShiftF12,
            "\x1b[9;132u": MoreKeys.AltShiftTab,
            "\x1b[27;132;9~": MoreKeys.AltShiftTab,
            "\x1b[13;132u": MoreKeys.AltShiftEnter,
            "\x1b[27;132;13~": MoreKeys.AltShiftEnter,
            "\x1b[27;132u": MoreKeys.AltShiftEscape,
            "\x1b[27;132;27~": MoreKeys.AltShiftEscape,
            "\x1b[32;132u": MoreKeys.AltShiftSpace,
            "\x1b[27;132;32~": MoreKeys.AltShiftSpace,
            "\x1b[126;132u": MoreKeys.AltShiftTilde,
            "\x1b[27;132;126~": MoreKeys.AltShiftTilde,
            "\x1b[125;132u": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[27;132;125~": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[124;132u": MoreKeys.AltShiftVerticalline,
            "\x1b[27;132;124~": MoreKeys.AltShiftVerticalline,
            "\x1b[123;132u": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[27;132;123~": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[122;132u": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[27;132;122~": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[121;132u": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[27;132;121~": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[120;132u": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[27;132;120~": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[119;132u": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[27;132;119~": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[118;132u": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[27;132;118~": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[117;132u": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[27;132;117~": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[116;132u": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[27;132;116~": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[115;132u": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[27;132;115~": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[114;132u": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[27;132;114~": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[113;132u": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[27;132;113~": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[112;132u": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[27;132;112~": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[111;132u": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[27;132;111~": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[110;132u": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[27;132;110~": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[109;132u": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[27;132;109~": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[108;132u": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[27;132;108~": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[107;132u": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[27;132;107~": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[106;132u": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[27;132;106~": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[105;132u": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[27;132;105~": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[104;132u": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[27;132;104~": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[103;132u": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[27;132;103~": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[102;132u": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[27;132;102~": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[101;132u": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[27;132;101~": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[100;132u": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[27;132;100~": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[99;132u": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[27;132;99~": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[98;132u": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[27;132;98~": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[97;132u": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[27;132;97~": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[96;132u": MoreKeys.AltShiftGraveaccent,
            "\x1b[27;132;96~": MoreKeys.AltShiftGraveaccent,
            "\x1b[95;132u": MoreKeys.AltShiftLowline,
            "\x1b[27;132;95~": MoreKeys.AltShiftLowline,
            "\x1b[94;132u": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[27;132;94~": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[93;132u": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[27;132;93~": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[92;132u": MoreKeys.AltShiftReversesolidus,
            "\x1b[27;132;92~": MoreKeys.AltShiftReversesolidus,
            "\x1b[91;132u": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[27;132;91~": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[90;132u": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[27;132;90~": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[89;132u": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[27;132;89~": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[88;132u": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[27;132;88~": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[87;132u": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[27;132;87~": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[86;132u": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[27;132;86~": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[85;132u": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[27;132;85~": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[84;132u": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[27;132;84~": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[83;132u": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[27;132;83~": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[82;132u": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[27;132;82~": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[81;132u": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[27;132;81~": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[80;132u": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[27;132;80~": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[79;132u": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[27;132;79~": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[78;132u": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[27;132;78~": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[77;132u": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[27;132;77~": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[76;132u": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[27;132;76~": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[75;132u": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[27;132;75~": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[74;132u": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[27;132;74~": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[73;132u": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[27;132;73~": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[72;132u": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[27;132;72~": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[71;132u": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[27;132;71~": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[70;132u": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[27;132;70~": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[69;132u": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[27;132;69~": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[68;132u": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[27;132;68~": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[67;132u": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[27;132;67~": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[66;132u": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[27;132;66~": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[65;132u": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[27;132;65~": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[64;132u": MoreKeys.AltShiftCommercialat,
            "\x1b[27;132;64~": MoreKeys.AltShiftCommercialat,
            "\x1b[63;132u": MoreKeys.AltShiftQuestionmark,
            "\x1b[27;132;63~": MoreKeys.AltShiftQuestionmark,
            "\x1b[62;132u": MoreKeys.AltShiftGreaterthansign,
            "\x1b[27;132;62~": MoreKeys.AltShiftGreaterthansign,
            "\x1b[61;132u": MoreKeys.AltShiftEqualssign,
            "\x1b[27;132;61~": MoreKeys.AltShiftEqualssign,
            "\x1b[60;132u": MoreKeys.AltShiftLessthansign,
            "\x1b[27;132;60~": MoreKeys.AltShiftLessthansign,
            "\x1b[59;132u": MoreKeys.AltShiftSemicolon,
            "\x1b[27;132;59~": MoreKeys.AltShiftSemicolon,
            "\x1b[58;132u": MoreKeys.AltShiftColon,
            "\x1b[27;132;58~": MoreKeys.AltShiftColon,
            "\x1b[57;132u": MoreKeys.AltShiftDigitnine,
            "\x1b[27;132;57~": MoreKeys.AltShiftDigitnine,
            "\x1b[56;132u": MoreKeys.AltShiftDigiteight,
            "\x1b[27;132;56~": MoreKeys.AltShiftDigiteight,
            "\x1b[55;132u": MoreKeys.AltShiftDigitseven,
            "\x1b[27;132;55~": MoreKeys.AltShiftDigitseven,
            "\x1b[54;132u": MoreKeys.AltShiftDigitsix,
            "\x1b[27;132;54~": MoreKeys.AltShiftDigitsix,
            "\x1b[53;132u": MoreKeys.AltShiftDigitfive,
            "\x1b[27;132;53~": MoreKeys.AltShiftDigitfive,
            "\x1b[52;132u": MoreKeys.AltShiftDigitfour,
            "\x1b[27;132;52~": MoreKeys.AltShiftDigitfour,
            "\x1b[51;132u": MoreKeys.AltShiftDigitthree,
            "\x1b[27;132;51~": MoreKeys.AltShiftDigitthree,
            "\x1b[50;132u": MoreKeys.AltShiftDigittwo,
            "\x1b[27;132;50~": MoreKeys.AltShiftDigittwo,
            "\x1b[49;132u": MoreKeys.AltShiftDigitone,
            "\x1b[27;132;49~": MoreKeys.AltShiftDigitone,
            "\x1b[48;132u": MoreKeys.AltShiftDigitzero,
            "\x1b[27;132;48~": MoreKeys.AltShiftDigitzero,
            "\x1b[47;132u": MoreKeys.AltShiftSolidus,
            "\x1b[27;132;47~": MoreKeys.AltShiftSolidus,
            "\x1b[46;132u": MoreKeys.AltShiftFullstop,
            "\x1b[27;132;46~": MoreKeys.AltShiftFullstop,
            "\x1b[45;132u": MoreKeys.AltShiftHyphenminus,
            "\x1b[27;132;45~": MoreKeys.AltShiftHyphenminus,
            "\x1b[44;132u": MoreKeys.AltShiftComma,
            "\x1b[27;132;44~": MoreKeys.AltShiftComma,
            "\x1b[43;132u": MoreKeys.AltShiftPlussign,
            "\x1b[27;132;43~": MoreKeys.AltShiftPlussign,
            "\x1b[42;132u": MoreKeys.AltShiftAsterisk,
            "\x1b[27;132;42~": MoreKeys.AltShiftAsterisk,
            "\x1b[41;132u": MoreKeys.AltShiftRightparenthesis,
            "\x1b[27;132;41~": MoreKeys.AltShiftRightparenthesis,
            "\x1b[40;132u": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[27;132;40~": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[39;132u": MoreKeys.AltShiftApostrophe,
            "\x1b[27;132;39~": MoreKeys.AltShiftApostrophe,
            "\x1b[38;132u": MoreKeys.AltShiftAmpersand,
            "\x1b[27;132;38~": MoreKeys.AltShiftAmpersand,
            "\x1b[37;132u": MoreKeys.AltShiftPercentsign,
            "\x1b[27;132;37~": MoreKeys.AltShiftPercentsign,
            "\x1b[36;132u": MoreKeys.AltShiftDollarsign,
            "\x1b[27;132;36~": MoreKeys.AltShiftDollarsign,
            "\x1b[35;132u": MoreKeys.AltShiftNumbersign,
            "\x1b[27;132;35~": MoreKeys.AltShiftNumbersign,
            "\x1b[34;132u": MoreKeys.AltShiftQuotationmark,
            "\x1b[27;132;34~": MoreKeys.AltShiftQuotationmark,
            "\x1b[33;132u": MoreKeys.AltShiftExclamationmark,
            "\x1b[27;132;33~": MoreKeys.AltShiftExclamationmark,
            "\x1b[127;132u": MoreKeys.AltShiftBackspace,
            "\x1b[27;132;127~": MoreKeys.AltShiftBackspace,
            "\x1b[2;132~": MoreKeys.AltShiftInsert,
            "\x1b[3;132~": MoreKeys.AltShiftDelete,
            "\x1b[5;132~": MoreKeys.AltShiftPageup,
            "\x1b[6;132~": MoreKeys.AltShiftPagedown,
            "\x1b[1;132A": MoreKeys.AltShiftUp,
            "\x1b[1;132B": MoreKeys.AltShiftDown,
            "\x1b[1;132C": MoreKeys.AltShiftRight,
            "\x1b[1;132D": MoreKeys.AltShiftLeft,
            "\x1b[1;132H": MoreKeys.AltShiftHome,
            "\x1b[1;132F": MoreKeys.AltShiftEnd,
            "\x1b[57361;132u": MoreKeys.AltShiftPrintscreen,
            "\x1b[57362;132u": MoreKeys.AltShiftPause,
            "\x1b[57363;132u": MoreKeys.AltShiftMenu,
            "\x1b[1;132P": MoreKeys.AltShiftF1,
            "\x1b[1;132Q": MoreKeys.AltShiftF2,
            "\x1b[13;132~": MoreKeys.AltShiftF3,
            "\x1b[1;132S": MoreKeys.AltShiftF4,
            "\x1b[15;132~": MoreKeys.AltShiftF5,
            "\x1b[17;132~": MoreKeys.AltShiftF6,
            "\x1b[18;132~": MoreKeys.AltShiftF7,
            "\x1b[19;132~": MoreKeys.AltShiftF8,
            "\x1b[20;132~": MoreKeys.AltShiftF9,
            "\x1b[21;132~": MoreKeys.AltShiftF10,
            "\x1b[23;132~": MoreKeys.AltShiftF11,
            "\x1b[24;132~": MoreKeys.AltShiftF12,
            "\x1b[9;70u": MoreKeys.ControlShiftTab,
            "\x1b[27;70;9~": MoreKeys.ControlShiftTab,
            "\x1b[13;70u": MoreKeys.ControlShiftEnter,
            "\x1b[27;70;13~": MoreKeys.ControlShiftEnter,
            "\x1b[27;70u": MoreKeys.ControlShiftEscape,
            "\x1b[27;70;27~": MoreKeys.ControlShiftEscape,
            "\x1b[32;70u": MoreKeys.ControlShiftSpace,
            "\x1b[27;70;32~": MoreKeys.ControlShiftSpace,
            "\x1b[126;70u": MoreKeys.ControlShiftTilde,
            "\x1b[27;70;126~": MoreKeys.ControlShiftTilde,
            "\x1b[125;70u": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[27;70;125~": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[124;70u": MoreKeys.ControlShiftVerticalline,
            "\x1b[27;70;124~": MoreKeys.ControlShiftVerticalline,
            "\x1b[123;70u": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[27;70;123~": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[122;70u": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[27;70;122~": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[121;70u": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[27;70;121~": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[120;70u": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[27;70;120~": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[119;70u": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[27;70;119~": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[118;70u": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[27;70;118~": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[117;70u": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[27;70;117~": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[116;70u": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[27;70;116~": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[115;70u": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[27;70;115~": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[114;70u": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[27;70;114~": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[113;70u": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[27;70;113~": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[112;70u": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[27;70;112~": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[111;70u": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[27;70;111~": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[110;70u": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[27;70;110~": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[109;70u": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[27;70;109~": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[108;70u": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[27;70;108~": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[107;70u": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[27;70;107~": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[106;70u": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[27;70;106~": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[105;70u": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[27;70;105~": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[104;70u": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[27;70;104~": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[103;70u": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[27;70;103~": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[102;70u": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[27;70;102~": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[101;70u": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[27;70;101~": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[100;70u": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[27;70;100~": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[99;70u": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[27;70;99~": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[98;70u": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[27;70;98~": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[97;70u": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[27;70;97~": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[96;70u": MoreKeys.ControlShiftGraveaccent,
            "\x1b[27;70;96~": MoreKeys.ControlShiftGraveaccent,
            "\x1b[95;70u": MoreKeys.ControlShiftLowline,
            "\x1b[27;70;95~": MoreKeys.ControlShiftLowline,
            "\x1b[94;70u": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[27;70;94~": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[93;70u": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[27;70;93~": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[92;70u": MoreKeys.ControlShiftReversesolidus,
            "\x1b[27;70;92~": MoreKeys.ControlShiftReversesolidus,
            "\x1b[91;70u": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[27;70;91~": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[90;70u": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[27;70;90~": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[89;70u": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[27;70;89~": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[88;70u": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[27;70;88~": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[87;70u": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[27;70;87~": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[86;70u": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[27;70;86~": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[85;70u": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[27;70;85~": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[84;70u": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[27;70;84~": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[83;70u": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[27;70;83~": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[82;70u": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[27;70;82~": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[81;70u": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[27;70;81~": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[80;70u": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[27;70;80~": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[79;70u": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[27;70;79~": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[78;70u": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[27;70;78~": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[77;70u": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[27;70;77~": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[76;70u": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[27;70;76~": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[75;70u": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[27;70;75~": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[74;70u": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[27;70;74~": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[73;70u": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[27;70;73~": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[72;70u": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[27;70;72~": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[71;70u": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[27;70;71~": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[70;70u": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[27;70;70~": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[69;70u": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[27;70;69~": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[68;70u": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[27;70;68~": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[67;70u": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[27;70;67~": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[66;70u": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[27;70;66~": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[65;70u": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[27;70;65~": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[64;70u": MoreKeys.ControlShiftCommercialat,
            "\x1b[27;70;64~": MoreKeys.ControlShiftCommercialat,
            "\x1b[63;70u": MoreKeys.ControlShiftQuestionmark,
            "\x1b[27;70;63~": MoreKeys.ControlShiftQuestionmark,
            "\x1b[62;70u": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[27;70;62~": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[61;70u": MoreKeys.ControlShiftEqualssign,
            "\x1b[27;70;61~": MoreKeys.ControlShiftEqualssign,
            "\x1b[60;70u": MoreKeys.ControlShiftLessthansign,
            "\x1b[27;70;60~": MoreKeys.ControlShiftLessthansign,
            "\x1b[59;70u": MoreKeys.ControlShiftSemicolon,
            "\x1b[27;70;59~": MoreKeys.ControlShiftSemicolon,
            "\x1b[58;70u": MoreKeys.ControlShiftColon,
            "\x1b[27;70;58~": MoreKeys.ControlShiftColon,
            "\x1b[57;70u": MoreKeys.ControlShiftDigitnine,
            "\x1b[27;70;57~": MoreKeys.ControlShiftDigitnine,
            "\x1b[56;70u": MoreKeys.ControlShiftDigiteight,
            "\x1b[27;70;56~": MoreKeys.ControlShiftDigiteight,
            "\x1b[55;70u": MoreKeys.ControlShiftDigitseven,
            "\x1b[27;70;55~": MoreKeys.ControlShiftDigitseven,
            "\x1b[54;70u": MoreKeys.ControlShiftDigitsix,
            "\x1b[27;70;54~": MoreKeys.ControlShiftDigitsix,
            "\x1b[53;70u": MoreKeys.ControlShiftDigitfive,
            "\x1b[27;70;53~": MoreKeys.ControlShiftDigitfive,
            "\x1b[52;70u": MoreKeys.ControlShiftDigitfour,
            "\x1b[27;70;52~": MoreKeys.ControlShiftDigitfour,
            "\x1b[51;70u": MoreKeys.ControlShiftDigitthree,
            "\x1b[27;70;51~": MoreKeys.ControlShiftDigitthree,
            "\x1b[50;70u": MoreKeys.ControlShiftDigittwo,
            "\x1b[27;70;50~": MoreKeys.ControlShiftDigittwo,
            "\x1b[49;70u": MoreKeys.ControlShiftDigitone,
            "\x1b[27;70;49~": MoreKeys.ControlShiftDigitone,
            "\x1b[48;70u": MoreKeys.ControlShiftDigitzero,
            "\x1b[27;70;48~": MoreKeys.ControlShiftDigitzero,
            "\x1b[47;70u": MoreKeys.ControlShiftSolidus,
            "\x1b[27;70;47~": MoreKeys.ControlShiftSolidus,
            "\x1b[46;70u": MoreKeys.ControlShiftFullstop,
            "\x1b[27;70;46~": MoreKeys.ControlShiftFullstop,
            "\x1b[45;70u": MoreKeys.ControlShiftHyphenminus,
            "\x1b[27;70;45~": MoreKeys.ControlShiftHyphenminus,
            "\x1b[44;70u": MoreKeys.ControlShiftComma,
            "\x1b[27;70;44~": MoreKeys.ControlShiftComma,
            "\x1b[43;70u": MoreKeys.ControlShiftPlussign,
            "\x1b[27;70;43~": MoreKeys.ControlShiftPlussign,
            "\x1b[42;70u": MoreKeys.ControlShiftAsterisk,
            "\x1b[27;70;42~": MoreKeys.ControlShiftAsterisk,
            "\x1b[41;70u": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[27;70;41~": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[40;70u": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[27;70;40~": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[39;70u": MoreKeys.ControlShiftApostrophe,
            "\x1b[27;70;39~": MoreKeys.ControlShiftApostrophe,
            "\x1b[38;70u": MoreKeys.ControlShiftAmpersand,
            "\x1b[27;70;38~": MoreKeys.ControlShiftAmpersand,
            "\x1b[37;70u": MoreKeys.ControlShiftPercentsign,
            "\x1b[27;70;37~": MoreKeys.ControlShiftPercentsign,
            "\x1b[36;70u": MoreKeys.ControlShiftDollarsign,
            "\x1b[27;70;36~": MoreKeys.ControlShiftDollarsign,
            "\x1b[35;70u": MoreKeys.ControlShiftNumbersign,
            "\x1b[27;70;35~": MoreKeys.ControlShiftNumbersign,
            "\x1b[34;70u": MoreKeys.ControlShiftQuotationmark,
            "\x1b[27;70;34~": MoreKeys.ControlShiftQuotationmark,
            "\x1b[33;70u": MoreKeys.ControlShiftExclamationmark,
            "\x1b[27;70;33~": MoreKeys.ControlShiftExclamationmark,
            "\x1b[127;70u": MoreKeys.ControlShiftBackspace,
            "\x1b[27;70;127~": MoreKeys.ControlShiftBackspace,
            "\x1b[2;70~": MoreKeys.ControlShiftInsert,
            "\x1b[3;70~": MoreKeys.ControlShiftDelete,
            "\x1b[5;70~": MoreKeys.ControlShiftPageup,
            "\x1b[6;70~": MoreKeys.ControlShiftPagedown,
            "\x1b[1;70A": MoreKeys.ControlShiftUp,
            "\x1b[1;70B": MoreKeys.ControlShiftDown,
            "\x1b[1;70C": MoreKeys.ControlShiftRight,
            "\x1b[1;70D": MoreKeys.ControlShiftLeft,
            "\x1b[1;70H": MoreKeys.ControlShiftHome,
            "\x1b[1;70F": MoreKeys.ControlShiftEnd,
            "\x1b[57361;70u": MoreKeys.ControlShiftPrintscreen,
            "\x1b[57362;70u": MoreKeys.ControlShiftPause,
            "\x1b[57363;70u": MoreKeys.ControlShiftMenu,
            "\x1b[1;70P": MoreKeys.ControlShiftF1,
            "\x1b[1;70Q": MoreKeys.ControlShiftF2,
            "\x1b[13;70~": MoreKeys.ControlShiftF3,
            "\x1b[1;70S": MoreKeys.ControlShiftF4,
            "\x1b[15;70~": MoreKeys.ControlShiftF5,
            "\x1b[17;70~": MoreKeys.ControlShiftF6,
            "\x1b[18;70~": MoreKeys.ControlShiftF7,
            "\x1b[19;70~": MoreKeys.ControlShiftF8,
            "\x1b[20;70~": MoreKeys.ControlShiftF9,
            "\x1b[21;70~": MoreKeys.ControlShiftF10,
            "\x1b[23;70~": MoreKeys.ControlShiftF11,
            "\x1b[24;70~": MoreKeys.ControlShiftF12,
            "\x1b[9;134u": MoreKeys.ControlShiftTab,
            "\x1b[27;134;9~": MoreKeys.ControlShiftTab,
            "\x1b[13;134u": MoreKeys.ControlShiftEnter,
            "\x1b[27;134;13~": MoreKeys.ControlShiftEnter,
            "\x1b[27;134u": MoreKeys.ControlShiftEscape,
            "\x1b[27;134;27~": MoreKeys.ControlShiftEscape,
            "\x1b[32;134u": MoreKeys.ControlShiftSpace,
            "\x1b[27;134;32~": MoreKeys.ControlShiftSpace,
            "\x1b[126;134u": MoreKeys.ControlShiftTilde,
            "\x1b[27;134;126~": MoreKeys.ControlShiftTilde,
            "\x1b[125;134u": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[27;134;125~": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[124;134u": MoreKeys.ControlShiftVerticalline,
            "\x1b[27;134;124~": MoreKeys.ControlShiftVerticalline,
            "\x1b[123;134u": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[27;134;123~": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[122;134u": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[27;134;122~": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[121;134u": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[27;134;121~": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[120;134u": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[27;134;120~": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[119;134u": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[27;134;119~": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[118;134u": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[27;134;118~": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[117;134u": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[27;134;117~": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[116;134u": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[27;134;116~": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[115;134u": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[27;134;115~": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[114;134u": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[27;134;114~": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[113;134u": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[27;134;113~": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[112;134u": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[27;134;112~": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[111;134u": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[27;134;111~": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[110;134u": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[27;134;110~": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[109;134u": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[27;134;109~": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[108;134u": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[27;134;108~": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[107;134u": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[27;134;107~": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[106;134u": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[27;134;106~": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[105;134u": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[27;134;105~": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[104;134u": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[27;134;104~": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[103;134u": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[27;134;103~": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[102;134u": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[27;134;102~": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[101;134u": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[27;134;101~": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[100;134u": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[27;134;100~": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[99;134u": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[27;134;99~": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[98;134u": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[27;134;98~": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[97;134u": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[27;134;97~": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[96;134u": MoreKeys.ControlShiftGraveaccent,
            "\x1b[27;134;96~": MoreKeys.ControlShiftGraveaccent,
            "\x1b[95;134u": MoreKeys.ControlShiftLowline,
            "\x1b[27;134;95~": MoreKeys.ControlShiftLowline,
            "\x1b[94;134u": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[27;134;94~": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[93;134u": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[27;134;93~": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[92;134u": MoreKeys.ControlShiftReversesolidus,
            "\x1b[27;134;92~": MoreKeys.ControlShiftReversesolidus,
            "\x1b[91;134u": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[27;134;91~": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[90;134u": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[27;134;90~": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[89;134u": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[27;134;89~": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[88;134u": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[27;134;88~": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[87;134u": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[27;134;87~": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[86;134u": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[27;134;86~": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[85;134u": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[27;134;85~": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[84;134u": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[27;134;84~": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[83;134u": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[27;134;83~": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[82;134u": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[27;134;82~": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[81;134u": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[27;134;81~": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[80;134u": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[27;134;80~": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[79;134u": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[27;134;79~": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[78;134u": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[27;134;78~": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[77;134u": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[27;134;77~": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[76;134u": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[27;134;76~": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[75;134u": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[27;134;75~": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[74;134u": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[27;134;74~": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[73;134u": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[27;134;73~": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[72;134u": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[27;134;72~": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[71;134u": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[27;134;71~": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[70;134u": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[27;134;70~": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[69;134u": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[27;134;69~": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[68;134u": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[27;134;68~": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[67;134u": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[27;134;67~": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[66;134u": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[27;134;66~": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[65;134u": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[27;134;65~": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[64;134u": MoreKeys.ControlShiftCommercialat,
            "\x1b[27;134;64~": MoreKeys.ControlShiftCommercialat,
            "\x1b[63;134u": MoreKeys.ControlShiftQuestionmark,
            "\x1b[27;134;63~": MoreKeys.ControlShiftQuestionmark,
            "\x1b[62;134u": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[27;134;62~": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[61;134u": MoreKeys.ControlShiftEqualssign,
            "\x1b[27;134;61~": MoreKeys.ControlShiftEqualssign,
            "\x1b[60;134u": MoreKeys.ControlShiftLessthansign,
            "\x1b[27;134;60~": MoreKeys.ControlShiftLessthansign,
            "\x1b[59;134u": MoreKeys.ControlShiftSemicolon,
            "\x1b[27;134;59~": MoreKeys.ControlShiftSemicolon,
            "\x1b[58;134u": MoreKeys.ControlShiftColon,
            "\x1b[27;134;58~": MoreKeys.ControlShiftColon,
            "\x1b[57;134u": MoreKeys.ControlShiftDigitnine,
            "\x1b[27;134;57~": MoreKeys.ControlShiftDigitnine,
            "\x1b[56;134u": MoreKeys.ControlShiftDigiteight,
            "\x1b[27;134;56~": MoreKeys.ControlShiftDigiteight,
            "\x1b[55;134u": MoreKeys.ControlShiftDigitseven,
            "\x1b[27;134;55~": MoreKeys.ControlShiftDigitseven,
            "\x1b[54;134u": MoreKeys.ControlShiftDigitsix,
            "\x1b[27;134;54~": MoreKeys.ControlShiftDigitsix,
            "\x1b[53;134u": MoreKeys.ControlShiftDigitfive,
            "\x1b[27;134;53~": MoreKeys.ControlShiftDigitfive,
            "\x1b[52;134u": MoreKeys.ControlShiftDigitfour,
            "\x1b[27;134;52~": MoreKeys.ControlShiftDigitfour,
            "\x1b[51;134u": MoreKeys.ControlShiftDigitthree,
            "\x1b[27;134;51~": MoreKeys.ControlShiftDigitthree,
            "\x1b[50;134u": MoreKeys.ControlShiftDigittwo,
            "\x1b[27;134;50~": MoreKeys.ControlShiftDigittwo,
            "\x1b[49;134u": MoreKeys.ControlShiftDigitone,
            "\x1b[27;134;49~": MoreKeys.ControlShiftDigitone,
            "\x1b[48;134u": MoreKeys.ControlShiftDigitzero,
            "\x1b[27;134;48~": MoreKeys.ControlShiftDigitzero,
            "\x1b[47;134u": MoreKeys.ControlShiftSolidus,
            "\x1b[27;134;47~": MoreKeys.ControlShiftSolidus,
            "\x1b[46;134u": MoreKeys.ControlShiftFullstop,
            "\x1b[27;134;46~": MoreKeys.ControlShiftFullstop,
            "\x1b[45;134u": MoreKeys.ControlShiftHyphenminus,
            "\x1b[27;134;45~": MoreKeys.ControlShiftHyphenminus,
            "\x1b[44;134u": MoreKeys.ControlShiftComma,
            "\x1b[27;134;44~": MoreKeys.ControlShiftComma,
            "\x1b[43;134u": MoreKeys.ControlShiftPlussign,
            "\x1b[27;134;43~": MoreKeys.ControlShiftPlussign,
            "\x1b[42;134u": MoreKeys.ControlShiftAsterisk,
            "\x1b[27;134;42~": MoreKeys.ControlShiftAsterisk,
            "\x1b[41;134u": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[27;134;41~": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[40;134u": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[27;134;40~": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[39;134u": MoreKeys.ControlShiftApostrophe,
            "\x1b[27;134;39~": MoreKeys.ControlShiftApostrophe,
            "\x1b[38;134u": MoreKeys.ControlShiftAmpersand,
            "\x1b[27;134;38~": MoreKeys.ControlShiftAmpersand,
            "\x1b[37;134u": MoreKeys.ControlShiftPercentsign,
            "\x1b[27;134;37~": MoreKeys.ControlShiftPercentsign,
            "\x1b[36;134u": MoreKeys.ControlShiftDollarsign,
            "\x1b[27;134;36~": MoreKeys.ControlShiftDollarsign,
            "\x1b[35;134u": MoreKeys.ControlShiftNumbersign,
            "\x1b[27;134;35~": MoreKeys.ControlShiftNumbersign,
            "\x1b[34;134u": MoreKeys.ControlShiftQuotationmark,
            "\x1b[27;134;34~": MoreKeys.ControlShiftQuotationmark,
            "\x1b[33;134u": MoreKeys.ControlShiftExclamationmark,
            "\x1b[27;134;33~": MoreKeys.ControlShiftExclamationmark,
            "\x1b[127;134u": MoreKeys.ControlShiftBackspace,
            "\x1b[27;134;127~": MoreKeys.ControlShiftBackspace,
            "\x1b[2;134~": MoreKeys.ControlShiftInsert,
            "\x1b[3;134~": MoreKeys.ControlShiftDelete,
            "\x1b[5;134~": MoreKeys.ControlShiftPageup,
            "\x1b[6;134~": MoreKeys.ControlShiftPagedown,
            "\x1b[1;134A": MoreKeys.ControlShiftUp,
            "\x1b[1;134B": MoreKeys.ControlShiftDown,
            "\x1b[1;134C": MoreKeys.ControlShiftRight,
            "\x1b[1;134D": MoreKeys.ControlShiftLeft,
            "\x1b[1;134H": MoreKeys.ControlShiftHome,
            "\x1b[1;134F": MoreKeys.ControlShiftEnd,
            "\x1b[57361;134u": MoreKeys.ControlShiftPrintscreen,
            "\x1b[57362;134u": MoreKeys.ControlShiftPause,
            "\x1b[57363;134u": MoreKeys.ControlShiftMenu,
            "\x1b[1;134P": MoreKeys.ControlShiftF1,
            "\x1b[1;134Q": MoreKeys.ControlShiftF2,
            "\x1b[13;134~": MoreKeys.ControlShiftF3,
            "\x1b[1;134S": MoreKeys.ControlShiftF4,
            "\x1b[15;134~": MoreKeys.ControlShiftF5,
            "\x1b[17;134~": MoreKeys.ControlShiftF6,
            "\x1b[18;134~": MoreKeys.ControlShiftF7,
            "\x1b[19;134~": MoreKeys.ControlShiftF8,
            "\x1b[20;134~": MoreKeys.ControlShiftF9,
            "\x1b[21;134~": MoreKeys.ControlShiftF10,
            "\x1b[23;134~": MoreKeys.ControlShiftF11,
            "\x1b[24;134~": MoreKeys.ControlShiftF12,
            "\x1b[9;194u": MoreKeys.ShiftTab,
            "\x1b[27;194;9~": MoreKeys.ShiftTab,
            "\x1b[13;194u": MoreKeys.ShiftEnter,
            "\x1b[27;194;13~": MoreKeys.ShiftEnter,
            "\x1b[27;194u": MoreKeys.ShiftEscape,
            "\x1b[27;194;27~": MoreKeys.ShiftEscape,
            "\x1b[32;194u": MoreKeys.ShiftSpace,
            "\x1b[27;194;32~": MoreKeys.ShiftSpace,
            "\x1b[126;194u": MoreKeys.ShiftTilde,
            "\x1b[27;194;126~": MoreKeys.ShiftTilde,
            "\x1b[125;194u": MoreKeys.ShiftRightcurlybracket,
            "\x1b[27;194;125~": MoreKeys.ShiftRightcurlybracket,
            "\x1b[124;194u": MoreKeys.ShiftVerticalline,
            "\x1b[27;194;124~": MoreKeys.ShiftVerticalline,
            "\x1b[123;194u": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[27;194;123~": MoreKeys.ShiftLeftcurlybracket,
            "\x1b[122;194u": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[27;194;122~": MoreKeys.ShiftLatinsmallletterz,
            "\x1b[121;194u": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[27;194;121~": MoreKeys.ShiftLatinsmalllettery,
            "\x1b[120;194u": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[27;194;120~": MoreKeys.ShiftLatinsmallletterx,
            "\x1b[119;194u": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[27;194;119~": MoreKeys.ShiftLatinsmallletterw,
            "\x1b[118;194u": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[27;194;118~": MoreKeys.ShiftLatinsmallletterv,
            "\x1b[117;194u": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[27;194;117~": MoreKeys.ShiftLatinsmallletteru,
            "\x1b[116;194u": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[27;194;116~": MoreKeys.ShiftLatinsmalllettert,
            "\x1b[115;194u": MoreKeys.ShiftLatinsmallletters,
            "\x1b[27;194;115~": MoreKeys.ShiftLatinsmallletters,
            "\x1b[114;194u": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[27;194;114~": MoreKeys.ShiftLatinsmallletterr,
            "\x1b[113;194u": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[27;194;113~": MoreKeys.ShiftLatinsmallletterq,
            "\x1b[112;194u": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[27;194;112~": MoreKeys.ShiftLatinsmallletterp,
            "\x1b[111;194u": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[27;194;111~": MoreKeys.ShiftLatinsmalllettero,
            "\x1b[110;194u": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[27;194;110~": MoreKeys.ShiftLatinsmalllettern,
            "\x1b[109;194u": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[27;194;109~": MoreKeys.ShiftLatinsmallletterm,
            "\x1b[108;194u": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[27;194;108~": MoreKeys.ShiftLatinsmallletterl,
            "\x1b[107;194u": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[27;194;107~": MoreKeys.ShiftLatinsmallletterk,
            "\x1b[106;194u": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[27;194;106~": MoreKeys.ShiftLatinsmallletterj,
            "\x1b[105;194u": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[27;194;105~": MoreKeys.ShiftLatinsmallletteri,
            "\x1b[104;194u": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[27;194;104~": MoreKeys.ShiftLatinsmallletterh,
            "\x1b[103;194u": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[27;194;103~": MoreKeys.ShiftLatinsmallletterg,
            "\x1b[102;194u": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[27;194;102~": MoreKeys.ShiftLatinsmallletterf,
            "\x1b[101;194u": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[27;194;101~": MoreKeys.ShiftLatinsmalllettere,
            "\x1b[100;194u": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[27;194;100~": MoreKeys.ShiftLatinsmallletterd,
            "\x1b[99;194u": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[27;194;99~": MoreKeys.ShiftLatinsmallletterc,
            "\x1b[98;194u": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[27;194;98~": MoreKeys.ShiftLatinsmallletterb,
            "\x1b[97;194u": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[27;194;97~": MoreKeys.ShiftLatinsmalllettera,
            "\x1b[96;194u": MoreKeys.ShiftGraveaccent,
            "\x1b[27;194;96~": MoreKeys.ShiftGraveaccent,
            "\x1b[95;194u": MoreKeys.ShiftLowline,
            "\x1b[27;194;95~": MoreKeys.ShiftLowline,
            "\x1b[94;194u": MoreKeys.ShiftCircumflexaccent,
            "\x1b[27;194;94~": MoreKeys.ShiftCircumflexaccent,
            "\x1b[93;194u": MoreKeys.ShiftRightsquarebracket,
            "\x1b[27;194;93~": MoreKeys.ShiftRightsquarebracket,
            "\x1b[92;194u": MoreKeys.ShiftReversesolidus,
            "\x1b[27;194;92~": MoreKeys.ShiftReversesolidus,
            "\x1b[91;194u": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[27;194;91~": MoreKeys.ShiftLeftsquarebracket,
            "\x1b[90;194u": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[27;194;90~": MoreKeys.ShiftLatincapitalletterz,
            "\x1b[89;194u": MoreKeys.ShiftLatincapitallettery,
            "\x1b[27;194;89~": MoreKeys.ShiftLatincapitallettery,
            "\x1b[88;194u": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[27;194;88~": MoreKeys.ShiftLatincapitalletterx,
            "\x1b[87;194u": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[27;194;87~": MoreKeys.ShiftLatincapitalletterw,
            "\x1b[86;194u": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[27;194;86~": MoreKeys.ShiftLatincapitalletterv,
            "\x1b[85;194u": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[27;194;85~": MoreKeys.ShiftLatincapitalletteru,
            "\x1b[84;194u": MoreKeys.ShiftLatincapitallettert,
            "\x1b[27;194;84~": MoreKeys.ShiftLatincapitallettert,
            "\x1b[83;194u": MoreKeys.ShiftLatincapitalletters,
            "\x1b[27;194;83~": MoreKeys.ShiftLatincapitalletters,
            "\x1b[82;194u": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[27;194;82~": MoreKeys.ShiftLatincapitalletterr,
            "\x1b[81;194u": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[27;194;81~": MoreKeys.ShiftLatincapitalletterq,
            "\x1b[80;194u": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[27;194;80~": MoreKeys.ShiftLatincapitalletterp,
            "\x1b[79;194u": MoreKeys.ShiftLatincapitallettero,
            "\x1b[27;194;79~": MoreKeys.ShiftLatincapitallettero,
            "\x1b[78;194u": MoreKeys.ShiftLatincapitallettern,
            "\x1b[27;194;78~": MoreKeys.ShiftLatincapitallettern,
            "\x1b[77;194u": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[27;194;77~": MoreKeys.ShiftLatincapitalletterm,
            "\x1b[76;194u": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[27;194;76~": MoreKeys.ShiftLatincapitalletterl,
            "\x1b[75;194u": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[27;194;75~": MoreKeys.ShiftLatincapitalletterk,
            "\x1b[74;194u": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[27;194;74~": MoreKeys.ShiftLatincapitalletterj,
            "\x1b[73;194u": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[27;194;73~": MoreKeys.ShiftLatincapitalletteri,
            "\x1b[72;194u": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[27;194;72~": MoreKeys.ShiftLatincapitalletterh,
            "\x1b[71;194u": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[27;194;71~": MoreKeys.ShiftLatincapitalletterg,
            "\x1b[70;194u": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[27;194;70~": MoreKeys.ShiftLatincapitalletterf,
            "\x1b[69;194u": MoreKeys.ShiftLatincapitallettere,
            "\x1b[27;194;69~": MoreKeys.ShiftLatincapitallettere,
            "\x1b[68;194u": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[27;194;68~": MoreKeys.ShiftLatincapitalletterd,
            "\x1b[67;194u": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[27;194;67~": MoreKeys.ShiftLatincapitalletterc,
            "\x1b[66;194u": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[27;194;66~": MoreKeys.ShiftLatincapitalletterb,
            "\x1b[65;194u": MoreKeys.ShiftLatincapitallettera,
            "\x1b[27;194;65~": MoreKeys.ShiftLatincapitallettera,
            "\x1b[64;194u": MoreKeys.ShiftCommercialat,
            "\x1b[27;194;64~": MoreKeys.ShiftCommercialat,
            "\x1b[63;194u": MoreKeys.ShiftQuestionmark,
            "\x1b[27;194;63~": MoreKeys.ShiftQuestionmark,
            "\x1b[62;194u": MoreKeys.ShiftGreaterthansign,
            "\x1b[27;194;62~": MoreKeys.ShiftGreaterthansign,
            "\x1b[61;194u": MoreKeys.ShiftEqualssign,
            "\x1b[27;194;61~": MoreKeys.ShiftEqualssign,
            "\x1b[60;194u": MoreKeys.ShiftLessthansign,
            "\x1b[27;194;60~": MoreKeys.ShiftLessthansign,
            "\x1b[59;194u": MoreKeys.ShiftSemicolon,
            "\x1b[27;194;59~": MoreKeys.ShiftSemicolon,
            "\x1b[58;194u": MoreKeys.ShiftColon,
            "\x1b[27;194;58~": MoreKeys.ShiftColon,
            "\x1b[57;194u": MoreKeys.ShiftDigitnine,
            "\x1b[27;194;57~": MoreKeys.ShiftDigitnine,
            "\x1b[56;194u": MoreKeys.ShiftDigiteight,
            "\x1b[27;194;56~": MoreKeys.ShiftDigiteight,
            "\x1b[55;194u": MoreKeys.ShiftDigitseven,
            "\x1b[27;194;55~": MoreKeys.ShiftDigitseven,
            "\x1b[54;194u": MoreKeys.ShiftDigitsix,
            "\x1b[27;194;54~": MoreKeys.ShiftDigitsix,
            "\x1b[53;194u": MoreKeys.ShiftDigitfive,
            "\x1b[27;194;53~": MoreKeys.ShiftDigitfive,
            "\x1b[52;194u": MoreKeys.ShiftDigitfour,
            "\x1b[27;194;52~": MoreKeys.ShiftDigitfour,
            "\x1b[51;194u": MoreKeys.ShiftDigitthree,
            "\x1b[27;194;51~": MoreKeys.ShiftDigitthree,
            "\x1b[50;194u": MoreKeys.ShiftDigittwo,
            "\x1b[27;194;50~": MoreKeys.ShiftDigittwo,
            "\x1b[49;194u": MoreKeys.ShiftDigitone,
            "\x1b[27;194;49~": MoreKeys.ShiftDigitone,
            "\x1b[48;194u": MoreKeys.ShiftDigitzero,
            "\x1b[27;194;48~": MoreKeys.ShiftDigitzero,
            "\x1b[47;194u": MoreKeys.ShiftSolidus,
            "\x1b[27;194;47~": MoreKeys.ShiftSolidus,
            "\x1b[46;194u": MoreKeys.ShiftFullstop,
            "\x1b[27;194;46~": MoreKeys.ShiftFullstop,
            "\x1b[45;194u": MoreKeys.ShiftHyphenminus,
            "\x1b[27;194;45~": MoreKeys.ShiftHyphenminus,
            "\x1b[44;194u": MoreKeys.ShiftComma,
            "\x1b[27;194;44~": MoreKeys.ShiftComma,
            "\x1b[43;194u": MoreKeys.ShiftPlussign,
            "\x1b[27;194;43~": MoreKeys.ShiftPlussign,
            "\x1b[42;194u": MoreKeys.ShiftAsterisk,
            "\x1b[27;194;42~": MoreKeys.ShiftAsterisk,
            "\x1b[41;194u": MoreKeys.ShiftRightparenthesis,
            "\x1b[27;194;41~": MoreKeys.ShiftRightparenthesis,
            "\x1b[40;194u": MoreKeys.ShiftLeftparenthesis,
            "\x1b[27;194;40~": MoreKeys.ShiftLeftparenthesis,
            "\x1b[39;194u": MoreKeys.ShiftApostrophe,
            "\x1b[27;194;39~": MoreKeys.ShiftApostrophe,
            "\x1b[38;194u": MoreKeys.ShiftAmpersand,
            "\x1b[27;194;38~": MoreKeys.ShiftAmpersand,
            "\x1b[37;194u": MoreKeys.ShiftPercentsign,
            "\x1b[27;194;37~": MoreKeys.ShiftPercentsign,
            "\x1b[36;194u": MoreKeys.ShiftDollarsign,
            "\x1b[27;194;36~": MoreKeys.ShiftDollarsign,
            "\x1b[35;194u": MoreKeys.ShiftNumbersign,
            "\x1b[27;194;35~": MoreKeys.ShiftNumbersign,
            "\x1b[34;194u": MoreKeys.ShiftQuotationmark,
            "\x1b[27;194;34~": MoreKeys.ShiftQuotationmark,
            "\x1b[33;194u": MoreKeys.ShiftExclamationmark,
            "\x1b[27;194;33~": MoreKeys.ShiftExclamationmark,
            "\x1b[127;194u": MoreKeys.ShiftBackspace,
            "\x1b[27;194;127~": MoreKeys.ShiftBackspace,
            "\x1b[2;194~": MoreKeys.ShiftInsert,
            "\x1b[3;194~": MoreKeys.ShiftDelete,
            "\x1b[5;194~": MoreKeys.ShiftPageup,
            "\x1b[6;194~": MoreKeys.ShiftPagedown,
            "\x1b[1;194A": MoreKeys.ShiftUp,
            "\x1b[1;194B": MoreKeys.ShiftDown,
            "\x1b[1;194C": MoreKeys.ShiftRight,
            "\x1b[1;194D": MoreKeys.ShiftLeft,
            "\x1b[1;194H": MoreKeys.ShiftHome,
            "\x1b[1;194F": MoreKeys.ShiftEnd,
            "\x1b[57361;194u": MoreKeys.ShiftPrintscreen,
            "\x1b[57362;194u": MoreKeys.ShiftPause,
            "\x1b[57363;194u": MoreKeys.ShiftMenu,
            "\x1b[1;194P": MoreKeys.ShiftF1,
            "\x1b[1;194Q": MoreKeys.ShiftF2,
            "\x1b[13;194~": MoreKeys.ShiftF3,
            "\x1b[1;194S": MoreKeys.ShiftF4,
            "\x1b[15;194~": MoreKeys.ShiftF5,
            "\x1b[17;194~": MoreKeys.ShiftF6,
            "\x1b[18;194~": MoreKeys.ShiftF7,
            "\x1b[19;194~": MoreKeys.ShiftF8,
            "\x1b[20;194~": MoreKeys.ShiftF9,
            "\x1b[21;194~": MoreKeys.ShiftF10,
            "\x1b[23;194~": MoreKeys.ShiftF11,
            "\x1b[24;194~": MoreKeys.ShiftF12,
            "\x1b[9;71u": MoreKeys.ControlAltTab,
            "\x1b[27;71;9~": MoreKeys.ControlAltTab,
            "\x1b[13;71u": MoreKeys.ControlAltEnter,
            "\x1b[27;71;13~": MoreKeys.ControlAltEnter,
            "\x1b[27;71u": MoreKeys.ControlAltEscape,
            "\x1b[27;71;27~": MoreKeys.ControlAltEscape,
            "\x1b[32;71u": MoreKeys.ControlAltSpace,
            "\x1b[27;71;32~": MoreKeys.ControlAltSpace,
            "\x1b[126;71u": MoreKeys.ControlAltTilde,
            "\x1b[27;71;126~": MoreKeys.ControlAltTilde,
            "\x1b[125;71u": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[27;71;125~": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[124;71u": MoreKeys.ControlAltVerticalline,
            "\x1b[27;71;124~": MoreKeys.ControlAltVerticalline,
            "\x1b[123;71u": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[27;71;123~": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[122;71u": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[27;71;122~": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[121;71u": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[27;71;121~": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[120;71u": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[27;71;120~": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[119;71u": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[27;71;119~": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[118;71u": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[27;71;118~": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[117;71u": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[27;71;117~": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[116;71u": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[27;71;116~": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[115;71u": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[27;71;115~": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[114;71u": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[27;71;114~": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[113;71u": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[27;71;113~": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[112;71u": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[27;71;112~": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[111;71u": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[27;71;111~": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[110;71u": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[27;71;110~": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[109;71u": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[27;71;109~": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[108;71u": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[27;71;108~": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[107;71u": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[27;71;107~": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[106;71u": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[27;71;106~": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[105;71u": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[27;71;105~": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[104;71u": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[27;71;104~": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[103;71u": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[27;71;103~": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[102;71u": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[27;71;102~": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[101;71u": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[27;71;101~": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[100;71u": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[27;71;100~": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[99;71u": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[27;71;99~": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[98;71u": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[27;71;98~": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[97;71u": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[27;71;97~": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[96;71u": MoreKeys.ControlAltGraveaccent,
            "\x1b[27;71;96~": MoreKeys.ControlAltGraveaccent,
            "\x1b[95;71u": MoreKeys.ControlAltLowline,
            "\x1b[27;71;95~": MoreKeys.ControlAltLowline,
            "\x1b[94;71u": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[27;71;94~": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[93;71u": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[27;71;93~": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[92;71u": MoreKeys.ControlAltReversesolidus,
            "\x1b[27;71;92~": MoreKeys.ControlAltReversesolidus,
            "\x1b[91;71u": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[27;71;91~": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[90;71u": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[27;71;90~": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[89;71u": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[27;71;89~": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[88;71u": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[27;71;88~": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[87;71u": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[27;71;87~": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[86;71u": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[27;71;86~": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[85;71u": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[27;71;85~": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[84;71u": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[27;71;84~": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[83;71u": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[27;71;83~": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[82;71u": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[27;71;82~": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[81;71u": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[27;71;81~": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[80;71u": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[27;71;80~": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[79;71u": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[27;71;79~": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[78;71u": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[27;71;78~": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[77;71u": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[27;71;77~": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[76;71u": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[27;71;76~": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[75;71u": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[27;71;75~": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[74;71u": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[27;71;74~": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[73;71u": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[27;71;73~": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[72;71u": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[27;71;72~": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[71;71u": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[27;71;71~": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[70;71u": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[27;71;70~": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[69;71u": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[27;71;69~": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[68;71u": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[27;71;68~": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[67;71u": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[27;71;67~": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[66;71u": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[27;71;66~": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[65;71u": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[27;71;65~": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[64;71u": MoreKeys.ControlAltCommercialat,
            "\x1b[27;71;64~": MoreKeys.ControlAltCommercialat,
            "\x1b[63;71u": MoreKeys.ControlAltQuestionmark,
            "\x1b[27;71;63~": MoreKeys.ControlAltQuestionmark,
            "\x1b[62;71u": MoreKeys.ControlAltGreaterthansign,
            "\x1b[27;71;62~": MoreKeys.ControlAltGreaterthansign,
            "\x1b[61;71u": MoreKeys.ControlAltEqualssign,
            "\x1b[27;71;61~": MoreKeys.ControlAltEqualssign,
            "\x1b[60;71u": MoreKeys.ControlAltLessthansign,
            "\x1b[27;71;60~": MoreKeys.ControlAltLessthansign,
            "\x1b[59;71u": MoreKeys.ControlAltSemicolon,
            "\x1b[27;71;59~": MoreKeys.ControlAltSemicolon,
            "\x1b[58;71u": MoreKeys.ControlAltColon,
            "\x1b[27;71;58~": MoreKeys.ControlAltColon,
            "\x1b[57;71u": MoreKeys.ControlAltDigitnine,
            "\x1b[27;71;57~": MoreKeys.ControlAltDigitnine,
            "\x1b[56;71u": MoreKeys.ControlAltDigiteight,
            "\x1b[27;71;56~": MoreKeys.ControlAltDigiteight,
            "\x1b[55;71u": MoreKeys.ControlAltDigitseven,
            "\x1b[27;71;55~": MoreKeys.ControlAltDigitseven,
            "\x1b[54;71u": MoreKeys.ControlAltDigitsix,
            "\x1b[27;71;54~": MoreKeys.ControlAltDigitsix,
            "\x1b[53;71u": MoreKeys.ControlAltDigitfive,
            "\x1b[27;71;53~": MoreKeys.ControlAltDigitfive,
            "\x1b[52;71u": MoreKeys.ControlAltDigitfour,
            "\x1b[27;71;52~": MoreKeys.ControlAltDigitfour,
            "\x1b[51;71u": MoreKeys.ControlAltDigitthree,
            "\x1b[27;71;51~": MoreKeys.ControlAltDigitthree,
            "\x1b[50;71u": MoreKeys.ControlAltDigittwo,
            "\x1b[27;71;50~": MoreKeys.ControlAltDigittwo,
            "\x1b[49;71u": MoreKeys.ControlAltDigitone,
            "\x1b[27;71;49~": MoreKeys.ControlAltDigitone,
            "\x1b[48;71u": MoreKeys.ControlAltDigitzero,
            "\x1b[27;71;48~": MoreKeys.ControlAltDigitzero,
            "\x1b[47;71u": MoreKeys.ControlAltSolidus,
            "\x1b[27;71;47~": MoreKeys.ControlAltSolidus,
            "\x1b[46;71u": MoreKeys.ControlAltFullstop,
            "\x1b[27;71;46~": MoreKeys.ControlAltFullstop,
            "\x1b[45;71u": MoreKeys.ControlAltHyphenminus,
            "\x1b[27;71;45~": MoreKeys.ControlAltHyphenminus,
            "\x1b[44;71u": MoreKeys.ControlAltComma,
            "\x1b[27;71;44~": MoreKeys.ControlAltComma,
            "\x1b[43;71u": MoreKeys.ControlAltPlussign,
            "\x1b[27;71;43~": MoreKeys.ControlAltPlussign,
            "\x1b[42;71u": MoreKeys.ControlAltAsterisk,
            "\x1b[27;71;42~": MoreKeys.ControlAltAsterisk,
            "\x1b[41;71u": MoreKeys.ControlAltRightparenthesis,
            "\x1b[27;71;41~": MoreKeys.ControlAltRightparenthesis,
            "\x1b[40;71u": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[27;71;40~": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[39;71u": MoreKeys.ControlAltApostrophe,
            "\x1b[27;71;39~": MoreKeys.ControlAltApostrophe,
            "\x1b[38;71u": MoreKeys.ControlAltAmpersand,
            "\x1b[27;71;38~": MoreKeys.ControlAltAmpersand,
            "\x1b[37;71u": MoreKeys.ControlAltPercentsign,
            "\x1b[27;71;37~": MoreKeys.ControlAltPercentsign,
            "\x1b[36;71u": MoreKeys.ControlAltDollarsign,
            "\x1b[27;71;36~": MoreKeys.ControlAltDollarsign,
            "\x1b[35;71u": MoreKeys.ControlAltNumbersign,
            "\x1b[27;71;35~": MoreKeys.ControlAltNumbersign,
            "\x1b[34;71u": MoreKeys.ControlAltQuotationmark,
            "\x1b[27;71;34~": MoreKeys.ControlAltQuotationmark,
            "\x1b[33;71u": MoreKeys.ControlAltExclamationmark,
            "\x1b[27;71;33~": MoreKeys.ControlAltExclamationmark,
            "\x1b[127;71u": MoreKeys.ControlAltBackspace,
            "\x1b[27;71;127~": MoreKeys.ControlAltBackspace,
            "\x1b[2;71~": MoreKeys.ControlAltInsert,
            "\x1b[3;71~": MoreKeys.ControlAltDelete,
            "\x1b[5;71~": MoreKeys.ControlAltPageup,
            "\x1b[6;71~": MoreKeys.ControlAltPagedown,
            "\x1b[1;71A": MoreKeys.ControlAltUp,
            "\x1b[1;71B": MoreKeys.ControlAltDown,
            "\x1b[1;71C": MoreKeys.ControlAltRight,
            "\x1b[1;71D": MoreKeys.ControlAltLeft,
            "\x1b[1;71H": MoreKeys.ControlAltHome,
            "\x1b[1;71F": MoreKeys.ControlAltEnd,
            "\x1b[57361;71u": MoreKeys.ControlAltPrintscreen,
            "\x1b[57362;71u": MoreKeys.ControlAltPause,
            "\x1b[57363;71u": MoreKeys.ControlAltMenu,
            "\x1b[1;71P": MoreKeys.ControlAltF1,
            "\x1b[1;71Q": MoreKeys.ControlAltF2,
            "\x1b[13;71~": MoreKeys.ControlAltF3,
            "\x1b[1;71S": MoreKeys.ControlAltF4,
            "\x1b[15;71~": MoreKeys.ControlAltF5,
            "\x1b[17;71~": MoreKeys.ControlAltF6,
            "\x1b[18;71~": MoreKeys.ControlAltF7,
            "\x1b[19;71~": MoreKeys.ControlAltF8,
            "\x1b[20;71~": MoreKeys.ControlAltF9,
            "\x1b[21;71~": MoreKeys.ControlAltF10,
            "\x1b[23;71~": MoreKeys.ControlAltF11,
            "\x1b[24;71~": MoreKeys.ControlAltF12,
            "\x1b[9;135u": MoreKeys.ControlAltTab,
            "\x1b[27;135;9~": MoreKeys.ControlAltTab,
            "\x1b[13;135u": MoreKeys.ControlAltEnter,
            "\x1b[27;135;13~": MoreKeys.ControlAltEnter,
            "\x1b[27;135u": MoreKeys.ControlAltEscape,
            "\x1b[27;135;27~": MoreKeys.ControlAltEscape,
            "\x1b[32;135u": MoreKeys.ControlAltSpace,
            "\x1b[27;135;32~": MoreKeys.ControlAltSpace,
            "\x1b[126;135u": MoreKeys.ControlAltTilde,
            "\x1b[27;135;126~": MoreKeys.ControlAltTilde,
            "\x1b[125;135u": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[27;135;125~": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[124;135u": MoreKeys.ControlAltVerticalline,
            "\x1b[27;135;124~": MoreKeys.ControlAltVerticalline,
            "\x1b[123;135u": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[27;135;123~": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[122;135u": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[27;135;122~": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[121;135u": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[27;135;121~": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[120;135u": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[27;135;120~": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[119;135u": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[27;135;119~": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[118;135u": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[27;135;118~": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[117;135u": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[27;135;117~": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[116;135u": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[27;135;116~": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[115;135u": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[27;135;115~": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[114;135u": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[27;135;114~": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[113;135u": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[27;135;113~": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[112;135u": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[27;135;112~": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[111;135u": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[27;135;111~": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[110;135u": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[27;135;110~": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[109;135u": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[27;135;109~": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[108;135u": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[27;135;108~": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[107;135u": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[27;135;107~": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[106;135u": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[27;135;106~": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[105;135u": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[27;135;105~": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[104;135u": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[27;135;104~": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[103;135u": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[27;135;103~": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[102;135u": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[27;135;102~": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[101;135u": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[27;135;101~": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[100;135u": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[27;135;100~": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[99;135u": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[27;135;99~": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[98;135u": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[27;135;98~": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[97;135u": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[27;135;97~": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[96;135u": MoreKeys.ControlAltGraveaccent,
            "\x1b[27;135;96~": MoreKeys.ControlAltGraveaccent,
            "\x1b[95;135u": MoreKeys.ControlAltLowline,
            "\x1b[27;135;95~": MoreKeys.ControlAltLowline,
            "\x1b[94;135u": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[27;135;94~": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[93;135u": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[27;135;93~": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[92;135u": MoreKeys.ControlAltReversesolidus,
            "\x1b[27;135;92~": MoreKeys.ControlAltReversesolidus,
            "\x1b[91;135u": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[27;135;91~": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[90;135u": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[27;135;90~": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[89;135u": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[27;135;89~": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[88;135u": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[27;135;88~": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[87;135u": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[27;135;87~": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[86;135u": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[27;135;86~": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[85;135u": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[27;135;85~": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[84;135u": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[27;135;84~": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[83;135u": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[27;135;83~": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[82;135u": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[27;135;82~": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[81;135u": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[27;135;81~": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[80;135u": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[27;135;80~": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[79;135u": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[27;135;79~": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[78;135u": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[27;135;78~": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[77;135u": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[27;135;77~": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[76;135u": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[27;135;76~": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[75;135u": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[27;135;75~": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[74;135u": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[27;135;74~": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[73;135u": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[27;135;73~": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[72;135u": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[27;135;72~": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[71;135u": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[27;135;71~": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[70;135u": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[27;135;70~": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[69;135u": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[27;135;69~": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[68;135u": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[27;135;68~": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[67;135u": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[27;135;67~": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[66;135u": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[27;135;66~": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[65;135u": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[27;135;65~": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[64;135u": MoreKeys.ControlAltCommercialat,
            "\x1b[27;135;64~": MoreKeys.ControlAltCommercialat,
            "\x1b[63;135u": MoreKeys.ControlAltQuestionmark,
            "\x1b[27;135;63~": MoreKeys.ControlAltQuestionmark,
            "\x1b[62;135u": MoreKeys.ControlAltGreaterthansign,
            "\x1b[27;135;62~": MoreKeys.ControlAltGreaterthansign,
            "\x1b[61;135u": MoreKeys.ControlAltEqualssign,
            "\x1b[27;135;61~": MoreKeys.ControlAltEqualssign,
            "\x1b[60;135u": MoreKeys.ControlAltLessthansign,
            "\x1b[27;135;60~": MoreKeys.ControlAltLessthansign,
            "\x1b[59;135u": MoreKeys.ControlAltSemicolon,
            "\x1b[27;135;59~": MoreKeys.ControlAltSemicolon,
            "\x1b[58;135u": MoreKeys.ControlAltColon,
            "\x1b[27;135;58~": MoreKeys.ControlAltColon,
            "\x1b[57;135u": MoreKeys.ControlAltDigitnine,
            "\x1b[27;135;57~": MoreKeys.ControlAltDigitnine,
            "\x1b[56;135u": MoreKeys.ControlAltDigiteight,
            "\x1b[27;135;56~": MoreKeys.ControlAltDigiteight,
            "\x1b[55;135u": MoreKeys.ControlAltDigitseven,
            "\x1b[27;135;55~": MoreKeys.ControlAltDigitseven,
            "\x1b[54;135u": MoreKeys.ControlAltDigitsix,
            "\x1b[27;135;54~": MoreKeys.ControlAltDigitsix,
            "\x1b[53;135u": MoreKeys.ControlAltDigitfive,
            "\x1b[27;135;53~": MoreKeys.ControlAltDigitfive,
            "\x1b[52;135u": MoreKeys.ControlAltDigitfour,
            "\x1b[27;135;52~": MoreKeys.ControlAltDigitfour,
            "\x1b[51;135u": MoreKeys.ControlAltDigitthree,
            "\x1b[27;135;51~": MoreKeys.ControlAltDigitthree,
            "\x1b[50;135u": MoreKeys.ControlAltDigittwo,
            "\x1b[27;135;50~": MoreKeys.ControlAltDigittwo,
            "\x1b[49;135u": MoreKeys.ControlAltDigitone,
            "\x1b[27;135;49~": MoreKeys.ControlAltDigitone,
            "\x1b[48;135u": MoreKeys.ControlAltDigitzero,
            "\x1b[27;135;48~": MoreKeys.ControlAltDigitzero,
            "\x1b[47;135u": MoreKeys.ControlAltSolidus,
            "\x1b[27;135;47~": MoreKeys.ControlAltSolidus,
            "\x1b[46;135u": MoreKeys.ControlAltFullstop,
            "\x1b[27;135;46~": MoreKeys.ControlAltFullstop,
            "\x1b[45;135u": MoreKeys.ControlAltHyphenminus,
            "\x1b[27;135;45~": MoreKeys.ControlAltHyphenminus,
            "\x1b[44;135u": MoreKeys.ControlAltComma,
            "\x1b[27;135;44~": MoreKeys.ControlAltComma,
            "\x1b[43;135u": MoreKeys.ControlAltPlussign,
            "\x1b[27;135;43~": MoreKeys.ControlAltPlussign,
            "\x1b[42;135u": MoreKeys.ControlAltAsterisk,
            "\x1b[27;135;42~": MoreKeys.ControlAltAsterisk,
            "\x1b[41;135u": MoreKeys.ControlAltRightparenthesis,
            "\x1b[27;135;41~": MoreKeys.ControlAltRightparenthesis,
            "\x1b[40;135u": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[27;135;40~": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[39;135u": MoreKeys.ControlAltApostrophe,
            "\x1b[27;135;39~": MoreKeys.ControlAltApostrophe,
            "\x1b[38;135u": MoreKeys.ControlAltAmpersand,
            "\x1b[27;135;38~": MoreKeys.ControlAltAmpersand,
            "\x1b[37;135u": MoreKeys.ControlAltPercentsign,
            "\x1b[27;135;37~": MoreKeys.ControlAltPercentsign,
            "\x1b[36;135u": MoreKeys.ControlAltDollarsign,
            "\x1b[27;135;36~": MoreKeys.ControlAltDollarsign,
            "\x1b[35;135u": MoreKeys.ControlAltNumbersign,
            "\x1b[27;135;35~": MoreKeys.ControlAltNumbersign,
            "\x1b[34;135u": MoreKeys.ControlAltQuotationmark,
            "\x1b[27;135;34~": MoreKeys.ControlAltQuotationmark,
            "\x1b[33;135u": MoreKeys.ControlAltExclamationmark,
            "\x1b[27;135;33~": MoreKeys.ControlAltExclamationmark,
            "\x1b[127;135u": MoreKeys.ControlAltBackspace,
            "\x1b[27;135;127~": MoreKeys.ControlAltBackspace,
            "\x1b[2;135~": MoreKeys.ControlAltInsert,
            "\x1b[3;135~": MoreKeys.ControlAltDelete,
            "\x1b[5;135~": MoreKeys.ControlAltPageup,
            "\x1b[6;135~": MoreKeys.ControlAltPagedown,
            "\x1b[1;135A": MoreKeys.ControlAltUp,
            "\x1b[1;135B": MoreKeys.ControlAltDown,
            "\x1b[1;135C": MoreKeys.ControlAltRight,
            "\x1b[1;135D": MoreKeys.ControlAltLeft,
            "\x1b[1;135H": MoreKeys.ControlAltHome,
            "\x1b[1;135F": MoreKeys.ControlAltEnd,
            "\x1b[57361;135u": MoreKeys.ControlAltPrintscreen,
            "\x1b[57362;135u": MoreKeys.ControlAltPause,
            "\x1b[57363;135u": MoreKeys.ControlAltMenu,
            "\x1b[1;135P": MoreKeys.ControlAltF1,
            "\x1b[1;135Q": MoreKeys.ControlAltF2,
            "\x1b[13;135~": MoreKeys.ControlAltF3,
            "\x1b[1;135S": MoreKeys.ControlAltF4,
            "\x1b[15;135~": MoreKeys.ControlAltF5,
            "\x1b[17;135~": MoreKeys.ControlAltF6,
            "\x1b[18;135~": MoreKeys.ControlAltF7,
            "\x1b[19;135~": MoreKeys.ControlAltF8,
            "\x1b[20;135~": MoreKeys.ControlAltF9,
            "\x1b[21;135~": MoreKeys.ControlAltF10,
            "\x1b[23;135~": MoreKeys.ControlAltF11,
            "\x1b[24;135~": MoreKeys.ControlAltF12,
            "\x1b[9;195u": MoreKeys.AltTab,
            "\x1b[27;195;9~": MoreKeys.AltTab,
            "\x1b[13;195u": MoreKeys.AltEnter,
            "\x1b[27;195;13~": MoreKeys.AltEnter,
            "\x1b[27;195u": MoreKeys.AltEscape,
            "\x1b[27;195;27~": MoreKeys.AltEscape,
            "\x1b[32;195u": MoreKeys.AltSpace,
            "\x1b[27;195;32~": MoreKeys.AltSpace,
            "\x1b[126;195u": MoreKeys.AltTilde,
            "\x1b[27;195;126~": MoreKeys.AltTilde,
            "\x1b[125;195u": MoreKeys.AltRightcurlybracket,
            "\x1b[27;195;125~": MoreKeys.AltRightcurlybracket,
            "\x1b[124;195u": MoreKeys.AltVerticalline,
            "\x1b[27;195;124~": MoreKeys.AltVerticalline,
            "\x1b[123;195u": MoreKeys.AltLeftcurlybracket,
            "\x1b[27;195;123~": MoreKeys.AltLeftcurlybracket,
            "\x1b[122;195u": MoreKeys.AltLatinsmallletterz,
            "\x1b[27;195;122~": MoreKeys.AltLatinsmallletterz,
            "\x1b[121;195u": MoreKeys.AltLatinsmalllettery,
            "\x1b[27;195;121~": MoreKeys.AltLatinsmalllettery,
            "\x1b[120;195u": MoreKeys.AltLatinsmallletterx,
            "\x1b[27;195;120~": MoreKeys.AltLatinsmallletterx,
            "\x1b[119;195u": MoreKeys.AltLatinsmallletterw,
            "\x1b[27;195;119~": MoreKeys.AltLatinsmallletterw,
            "\x1b[118;195u": MoreKeys.AltLatinsmallletterv,
            "\x1b[27;195;118~": MoreKeys.AltLatinsmallletterv,
            "\x1b[117;195u": MoreKeys.AltLatinsmallletteru,
            "\x1b[27;195;117~": MoreKeys.AltLatinsmallletteru,
            "\x1b[116;195u": MoreKeys.AltLatinsmalllettert,
            "\x1b[27;195;116~": MoreKeys.AltLatinsmalllettert,
            "\x1b[115;195u": MoreKeys.AltLatinsmallletters,
            "\x1b[27;195;115~": MoreKeys.AltLatinsmallletters,
            "\x1b[114;195u": MoreKeys.AltLatinsmallletterr,
            "\x1b[27;195;114~": MoreKeys.AltLatinsmallletterr,
            "\x1b[113;195u": MoreKeys.AltLatinsmallletterq,
            "\x1b[27;195;113~": MoreKeys.AltLatinsmallletterq,
            "\x1b[112;195u": MoreKeys.AltLatinsmallletterp,
            "\x1b[27;195;112~": MoreKeys.AltLatinsmallletterp,
            "\x1b[111;195u": MoreKeys.AltLatinsmalllettero,
            "\x1b[27;195;111~": MoreKeys.AltLatinsmalllettero,
            "\x1b[110;195u": MoreKeys.AltLatinsmalllettern,
            "\x1b[27;195;110~": MoreKeys.AltLatinsmalllettern,
            "\x1b[109;195u": MoreKeys.AltLatinsmallletterm,
            "\x1b[27;195;109~": MoreKeys.AltLatinsmallletterm,
            "\x1b[108;195u": MoreKeys.AltLatinsmallletterl,
            "\x1b[27;195;108~": MoreKeys.AltLatinsmallletterl,
            "\x1b[107;195u": MoreKeys.AltLatinsmallletterk,
            "\x1b[27;195;107~": MoreKeys.AltLatinsmallletterk,
            "\x1b[106;195u": MoreKeys.AltLatinsmallletterj,
            "\x1b[27;195;106~": MoreKeys.AltLatinsmallletterj,
            "\x1b[105;195u": MoreKeys.AltLatinsmallletteri,
            "\x1b[27;195;105~": MoreKeys.AltLatinsmallletteri,
            "\x1b[104;195u": MoreKeys.AltLatinsmallletterh,
            "\x1b[27;195;104~": MoreKeys.AltLatinsmallletterh,
            "\x1b[103;195u": MoreKeys.AltLatinsmallletterg,
            "\x1b[27;195;103~": MoreKeys.AltLatinsmallletterg,
            "\x1b[102;195u": MoreKeys.AltLatinsmallletterf,
            "\x1b[27;195;102~": MoreKeys.AltLatinsmallletterf,
            "\x1b[101;195u": MoreKeys.AltLatinsmalllettere,
            "\x1b[27;195;101~": MoreKeys.AltLatinsmalllettere,
            "\x1b[100;195u": MoreKeys.AltLatinsmallletterd,
            "\x1b[27;195;100~": MoreKeys.AltLatinsmallletterd,
            "\x1b[99;195u": MoreKeys.AltLatinsmallletterc,
            "\x1b[27;195;99~": MoreKeys.AltLatinsmallletterc,
            "\x1b[98;195u": MoreKeys.AltLatinsmallletterb,
            "\x1b[27;195;98~": MoreKeys.AltLatinsmallletterb,
            "\x1b[97;195u": MoreKeys.AltLatinsmalllettera,
            "\x1b[27;195;97~": MoreKeys.AltLatinsmalllettera,
            "\x1b[96;195u": MoreKeys.AltGraveaccent,
            "\x1b[27;195;96~": MoreKeys.AltGraveaccent,
            "\x1b[95;195u": MoreKeys.AltLowline,
            "\x1b[27;195;95~": MoreKeys.AltLowline,
            "\x1b[94;195u": MoreKeys.AltCircumflexaccent,
            "\x1b[27;195;94~": MoreKeys.AltCircumflexaccent,
            "\x1b[93;195u": MoreKeys.AltRightsquarebracket,
            "\x1b[27;195;93~": MoreKeys.AltRightsquarebracket,
            "\x1b[92;195u": MoreKeys.AltReversesolidus,
            "\x1b[27;195;92~": MoreKeys.AltReversesolidus,
            "\x1b[91;195u": MoreKeys.AltLeftsquarebracket,
            "\x1b[27;195;91~": MoreKeys.AltLeftsquarebracket,
            "\x1b[90;195u": MoreKeys.AltLatincapitalletterz,
            "\x1b[27;195;90~": MoreKeys.AltLatincapitalletterz,
            "\x1b[89;195u": MoreKeys.AltLatincapitallettery,
            "\x1b[27;195;89~": MoreKeys.AltLatincapitallettery,
            "\x1b[88;195u": MoreKeys.AltLatincapitalletterx,
            "\x1b[27;195;88~": MoreKeys.AltLatincapitalletterx,
            "\x1b[87;195u": MoreKeys.AltLatincapitalletterw,
            "\x1b[27;195;87~": MoreKeys.AltLatincapitalletterw,
            "\x1b[86;195u": MoreKeys.AltLatincapitalletterv,
            "\x1b[27;195;86~": MoreKeys.AltLatincapitalletterv,
            "\x1b[85;195u": MoreKeys.AltLatincapitalletteru,
            "\x1b[27;195;85~": MoreKeys.AltLatincapitalletteru,
            "\x1b[84;195u": MoreKeys.AltLatincapitallettert,
            "\x1b[27;195;84~": MoreKeys.AltLatincapitallettert,
            "\x1b[83;195u": MoreKeys.AltLatincapitalletters,
            "\x1b[27;195;83~": MoreKeys.AltLatincapitalletters,
            "\x1b[82;195u": MoreKeys.AltLatincapitalletterr,
            "\x1b[27;195;82~": MoreKeys.AltLatincapitalletterr,
            "\x1b[81;195u": MoreKeys.AltLatincapitalletterq,
            "\x1b[27;195;81~": MoreKeys.AltLatincapitalletterq,
            "\x1b[80;195u": MoreKeys.AltLatincapitalletterp,
            "\x1b[27;195;80~": MoreKeys.AltLatincapitalletterp,
            "\x1b[79;195u": MoreKeys.AltLatincapitallettero,
            "\x1b[27;195;79~": MoreKeys.AltLatincapitallettero,
            "\x1b[78;195u": MoreKeys.AltLatincapitallettern,
            "\x1b[27;195;78~": MoreKeys.AltLatincapitallettern,
            "\x1b[77;195u": MoreKeys.AltLatincapitalletterm,
            "\x1b[27;195;77~": MoreKeys.AltLatincapitalletterm,
            "\x1b[76;195u": MoreKeys.AltLatincapitalletterl,
            "\x1b[27;195;76~": MoreKeys.AltLatincapitalletterl,
            "\x1b[75;195u": MoreKeys.AltLatincapitalletterk,
            "\x1b[27;195;75~": MoreKeys.AltLatincapitalletterk,
            "\x1b[74;195u": MoreKeys.AltLatincapitalletterj,
            "\x1b[27;195;74~": MoreKeys.AltLatincapitalletterj,
            "\x1b[73;195u": MoreKeys.AltLatincapitalletteri,
            "\x1b[27;195;73~": MoreKeys.AltLatincapitalletteri,
            "\x1b[72;195u": MoreKeys.AltLatincapitalletterh,
            "\x1b[27;195;72~": MoreKeys.AltLatincapitalletterh,
            "\x1b[71;195u": MoreKeys.AltLatincapitalletterg,
            "\x1b[27;195;71~": MoreKeys.AltLatincapitalletterg,
            "\x1b[70;195u": MoreKeys.AltLatincapitalletterf,
            "\x1b[27;195;70~": MoreKeys.AltLatincapitalletterf,
            "\x1b[69;195u": MoreKeys.AltLatincapitallettere,
            "\x1b[27;195;69~": MoreKeys.AltLatincapitallettere,
            "\x1b[68;195u": MoreKeys.AltLatincapitalletterd,
            "\x1b[27;195;68~": MoreKeys.AltLatincapitalletterd,
            "\x1b[67;195u": MoreKeys.AltLatincapitalletterc,
            "\x1b[27;195;67~": MoreKeys.AltLatincapitalletterc,
            "\x1b[66;195u": MoreKeys.AltLatincapitalletterb,
            "\x1b[27;195;66~": MoreKeys.AltLatincapitalletterb,
            "\x1b[65;195u": MoreKeys.AltLatincapitallettera,
            "\x1b[27;195;65~": MoreKeys.AltLatincapitallettera,
            "\x1b[64;195u": MoreKeys.AltCommercialat,
            "\x1b[27;195;64~": MoreKeys.AltCommercialat,
            "\x1b[63;195u": MoreKeys.AltQuestionmark,
            "\x1b[27;195;63~": MoreKeys.AltQuestionmark,
            "\x1b[62;195u": MoreKeys.AltGreaterthansign,
            "\x1b[27;195;62~": MoreKeys.AltGreaterthansign,
            "\x1b[61;195u": MoreKeys.AltEqualssign,
            "\x1b[27;195;61~": MoreKeys.AltEqualssign,
            "\x1b[60;195u": MoreKeys.AltLessthansign,
            "\x1b[27;195;60~": MoreKeys.AltLessthansign,
            "\x1b[59;195u": MoreKeys.AltSemicolon,
            "\x1b[27;195;59~": MoreKeys.AltSemicolon,
            "\x1b[58;195u": MoreKeys.AltColon,
            "\x1b[27;195;58~": MoreKeys.AltColon,
            "\x1b[57;195u": MoreKeys.AltDigitnine,
            "\x1b[27;195;57~": MoreKeys.AltDigitnine,
            "\x1b[56;195u": MoreKeys.AltDigiteight,
            "\x1b[27;195;56~": MoreKeys.AltDigiteight,
            "\x1b[55;195u": MoreKeys.AltDigitseven,
            "\x1b[27;195;55~": MoreKeys.AltDigitseven,
            "\x1b[54;195u": MoreKeys.AltDigitsix,
            "\x1b[27;195;54~": MoreKeys.AltDigitsix,
            "\x1b[53;195u": MoreKeys.AltDigitfive,
            "\x1b[27;195;53~": MoreKeys.AltDigitfive,
            "\x1b[52;195u": MoreKeys.AltDigitfour,
            "\x1b[27;195;52~": MoreKeys.AltDigitfour,
            "\x1b[51;195u": MoreKeys.AltDigitthree,
            "\x1b[27;195;51~": MoreKeys.AltDigitthree,
            "\x1b[50;195u": MoreKeys.AltDigittwo,
            "\x1b[27;195;50~": MoreKeys.AltDigittwo,
            "\x1b[49;195u": MoreKeys.AltDigitone,
            "\x1b[27;195;49~": MoreKeys.AltDigitone,
            "\x1b[48;195u": MoreKeys.AltDigitzero,
            "\x1b[27;195;48~": MoreKeys.AltDigitzero,
            "\x1b[47;195u": MoreKeys.AltSolidus,
            "\x1b[27;195;47~": MoreKeys.AltSolidus,
            "\x1b[46;195u": MoreKeys.AltFullstop,
            "\x1b[27;195;46~": MoreKeys.AltFullstop,
            "\x1b[45;195u": MoreKeys.AltHyphenminus,
            "\x1b[27;195;45~": MoreKeys.AltHyphenminus,
            "\x1b[44;195u": MoreKeys.AltComma,
            "\x1b[27;195;44~": MoreKeys.AltComma,
            "\x1b[43;195u": MoreKeys.AltPlussign,
            "\x1b[27;195;43~": MoreKeys.AltPlussign,
            "\x1b[42;195u": MoreKeys.AltAsterisk,
            "\x1b[27;195;42~": MoreKeys.AltAsterisk,
            "\x1b[41;195u": MoreKeys.AltRightparenthesis,
            "\x1b[27;195;41~": MoreKeys.AltRightparenthesis,
            "\x1b[40;195u": MoreKeys.AltLeftparenthesis,
            "\x1b[27;195;40~": MoreKeys.AltLeftparenthesis,
            "\x1b[39;195u": MoreKeys.AltApostrophe,
            "\x1b[27;195;39~": MoreKeys.AltApostrophe,
            "\x1b[38;195u": MoreKeys.AltAmpersand,
            "\x1b[27;195;38~": MoreKeys.AltAmpersand,
            "\x1b[37;195u": MoreKeys.AltPercentsign,
            "\x1b[27;195;37~": MoreKeys.AltPercentsign,
            "\x1b[36;195u": MoreKeys.AltDollarsign,
            "\x1b[27;195;36~": MoreKeys.AltDollarsign,
            "\x1b[35;195u": MoreKeys.AltNumbersign,
            "\x1b[27;195;35~": MoreKeys.AltNumbersign,
            "\x1b[34;195u": MoreKeys.AltQuotationmark,
            "\x1b[27;195;34~": MoreKeys.AltQuotationmark,
            "\x1b[33;195u": MoreKeys.AltExclamationmark,
            "\x1b[27;195;33~": MoreKeys.AltExclamationmark,
            "\x1b[127;195u": MoreKeys.AltBackspace,
            "\x1b[27;195;127~": MoreKeys.AltBackspace,
            "\x1b[2;195~": MoreKeys.AltInsert,
            "\x1b[3;195~": MoreKeys.AltDelete,
            "\x1b[5;195~": MoreKeys.AltPageup,
            "\x1b[6;195~": MoreKeys.AltPagedown,
            "\x1b[1;195A": MoreKeys.AltUp,
            "\x1b[1;195B": MoreKeys.AltDown,
            "\x1b[1;195C": MoreKeys.AltRight,
            "\x1b[1;195D": MoreKeys.AltLeft,
            "\x1b[1;195H": MoreKeys.AltHome,
            "\x1b[1;195F": MoreKeys.AltEnd,
            "\x1b[57361;195u": MoreKeys.AltPrintscreen,
            "\x1b[57362;195u": MoreKeys.AltPause,
            "\x1b[57363;195u": MoreKeys.AltMenu,
            "\x1b[1;195P": MoreKeys.AltF1,
            "\x1b[1;195Q": MoreKeys.AltF2,
            "\x1b[13;195~": MoreKeys.AltF3,
            "\x1b[1;195S": MoreKeys.AltF4,
            "\x1b[15;195~": MoreKeys.AltF5,
            "\x1b[17;195~": MoreKeys.AltF6,
            "\x1b[18;195~": MoreKeys.AltF7,
            "\x1b[19;195~": MoreKeys.AltF8,
            "\x1b[20;195~": MoreKeys.AltF9,
            "\x1b[21;195~": MoreKeys.AltF10,
            "\x1b[23;195~": MoreKeys.AltF11,
            "\x1b[24;195~": MoreKeys.AltF12,
            "\x1b[9;197u": MoreKeys.ControlTab,
            "\x1b[27;197;9~": MoreKeys.ControlTab,
            "\x1b[13;197u": MoreKeys.ControlEnter,
            "\x1b[27;197;13~": MoreKeys.ControlEnter,
            "\x1b[27;197u": MoreKeys.ControlEscape,
            "\x1b[27;197;27~": MoreKeys.ControlEscape,
            "\x1b[32;197u": MoreKeys.ControlSpace,
            "\x1b[27;197;32~": MoreKeys.ControlSpace,
            "\x1b[126;197u": MoreKeys.ControlTilde,
            "\x1b[27;197;126~": MoreKeys.ControlTilde,
            "\x1b[125;197u": MoreKeys.ControlRightcurlybracket,
            "\x1b[27;197;125~": MoreKeys.ControlRightcurlybracket,
            "\x1b[124;197u": MoreKeys.ControlVerticalline,
            "\x1b[27;197;124~": MoreKeys.ControlVerticalline,
            "\x1b[123;197u": MoreKeys.ControlLeftcurlybracket,
            "\x1b[27;197;123~": MoreKeys.ControlLeftcurlybracket,
            "\x1b[122;197u": MoreKeys.ControlLatinsmallletterz,
            "\x1b[27;197;122~": MoreKeys.ControlLatinsmallletterz,
            "\x1b[121;197u": MoreKeys.ControlLatinsmalllettery,
            "\x1b[27;197;121~": MoreKeys.ControlLatinsmalllettery,
            "\x1b[120;197u": MoreKeys.ControlLatinsmallletterx,
            "\x1b[27;197;120~": MoreKeys.ControlLatinsmallletterx,
            "\x1b[119;197u": MoreKeys.ControlLatinsmallletterw,
            "\x1b[27;197;119~": MoreKeys.ControlLatinsmallletterw,
            "\x1b[118;197u": MoreKeys.ControlLatinsmallletterv,
            "\x1b[27;197;118~": MoreKeys.ControlLatinsmallletterv,
            "\x1b[117;197u": MoreKeys.ControlLatinsmallletteru,
            "\x1b[27;197;117~": MoreKeys.ControlLatinsmallletteru,
            "\x1b[116;197u": MoreKeys.ControlLatinsmalllettert,
            "\x1b[27;197;116~": MoreKeys.ControlLatinsmalllettert,
            "\x1b[115;197u": MoreKeys.ControlLatinsmallletters,
            "\x1b[27;197;115~": MoreKeys.ControlLatinsmallletters,
            "\x1b[114;197u": MoreKeys.ControlLatinsmallletterr,
            "\x1b[27;197;114~": MoreKeys.ControlLatinsmallletterr,
            "\x1b[113;197u": MoreKeys.ControlLatinsmallletterq,
            "\x1b[27;197;113~": MoreKeys.ControlLatinsmallletterq,
            "\x1b[112;197u": MoreKeys.ControlLatinsmallletterp,
            "\x1b[27;197;112~": MoreKeys.ControlLatinsmallletterp,
            "\x1b[111;197u": MoreKeys.ControlLatinsmalllettero,
            "\x1b[27;197;111~": MoreKeys.ControlLatinsmalllettero,
            "\x1b[110;197u": MoreKeys.ControlLatinsmalllettern,
            "\x1b[27;197;110~": MoreKeys.ControlLatinsmalllettern,
            "\x1b[109;197u": MoreKeys.ControlLatinsmallletterm,
            "\x1b[27;197;109~": MoreKeys.ControlLatinsmallletterm,
            "\x1b[108;197u": MoreKeys.ControlLatinsmallletterl,
            "\x1b[27;197;108~": MoreKeys.ControlLatinsmallletterl,
            "\x1b[107;197u": MoreKeys.ControlLatinsmallletterk,
            "\x1b[27;197;107~": MoreKeys.ControlLatinsmallletterk,
            "\x1b[106;197u": MoreKeys.ControlLatinsmallletterj,
            "\x1b[27;197;106~": MoreKeys.ControlLatinsmallletterj,
            "\x1b[105;197u": MoreKeys.ControlLatinsmallletteri,
            "\x1b[27;197;105~": MoreKeys.ControlLatinsmallletteri,
            "\x1b[104;197u": MoreKeys.ControlLatinsmallletterh,
            "\x1b[27;197;104~": MoreKeys.ControlLatinsmallletterh,
            "\x1b[103;197u": MoreKeys.ControlLatinsmallletterg,
            "\x1b[27;197;103~": MoreKeys.ControlLatinsmallletterg,
            "\x1b[102;197u": MoreKeys.ControlLatinsmallletterf,
            "\x1b[27;197;102~": MoreKeys.ControlLatinsmallletterf,
            "\x1b[101;197u": MoreKeys.ControlLatinsmalllettere,
            "\x1b[27;197;101~": MoreKeys.ControlLatinsmalllettere,
            "\x1b[100;197u": MoreKeys.ControlLatinsmallletterd,
            "\x1b[27;197;100~": MoreKeys.ControlLatinsmallletterd,
            "\x1b[99;197u": MoreKeys.ControlLatinsmallletterc,
            "\x1b[27;197;99~": MoreKeys.ControlLatinsmallletterc,
            "\x1b[98;197u": MoreKeys.ControlLatinsmallletterb,
            "\x1b[27;197;98~": MoreKeys.ControlLatinsmallletterb,
            "\x1b[97;197u": MoreKeys.ControlLatinsmalllettera,
            "\x1b[27;197;97~": MoreKeys.ControlLatinsmalllettera,
            "\x1b[96;197u": MoreKeys.ControlGraveaccent,
            "\x1b[27;197;96~": MoreKeys.ControlGraveaccent,
            "\x1b[95;197u": MoreKeys.ControlLowline,
            "\x1b[27;197;95~": MoreKeys.ControlLowline,
            "\x1b[94;197u": MoreKeys.ControlCircumflexaccent,
            "\x1b[27;197;94~": MoreKeys.ControlCircumflexaccent,
            "\x1b[93;197u": MoreKeys.ControlRightsquarebracket,
            "\x1b[27;197;93~": MoreKeys.ControlRightsquarebracket,
            "\x1b[92;197u": MoreKeys.ControlReversesolidus,
            "\x1b[27;197;92~": MoreKeys.ControlReversesolidus,
            "\x1b[91;197u": MoreKeys.ControlLeftsquarebracket,
            "\x1b[27;197;91~": MoreKeys.ControlLeftsquarebracket,
            "\x1b[90;197u": MoreKeys.ControlLatincapitalletterz,
            "\x1b[27;197;90~": MoreKeys.ControlLatincapitalletterz,
            "\x1b[89;197u": MoreKeys.ControlLatincapitallettery,
            "\x1b[27;197;89~": MoreKeys.ControlLatincapitallettery,
            "\x1b[88;197u": MoreKeys.ControlLatincapitalletterx,
            "\x1b[27;197;88~": MoreKeys.ControlLatincapitalletterx,
            "\x1b[87;197u": MoreKeys.ControlLatincapitalletterw,
            "\x1b[27;197;87~": MoreKeys.ControlLatincapitalletterw,
            "\x1b[86;197u": MoreKeys.ControlLatincapitalletterv,
            "\x1b[27;197;86~": MoreKeys.ControlLatincapitalletterv,
            "\x1b[85;197u": MoreKeys.ControlLatincapitalletteru,
            "\x1b[27;197;85~": MoreKeys.ControlLatincapitalletteru,
            "\x1b[84;197u": MoreKeys.ControlLatincapitallettert,
            "\x1b[27;197;84~": MoreKeys.ControlLatincapitallettert,
            "\x1b[83;197u": MoreKeys.ControlLatincapitalletters,
            "\x1b[27;197;83~": MoreKeys.ControlLatincapitalletters,
            "\x1b[82;197u": MoreKeys.ControlLatincapitalletterr,
            "\x1b[27;197;82~": MoreKeys.ControlLatincapitalletterr,
            "\x1b[81;197u": MoreKeys.ControlLatincapitalletterq,
            "\x1b[27;197;81~": MoreKeys.ControlLatincapitalletterq,
            "\x1b[80;197u": MoreKeys.ControlLatincapitalletterp,
            "\x1b[27;197;80~": MoreKeys.ControlLatincapitalletterp,
            "\x1b[79;197u": MoreKeys.ControlLatincapitallettero,
            "\x1b[27;197;79~": MoreKeys.ControlLatincapitallettero,
            "\x1b[78;197u": MoreKeys.ControlLatincapitallettern,
            "\x1b[27;197;78~": MoreKeys.ControlLatincapitallettern,
            "\x1b[77;197u": MoreKeys.ControlLatincapitalletterm,
            "\x1b[27;197;77~": MoreKeys.ControlLatincapitalletterm,
            "\x1b[76;197u": MoreKeys.ControlLatincapitalletterl,
            "\x1b[27;197;76~": MoreKeys.ControlLatincapitalletterl,
            "\x1b[75;197u": MoreKeys.ControlLatincapitalletterk,
            "\x1b[27;197;75~": MoreKeys.ControlLatincapitalletterk,
            "\x1b[74;197u": MoreKeys.ControlLatincapitalletterj,
            "\x1b[27;197;74~": MoreKeys.ControlLatincapitalletterj,
            "\x1b[73;197u": MoreKeys.ControlLatincapitalletteri,
            "\x1b[27;197;73~": MoreKeys.ControlLatincapitalletteri,
            "\x1b[72;197u": MoreKeys.ControlLatincapitalletterh,
            "\x1b[27;197;72~": MoreKeys.ControlLatincapitalletterh,
            "\x1b[71;197u": MoreKeys.ControlLatincapitalletterg,
            "\x1b[27;197;71~": MoreKeys.ControlLatincapitalletterg,
            "\x1b[70;197u": MoreKeys.ControlLatincapitalletterf,
            "\x1b[27;197;70~": MoreKeys.ControlLatincapitalletterf,
            "\x1b[69;197u": MoreKeys.ControlLatincapitallettere,
            "\x1b[27;197;69~": MoreKeys.ControlLatincapitallettere,
            "\x1b[68;197u": MoreKeys.ControlLatincapitalletterd,
            "\x1b[27;197;68~": MoreKeys.ControlLatincapitalletterd,
            "\x1b[67;197u": MoreKeys.ControlLatincapitalletterc,
            "\x1b[27;197;67~": MoreKeys.ControlLatincapitalletterc,
            "\x1b[66;197u": MoreKeys.ControlLatincapitalletterb,
            "\x1b[27;197;66~": MoreKeys.ControlLatincapitalletterb,
            "\x1b[65;197u": MoreKeys.ControlLatincapitallettera,
            "\x1b[27;197;65~": MoreKeys.ControlLatincapitallettera,
            "\x1b[64;197u": MoreKeys.ControlCommercialat,
            "\x1b[27;197;64~": MoreKeys.ControlCommercialat,
            "\x1b[63;197u": MoreKeys.ControlQuestionmark,
            "\x1b[27;197;63~": MoreKeys.ControlQuestionmark,
            "\x1b[62;197u": MoreKeys.ControlGreaterthansign,
            "\x1b[27;197;62~": MoreKeys.ControlGreaterthansign,
            "\x1b[61;197u": MoreKeys.ControlEqualssign,
            "\x1b[27;197;61~": MoreKeys.ControlEqualssign,
            "\x1b[60;197u": MoreKeys.ControlLessthansign,
            "\x1b[27;197;60~": MoreKeys.ControlLessthansign,
            "\x1b[59;197u": MoreKeys.ControlSemicolon,
            "\x1b[27;197;59~": MoreKeys.ControlSemicolon,
            "\x1b[58;197u": MoreKeys.ControlColon,
            "\x1b[27;197;58~": MoreKeys.ControlColon,
            "\x1b[57;197u": MoreKeys.ControlDigitnine,
            "\x1b[27;197;57~": MoreKeys.ControlDigitnine,
            "\x1b[56;197u": MoreKeys.ControlDigiteight,
            "\x1b[27;197;56~": MoreKeys.ControlDigiteight,
            "\x1b[55;197u": MoreKeys.ControlDigitseven,
            "\x1b[27;197;55~": MoreKeys.ControlDigitseven,
            "\x1b[54;197u": MoreKeys.ControlDigitsix,
            "\x1b[27;197;54~": MoreKeys.ControlDigitsix,
            "\x1b[53;197u": MoreKeys.ControlDigitfive,
            "\x1b[27;197;53~": MoreKeys.ControlDigitfive,
            "\x1b[52;197u": MoreKeys.ControlDigitfour,
            "\x1b[27;197;52~": MoreKeys.ControlDigitfour,
            "\x1b[51;197u": MoreKeys.ControlDigitthree,
            "\x1b[27;197;51~": MoreKeys.ControlDigitthree,
            "\x1b[50;197u": MoreKeys.ControlDigittwo,
            "\x1b[27;197;50~": MoreKeys.ControlDigittwo,
            "\x1b[49;197u": MoreKeys.ControlDigitone,
            "\x1b[27;197;49~": MoreKeys.ControlDigitone,
            "\x1b[48;197u": MoreKeys.ControlDigitzero,
            "\x1b[27;197;48~": MoreKeys.ControlDigitzero,
            "\x1b[47;197u": MoreKeys.ControlSolidus,
            "\x1b[27;197;47~": MoreKeys.ControlSolidus,
            "\x1b[46;197u": MoreKeys.ControlFullstop,
            "\x1b[27;197;46~": MoreKeys.ControlFullstop,
            "\x1b[45;197u": MoreKeys.ControlHyphenminus,
            "\x1b[27;197;45~": MoreKeys.ControlHyphenminus,
            "\x1b[44;197u": MoreKeys.ControlComma,
            "\x1b[27;197;44~": MoreKeys.ControlComma,
            "\x1b[43;197u": MoreKeys.ControlPlussign,
            "\x1b[27;197;43~": MoreKeys.ControlPlussign,
            "\x1b[42;197u": MoreKeys.ControlAsterisk,
            "\x1b[27;197;42~": MoreKeys.ControlAsterisk,
            "\x1b[41;197u": MoreKeys.ControlRightparenthesis,
            "\x1b[27;197;41~": MoreKeys.ControlRightparenthesis,
            "\x1b[40;197u": MoreKeys.ControlLeftparenthesis,
            "\x1b[27;197;40~": MoreKeys.ControlLeftparenthesis,
            "\x1b[39;197u": MoreKeys.ControlApostrophe,
            "\x1b[27;197;39~": MoreKeys.ControlApostrophe,
            "\x1b[38;197u": MoreKeys.ControlAmpersand,
            "\x1b[27;197;38~": MoreKeys.ControlAmpersand,
            "\x1b[37;197u": MoreKeys.ControlPercentsign,
            "\x1b[27;197;37~": MoreKeys.ControlPercentsign,
            "\x1b[36;197u": MoreKeys.ControlDollarsign,
            "\x1b[27;197;36~": MoreKeys.ControlDollarsign,
            "\x1b[35;197u": MoreKeys.ControlNumbersign,
            "\x1b[27;197;35~": MoreKeys.ControlNumbersign,
            "\x1b[34;197u": MoreKeys.ControlQuotationmark,
            "\x1b[27;197;34~": MoreKeys.ControlQuotationmark,
            "\x1b[33;197u": MoreKeys.ControlExclamationmark,
            "\x1b[27;197;33~": MoreKeys.ControlExclamationmark,
            "\x1b[127;197u": MoreKeys.ControlBackspace,
            "\x1b[27;197;127~": MoreKeys.ControlBackspace,
            "\x1b[2;197~": MoreKeys.ControlInsert,
            "\x1b[3;197~": MoreKeys.ControlDelete,
            "\x1b[5;197~": MoreKeys.ControlPageup,
            "\x1b[6;197~": MoreKeys.ControlPagedown,
            "\x1b[1;197A": MoreKeys.ControlUp,
            "\x1b[1;197B": MoreKeys.ControlDown,
            "\x1b[1;197C": MoreKeys.ControlRight,
            "\x1b[1;197D": MoreKeys.ControlLeft,
            "\x1b[1;197H": MoreKeys.ControlHome,
            "\x1b[1;197F": MoreKeys.ControlEnd,
            "\x1b[57361;197u": MoreKeys.ControlPrintscreen,
            "\x1b[57362;197u": MoreKeys.ControlPause,
            "\x1b[57363;197u": MoreKeys.ControlMenu,
            "\x1b[1;197P": MoreKeys.ControlF1,
            "\x1b[1;197Q": MoreKeys.ControlF2,
            "\x1b[13;197~": MoreKeys.ControlF3,
            "\x1b[1;197S": MoreKeys.ControlF4,
            "\x1b[15;197~": MoreKeys.ControlF5,
            "\x1b[17;197~": MoreKeys.ControlF6,
            "\x1b[18;197~": MoreKeys.ControlF7,
            "\x1b[19;197~": MoreKeys.ControlF8,
            "\x1b[20;197~": MoreKeys.ControlF9,
            "\x1b[21;197~": MoreKeys.ControlF10,
            "\x1b[23;197~": MoreKeys.ControlF11,
            "\x1b[24;197~": MoreKeys.ControlF12,
            "\x1b[9;72u": MoreKeys.ControlAltShiftTab,
            "\x1b[27;72;9~": MoreKeys.ControlAltShiftTab,
            "\x1b[13;72u": MoreKeys.ControlAltShiftEnter,
            "\x1b[27;72;13~": MoreKeys.ControlAltShiftEnter,
            "\x1b[27;72u": MoreKeys.ControlAltShiftEscape,
            "\x1b[27;72;27~": MoreKeys.ControlAltShiftEscape,
            "\x1b[32;72u": MoreKeys.ControlAltShiftSpace,
            "\x1b[27;72;32~": MoreKeys.ControlAltShiftSpace,
            "\x1b[126;72u": MoreKeys.ControlAltShiftTilde,
            "\x1b[27;72;126~": MoreKeys.ControlAltShiftTilde,
            "\x1b[125;72u": MoreKeys.ControlAltShiftRightcurlybracket,
            "\x1b[27;72;125~": MoreKeys.ControlAltShiftRightcurlybracket,
            "\x1b[124;72u": MoreKeys.ControlAltShiftVerticalline,
            "\x1b[27;72;124~": MoreKeys.ControlAltShiftVerticalline,
            "\x1b[123;72u": MoreKeys.ControlAltShiftLeftcurlybracket,
            "\x1b[27;72;123~": MoreKeys.ControlAltShiftLeftcurlybracket,
            "\x1b[122;72u": MoreKeys.ControlAltShiftLatinsmallletterz,
            "\x1b[27;72;122~": MoreKeys.ControlAltShiftLatinsmallletterz,
            "\x1b[121;72u": MoreKeys.ControlAltShiftLatinsmalllettery,
            "\x1b[27;72;121~": MoreKeys.ControlAltShiftLatinsmalllettery,
            "\x1b[120;72u": MoreKeys.ControlAltShiftLatinsmallletterx,
            "\x1b[27;72;120~": MoreKeys.ControlAltShiftLatinsmallletterx,
            "\x1b[119;72u": MoreKeys.ControlAltShiftLatinsmallletterw,
            "\x1b[27;72;119~": MoreKeys.ControlAltShiftLatinsmallletterw,
            "\x1b[118;72u": MoreKeys.ControlAltShiftLatinsmallletterv,
            "\x1b[27;72;118~": MoreKeys.ControlAltShiftLatinsmallletterv,
            "\x1b[117;72u": MoreKeys.ControlAltShiftLatinsmallletteru,
            "\x1b[27;72;117~": MoreKeys.ControlAltShiftLatinsmallletteru,
            "\x1b[116;72u": MoreKeys.ControlAltShiftLatinsmalllettert,
            "\x1b[27;72;116~": MoreKeys.ControlAltShiftLatinsmalllettert,
            "\x1b[115;72u": MoreKeys.ControlAltShiftLatinsmallletters,
            "\x1b[27;72;115~": MoreKeys.ControlAltShiftLatinsmallletters,
            "\x1b[114;72u": MoreKeys.ControlAltShiftLatinsmallletterr,
            "\x1b[27;72;114~": MoreKeys.ControlAltShiftLatinsmallletterr,
            "\x1b[113;72u": MoreKeys.ControlAltShiftLatinsmallletterq,
            "\x1b[27;72;113~": MoreKeys.ControlAltShiftLatinsmallletterq,
            "\x1b[112;72u": MoreKeys.ControlAltShiftLatinsmallletterp,
            "\x1b[27;72;112~": MoreKeys.ControlAltShiftLatinsmallletterp,
            "\x1b[111;72u": MoreKeys.ControlAltShiftLatinsmalllettero,
            "\x1b[27;72;111~": MoreKeys.ControlAltShiftLatinsmalllettero,
            "\x1b[110;72u": MoreKeys.ControlAltShiftLatinsmalllettern,
            "\x1b[27;72;110~": MoreKeys.ControlAltShiftLatinsmalllettern,
            "\x1b[109;72u": MoreKeys.ControlAltShiftLatinsmallletterm,
            "\x1b[27;72;109~": MoreKeys.ControlAltShiftLatinsmallletterm,
            "\x1b[108;72u": MoreKeys.ControlAltShiftLatinsmallletterl,
            "\x1b[27;72;108~": MoreKeys.ControlAltShiftLatinsmallletterl,
            "\x1b[107;72u": MoreKeys.ControlAltShiftLatinsmallletterk,
            "\x1b[27;72;107~": MoreKeys.ControlAltShiftLatinsmallletterk,
            "\x1b[106;72u": MoreKeys.ControlAltShiftLatinsmallletterj,
            "\x1b[27;72;106~": MoreKeys.ControlAltShiftLatinsmallletterj,
            "\x1b[105;72u": MoreKeys.ControlAltShiftLatinsmallletteri,
            "\x1b[27;72;105~": MoreKeys.ControlAltShiftLatinsmallletteri,
            "\x1b[104;72u": MoreKeys.ControlAltShiftLatinsmallletterh,
            "\x1b[27;72;104~": MoreKeys.ControlAltShiftLatinsmallletterh,
            "\x1b[103;72u": MoreKeys.ControlAltShiftLatinsmallletterg,
            "\x1b[27;72;103~": MoreKeys.ControlAltShiftLatinsmallletterg,
            "\x1b[102;72u": MoreKeys.ControlAltShiftLatinsmallletterf,
            "\x1b[27;72;102~": MoreKeys.ControlAltShiftLatinsmallletterf,
            "\x1b[101;72u": MoreKeys.ControlAltShiftLatinsmalllettere,
            "\x1b[27;72;101~": MoreKeys.ControlAltShiftLatinsmalllettere,
            "\x1b[100;72u": MoreKeys.ControlAltShiftLatinsmallletterd,
            "\x1b[27;72;100~": MoreKeys.ControlAltShiftLatinsmallletterd,
            "\x1b[99;72u": MoreKeys.ControlAltShiftLatinsmallletterc,
            "\x1b[27;72;99~": MoreKeys.ControlAltShiftLatinsmallletterc,
            "\x1b[98;72u": MoreKeys.ControlAltShiftLatinsmallletterb,
            "\x1b[27;72;98~": MoreKeys.ControlAltShiftLatinsmallletterb,
            "\x1b[97;72u": MoreKeys.ControlAltShiftLatinsmalllettera,
            "\x1b[27;72;97~": MoreKeys.ControlAltShiftLatinsmalllettera,
            "\x1b[96;72u": MoreKeys.ControlAltShiftGraveaccent,
            "\x1b[27;72;96~": MoreKeys.ControlAltShiftGraveaccent,
            "\x1b[95;72u": MoreKeys.ControlAltShiftLowline,
            "\x1b[27;72;95~": MoreKeys.ControlAltShiftLowline,
            "\x1b[94;72u": MoreKeys.ControlAltShiftCircumflexaccent,
            "\x1b[27;72;94~": MoreKeys.ControlAltShiftCircumflexaccent,
            "\x1b[93;72u": MoreKeys.ControlAltShiftRightsquarebracket,
            "\x1b[27;72;93~": MoreKeys.ControlAltShiftRightsquarebracket,
            "\x1b[92;72u": MoreKeys.ControlAltShiftReversesolidus,
            "\x1b[27;72;92~": MoreKeys.ControlAltShiftReversesolidus,
            "\x1b[91;72u": MoreKeys.ControlAltShiftLeftsquarebracket,
            "\x1b[27;72;91~": MoreKeys.ControlAltShiftLeftsquarebracket,
            "\x1b[90;72u": MoreKeys.ControlAltShiftLatincapitalletterz,
            "\x1b[27;72;90~": MoreKeys.ControlAltShiftLatincapitalletterz,
            "\x1b[89;72u": MoreKeys.ControlAltShiftLatincapitallettery,
            "\x1b[27;72;89~": MoreKeys.ControlAltShiftLatincapitallettery,
            "\x1b[88;72u": MoreKeys.ControlAltShiftLatincapitalletterx,
            "\x1b[27;72;88~": MoreKeys.ControlAltShiftLatincapitalletterx,
            "\x1b[87;72u": MoreKeys.ControlAltShiftLatincapitalletterw,
            "\x1b[27;72;87~": MoreKeys.ControlAltShiftLatincapitalletterw,
            "\x1b[86;72u": MoreKeys.ControlAltShiftLatincapitalletterv,
            "\x1b[27;72;86~": MoreKeys.ControlAltShiftLatincapitalletterv,
            "\x1b[85;72u": MoreKeys.ControlAltShiftLatincapitalletteru,
            "\x1b[27;72;85~": MoreKeys.ControlAltShiftLatincapitalletteru,
            "\x1b[84;72u": MoreKeys.ControlAltShiftLatincapitallettert,
            "\x1b[27;72;84~": MoreKeys.ControlAltShiftLatincapitallettert,
            "\x1b[83;72u": MoreKeys.ControlAltShiftLatincapitalletters,
            "\x1b[27;72;83~": MoreKeys.ControlAltShiftLatincapitalletters,
            "\x1b[82;72u": MoreKeys.ControlAltShiftLatincapitalletterr,
            "\x1b[27;72;82~": MoreKeys.ControlAltShiftLatincapitalletterr,
            "\x1b[81;72u": MoreKeys.ControlAltShiftLatincapitalletterq,
            "\x1b[27;72;81~": MoreKeys.ControlAltShiftLatincapitalletterq,
            "\x1b[80;72u": MoreKeys.ControlAltShiftLatincapitalletterp,
            "\x1b[27;72;80~": MoreKeys.ControlAltShiftLatincapitalletterp,
            "\x1b[79;72u": MoreKeys.ControlAltShiftLatincapitallettero,
            "\x1b[27;72;79~": MoreKeys.ControlAltShiftLatincapitallettero,
            "\x1b[78;72u": MoreKeys.ControlAltShiftLatincapitallettern,
            "\x1b[27;72;78~": MoreKeys.ControlAltShiftLatincapitallettern,
            "\x1b[77;72u": MoreKeys.ControlAltShiftLatincapitalletterm,
            "\x1b[27;72;77~": MoreKeys.ControlAltShiftLatincapitalletterm,
            "\x1b[76;72u": MoreKeys.ControlAltShiftLatincapitalletterl,
            "\x1b[27;72;76~": MoreKeys.ControlAltShiftLatincapitalletterl,
            "\x1b[75;72u": MoreKeys.ControlAltShiftLatincapitalletterk,
            "\x1b[27;72;75~": MoreKeys.ControlAltShiftLatincapitalletterk,
            "\x1b[74;72u": MoreKeys.ControlAltShiftLatincapitalletterj,
            "\x1b[27;72;74~": MoreKeys.ControlAltShiftLatincapitalletterj,
            "\x1b[73;72u": MoreKeys.ControlAltShiftLatincapitalletteri,
            "\x1b[27;72;73~": MoreKeys.ControlAltShiftLatincapitalletteri,
            "\x1b[72;72u": MoreKeys.ControlAltShiftLatincapitalletterh,
            "\x1b[27;72;72~": MoreKeys.ControlAltShiftLatincapitalletterh,
            "\x1b[71;72u": MoreKeys.ControlAltShiftLatincapitalletterg,
            "\x1b[27;72;71~": MoreKeys.ControlAltShiftLatincapitalletterg,
            "\x1b[70;72u": MoreKeys.ControlAltShiftLatincapitalletterf,
            "\x1b[27;72;70~": MoreKeys.ControlAltShiftLatincapitalletterf,
            "\x1b[69;72u": MoreKeys.ControlAltShiftLatincapitallettere,
            "\x1b[27;72;69~": MoreKeys.ControlAltShiftLatincapitallettere,
            "\x1b[68;72u": MoreKeys.ControlAltShiftLatincapitalletterd,
            "\x1b[27;72;68~": MoreKeys.ControlAltShiftLatincapitalletterd,
            "\x1b[67;72u": MoreKeys.ControlAltShiftLatincapitalletterc,
            "\x1b[27;72;67~": MoreKeys.ControlAltShiftLatincapitalletterc,
            "\x1b[66;72u": MoreKeys.ControlAltShiftLatincapitalletterb,
            "\x1b[27;72;66~": MoreKeys.ControlAltShiftLatincapitalletterb,
            "\x1b[65;72u": MoreKeys.ControlAltShiftLatincapitallettera,
            "\x1b[27;72;65~": MoreKeys.ControlAltShiftLatincapitallettera,
            "\x1b[64;72u": MoreKeys.ControlAltShiftCommercialat,
            "\x1b[27;72;64~": MoreKeys.ControlAltShiftCommercialat,
            "\x1b[63;72u": MoreKeys.ControlAltShiftQuestionmark,
            "\x1b[27;72;63~": MoreKeys.ControlAltShiftQuestionmark,
            "\x1b[62;72u": MoreKeys.ControlAltShiftGreaterthansign,
            "\x1b[27;72;62~": MoreKeys.ControlAltShiftGreaterthansign,
            "\x1b[61;72u": MoreKeys.ControlAltShiftEqualssign,
            "\x1b[27;72;61~": MoreKeys.ControlAltShiftEqualssign,
            "\x1b[60;72u": MoreKeys.ControlAltShiftLessthansign,
            "\x1b[27;72;60~": MoreKeys.ControlAltShiftLessthansign,
            "\x1b[59;72u": MoreKeys.ControlAltShiftSemicolon,
            "\x1b[27;72;59~": MoreKeys.ControlAltShiftSemicolon,
            "\x1b[58;72u": MoreKeys.ControlAltShiftColon,
            "\x1b[27;72;58~": MoreKeys.ControlAltShiftColon,
            "\x1b[57;72u": MoreKeys.ControlAltShiftDigitnine,
            "\x1b[27;72;57~": MoreKeys.ControlAltShiftDigitnine,
            "\x1b[56;72u": MoreKeys.ControlAltShiftDigiteight,
            "\x1b[27;72;56~": MoreKeys.ControlAltShiftDigiteight,
            "\x1b[55;72u": MoreKeys.ControlAltShiftDigitseven,
            "\x1b[27;72;55~": MoreKeys.ControlAltShiftDigitseven,
            "\x1b[54;72u": MoreKeys.ControlAltShiftDigitsix,
            "\x1b[27;72;54~": MoreKeys.ControlAltShiftDigitsix,
            "\x1b[53;72u": MoreKeys.ControlAltShiftDigitfive,
            "\x1b[27;72;53~": MoreKeys.ControlAltShiftDigitfive,
            "\x1b[52;72u": MoreKeys.ControlAltShiftDigitfour,
            "\x1b[27;72;52~": MoreKeys.ControlAltShiftDigitfour,
            "\x1b[51;72u": MoreKeys.ControlAltShiftDigitthree,
            "\x1b[27;72;51~": MoreKeys.ControlAltShiftDigitthree,
            "\x1b[50;72u": MoreKeys.ControlAltShiftDigittwo,
            "\x1b[27;72;50~": MoreKeys.ControlAltShiftDigittwo,
            "\x1b[49;72u": MoreKeys.ControlAltShiftDigitone,
            "\x1b[27;72;49~": MoreKeys.ControlAltShiftDigitone,
            "\x1b[48;72u": MoreKeys.ControlAltShiftDigitzero,
            "\x1b[27;72;48~": MoreKeys.ControlAltShiftDigitzero,
            "\x1b[47;72u": MoreKeys.ControlAltShiftSolidus,
            "\x1b[27;72;47~": MoreKeys.ControlAltShiftSolidus,
            "\x1b[46;72u": MoreKeys.ControlAltShiftFullstop,
            "\x1b[27;72;46~": MoreKeys.ControlAltShiftFullstop,
            "\x1b[45;72u": MoreKeys.ControlAltShiftHyphenminus,
            "\x1b[27;72;45~": MoreKeys.ControlAltShiftHyphenminus,
            "\x1b[44;72u": MoreKeys.ControlAltShiftComma,
            "\x1b[27;72;44~": MoreKeys.ControlAltShiftComma,
            "\x1b[43;72u": MoreKeys.ControlAltShiftPlussign,
            "\x1b[27;72;43~": MoreKeys.ControlAltShiftPlussign,
            "\x1b[42;72u": MoreKeys.ControlAltShiftAsterisk,
            "\x1b[27;72;42~": MoreKeys.ControlAltShiftAsterisk,
            "\x1b[41;72u": MoreKeys.ControlAltShiftRightparenthesis,
            "\x1b[27;72;41~": MoreKeys.ControlAltShiftRightparenthesis,
            "\x1b[40;72u": MoreKeys.ControlAltShiftLeftparenthesis,
            "\x1b[27;72;40~": MoreKeys.ControlAltShiftLeftparenthesis,
            "\x1b[39;72u": MoreKeys.ControlAltShiftApostrophe,
            "\x1b[27;72;39~": MoreKeys.ControlAltShiftApostrophe,
            "\x1b[38;72u": MoreKeys.ControlAltShiftAmpersand,
            "\x1b[27;72;38~": MoreKeys.ControlAltShiftAmpersand,
            "\x1b[37;72u": MoreKeys.ControlAltShiftPercentsign,
            "\x1b[27;72;37~": MoreKeys.ControlAltShiftPercentsign,
            "\x1b[36;72u": MoreKeys.ControlAltShiftDollarsign,
            "\x1b[27;72;36~": MoreKeys.ControlAltShiftDollarsign,
            "\x1b[35;72u": MoreKeys.ControlAltShiftNumbersign,
            "\x1b[27;72;35~": MoreKeys.ControlAltShiftNumbersign,
            "\x1b[34;72u": MoreKeys.ControlAltShiftQuotationmark,
            "\x1b[27;72;34~": MoreKeys.ControlAltShiftQuotationmark,
            "\x1b[33;72u": MoreKeys.ControlAltShiftExclamationmark,
            "\x1b[27;72;33~": MoreKeys.ControlAltShiftExclamationmark,
            "\x1b[127;72u": MoreKeys.ControlAltShiftBackspace,
            "\x1b[27;72;127~": MoreKeys.ControlAltShiftBackspace,
            "\x1b[2;72~": MoreKeys.ControlAltShiftInsert,
            "\x1b[3;72~": MoreKeys.ControlAltShiftDelete,
            "\x1b[5;72~": MoreKeys.ControlAltShiftPageup,
            "\x1b[6;72~": MoreKeys.ControlAltShiftPagedown,
            "\x1b[1;72A": MoreKeys.ControlAltShiftUp,
            "\x1b[1;72B": MoreKeys.ControlAltShiftDown,
            "\x1b[1;72C": MoreKeys.ControlAltShiftRight,
            "\x1b[1;72D": MoreKeys.ControlAltShiftLeft,
            "\x1b[1;72H": MoreKeys.ControlAltShiftHome,
            "\x1b[1;72F": MoreKeys.ControlAltShiftEnd,
            "\x1b[57361;72u": MoreKeys.ControlAltShiftPrintscreen,
            "\x1b[57362;72u": MoreKeys.ControlAltShiftPause,
            "\x1b[57363;72u": MoreKeys.ControlAltShiftMenu,
            "\x1b[1;72P": MoreKeys.ControlAltShiftF1,
            "\x1b[1;72Q": MoreKeys.ControlAltShiftF2,
            "\x1b[13;72~": MoreKeys.ControlAltShiftF3,
            "\x1b[1;72S": MoreKeys.ControlAltShiftF4,
            "\x1b[15;72~": MoreKeys.ControlAltShiftF5,
            "\x1b[17;72~": MoreKeys.ControlAltShiftF6,
            "\x1b[18;72~": MoreKeys.ControlAltShiftF7,
            "\x1b[19;72~": MoreKeys.ControlAltShiftF8,
            "\x1b[20;72~": MoreKeys.ControlAltShiftF9,
            "\x1b[21;72~": MoreKeys.ControlAltShiftF10,
            "\x1b[23;72~": MoreKeys.ControlAltShiftF11,
            "\x1b[24;72~": MoreKeys.ControlAltShiftF12,
            "\x1b[9;136u": MoreKeys.ControlAltShiftTab,
            "\x1b[27;136;9~": MoreKeys.ControlAltShiftTab,
            "\x1b[13;136u": MoreKeys.ControlAltShiftEnter,
            "\x1b[27;136;13~": MoreKeys.ControlAltShiftEnter,
            "\x1b[27;136u": MoreKeys.ControlAltShiftEscape,
            "\x1b[27;136;27~": MoreKeys.ControlAltShiftEscape,
            "\x1b[32;136u": MoreKeys.ControlAltShiftSpace,
            "\x1b[27;136;32~": MoreKeys.ControlAltShiftSpace,
            "\x1b[126;136u": MoreKeys.ControlAltShiftTilde,
            "\x1b[27;136;126~": MoreKeys.ControlAltShiftTilde,
            "\x1b[125;136u": MoreKeys.ControlAltShiftRightcurlybracket,
            "\x1b[27;136;125~": MoreKeys.ControlAltShiftRightcurlybracket,
            "\x1b[124;136u": MoreKeys.ControlAltShiftVerticalline,
            "\x1b[27;136;124~": MoreKeys.ControlAltShiftVerticalline,
            "\x1b[123;136u": MoreKeys.ControlAltShiftLeftcurlybracket,
            "\x1b[27;136;123~": MoreKeys.ControlAltShiftLeftcurlybracket,
            "\x1b[122;136u": MoreKeys.ControlAltShiftLatinsmallletterz,
            "\x1b[27;136;122~": MoreKeys.ControlAltShiftLatinsmallletterz,
            "\x1b[121;136u": MoreKeys.ControlAltShiftLatinsmalllettery,
            "\x1b[27;136;121~": MoreKeys.ControlAltShiftLatinsmalllettery,
            "\x1b[120;136u": MoreKeys.ControlAltShiftLatinsmallletterx,
            "\x1b[27;136;120~": MoreKeys.ControlAltShiftLatinsmallletterx,
            "\x1b[119;136u": MoreKeys.ControlAltShiftLatinsmallletterw,
            "\x1b[27;136;119~": MoreKeys.ControlAltShiftLatinsmallletterw,
            "\x1b[118;136u": MoreKeys.ControlAltShiftLatinsmallletterv,
            "\x1b[27;136;118~": MoreKeys.ControlAltShiftLatinsmallletterv,
            "\x1b[117;136u": MoreKeys.ControlAltShiftLatinsmallletteru,
            "\x1b[27;136;117~": MoreKeys.ControlAltShiftLatinsmallletteru,
            "\x1b[116;136u": MoreKeys.ControlAltShiftLatinsmalllettert,
            "\x1b[27;136;116~": MoreKeys.ControlAltShiftLatinsmalllettert,
            "\x1b[115;136u": MoreKeys.ControlAltShiftLatinsmallletters,
            "\x1b[27;136;115~": MoreKeys.ControlAltShiftLatinsmallletters,
            "\x1b[114;136u": MoreKeys.ControlAltShiftLatinsmallletterr,
            "\x1b[27;136;114~": MoreKeys.ControlAltShiftLatinsmallletterr,
            "\x1b[113;136u": MoreKeys.ControlAltShiftLatinsmallletterq,
            "\x1b[27;136;113~": MoreKeys.ControlAltShiftLatinsmallletterq,
            "\x1b[112;136u": MoreKeys.ControlAltShiftLatinsmallletterp,
            "\x1b[27;136;112~": MoreKeys.ControlAltShiftLatinsmallletterp,
            "\x1b[111;136u": MoreKeys.ControlAltShiftLatinsmalllettero,
            "\x1b[27;136;111~": MoreKeys.ControlAltShiftLatinsmalllettero,
            "\x1b[110;136u": MoreKeys.ControlAltShiftLatinsmalllettern,
            "\x1b[27;136;110~": MoreKeys.ControlAltShiftLatinsmalllettern,
            "\x1b[109;136u": MoreKeys.ControlAltShiftLatinsmallletterm,
            "\x1b[27;136;109~": MoreKeys.ControlAltShiftLatinsmallletterm,
            "\x1b[108;136u": MoreKeys.ControlAltShiftLatinsmallletterl,
            "\x1b[27;136;108~": MoreKeys.ControlAltShiftLatinsmallletterl,
            "\x1b[107;136u": MoreKeys.ControlAltShiftLatinsmallletterk,
            "\x1b[27;136;107~": MoreKeys.ControlAltShiftLatinsmallletterk,
            "\x1b[106;136u": MoreKeys.ControlAltShiftLatinsmallletterj,
            "\x1b[27;136;106~": MoreKeys.ControlAltShiftLatinsmallletterj,
            "\x1b[105;136u": MoreKeys.ControlAltShiftLatinsmallletteri,
            "\x1b[27;136;105~": MoreKeys.ControlAltShiftLatinsmallletteri,
            "\x1b[104;136u": MoreKeys.ControlAltShiftLatinsmallletterh,
            "\x1b[27;136;104~": MoreKeys.ControlAltShiftLatinsmallletterh,
            "\x1b[103;136u": MoreKeys.ControlAltShiftLatinsmallletterg,
            "\x1b[27;136;103~": MoreKeys.ControlAltShiftLatinsmallletterg,
            "\x1b[102;136u": MoreKeys.ControlAltShiftLatinsmallletterf,
            "\x1b[27;136;102~": MoreKeys.ControlAltShiftLatinsmallletterf,
            "\x1b[101;136u": MoreKeys.ControlAltShiftLatinsmalllettere,
            "\x1b[27;136;101~": MoreKeys.ControlAltShiftLatinsmalllettere,
            "\x1b[100;136u": MoreKeys.ControlAltShiftLatinsmallletterd,
            "\x1b[27;136;100~": MoreKeys.ControlAltShiftLatinsmallletterd,
            "\x1b[99;136u": MoreKeys.ControlAltShiftLatinsmallletterc,
            "\x1b[27;136;99~": MoreKeys.ControlAltShiftLatinsmallletterc,
            "\x1b[98;136u": MoreKeys.ControlAltShiftLatinsmallletterb,
            "\x1b[27;136;98~": MoreKeys.ControlAltShiftLatinsmallletterb,
            "\x1b[97;136u": MoreKeys.ControlAltShiftLatinsmalllettera,
            "\x1b[27;136;97~": MoreKeys.ControlAltShiftLatinsmalllettera,
            "\x1b[96;136u": MoreKeys.ControlAltShiftGraveaccent,
            "\x1b[27;136;96~": MoreKeys.ControlAltShiftGraveaccent,
            "\x1b[95;136u": MoreKeys.ControlAltShiftLowline,
            "\x1b[27;136;95~": MoreKeys.ControlAltShiftLowline,
            "\x1b[94;136u": MoreKeys.ControlAltShiftCircumflexaccent,
            "\x1b[27;136;94~": MoreKeys.ControlAltShiftCircumflexaccent,
            "\x1b[93;136u": MoreKeys.ControlAltShiftRightsquarebracket,
            "\x1b[27;136;93~": MoreKeys.ControlAltShiftRightsquarebracket,
            "\x1b[92;136u": MoreKeys.ControlAltShiftReversesolidus,
            "\x1b[27;136;92~": MoreKeys.ControlAltShiftReversesolidus,
            "\x1b[91;136u": MoreKeys.ControlAltShiftLeftsquarebracket,
            "\x1b[27;136;91~": MoreKeys.ControlAltShiftLeftsquarebracket,
            "\x1b[90;136u": MoreKeys.ControlAltShiftLatincapitalletterz,
            "\x1b[27;136;90~": MoreKeys.ControlAltShiftLatincapitalletterz,
            "\x1b[89;136u": MoreKeys.ControlAltShiftLatincapitallettery,
            "\x1b[27;136;89~": MoreKeys.ControlAltShiftLatincapitallettery,
            "\x1b[88;136u": MoreKeys.ControlAltShiftLatincapitalletterx,
            "\x1b[27;136;88~": MoreKeys.ControlAltShiftLatincapitalletterx,
            "\x1b[87;136u": MoreKeys.ControlAltShiftLatincapitalletterw,
            "\x1b[27;136;87~": MoreKeys.ControlAltShiftLatincapitalletterw,
            "\x1b[86;136u": MoreKeys.ControlAltShiftLatincapitalletterv,
            "\x1b[27;136;86~": MoreKeys.ControlAltShiftLatincapitalletterv,
            "\x1b[85;136u": MoreKeys.ControlAltShiftLatincapitalletteru,
            "\x1b[27;136;85~": MoreKeys.ControlAltShiftLatincapitalletteru,
            "\x1b[84;136u": MoreKeys.ControlAltShiftLatincapitallettert,
            "\x1b[27;136;84~": MoreKeys.ControlAltShiftLatincapitallettert,
            "\x1b[83;136u": MoreKeys.ControlAltShiftLatincapitalletters,
            "\x1b[27;136;83~": MoreKeys.ControlAltShiftLatincapitalletters,
            "\x1b[82;136u": MoreKeys.ControlAltShiftLatincapitalletterr,
            "\x1b[27;136;82~": MoreKeys.ControlAltShiftLatincapitalletterr,
            "\x1b[81;136u": MoreKeys.ControlAltShiftLatincapitalletterq,
            "\x1b[27;136;81~": MoreKeys.ControlAltShiftLatincapitalletterq,
            "\x1b[80;136u": MoreKeys.ControlAltShiftLatincapitalletterp,
            "\x1b[27;136;80~": MoreKeys.ControlAltShiftLatincapitalletterp,
            "\x1b[79;136u": MoreKeys.ControlAltShiftLatincapitallettero,
            "\x1b[27;136;79~": MoreKeys.ControlAltShiftLatincapitallettero,
            "\x1b[78;136u": MoreKeys.ControlAltShiftLatincapitallettern,
            "\x1b[27;136;78~": MoreKeys.ControlAltShiftLatincapitallettern,
            "\x1b[77;136u": MoreKeys.ControlAltShiftLatincapitalletterm,
            "\x1b[27;136;77~": MoreKeys.ControlAltShiftLatincapitalletterm,
            "\x1b[76;136u": MoreKeys.ControlAltShiftLatincapitalletterl,
            "\x1b[27;136;76~": MoreKeys.ControlAltShiftLatincapitalletterl,
            "\x1b[75;136u": MoreKeys.ControlAltShiftLatincapitalletterk,
            "\x1b[27;136;75~": MoreKeys.ControlAltShiftLatincapitalletterk,
            "\x1b[74;136u": MoreKeys.ControlAltShiftLatincapitalletterj,
            "\x1b[27;136;74~": MoreKeys.ControlAltShiftLatincapitalletterj,
            "\x1b[73;136u": MoreKeys.ControlAltShiftLatincapitalletteri,
            "\x1b[27;136;73~": MoreKeys.ControlAltShiftLatincapitalletteri,
            "\x1b[72;136u": MoreKeys.ControlAltShiftLatincapitalletterh,
            "\x1b[27;136;72~": MoreKeys.ControlAltShiftLatincapitalletterh,
            "\x1b[71;136u": MoreKeys.ControlAltShiftLatincapitalletterg,
            "\x1b[27;136;71~": MoreKeys.ControlAltShiftLatincapitalletterg,
            "\x1b[70;136u": MoreKeys.ControlAltShiftLatincapitalletterf,
            "\x1b[27;136;70~": MoreKeys.ControlAltShiftLatincapitalletterf,
            "\x1b[69;136u": MoreKeys.ControlAltShiftLatincapitallettere,
            "\x1b[27;136;69~": MoreKeys.ControlAltShiftLatincapitallettere,
            "\x1b[68;136u": MoreKeys.ControlAltShiftLatincapitalletterd,
            "\x1b[27;136;68~": MoreKeys.ControlAltShiftLatincapitalletterd,
            "\x1b[67;136u": MoreKeys.ControlAltShiftLatincapitalletterc,
            "\x1b[27;136;67~": MoreKeys.ControlAltShiftLatincapitalletterc,
            "\x1b[66;136u": MoreKeys.ControlAltShiftLatincapitalletterb,
            "\x1b[27;136;66~": MoreKeys.ControlAltShiftLatincapitalletterb,
            "\x1b[65;136u": MoreKeys.ControlAltShiftLatincapitallettera,
            "\x1b[27;136;65~": MoreKeys.ControlAltShiftLatincapitallettera,
            "\x1b[64;136u": MoreKeys.ControlAltShiftCommercialat,
            "\x1b[27;136;64~": MoreKeys.ControlAltShiftCommercialat,
            "\x1b[63;136u": MoreKeys.ControlAltShiftQuestionmark,
            "\x1b[27;136;63~": MoreKeys.ControlAltShiftQuestionmark,
            "\x1b[62;136u": MoreKeys.ControlAltShiftGreaterthansign,
            "\x1b[27;136;62~": MoreKeys.ControlAltShiftGreaterthansign,
            "\x1b[61;136u": MoreKeys.ControlAltShiftEqualssign,
            "\x1b[27;136;61~": MoreKeys.ControlAltShiftEqualssign,
            "\x1b[60;136u": MoreKeys.ControlAltShiftLessthansign,
            "\x1b[27;136;60~": MoreKeys.ControlAltShiftLessthansign,
            "\x1b[59;136u": MoreKeys.ControlAltShiftSemicolon,
            "\x1b[27;136;59~": MoreKeys.ControlAltShiftSemicolon,
            "\x1b[58;136u": MoreKeys.ControlAltShiftColon,
            "\x1b[27;136;58~": MoreKeys.ControlAltShiftColon,
            "\x1b[57;136u": MoreKeys.ControlAltShiftDigitnine,
            "\x1b[27;136;57~": MoreKeys.ControlAltShiftDigitnine,
            "\x1b[56;136u": MoreKeys.ControlAltShiftDigiteight,
            "\x1b[27;136;56~": MoreKeys.ControlAltShiftDigiteight,
            "\x1b[55;136u": MoreKeys.ControlAltShiftDigitseven,
            "\x1b[27;136;55~": MoreKeys.ControlAltShiftDigitseven,
            "\x1b[54;136u": MoreKeys.ControlAltShiftDigitsix,
            "\x1b[27;136;54~": MoreKeys.ControlAltShiftDigitsix,
            "\x1b[53;136u": MoreKeys.ControlAltShiftDigitfive,
            "\x1b[27;136;53~": MoreKeys.ControlAltShiftDigitfive,
            "\x1b[52;136u": MoreKeys.ControlAltShiftDigitfour,
            "\x1b[27;136;52~": MoreKeys.ControlAltShiftDigitfour,
            "\x1b[51;136u": MoreKeys.ControlAltShiftDigitthree,
            "\x1b[27;136;51~": MoreKeys.ControlAltShiftDigitthree,
            "\x1b[50;136u": MoreKeys.ControlAltShiftDigittwo,
            "\x1b[27;136;50~": MoreKeys.ControlAltShiftDigittwo,
            "\x1b[49;136u": MoreKeys.ControlAltShiftDigitone,
            "\x1b[27;136;49~": MoreKeys.ControlAltShiftDigitone,
            "\x1b[48;136u": MoreKeys.ControlAltShiftDigitzero,
            "\x1b[27;136;48~": MoreKeys.ControlAltShiftDigitzero,
            "\x1b[47;136u": MoreKeys.ControlAltShiftSolidus,
            "\x1b[27;136;47~": MoreKeys.ControlAltShiftSolidus,
            "\x1b[46;136u": MoreKeys.ControlAltShiftFullstop,
            "\x1b[27;136;46~": MoreKeys.ControlAltShiftFullstop,
            "\x1b[45;136u": MoreKeys.ControlAltShiftHyphenminus,
            "\x1b[27;136;45~": MoreKeys.ControlAltShiftHyphenminus,
            "\x1b[44;136u": MoreKeys.ControlAltShiftComma,
            "\x1b[27;136;44~": MoreKeys.ControlAltShiftComma,
            "\x1b[43;136u": MoreKeys.ControlAltShiftPlussign,
            "\x1b[27;136;43~": MoreKeys.ControlAltShiftPlussign,
            "\x1b[42;136u": MoreKeys.ControlAltShiftAsterisk,
            "\x1b[27;136;42~": MoreKeys.ControlAltShiftAsterisk,
            "\x1b[41;136u": MoreKeys.ControlAltShiftRightparenthesis,
            "\x1b[27;136;41~": MoreKeys.ControlAltShiftRightparenthesis,
            "\x1b[40;136u": MoreKeys.ControlAltShiftLeftparenthesis,
            "\x1b[27;136;40~": MoreKeys.ControlAltShiftLeftparenthesis,
            "\x1b[39;136u": MoreKeys.ControlAltShiftApostrophe,
            "\x1b[27;136;39~": MoreKeys.ControlAltShiftApostrophe,
            "\x1b[38;136u": MoreKeys.ControlAltShiftAmpersand,
            "\x1b[27;136;38~": MoreKeys.ControlAltShiftAmpersand,
            "\x1b[37;136u": MoreKeys.ControlAltShiftPercentsign,
            "\x1b[27;136;37~": MoreKeys.ControlAltShiftPercentsign,
            "\x1b[36;136u": MoreKeys.ControlAltShiftDollarsign,
            "\x1b[27;136;36~": MoreKeys.ControlAltShiftDollarsign,
            "\x1b[35;136u": MoreKeys.ControlAltShiftNumbersign,
            "\x1b[27;136;35~": MoreKeys.ControlAltShiftNumbersign,
            "\x1b[34;136u": MoreKeys.ControlAltShiftQuotationmark,
            "\x1b[27;136;34~": MoreKeys.ControlAltShiftQuotationmark,
            "\x1b[33;136u": MoreKeys.ControlAltShiftExclamationmark,
            "\x1b[27;136;33~": MoreKeys.ControlAltShiftExclamationmark,
            "\x1b[127;136u": MoreKeys.ControlAltShiftBackspace,
            "\x1b[27;136;127~": MoreKeys.ControlAltShiftBackspace,
            "\x1b[2;136~": MoreKeys.ControlAltShiftInsert,
            "\x1b[3;136~": MoreKeys.ControlAltShiftDelete,
            "\x1b[5;136~": MoreKeys.ControlAltShiftPageup,
            "\x1b[6;136~": MoreKeys.ControlAltShiftPagedown,
            "\x1b[1;136A": MoreKeys.ControlAltShiftUp,
            "\x1b[1;136B": MoreKeys.ControlAltShiftDown,
            "\x1b[1;136C": MoreKeys.ControlAltShiftRight,
            "\x1b[1;136D": MoreKeys.ControlAltShiftLeft,
            "\x1b[1;136H": MoreKeys.ControlAltShiftHome,
            "\x1b[1;136F": MoreKeys.ControlAltShiftEnd,
            "\x1b[57361;136u": MoreKeys.ControlAltShiftPrintscreen,
            "\x1b[57362;136u": MoreKeys.ControlAltShiftPause,
            "\x1b[57363;136u": MoreKeys.ControlAltShiftMenu,
            "\x1b[1;136P": MoreKeys.ControlAltShiftF1,
            "\x1b[1;136Q": MoreKeys.ControlAltShiftF2,
            "\x1b[13;136~": MoreKeys.ControlAltShiftF3,
            "\x1b[1;136S": MoreKeys.ControlAltShiftF4,
            "\x1b[15;136~": MoreKeys.ControlAltShiftF5,
            "\x1b[17;136~": MoreKeys.ControlAltShiftF6,
            "\x1b[18;136~": MoreKeys.ControlAltShiftF7,
            "\x1b[19;136~": MoreKeys.ControlAltShiftF8,
            "\x1b[20;136~": MoreKeys.ControlAltShiftF9,
            "\x1b[21;136~": MoreKeys.ControlAltShiftF10,
            "\x1b[23;136~": MoreKeys.ControlAltShiftF11,
            "\x1b[24;136~": MoreKeys.ControlAltShiftF12,
            "\x1b[9;196u": MoreKeys.AltShiftTab,
            "\x1b[27;196;9~": MoreKeys.AltShiftTab,
            "\x1b[13;196u": MoreKeys.AltShiftEnter,
            "\x1b[27;196;13~": MoreKeys.AltShiftEnter,
            "\x1b[27;196u": MoreKeys.AltShiftEscape,
            "\x1b[27;196;27~": MoreKeys.AltShiftEscape,
            "\x1b[32;196u": MoreKeys.AltShiftSpace,
            "\x1b[27;196;32~": MoreKeys.AltShiftSpace,
            "\x1b[126;196u": MoreKeys.AltShiftTilde,
            "\x1b[27;196;126~": MoreKeys.AltShiftTilde,
            "\x1b[125;196u": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[27;196;125~": MoreKeys.AltShiftRightcurlybracket,
            "\x1b[124;196u": MoreKeys.AltShiftVerticalline,
            "\x1b[27;196;124~": MoreKeys.AltShiftVerticalline,
            "\x1b[123;196u": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[27;196;123~": MoreKeys.AltShiftLeftcurlybracket,
            "\x1b[122;196u": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[27;196;122~": MoreKeys.AltShiftLatinsmallletterz,
            "\x1b[121;196u": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[27;196;121~": MoreKeys.AltShiftLatinsmalllettery,
            "\x1b[120;196u": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[27;196;120~": MoreKeys.AltShiftLatinsmallletterx,
            "\x1b[119;196u": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[27;196;119~": MoreKeys.AltShiftLatinsmallletterw,
            "\x1b[118;196u": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[27;196;118~": MoreKeys.AltShiftLatinsmallletterv,
            "\x1b[117;196u": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[27;196;117~": MoreKeys.AltShiftLatinsmallletteru,
            "\x1b[116;196u": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[27;196;116~": MoreKeys.AltShiftLatinsmalllettert,
            "\x1b[115;196u": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[27;196;115~": MoreKeys.AltShiftLatinsmallletters,
            "\x1b[114;196u": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[27;196;114~": MoreKeys.AltShiftLatinsmallletterr,
            "\x1b[113;196u": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[27;196;113~": MoreKeys.AltShiftLatinsmallletterq,
            "\x1b[112;196u": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[27;196;112~": MoreKeys.AltShiftLatinsmallletterp,
            "\x1b[111;196u": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[27;196;111~": MoreKeys.AltShiftLatinsmalllettero,
            "\x1b[110;196u": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[27;196;110~": MoreKeys.AltShiftLatinsmalllettern,
            "\x1b[109;196u": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[27;196;109~": MoreKeys.AltShiftLatinsmallletterm,
            "\x1b[108;196u": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[27;196;108~": MoreKeys.AltShiftLatinsmallletterl,
            "\x1b[107;196u": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[27;196;107~": MoreKeys.AltShiftLatinsmallletterk,
            "\x1b[106;196u": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[27;196;106~": MoreKeys.AltShiftLatinsmallletterj,
            "\x1b[105;196u": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[27;196;105~": MoreKeys.AltShiftLatinsmallletteri,
            "\x1b[104;196u": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[27;196;104~": MoreKeys.AltShiftLatinsmallletterh,
            "\x1b[103;196u": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[27;196;103~": MoreKeys.AltShiftLatinsmallletterg,
            "\x1b[102;196u": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[27;196;102~": MoreKeys.AltShiftLatinsmallletterf,
            "\x1b[101;196u": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[27;196;101~": MoreKeys.AltShiftLatinsmalllettere,
            "\x1b[100;196u": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[27;196;100~": MoreKeys.AltShiftLatinsmallletterd,
            "\x1b[99;196u": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[27;196;99~": MoreKeys.AltShiftLatinsmallletterc,
            "\x1b[98;196u": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[27;196;98~": MoreKeys.AltShiftLatinsmallletterb,
            "\x1b[97;196u": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[27;196;97~": MoreKeys.AltShiftLatinsmalllettera,
            "\x1b[96;196u": MoreKeys.AltShiftGraveaccent,
            "\x1b[27;196;96~": MoreKeys.AltShiftGraveaccent,
            "\x1b[95;196u": MoreKeys.AltShiftLowline,
            "\x1b[27;196;95~": MoreKeys.AltShiftLowline,
            "\x1b[94;196u": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[27;196;94~": MoreKeys.AltShiftCircumflexaccent,
            "\x1b[93;196u": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[27;196;93~": MoreKeys.AltShiftRightsquarebracket,
            "\x1b[92;196u": MoreKeys.AltShiftReversesolidus,
            "\x1b[27;196;92~": MoreKeys.AltShiftReversesolidus,
            "\x1b[91;196u": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[27;196;91~": MoreKeys.AltShiftLeftsquarebracket,
            "\x1b[90;196u": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[27;196;90~": MoreKeys.AltShiftLatincapitalletterz,
            "\x1b[89;196u": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[27;196;89~": MoreKeys.AltShiftLatincapitallettery,
            "\x1b[88;196u": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[27;196;88~": MoreKeys.AltShiftLatincapitalletterx,
            "\x1b[87;196u": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[27;196;87~": MoreKeys.AltShiftLatincapitalletterw,
            "\x1b[86;196u": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[27;196;86~": MoreKeys.AltShiftLatincapitalletterv,
            "\x1b[85;196u": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[27;196;85~": MoreKeys.AltShiftLatincapitalletteru,
            "\x1b[84;196u": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[27;196;84~": MoreKeys.AltShiftLatincapitallettert,
            "\x1b[83;196u": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[27;196;83~": MoreKeys.AltShiftLatincapitalletters,
            "\x1b[82;196u": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[27;196;82~": MoreKeys.AltShiftLatincapitalletterr,
            "\x1b[81;196u": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[27;196;81~": MoreKeys.AltShiftLatincapitalletterq,
            "\x1b[80;196u": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[27;196;80~": MoreKeys.AltShiftLatincapitalletterp,
            "\x1b[79;196u": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[27;196;79~": MoreKeys.AltShiftLatincapitallettero,
            "\x1b[78;196u": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[27;196;78~": MoreKeys.AltShiftLatincapitallettern,
            "\x1b[77;196u": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[27;196;77~": MoreKeys.AltShiftLatincapitalletterm,
            "\x1b[76;196u": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[27;196;76~": MoreKeys.AltShiftLatincapitalletterl,
            "\x1b[75;196u": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[27;196;75~": MoreKeys.AltShiftLatincapitalletterk,
            "\x1b[74;196u": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[27;196;74~": MoreKeys.AltShiftLatincapitalletterj,
            "\x1b[73;196u": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[27;196;73~": MoreKeys.AltShiftLatincapitalletteri,
            "\x1b[72;196u": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[27;196;72~": MoreKeys.AltShiftLatincapitalletterh,
            "\x1b[71;196u": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[27;196;71~": MoreKeys.AltShiftLatincapitalletterg,
            "\x1b[70;196u": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[27;196;70~": MoreKeys.AltShiftLatincapitalletterf,
            "\x1b[69;196u": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[27;196;69~": MoreKeys.AltShiftLatincapitallettere,
            "\x1b[68;196u": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[27;196;68~": MoreKeys.AltShiftLatincapitalletterd,
            "\x1b[67;196u": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[27;196;67~": MoreKeys.AltShiftLatincapitalletterc,
            "\x1b[66;196u": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[27;196;66~": MoreKeys.AltShiftLatincapitalletterb,
            "\x1b[65;196u": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[27;196;65~": MoreKeys.AltShiftLatincapitallettera,
            "\x1b[64;196u": MoreKeys.AltShiftCommercialat,
            "\x1b[27;196;64~": MoreKeys.AltShiftCommercialat,
            "\x1b[63;196u": MoreKeys.AltShiftQuestionmark,
            "\x1b[27;196;63~": MoreKeys.AltShiftQuestionmark,
            "\x1b[62;196u": MoreKeys.AltShiftGreaterthansign,
            "\x1b[27;196;62~": MoreKeys.AltShiftGreaterthansign,
            "\x1b[61;196u": MoreKeys.AltShiftEqualssign,
            "\x1b[27;196;61~": MoreKeys.AltShiftEqualssign,
            "\x1b[60;196u": MoreKeys.AltShiftLessthansign,
            "\x1b[27;196;60~": MoreKeys.AltShiftLessthansign,
            "\x1b[59;196u": MoreKeys.AltShiftSemicolon,
            "\x1b[27;196;59~": MoreKeys.AltShiftSemicolon,
            "\x1b[58;196u": MoreKeys.AltShiftColon,
            "\x1b[27;196;58~": MoreKeys.AltShiftColon,
            "\x1b[57;196u": MoreKeys.AltShiftDigitnine,
            "\x1b[27;196;57~": MoreKeys.AltShiftDigitnine,
            "\x1b[56;196u": MoreKeys.AltShiftDigiteight,
            "\x1b[27;196;56~": MoreKeys.AltShiftDigiteight,
            "\x1b[55;196u": MoreKeys.AltShiftDigitseven,
            "\x1b[27;196;55~": MoreKeys.AltShiftDigitseven,
            "\x1b[54;196u": MoreKeys.AltShiftDigitsix,
            "\x1b[27;196;54~": MoreKeys.AltShiftDigitsix,
            "\x1b[53;196u": MoreKeys.AltShiftDigitfive,
            "\x1b[27;196;53~": MoreKeys.AltShiftDigitfive,
            "\x1b[52;196u": MoreKeys.AltShiftDigitfour,
            "\x1b[27;196;52~": MoreKeys.AltShiftDigitfour,
            "\x1b[51;196u": MoreKeys.AltShiftDigitthree,
            "\x1b[27;196;51~": MoreKeys.AltShiftDigitthree,
            "\x1b[50;196u": MoreKeys.AltShiftDigittwo,
            "\x1b[27;196;50~": MoreKeys.AltShiftDigittwo,
            "\x1b[49;196u": MoreKeys.AltShiftDigitone,
            "\x1b[27;196;49~": MoreKeys.AltShiftDigitone,
            "\x1b[48;196u": MoreKeys.AltShiftDigitzero,
            "\x1b[27;196;48~": MoreKeys.AltShiftDigitzero,
            "\x1b[47;196u": MoreKeys.AltShiftSolidus,
            "\x1b[27;196;47~": MoreKeys.AltShiftSolidus,
            "\x1b[46;196u": MoreKeys.AltShiftFullstop,
            "\x1b[27;196;46~": MoreKeys.AltShiftFullstop,
            "\x1b[45;196u": MoreKeys.AltShiftHyphenminus,
            "\x1b[27;196;45~": MoreKeys.AltShiftHyphenminus,
            "\x1b[44;196u": MoreKeys.AltShiftComma,
            "\x1b[27;196;44~": MoreKeys.AltShiftComma,
            "\x1b[43;196u": MoreKeys.AltShiftPlussign,
            "\x1b[27;196;43~": MoreKeys.AltShiftPlussign,
            "\x1b[42;196u": MoreKeys.AltShiftAsterisk,
            "\x1b[27;196;42~": MoreKeys.AltShiftAsterisk,
            "\x1b[41;196u": MoreKeys.AltShiftRightparenthesis,
            "\x1b[27;196;41~": MoreKeys.AltShiftRightparenthesis,
            "\x1b[40;196u": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[27;196;40~": MoreKeys.AltShiftLeftparenthesis,
            "\x1b[39;196u": MoreKeys.AltShiftApostrophe,
            "\x1b[27;196;39~": MoreKeys.AltShiftApostrophe,
            "\x1b[38;196u": MoreKeys.AltShiftAmpersand,
            "\x1b[27;196;38~": MoreKeys.AltShiftAmpersand,
            "\x1b[37;196u": MoreKeys.AltShiftPercentsign,
            "\x1b[27;196;37~": MoreKeys.AltShiftPercentsign,
            "\x1b[36;196u": MoreKeys.AltShiftDollarsign,
            "\x1b[27;196;36~": MoreKeys.AltShiftDollarsign,
            "\x1b[35;196u": MoreKeys.AltShiftNumbersign,
            "\x1b[27;196;35~": MoreKeys.AltShiftNumbersign,
            "\x1b[34;196u": MoreKeys.AltShiftQuotationmark,
            "\x1b[27;196;34~": MoreKeys.AltShiftQuotationmark,
            "\x1b[33;196u": MoreKeys.AltShiftExclamationmark,
            "\x1b[27;196;33~": MoreKeys.AltShiftExclamationmark,
            "\x1b[127;196u": MoreKeys.AltShiftBackspace,
            "\x1b[27;196;127~": MoreKeys.AltShiftBackspace,
            "\x1b[2;196~": MoreKeys.AltShiftInsert,
            "\x1b[3;196~": MoreKeys.AltShiftDelete,
            "\x1b[5;196~": MoreKeys.AltShiftPageup,
            "\x1b[6;196~": MoreKeys.AltShiftPagedown,
            "\x1b[1;196A": MoreKeys.AltShiftUp,
            "\x1b[1;196B": MoreKeys.AltShiftDown,
            "\x1b[1;196C": MoreKeys.AltShiftRight,
            "\x1b[1;196D": MoreKeys.AltShiftLeft,
            "\x1b[1;196H": MoreKeys.AltShiftHome,
            "\x1b[1;196F": MoreKeys.AltShiftEnd,
            "\x1b[57361;196u": MoreKeys.AltShiftPrintscreen,
            "\x1b[57362;196u": MoreKeys.AltShiftPause,
            "\x1b[57363;196u": MoreKeys.AltShiftMenu,
            "\x1b[1;196P": MoreKeys.AltShiftF1,
            "\x1b[1;196Q": MoreKeys.AltShiftF2,
            "\x1b[13;196~": MoreKeys.AltShiftF3,
            "\x1b[1;196S": MoreKeys.AltShiftF4,
            "\x1b[15;196~": MoreKeys.AltShiftF5,
            "\x1b[17;196~": MoreKeys.AltShiftF6,
            "\x1b[18;196~": MoreKeys.AltShiftF7,
            "\x1b[19;196~": MoreKeys.AltShiftF8,
            "\x1b[20;196~": MoreKeys.AltShiftF9,
            "\x1b[21;196~": MoreKeys.AltShiftF10,
            "\x1b[23;196~": MoreKeys.AltShiftF11,
            "\x1b[24;196~": MoreKeys.AltShiftF12,
            "\x1b[9;198u": MoreKeys.ControlShiftTab,
            "\x1b[27;198;9~": MoreKeys.ControlShiftTab,
            "\x1b[13;198u": MoreKeys.ControlShiftEnter,
            "\x1b[27;198;13~": MoreKeys.ControlShiftEnter,
            "\x1b[27;198u": MoreKeys.ControlShiftEscape,
            "\x1b[27;198;27~": MoreKeys.ControlShiftEscape,
            "\x1b[32;198u": MoreKeys.ControlShiftSpace,
            "\x1b[27;198;32~": MoreKeys.ControlShiftSpace,
            "\x1b[126;198u": MoreKeys.ControlShiftTilde,
            "\x1b[27;198;126~": MoreKeys.ControlShiftTilde,
            "\x1b[125;198u": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[27;198;125~": MoreKeys.ControlShiftRightcurlybracket,
            "\x1b[124;198u": MoreKeys.ControlShiftVerticalline,
            "\x1b[27;198;124~": MoreKeys.ControlShiftVerticalline,
            "\x1b[123;198u": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[27;198;123~": MoreKeys.ControlShiftLeftcurlybracket,
            "\x1b[122;198u": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[27;198;122~": MoreKeys.ControlShiftLatinsmallletterz,
            "\x1b[121;198u": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[27;198;121~": MoreKeys.ControlShiftLatinsmalllettery,
            "\x1b[120;198u": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[27;198;120~": MoreKeys.ControlShiftLatinsmallletterx,
            "\x1b[119;198u": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[27;198;119~": MoreKeys.ControlShiftLatinsmallletterw,
            "\x1b[118;198u": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[27;198;118~": MoreKeys.ControlShiftLatinsmallletterv,
            "\x1b[117;198u": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[27;198;117~": MoreKeys.ControlShiftLatinsmallletteru,
            "\x1b[116;198u": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[27;198;116~": MoreKeys.ControlShiftLatinsmalllettert,
            "\x1b[115;198u": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[27;198;115~": MoreKeys.ControlShiftLatinsmallletters,
            "\x1b[114;198u": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[27;198;114~": MoreKeys.ControlShiftLatinsmallletterr,
            "\x1b[113;198u": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[27;198;113~": MoreKeys.ControlShiftLatinsmallletterq,
            "\x1b[112;198u": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[27;198;112~": MoreKeys.ControlShiftLatinsmallletterp,
            "\x1b[111;198u": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[27;198;111~": MoreKeys.ControlShiftLatinsmalllettero,
            "\x1b[110;198u": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[27;198;110~": MoreKeys.ControlShiftLatinsmalllettern,
            "\x1b[109;198u": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[27;198;109~": MoreKeys.ControlShiftLatinsmallletterm,
            "\x1b[108;198u": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[27;198;108~": MoreKeys.ControlShiftLatinsmallletterl,
            "\x1b[107;198u": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[27;198;107~": MoreKeys.ControlShiftLatinsmallletterk,
            "\x1b[106;198u": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[27;198;106~": MoreKeys.ControlShiftLatinsmallletterj,
            "\x1b[105;198u": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[27;198;105~": MoreKeys.ControlShiftLatinsmallletteri,
            "\x1b[104;198u": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[27;198;104~": MoreKeys.ControlShiftLatinsmallletterh,
            "\x1b[103;198u": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[27;198;103~": MoreKeys.ControlShiftLatinsmallletterg,
            "\x1b[102;198u": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[27;198;102~": MoreKeys.ControlShiftLatinsmallletterf,
            "\x1b[101;198u": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[27;198;101~": MoreKeys.ControlShiftLatinsmalllettere,
            "\x1b[100;198u": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[27;198;100~": MoreKeys.ControlShiftLatinsmallletterd,
            "\x1b[99;198u": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[27;198;99~": MoreKeys.ControlShiftLatinsmallletterc,
            "\x1b[98;198u": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[27;198;98~": MoreKeys.ControlShiftLatinsmallletterb,
            "\x1b[97;198u": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[27;198;97~": MoreKeys.ControlShiftLatinsmalllettera,
            "\x1b[96;198u": MoreKeys.ControlShiftGraveaccent,
            "\x1b[27;198;96~": MoreKeys.ControlShiftGraveaccent,
            "\x1b[95;198u": MoreKeys.ControlShiftLowline,
            "\x1b[27;198;95~": MoreKeys.ControlShiftLowline,
            "\x1b[94;198u": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[27;198;94~": MoreKeys.ControlShiftCircumflexaccent,
            "\x1b[93;198u": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[27;198;93~": MoreKeys.ControlShiftRightsquarebracket,
            "\x1b[92;198u": MoreKeys.ControlShiftReversesolidus,
            "\x1b[27;198;92~": MoreKeys.ControlShiftReversesolidus,
            "\x1b[91;198u": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[27;198;91~": MoreKeys.ControlShiftLeftsquarebracket,
            "\x1b[90;198u": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[27;198;90~": MoreKeys.ControlShiftLatincapitalletterz,
            "\x1b[89;198u": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[27;198;89~": MoreKeys.ControlShiftLatincapitallettery,
            "\x1b[88;198u": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[27;198;88~": MoreKeys.ControlShiftLatincapitalletterx,
            "\x1b[87;198u": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[27;198;87~": MoreKeys.ControlShiftLatincapitalletterw,
            "\x1b[86;198u": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[27;198;86~": MoreKeys.ControlShiftLatincapitalletterv,
            "\x1b[85;198u": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[27;198;85~": MoreKeys.ControlShiftLatincapitalletteru,
            "\x1b[84;198u": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[27;198;84~": MoreKeys.ControlShiftLatincapitallettert,
            "\x1b[83;198u": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[27;198;83~": MoreKeys.ControlShiftLatincapitalletters,
            "\x1b[82;198u": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[27;198;82~": MoreKeys.ControlShiftLatincapitalletterr,
            "\x1b[81;198u": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[27;198;81~": MoreKeys.ControlShiftLatincapitalletterq,
            "\x1b[80;198u": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[27;198;80~": MoreKeys.ControlShiftLatincapitalletterp,
            "\x1b[79;198u": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[27;198;79~": MoreKeys.ControlShiftLatincapitallettero,
            "\x1b[78;198u": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[27;198;78~": MoreKeys.ControlShiftLatincapitallettern,
            "\x1b[77;198u": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[27;198;77~": MoreKeys.ControlShiftLatincapitalletterm,
            "\x1b[76;198u": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[27;198;76~": MoreKeys.ControlShiftLatincapitalletterl,
            "\x1b[75;198u": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[27;198;75~": MoreKeys.ControlShiftLatincapitalletterk,
            "\x1b[74;198u": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[27;198;74~": MoreKeys.ControlShiftLatincapitalletterj,
            "\x1b[73;198u": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[27;198;73~": MoreKeys.ControlShiftLatincapitalletteri,
            "\x1b[72;198u": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[27;198;72~": MoreKeys.ControlShiftLatincapitalletterh,
            "\x1b[71;198u": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[27;198;71~": MoreKeys.ControlShiftLatincapitalletterg,
            "\x1b[70;198u": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[27;198;70~": MoreKeys.ControlShiftLatincapitalletterf,
            "\x1b[69;198u": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[27;198;69~": MoreKeys.ControlShiftLatincapitallettere,
            "\x1b[68;198u": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[27;198;68~": MoreKeys.ControlShiftLatincapitalletterd,
            "\x1b[67;198u": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[27;198;67~": MoreKeys.ControlShiftLatincapitalletterc,
            "\x1b[66;198u": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[27;198;66~": MoreKeys.ControlShiftLatincapitalletterb,
            "\x1b[65;198u": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[27;198;65~": MoreKeys.ControlShiftLatincapitallettera,
            "\x1b[64;198u": MoreKeys.ControlShiftCommercialat,
            "\x1b[27;198;64~": MoreKeys.ControlShiftCommercialat,
            "\x1b[63;198u": MoreKeys.ControlShiftQuestionmark,
            "\x1b[27;198;63~": MoreKeys.ControlShiftQuestionmark,
            "\x1b[62;198u": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[27;198;62~": MoreKeys.ControlShiftGreaterthansign,
            "\x1b[61;198u": MoreKeys.ControlShiftEqualssign,
            "\x1b[27;198;61~": MoreKeys.ControlShiftEqualssign,
            "\x1b[60;198u": MoreKeys.ControlShiftLessthansign,
            "\x1b[27;198;60~": MoreKeys.ControlShiftLessthansign,
            "\x1b[59;198u": MoreKeys.ControlShiftSemicolon,
            "\x1b[27;198;59~": MoreKeys.ControlShiftSemicolon,
            "\x1b[58;198u": MoreKeys.ControlShiftColon,
            "\x1b[27;198;58~": MoreKeys.ControlShiftColon,
            "\x1b[57;198u": MoreKeys.ControlShiftDigitnine,
            "\x1b[27;198;57~": MoreKeys.ControlShiftDigitnine,
            "\x1b[56;198u": MoreKeys.ControlShiftDigiteight,
            "\x1b[27;198;56~": MoreKeys.ControlShiftDigiteight,
            "\x1b[55;198u": MoreKeys.ControlShiftDigitseven,
            "\x1b[27;198;55~": MoreKeys.ControlShiftDigitseven,
            "\x1b[54;198u": MoreKeys.ControlShiftDigitsix,
            "\x1b[27;198;54~": MoreKeys.ControlShiftDigitsix,
            "\x1b[53;198u": MoreKeys.ControlShiftDigitfive,
            "\x1b[27;198;53~": MoreKeys.ControlShiftDigitfive,
            "\x1b[52;198u": MoreKeys.ControlShiftDigitfour,
            "\x1b[27;198;52~": MoreKeys.ControlShiftDigitfour,
            "\x1b[51;198u": MoreKeys.ControlShiftDigitthree,
            "\x1b[27;198;51~": MoreKeys.ControlShiftDigitthree,
            "\x1b[50;198u": MoreKeys.ControlShiftDigittwo,
            "\x1b[27;198;50~": MoreKeys.ControlShiftDigittwo,
            "\x1b[49;198u": MoreKeys.ControlShiftDigitone,
            "\x1b[27;198;49~": MoreKeys.ControlShiftDigitone,
            "\x1b[48;198u": MoreKeys.ControlShiftDigitzero,
            "\x1b[27;198;48~": MoreKeys.ControlShiftDigitzero,
            "\x1b[47;198u": MoreKeys.ControlShiftSolidus,
            "\x1b[27;198;47~": MoreKeys.ControlShiftSolidus,
            "\x1b[46;198u": MoreKeys.ControlShiftFullstop,
            "\x1b[27;198;46~": MoreKeys.ControlShiftFullstop,
            "\x1b[45;198u": MoreKeys.ControlShiftHyphenminus,
            "\x1b[27;198;45~": MoreKeys.ControlShiftHyphenminus,
            "\x1b[44;198u": MoreKeys.ControlShiftComma,
            "\x1b[27;198;44~": MoreKeys.ControlShiftComma,
            "\x1b[43;198u": MoreKeys.ControlShiftPlussign,
            "\x1b[27;198;43~": MoreKeys.ControlShiftPlussign,
            "\x1b[42;198u": MoreKeys.ControlShiftAsterisk,
            "\x1b[27;198;42~": MoreKeys.ControlShiftAsterisk,
            "\x1b[41;198u": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[27;198;41~": MoreKeys.ControlShiftRightparenthesis,
            "\x1b[40;198u": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[27;198;40~": MoreKeys.ControlShiftLeftparenthesis,
            "\x1b[39;198u": MoreKeys.ControlShiftApostrophe,
            "\x1b[27;198;39~": MoreKeys.ControlShiftApostrophe,
            "\x1b[38;198u": MoreKeys.ControlShiftAmpersand,
            "\x1b[27;198;38~": MoreKeys.ControlShiftAmpersand,
            "\x1b[37;198u": MoreKeys.ControlShiftPercentsign,
            "\x1b[27;198;37~": MoreKeys.ControlShiftPercentsign,
            "\x1b[36;198u": MoreKeys.ControlShiftDollarsign,
            "\x1b[27;198;36~": MoreKeys.ControlShiftDollarsign,
            "\x1b[35;198u": MoreKeys.ControlShiftNumbersign,
            "\x1b[27;198;35~": MoreKeys.ControlShiftNumbersign,
            "\x1b[34;198u": MoreKeys.ControlShiftQuotationmark,
            "\x1b[27;198;34~": MoreKeys.ControlShiftQuotationmark,
            "\x1b[33;198u": MoreKeys.ControlShiftExclamationmark,
            "\x1b[27;198;33~": MoreKeys.ControlShiftExclamationmark,
            "\x1b[127;198u": MoreKeys.ControlShiftBackspace,
            "\x1b[27;198;127~": MoreKeys.ControlShiftBackspace,
            "\x1b[2;198~": MoreKeys.ControlShiftInsert,
            "\x1b[3;198~": MoreKeys.ControlShiftDelete,
            "\x1b[5;198~": MoreKeys.ControlShiftPageup,
            "\x1b[6;198~": MoreKeys.ControlShiftPagedown,
            "\x1b[1;198A": MoreKeys.ControlShiftUp,
            "\x1b[1;198B": MoreKeys.ControlShiftDown,
            "\x1b[1;198C": MoreKeys.ControlShiftRight,
            "\x1b[1;198D": MoreKeys.ControlShiftLeft,
            "\x1b[1;198H": MoreKeys.ControlShiftHome,
            "\x1b[1;198F": MoreKeys.ControlShiftEnd,
            "\x1b[57361;198u": MoreKeys.ControlShiftPrintscreen,
            "\x1b[57362;198u": MoreKeys.ControlShiftPause,
            "\x1b[57363;198u": MoreKeys.ControlShiftMenu,
            "\x1b[1;198P": MoreKeys.ControlShiftF1,
            "\x1b[1;198Q": MoreKeys.ControlShiftF2,
            "\x1b[13;198~": MoreKeys.ControlShiftF3,
            "\x1b[1;198S": MoreKeys.ControlShiftF4,
            "\x1b[15;198~": MoreKeys.ControlShiftF5,
            "\x1b[17;198~": MoreKeys.ControlShiftF6,
            "\x1b[18;198~": MoreKeys.ControlShiftF7,
            "\x1b[19;198~": MoreKeys.ControlShiftF8,
            "\x1b[20;198~": MoreKeys.ControlShiftF9,
            "\x1b[21;198~": MoreKeys.ControlShiftF10,
            "\x1b[23;198~": MoreKeys.ControlShiftF11,
            "\x1b[24;198~": MoreKeys.ControlShiftF12,
            "\x1b[9;199u": MoreKeys.ControlAltTab,
            "\x1b[27;199;9~": MoreKeys.ControlAltTab,
            "\x1b[13;199u": MoreKeys.ControlAltEnter,
            "\x1b[27;199;13~": MoreKeys.ControlAltEnter,
            "\x1b[27;199u": MoreKeys.ControlAltEscape,
            "\x1b[27;199;27~": MoreKeys.ControlAltEscape,
            "\x1b[32;199u": MoreKeys.ControlAltSpace,
            "\x1b[27;199;32~": MoreKeys.ControlAltSpace,
            "\x1b[126;199u": MoreKeys.ControlAltTilde,
            "\x1b[27;199;126~": MoreKeys.ControlAltTilde,
            "\x1b[125;199u": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[27;199;125~": MoreKeys.ControlAltRightcurlybracket,
            "\x1b[124;199u": MoreKeys.ControlAltVerticalline,
            "\x1b[27;199;124~": MoreKeys.ControlAltVerticalline,
            "\x1b[123;199u": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[27;199;123~": MoreKeys.ControlAltLeftcurlybracket,
            "\x1b[122;199u": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[27;199;122~": MoreKeys.ControlAltLatinsmallletterz,
            "\x1b[121;199u": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[27;199;121~": MoreKeys.ControlAltLatinsmalllettery,
            "\x1b[120;199u": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[27;199;120~": MoreKeys.ControlAltLatinsmallletterx,
            "\x1b[119;199u": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[27;199;119~": MoreKeys.ControlAltLatinsmallletterw,
            "\x1b[118;199u": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[27;199;118~": MoreKeys.ControlAltLatinsmallletterv,
            "\x1b[117;199u": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[27;199;117~": MoreKeys.ControlAltLatinsmallletteru,
            "\x1b[116;199u": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[27;199;116~": MoreKeys.ControlAltLatinsmalllettert,
            "\x1b[115;199u": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[27;199;115~": MoreKeys.ControlAltLatinsmallletters,
            "\x1b[114;199u": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[27;199;114~": MoreKeys.ControlAltLatinsmallletterr,
            "\x1b[113;199u": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[27;199;113~": MoreKeys.ControlAltLatinsmallletterq,
            "\x1b[112;199u": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[27;199;112~": MoreKeys.ControlAltLatinsmallletterp,
            "\x1b[111;199u": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[27;199;111~": MoreKeys.ControlAltLatinsmalllettero,
            "\x1b[110;199u": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[27;199;110~": MoreKeys.ControlAltLatinsmalllettern,
            "\x1b[109;199u": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[27;199;109~": MoreKeys.ControlAltLatinsmallletterm,
            "\x1b[108;199u": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[27;199;108~": MoreKeys.ControlAltLatinsmallletterl,
            "\x1b[107;199u": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[27;199;107~": MoreKeys.ControlAltLatinsmallletterk,
            "\x1b[106;199u": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[27;199;106~": MoreKeys.ControlAltLatinsmallletterj,
            "\x1b[105;199u": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[27;199;105~": MoreKeys.ControlAltLatinsmallletteri,
            "\x1b[104;199u": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[27;199;104~": MoreKeys.ControlAltLatinsmallletterh,
            "\x1b[103;199u": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[27;199;103~": MoreKeys.ControlAltLatinsmallletterg,
            "\x1b[102;199u": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[27;199;102~": MoreKeys.ControlAltLatinsmallletterf,
            "\x1b[101;199u": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[27;199;101~": MoreKeys.ControlAltLatinsmalllettere,
            "\x1b[100;199u": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[27;199;100~": MoreKeys.ControlAltLatinsmallletterd,
            "\x1b[99;199u": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[27;199;99~": MoreKeys.ControlAltLatinsmallletterc,
            "\x1b[98;199u": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[27;199;98~": MoreKeys.ControlAltLatinsmallletterb,
            "\x1b[97;199u": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[27;199;97~": MoreKeys.ControlAltLatinsmalllettera,
            "\x1b[96;199u": MoreKeys.ControlAltGraveaccent,
            "\x1b[27;199;96~": MoreKeys.ControlAltGraveaccent,
            "\x1b[95;199u": MoreKeys.ControlAltLowline,
            "\x1b[27;199;95~": MoreKeys.ControlAltLowline,
            "\x1b[94;199u": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[27;199;94~": MoreKeys.ControlAltCircumflexaccent,
            "\x1b[93;199u": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[27;199;93~": MoreKeys.ControlAltRightsquarebracket,
            "\x1b[92;199u": MoreKeys.ControlAltReversesolidus,
            "\x1b[27;199;92~": MoreKeys.ControlAltReversesolidus,
            "\x1b[91;199u": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[27;199;91~": MoreKeys.ControlAltLeftsquarebracket,
            "\x1b[90;199u": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[27;199;90~": MoreKeys.ControlAltLatincapitalletterz,
            "\x1b[89;199u": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[27;199;89~": MoreKeys.ControlAltLatincapitallettery,
            "\x1b[88;199u": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[27;199;88~": MoreKeys.ControlAltLatincapitalletterx,
            "\x1b[87;199u": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[27;199;87~": MoreKeys.ControlAltLatincapitalletterw,
            "\x1b[86;199u": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[27;199;86~": MoreKeys.ControlAltLatincapitalletterv,
            "\x1b[85;199u": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[27;199;85~": MoreKeys.ControlAltLatincapitalletteru,
            "\x1b[84;199u": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[27;199;84~": MoreKeys.ControlAltLatincapitallettert,
            "\x1b[83;199u": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[27;199;83~": MoreKeys.ControlAltLatincapitalletters,
            "\x1b[82;199u": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[27;199;82~": MoreKeys.ControlAltLatincapitalletterr,
            "\x1b[81;199u": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[27;199;81~": MoreKeys.ControlAltLatincapitalletterq,
            "\x1b[80;199u": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[27;199;80~": MoreKeys.ControlAltLatincapitalletterp,
            "\x1b[79;199u": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[27;199;79~": MoreKeys.ControlAltLatincapitallettero,
            "\x1b[78;199u": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[27;199;78~": MoreKeys.ControlAltLatincapitallettern,
            "\x1b[77;199u": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[27;199;77~": MoreKeys.ControlAltLatincapitalletterm,
            "\x1b[76;199u": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[27;199;76~": MoreKeys.ControlAltLatincapitalletterl,
            "\x1b[75;199u": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[27;199;75~": MoreKeys.ControlAltLatincapitalletterk,
            "\x1b[74;199u": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[27;199;74~": MoreKeys.ControlAltLatincapitalletterj,
            "\x1b[73;199u": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[27;199;73~": MoreKeys.ControlAltLatincapitalletteri,
            "\x1b[72;199u": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[27;199;72~": MoreKeys.ControlAltLatincapitalletterh,
            "\x1b[71;199u": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[27;199;71~": MoreKeys.ControlAltLatincapitalletterg,
            "\x1b[70;199u": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[27;199;70~": MoreKeys.ControlAltLatincapitalletterf,
            "\x1b[69;199u": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[27;199;69~": MoreKeys.ControlAltLatincapitallettere,
            "\x1b[68;199u": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[27;199;68~": MoreKeys.ControlAltLatincapitalletterd,
            "\x1b[67;199u": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[27;199;67~": MoreKeys.ControlAltLatincapitalletterc,
            "\x1b[66;199u": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[27;199;66~": MoreKeys.ControlAltLatincapitalletterb,
            "\x1b[65;199u": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[27;199;65~": MoreKeys.ControlAltLatincapitallettera,
            "\x1b[64;199u": MoreKeys.ControlAltCommercialat,
            "\x1b[27;199;64~": MoreKeys.ControlAltCommercialat,
            "\x1b[63;199u": MoreKeys.ControlAltQuestionmark,
            "\x1b[27;199;63~": MoreKeys.ControlAltQuestionmark,
            "\x1b[62;199u": MoreKeys.ControlAltGreaterthansign,
            "\x1b[27;199;62~": MoreKeys.ControlAltGreaterthansign,
            "\x1b[61;199u": MoreKeys.ControlAltEqualssign,
            "\x1b[27;199;61~": MoreKeys.ControlAltEqualssign,
            "\x1b[60;199u": MoreKeys.ControlAltLessthansign,
            "\x1b[27;199;60~": MoreKeys.ControlAltLessthansign,
            "\x1b[59;199u": MoreKeys.ControlAltSemicolon,
            "\x1b[27;199;59~": MoreKeys.ControlAltSemicolon,
            "\x1b[58;199u": MoreKeys.ControlAltColon,
            "\x1b[27;199;58~": MoreKeys.ControlAltColon,
            "\x1b[57;199u": MoreKeys.ControlAltDigitnine,
            "\x1b[27;199;57~": MoreKeys.ControlAltDigitnine,
            "\x1b[56;199u": MoreKeys.ControlAltDigiteight,
            "\x1b[27;199;56~": MoreKeys.ControlAltDigiteight,
            "\x1b[55;199u": MoreKeys.ControlAltDigitseven,
            "\x1b[27;199;55~": MoreKeys.ControlAltDigitseven,
            "\x1b[54;199u": MoreKeys.ControlAltDigitsix,
            "\x1b[27;199;54~": MoreKeys.ControlAltDigitsix,
            "\x1b[53;199u": MoreKeys.ControlAltDigitfive,
            "\x1b[27;199;53~": MoreKeys.ControlAltDigitfive,
            "\x1b[52;199u": MoreKeys.ControlAltDigitfour,
            "\x1b[27;199;52~": MoreKeys.ControlAltDigitfour,
            "\x1b[51;199u": MoreKeys.ControlAltDigitthree,
            "\x1b[27;199;51~": MoreKeys.ControlAltDigitthree,
            "\x1b[50;199u": MoreKeys.ControlAltDigittwo,
            "\x1b[27;199;50~": MoreKeys.ControlAltDigittwo,
            "\x1b[49;199u": MoreKeys.ControlAltDigitone,
            "\x1b[27;199;49~": MoreKeys.ControlAltDigitone,
            "\x1b[48;199u": MoreKeys.ControlAltDigitzero,
            "\x1b[27;199;48~": MoreKeys.ControlAltDigitzero,
            "\x1b[47;199u": MoreKeys.ControlAltSolidus,
            "\x1b[27;199;47~": MoreKeys.ControlAltSolidus,
            "\x1b[46;199u": MoreKeys.ControlAltFullstop,
            "\x1b[27;199;46~": MoreKeys.ControlAltFullstop,
            "\x1b[45;199u": MoreKeys.ControlAltHyphenminus,
            "\x1b[27;199;45~": MoreKeys.ControlAltHyphenminus,
            "\x1b[44;199u": MoreKeys.ControlAltComma,
            "\x1b[27;199;44~": MoreKeys.ControlAltComma,
            "\x1b[43;199u": MoreKeys.ControlAltPlussign,
            "\x1b[27;199;43~": MoreKeys.ControlAltPlussign,
            "\x1b[42;199u": MoreKeys.ControlAltAsterisk,
            "\x1b[27;199;42~": MoreKeys.ControlAltAsterisk,
            "\x1b[41;199u": MoreKeys.ControlAltRightparenthesis,
            "\x1b[27;199;41~": MoreKeys.ControlAltRightparenthesis,
            "\x1b[40;199u": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[27;199;40~": MoreKeys.ControlAltLeftparenthesis,
            "\x1b[39;199u": MoreKeys.ControlAltApostrophe,
            "\x1b[27;199;39~": MoreKeys.ControlAltApostrophe,
            "\x1b[38;199u": MoreKeys.ControlAltAmpersand,
            "\x1b[27;199;38~": MoreKeys.ControlAltAmpersand,
            "\x1b[37;199u": MoreKeys.ControlAltPercentsign,
            "\x1b[27;199;37~": MoreKeys.ControlAltPercentsign,
            "\x1b[36;199u": MoreKeys.ControlAltDollarsign,
            "\x1b[27;199;36~": MoreKeys.ControlAltDollarsign,
            "\x1b[35;199u": MoreKeys.ControlAltNumbersign,
            "\x1b[27;199;35~": MoreKeys.ControlAltNumbersign,
            "\x1b[34;199u": MoreKeys.ControlAltQuotationmark,
            "\x1b[27;199;34~": MoreKeys.ControlAltQuotationmark,
            "\x1b[33;199u": MoreKeys.ControlAltExclamationmark,
            "\x1b[27;199;33~": MoreKeys.ControlAltExclamationmark,
            "\x1b[127;199u": MoreKeys.ControlAltBackspace,
            "\x1b[27;199;127~": MoreKeys.ControlAltBackspace,
            "\x1b[2;199~": MoreKeys.ControlAltInsert,
            "\x1b[3;199~": MoreKeys.ControlAltDelete,
            "\x1b[5;199~": MoreKeys.ControlAltPageup,
            "\x1b[6;199~": MoreKeys.ControlAltPagedown,
            "\x1b[1;199A": MoreKeys.ControlAltUp,
            "\x1b[1;199B": MoreKeys.ControlAltDown,
            "\x1b[1;199C": MoreKeys.ControlAltRight,
            "\x1b[1;199D": MoreKeys.ControlAltLeft,
            "\x1b[1;199H": MoreKeys.ControlAltHome,
            "\x1b[1;199F": MoreKeys.ControlAltEnd,
            "\x1b[57361;199u": MoreKeys.ControlAltPrintscreen,
            "\x1b[57362;199u": MoreKeys.ControlAltPause,
            "\x1b[57363;199u": MoreKeys.ControlAltMenu,
            "\x1b[1;199P": MoreKeys.ControlAltF1,
            "\x1b[1;199Q": MoreKeys.ControlAltF2,
            "\x1b[13;199~": MoreKeys.ControlAltF3,
            "\x1b[1;199S": MoreKeys.ControlAltF4,
            "\x1b[15;199~": MoreKeys.ControlAltF5,
            "\x1b[17;199~": MoreKeys.ControlAltF6,
            "\x1b[18;199~": MoreKeys.ControlAltF7,
            "\x1b[19;199~": MoreKeys.ControlAltF8,
            "\x1b[20;199~": MoreKeys.ControlAltF9,
            "\x1b[21;199~": MoreKeys.ControlAltF10,
            "\x1b[23;199~": MoreKeys.ControlAltF11,
            "\x1b[24;199~": MoreKeys.ControlAltF12,
        },
    )
)
